/*
 * Copyright (c) 2021-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * Copyright (c) 2019-2020 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *	- Redistributions of source code must retain the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer.
 *
 *	- Redistributions in binary form must reproduce the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer in the documentation and/or other materials
 *	  provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#define __STDC_FORMAT_MACROS
#include <inttypes.h>

/* osm_config.h needs to be before all other includes */
#include <opensm/osm_config.h>
#include <errno.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/time.h>

#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_AR_KDOR_ALGORITHM_CPP

#include <opensm/osm_sm.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_base.h>
#include <complib/cl_qmap.h>

#include <algorithm>

#include <iostream>

#include "opensm/osm_ar_kdor_algorithm.hpp"
#include "osm_armgr.hpp"

extern "C" int osm_ucast_dor_build_lfts(void *context);
static osm_log_t *g_osm_log = NULL;
uint16_t CoordinateData::sm_total_set_coordinates = 0;

int ArKdorAlgorithm::Preprocess()
{
	OSM_AR_LOG_ENTER(m_p_osm_log);
	int rc = 0;

	rc = Init();
	if (rc) {
		OSM_AR_LOG_RETURN(m_p_osm_log, rc);
	}

	/* build static lft tables */
	rc = osm_ucast_dor_build_lfts(&m_ar_mgr_.m_p_osm->sm.ucast_mgr);
	if (rc) {
		OSM_AR_LOG_RETURN(m_p_osm_log, rc);
	}

	/* TODO: check if can be done before Init */
	if (!SetCapable()) {
		OSM_AR_LOG_RETURN(m_p_osm_log, 1);
	}

	int unsupported_num = m_ar_mgr_.ARInfoGetGroupCapProcess();
	if (unsupported_num) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_INFO,
			   "%d switches do not support kDOR Cap.\n" ,
			   unsupported_num);
	}

	/* Validate switch capabilities */
	rc = m_ar_mgr_.ValidateSwitchesCapabilities();
	if (rc) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 9201: Failed to validate the required ar capabilities "
			   "for advanced routing configuration\n");
		OSM_AR_LOG_RETURN(m_p_osm_log, rc);
	}

	/*
	 * Calls osm_ar_prepare_tables,
	 * therefore it must be called after ValidateSwitchesCapabilities
	 */
	SetPlftConfiguration();

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
		   "ArKdorAlgorithm::Preprocess Ended. rc:%d\n", rc);
	OSM_AR_LOG_RETURN(m_p_osm_log, rc);
}

int ArKdorAlgorithm::RunCycle()
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	int rc = 0;

	for (int i=0 ; i<1 ; i++) {

		if (Preprocess()) {
			rc = -1;
			break;
		}

		rc = BuildStaticRouteInfo();
		if (rc) {
			m_ar_mgr_.m_is_permanent_error = true;
			break;
		}

		rc = CalculatePortGroups();
		if (rc) {
			break;
		}

		BuildSl2VlPerOpVl(m_ar_mgr_.m_master_db.m_en_sl_mask);

		OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
			   "ArKdorAlgorithm cycle calculation Ended.\n");

		int unsupported_num = m_ar_mgr_.ARInfoSetProcess();
		if (unsupported_num) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 9202: %d switches do not support AR Set.\n" ,
				   unsupported_num);
			m_ar_mgr_.m_is_temporary_error = true;
		}

		SetPlftMap();
	}

	CycleEnd(rc);

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
		   "ArKdorAlgorithm::RunCycle Ended. rc:%d\n", rc);
	OSM_AR_LOG_RETURN(m_p_osm_log, rc);
}

int ArKdorAlgorithm::Init()
{
	OSM_AR_LOG_ENTER(m_p_osm_log);
	OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE, "Init kDOR Algorithm.\n");

	g_osm_log = m_p_osm_log;
	m_ar_mgr_.m_is_permanent_error = false;
	m_ar_mgr_.m_is_temporary_error = false;

	memset(m_sw_lid_to_kdor_data_, 0, sizeof(m_sw_lid_to_kdor_data_));

	uint32_t sw_idx = 0;
	uint32_t switches_number = (uint32_t)m_sw_map_.size();
	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_map_.begin();
	for (; sw_node_it != m_sw_map_.end();  ++sw_node_it) {
		if (sw_node_it->second.m_kdor_data == NULL) {
			/* Allocate KDOR Data for all switches */
			sw_node_it->second.m_kdor_data = new KdorSwData(
				sw_node_it->second, &m_ar_mgr_.m_port_groups_calculator);
			if (sw_node_it->second.m_kdor_data == NULL) {
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
					   "ERR 9203: Failed to allocate kDOR Data.\n");
				OSM_AR_LOG_RETURN(m_p_osm_log, -1);
			}

			sw_node_it->second.m_kdor_data->Resize(
				sw_node_it->second.m_general_sw_info.m_p_osm_sw->p_node->physp_tbl_size);
		}

		int rc = sw_node_it->second.m_kdor_data->Init(
			*this, sw_idx, switches_number);

		if (rc) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 9204: Failed to init kDOR Data. kDOR failed.\n");
			OSM_AR_LOG_RETURN(m_p_osm_log, rc);
		}

		m_sw_lid_to_kdor_data_[sw_node_it->second.m_general_sw_info.m_lid] =
			sw_node_it->second.m_kdor_data;
		sw_idx++;
	}

	if (SetCoordinates()) {
		return 1;
	}

	/* Set Direction for all fabric connections */
	sw_node_it = m_sw_map_.begin();
	for (; sw_node_it != m_sw_map_.end();  ++sw_node_it) {
		if (sw_node_it->second.m_kdor_data->SetConnectionsDirection(*this)) {
			OSM_AR_LOG_RETURN(m_p_osm_log, 1);
		}
	}

	OSM_AR_LOG_RETURN(m_p_osm_log, 0);
}

/* SetCapable also sets calculated VL as SL */
bool ArKdorAlgorithm::SetCapable()
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,"SetCapable algorithm feature: %u\n",
		   m_algorithm_feature_);

	int unsupported_num = 0;
	int disable_by_tr_err = 0;
	bool new_dev = false;
	bool algorithm_supported = true;
	uint8_t configured_max_op_vl = m_ar_mgr_.m_p_osm_subn->opt.max_op_vls_sw ?
		m_ar_mgr_.m_p_osm_subn->opt.max_op_vls_sw :
		m_ar_mgr_.m_p_osm_subn->opt.max_op_vls;

	if (m_ar_mgr_.m_p_osm_subn->opt.qos != TRUE) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 9205: KDOR Algorithm not supported "
			   "together with disabled qos.\n");
		OSM_AR_LOG_RETURN(m_p_osm_log, false);
	}

	/* It's important to have enough VLs on switches */
	if (configured_max_op_vl == 0) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 9206: Unsupported max_op_vls configuration parameter "
			   "value equals zero for kDOR setup.\n");

		OSM_AR_LOG_RETURN(m_p_osm_log, false);
	}
	uint8_t num_vls = (1 << (configured_max_op_vl - 1));

	/* It's important to have enough VLs on switches */
	if (num_vls < m_min_vl_number_) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 9207: Invalid maximal operational VLs %u < %u "
			   "configuration for kDOR setup. max_op_vls %u max_op_vls_sw %u\n",
			   num_vls, m_min_vl_number_, m_ar_mgr_.m_p_osm_subn->opt.max_op_vls,
			   m_ar_mgr_.m_p_osm_subn->opt.max_op_vls_sw);

		OSM_AR_LOG_RETURN(m_p_osm_log, false);
	}

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_map_.begin();
	for (; sw_node_it != m_sw_map_.end(); ++sw_node_it) {

		if (sw_node_it->second.m_osm_update_needed) {
			sw_node_it->second.m_support[m_algorithm_feature_] = SUPPORT_UNKNOWN;
		} else if (sw_node_it->second.m_support[m_algorithm_feature_] == NOT_SUPPORTED) {
			unsupported_num++;
			continue;
		}

		if (!m_ar_mgr_.IsARActive(sw_node_it->second)) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Skip Switch GUID: 0x%" PRIx64 ", LID: %u - " \
				   "AR not supported or not enabled.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			unsupported_num++;
			continue;
		}
		if (!sw_node_it->second.m_general_sw_info.m_p_osm_sw->ar_data.plft_info_updated) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Skip Switch GUID: 0x%" PRIx64 ", LID: %u - " \
				   "PLFT Info is not supported or not enabled.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			unsupported_num++;
			sw_node_it->second.MarkSWNotSupport(m_algorithm_feature_, DF_PLFT_CAP_ZERO);
			continue;
		}

		if (sw_node_it->second.in_temporary_error)
			continue;

		if (!IsValidCoordinate(sw_node_it->second)) {
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
					   "ERR 9208: "
					   "Switch GUID: 0x%" PRIx64 ", LID: %u, "
					   "do not have coordinate. "
					   "Cannot guarantee credit loop free.\n",
					   sw_node_it->second.m_general_sw_info.m_guid,
					   sw_node_it->second.m_general_sw_info.m_lid);
		}

		if (sw_node_it->second.m_support[m_algorithm_feature_] == SUPPORTED)
			continue;

		new_dev = true;

		/*
		 * All switches should support vl2vl configuration
		 * ( ExtendedNodeInfo  SL2VLCap  bit 0 is on)
		 */
		osm_switch_t  *p_osm_sw = sw_node_it->second.m_general_sw_info.m_p_osm_sw;
		osm_node_t *p_osm_node = p_osm_sw->p_node;
		osm_advanced_routing_data_t *p_ar_data = &p_osm_sw->ar_data;

		if (!(p_osm_node->ext_node_info.sl_2_vl_cap & 0x1)) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 9209: "
				   "Switch GUID 0x%" PRIx64 ", LID %u, "  \
				   "does not support VL2VL. kDOR algorithm not supported.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			sw_node_it->second.MarkSWNotSupport(m_algorithm_feature_, DF_NO_VL2VL_CAP);
			algorithm_supported = false;
		} else {
			p_osm_node->new_ext_node_info.sl_2_vl_act = 1; /* Set VL as SL */
		}

		if ((m_ar_mgr_.m_master_db.m_en_tr_mask != AR_CFG_TRANSPORT_MASK_EN_ALL) &&
			!(p_ar_data->ar_info.enabled_and_cap_fields &
			  IB_AR_INFO_CAP_MASK_BY_TRANS_CAP)) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 920A: "
				   "Switch GUID: 0x%" PRIx64 ", LID: %u, "
				   "does not support by_transport_disable.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			disable_by_tr_err++;
		}

		/* All switches should support adaptive Routing checked by ARInfoProcess */
		if (!(p_ar_data->ar_info.enabled_and_cap_fields &
			  IB_AR_INFO_CAP_MASK_BY_SL_CAP)) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 920B: "
				   "Switch GUID 0x%" PRIx64 ", LID %u: " \
				   "BySLCap not supported.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			sw_node_it->second.MarkSWNotSupport(m_algorithm_feature_, DF_NO_AR_BY_SL_CAP);
			unsupported_num++;
			continue;
		}

		if (!(p_ar_data->ar_info.enabled_and_cap_fields &
			  IB_AR_INFO_CAP_MASK_GLB_GROUPS)) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 920C: "
				   "Switch GUID 0x%" PRIx64 ", LID %u: " \
				   "Only global AR groups supported.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			sw_node_it->second.MarkSWNotSupport(m_algorithm_feature_, DF_GLOBAL_AR_GROUP);
		}

		/*
		 * All links between switches should support at least m_min_vl_number_
		 * VLs in PortInfo VLCap
		 * We check VL cap and not operational VLs because configuring
		 * VLs comes after routing calculation
		 */
		osm_physp_t *p_physp_0 = osm_node_get_physp_ptr(p_osm_node, 0);
		if (!p_physp_0) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 920D: "
				   "Switch GUID: 0x%" PRIx64 ", LID: %u - "
				   "Failed to get phys port 0.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			sw_node_it->second.MarkSWNotSupport(m_algorithm_feature_, DB_ERR);
		}

		if (p_physp_0 &&
			!(p_physp_0->port_info.capability_mask & IB_PORT_CAP_HAS_SL_MAP)) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 920E: "
				   "Switch GUID: 0x%" PRIx64 ", LID: %u - "
				   "SL to VL map not supported. kDOR algorithm not supported.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			sw_node_it->second.MarkSWNotSupport(m_algorithm_feature_, DF_NO_SL_MAP_CAP);
			algorithm_supported = false;
		}

		for (u_int8_t port_num = 1; port_num <= p_osm_node->node_info.num_ports;
			port_num++) {

			osm_physp_t *p_physp = osm_node_get_physp_ptr(p_osm_node, port_num);
			if (!p_physp || !osm_link_is_healthy(p_physp))
				continue;

			osm_physp_t *p_remote_physp = osm_physp_get_remote(p_physp);
			if (!p_remote_physp)
				continue;

			u_int8_t remote_node_type = osm_node_get_type(p_remote_physp->p_node);
			if (remote_node_type != IB_NODE_TYPE_SWITCH) {
				continue;
			}

			u_int8_t vl_cap = ib_port_info_get_vl_cap(&p_physp->port_info);
			num_vls = (1 << (vl_cap - 1));

			if (num_vls < m_min_vl_number_) {
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
					   "ERR 920F: "
					   "Invalid number of VLs: %u < %u on "
					   "Switch GUID: 0x%" PRIx64 ", LID: %u, port: %u\n",
					   vl_cap, m_min_vl_number_,
					   sw_node_it->second.m_general_sw_info.m_guid,
					   sw_node_it->second.m_general_sw_info.m_lid,
					   port_num);

				sw_node_it->second.MarkSWNotSupport(m_algorithm_feature_,
								    DF_VL_CAP_LESS_THAN_MIN);
				break;

			}
		}

		if (sw_node_it->second.m_support[m_algorithm_feature_] ==
		    SUPPORT_UNKNOWN) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Mark Switch GUID: 0x%" PRIx64 ", LID: %u "
				   " support algorithm feature:%u \n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid,
				   m_algorithm_feature_);

			sw_node_it->second.MarkSWSupport(m_algorithm_feature_);
		}
	}

	if (unsupported_num || disable_by_tr_err) {
		if (unsupported_num) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_INFO,
				   "%d switches are not kDOR Capable.\n",
				   unsupported_num);
		}
		if (disable_by_tr_err) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_INFO,
				   "%d switches do not support by_transport_disable.\n",
				   disable_by_tr_err);
		}
	}

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,"SetCapable returns %u\n",
		   algorithm_supported);
	OSM_AR_LOG_RETURN(m_p_osm_log, algorithm_supported);
}

int ArKdorAlgorithm::BuildStaticRouteInfo()
{
	int rc = PlftBasedArAlgorithm::BuildStaticRouteInfo();

	if (m_max_vlinc_ >= m_planes_number_) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 9210: "
			   "Max required VLs per SL: %u "
			   "is higher than VL budget per SL: %u. "
			   "Cannot guarantee credit loop free.\n",
			   m_max_vlinc_ +1, m_planes_number_);
	}

	return rc;
}

/*
 * If flip_from_connection is true, the real from_connection is opposite
 * to the given one and therefore have an opposite from dim_sign
 *
 *   TURN_TYPE_0:  +-Low   -> +-High
 *              :   +Xi    ->  +Xi
 *              :   -Xi    ->  -Xi
 *   TURN_TYPE_1:  +-High  ->  +Low
 *   TURN_TYPE_2:  +-High  ->  -Low
 */
ar_algorithm_turn_t ArKdorAlgorithm::GetTurnType(
	const SwitchConnection &from_connection,
	const SwitchConnection &to_connection,
	bool flip_from_connection) const
{
	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
		   "GetTurnType from (dim_idx,dim_sign): (%u,%d) to "
		   "(dim_idx,dim_sign): (%u,%d)\n",
		   from_connection.m_dim_idx, from_connection.m_dim_sign,
		   to_connection.m_dim_idx, to_connection.m_dim_sign);

	if (from_connection.m_remote_switch == NULL ||
		to_connection.m_remote_switch == NULL) {
		/* from or to non switch port */
		return TURN_TYPE_0;
	}
	if (from_connection.m_dim_idx < to_connection.m_dim_idx) {
		return TURN_TYPE_0;
	}
	if ((from_connection.m_dim_idx == to_connection.m_dim_idx) &&
	    ((from_connection.m_dim_sign * (flip_from_connection ? -1 : 1)) ==
	     to_connection.m_dim_sign)) {
		return TURN_TYPE_0;
	}
	if (to_connection.m_dim_sign > 0) {
		return TURN_TYPE_1;
	}
	return TURN_TYPE_2;
}
/*
 * calculate new vl increase and turn type for given connection to sw
 *
 * Vl calculation if last turn type is T1:
 *   T1 | T2 | T1 | T2 | T1
 *   ----------------------
 *      |    |    |    | 0      total vl_inc = 0
 *   ----------------------
 *      |    |    | 1  | 2      total vl_inc = 2 (prev vl inc + 2)
 *   ----------------------
 *      |    |  0 | 1  | 2      total vl_inc = 2 (prev vl inc)
 *   ----------------------
 *      |  1 |  2 | 3  | 4      total vl_inc = 4 (prev vl inc + 2)
 *   ----------------------
 *    0 |  1 |  2 | 3  | 4      total vl_inc = 4 (prev vl inc)
 *
 * Vl calculation if last turn type is T2:
 *   T2 | T1 | T2 | T1 | T2
 *   ----------------------
 *      |    |    |    | 1      total vl_inc = 1
 *   ----------------------
 *      |    |    | 0  | 1      total vl_inc = 1 (prev vl inc)
 *   ----------------------
 *      |    |  1 | 2  | 3      total vl_inc = 2 (prev vl inc + 2)
 *   ----------------------
 *      |  0 |  1 | 2  | 3      total vl_inc = 4 (prev vl inc)
 *   ----------------------
 *    1 |  2 |  3 | 4  | 5      total vl_inc = 5 (prev vl inc + 2)
 */
void ArKdorAlgorithm::CalculateRouteInfo(SwitchConnection &connection,
					 RouteInfo &remote_route_info,
					 RouteInfo &route_info)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);
	route_info.m_connection = &connection;

	ar_algorithm_turn_t turn_type = TURN_TYPE_0;

	if (remote_route_info.m_connection != NULL) {
		/* if remote is destination remote_route_info.m_connection == NULL */
		turn_type = GetTurnType(connection,
					*remote_route_info.m_connection,
					false); /*real from connection*/
	}
	route_info.m_hops_count = remote_route_info.m_hops_count + 1;

	if (turn_type == TURN_TYPE_0) { /* TURN_TYPE_0 */
		route_info.m_vl_inc = remote_route_info.m_vl_inc;
		route_info.m_turn_type = remote_route_info.m_turn_type;
	} else if (turn_type == TURN_TYPE_1) { /* TURN_TYPE_1 */
		route_info.m_turn_type = TURN_TYPE_1;
		/*
		 * TURN_TYPE_1 starts on lowest vl (TURN_TYPE_2 VL -1) therefore it
		 * does not change vl_inc
		 */
		route_info.m_vl_inc = remote_route_info.m_vl_inc;

		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "TURN_TYPE_1, vl_inc remains: %u\n", route_info.m_vl_inc);
	} else { /* TURN_TYPE_2 */
		route_info.m_turn_type = TURN_TYPE_2;
		if (remote_route_info.m_turn_type == TURN_TYPE_0) {
			route_info.m_vl_inc = remote_route_info.m_vl_inc + 1;
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "TURN_TYPE_2 to TURN_TYPE_0, inc by 1 vl_inc: %u\n",
				   route_info.m_vl_inc);
		} else if (remote_route_info.m_turn_type == TURN_TYPE_1) {
			route_info.m_vl_inc = remote_route_info.m_vl_inc + 2;
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "TURN_TYPE_2 to TURN_TYPE_1, inc by 2 vl_inc: %u\n",
				   route_info.m_vl_inc);
		} else {
			route_info.m_vl_inc = remote_route_info.m_vl_inc;
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "TURN_TYPE_2 to same turn type, vl_inc remains: %u\n",
				   route_info.m_vl_inc);
		}
	}
	OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

void ArKdorAlgorithm::CalculateSwitchPortGroups(
	ARSWDataBaseEntry &sw_db_entry,
	LidMapping *p_lid_mapping,
	LidToConnectionArray &lid_to_connections,
	LidToPortArray &base_lid_to_port)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	/*
	 * Step 1.1 -
	 *  Ensure that this switch supports the algorithm.
	 */
	if (!IsFeatureActive(sw_db_entry, m_algorithm_feature_)) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Skip switch GUID: 0x%" PRIx64 ", LID: %u - "
			   "Algorithm not supported or not enabled.\n",
			   sw_db_entry.m_general_sw_info.m_guid,
			   sw_db_entry.m_general_sw_info.m_lid);

		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Algorithm %u not supported or not enabled. "
			   "sw_db_entry.m_support[SUPPORT_AR]: %u "
			   "sw_db_entry.m_support[m_algorithm_feature_]: %u "
			   "sw_db_entry.m_option_on: %u\n",
			   m_algorithm_feature_,
			   sw_db_entry.m_support[SUPPORT_AR],
			   sw_db_entry.m_support[m_algorithm_feature_],
			   sw_db_entry.m_option_on);

		OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
	}

	KdorAlgorithmData algorithm_data;

	/*
	 * Step 2 - Calculate Ar Groups
	 */
	CalculateArGroups(sw_db_entry,
			  p_lid_mapping->m_lid_to_sw_lid_mapping,
			  algorithm_data);

	/*
	 * Step 3 - Calculate Ar PLFTs
	 */
	CalculateArPlfts(sw_db_entry,
			 p_lid_mapping,
			 algorithm_data,
			 lid_to_connections,
			 base_lid_to_port);

	OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

void ArKdorAlgorithm::CalculateArGroups(ARSWDataBaseEntry &sw_db_entry,
					u_int16_t *lid_to_sw_lid_mapping,
					KdorAlgorithmData &algorithm_data)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
		   "Calculate group bitmask for GUID: 0x%" PRIx64 ", LID: %u: \n",
		   sw_db_entry.m_general_sw_info.m_guid,
		   sw_db_entry.m_general_sw_info.m_lid);

	/*
	 * Step 2.1 -
	 *  Iterate through all lids and calculate group bitmask for each lid.
	 */
	osm_switch_t *p_osm_sw = sw_db_entry.m_general_sw_info.m_p_osm_sw;
	uint16_t max_lid = p_osm_sw->max_lid_ho;
	uint8_t **hops = p_osm_sw->hops;
	/* last plft will include only static route (no AR) */
	GroupData group_data_array[MAX_KDOR_PLFT_NUMBER - 1];
    /* ports to neighbors with invalid dim */
    GroupData invalid_dim_group_data_array[MAX_KDOR_PLFT_NUMBER - 1];
	int group_data_array_size = MAX_KDOR_PLFT_NUMBER - 1;
	KdorGroupData kdor_group_data;
	RouteInfo route_info;
	RouteInfo ingress_route_info;
	SwitchConnection ingress_connection;
	ingress_connection.m_dim_idx = 0xFF; /* max index TODO: optimize to max dim */
	ingress_connection.m_remote_switch = p_osm_sw;

	uint16_t sw_lid_num;
	for (uint16_t lid_num = 1; lid_num <= max_lid; ++lid_num) {

		if (lid_to_sw_lid_mapping[lid_num] == lid_num) {
			/* this is sw lid */
			sw_lid_num = lid_num;
		} else {
			sw_lid_num = lid_to_sw_lid_mapping[lid_num];
			if (sw_lid_num == 0) {
				continue; /* lid_num is invalid lid */
			}
		}
		uint8_t least_hops = osm_switch_get_least_hops(p_osm_sw, sw_lid_num);
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "lid_num: %u sw_lid_num: %u least_hops: %u\n",
			   lid_num , sw_lid_num, least_hops);

		if (least_hops == OSM_NO_PATH || least_hops == 0) {
			if (!hops[sw_lid_num]) {
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG, "hops[lid:%u] == NULL \n", sw_lid_num);
			}
			continue;
		}

		LidToKdorGroupDataMapIter lid_to_kdor_group_iter =
			algorithm_data.m_lid_to_kdor_group_map.find(sw_lid_num);
		if (lid_to_kdor_group_iter != algorithm_data.m_lid_to_kdor_group_map.end()) {

			algorithm_data.AddLidToKdorGroupData(lid_num,
							     sw_lid_num,
							     lid_to_kdor_group_iter->second,
							     false,
							     m_p_osm_log);
			continue;
		}

		KdorSwData *p_dest_switch_data = m_sw_lid_to_kdor_data_[sw_lid_num];
		if (p_dest_switch_data == NULL) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 9212: "
				   "Failed to find kDOR data for switch lid %u\n",
				   sw_lid_num);
			continue;
		}

		for (int i = 0; i < group_data_array_size; ++i) {
			group_data_array[i].m_group_bitmask.reset();
			group_data_array[i].m_group_size = 0;
			invalid_dim_group_data_array[i].m_group_bitmask.reset();
			invalid_dim_group_data_array[i].m_group_size = 0;
		}

		/*
		 *  Calculate group bitmask.
		 */
		ConnectionVec &connections = sw_db_entry.m_kdor_data->GetConnectionsVec();
		int connections_number = sw_db_entry.m_kdor_data->GetConnectionsNumber();
		for (int connection_idx = 0; connection_idx < connections_number;
		     ++connection_idx) {

			SwitchConnection &connection = connections[connection_idx];

			if (m_p_osm_log->level & OSM_LOG_DEBUG) {
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
					   "Calculate AR group on Switch GUID: 0x%" PRIx64
					   ", LID: %u to destination switch LID: %u %s\n",
					   sw_db_entry.m_general_sw_info.m_guid,
					   sw_db_entry.m_general_sw_info.m_lid,
					   sw_lid_num,
					   connection.ToString().c_str());
			}

			if (connection.m_remote_switch == NULL) {
				continue; /* CAs connection */
			}
			/* first port */
			u_int8_t port_num = (*connection.m_ports.begin())->port_num;
			if (osm_switch_get_hop_count(p_osm_sw, sw_lid_num, port_num) !=
			    least_hops) {
				continue; /* not shortest path */
			}

			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "shortest path to LID: %u using port: %u \n",
				   sw_lid_num, port_num);

			ARSWDataBaseEntry *p_remote_db_entry =
				connection.GetRemoteSwitchDbEntry();
			KdorSwData *p_remote_switch_data = p_remote_db_entry->m_kdor_data;

			RouteInfo &remote_route_info =
				p_remote_switch_data->GetRouteInfo(p_dest_switch_data);

			/* calculate route info from curr sw */
			CalculateRouteInfo(connection,
					   remote_route_info,
					   route_info);

			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Route from remote SW LID: %u to destination SW LID: %u "
				   "vl_inc: %u turn_type: %u\n",
				   p_remote_db_entry->m_general_sw_info.m_lid,
				   sw_lid_num,
				   route_info.m_vl_inc,
				   route_info.m_turn_type);

			/* calculate worst case route info from ingress to destination */
			CalculateRouteInfo(ingress_connection,
							   route_info,
							   ingress_route_info);

			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Route from local SW LID: %u to destination SW LID: %u "
				   "vl_inc: %u turn_type: %u\n",
				   sw_db_entry.m_general_sw_info.m_lid,
				   sw_lid_num,
				   ingress_route_info.m_vl_inc,
				   ingress_route_info.m_turn_type);
			/*
			 *            -----------        ------------         -------------
			 * ---------> | curr sw | -----> | neighbor | ------> |destination|
			 *  ingress   |   (U)   |        |    (V)   |  static |    (D)    |
			 *            -----------        ------------         -------------
			 *
			 *
			 *
			 * The rout out of this port is valid if:
			 * ingress plane + (is invalid turn from ingress to neighbor) +
			 * min static vl increase to destination +
			 * is invalid turn from current sw to neighbor static egress <
			 * m_max_planes_
			 * therefore:
			 * plft + ingress_route_info.m_vl_inc < m_max_planes_
			 *
			 * The following table show the final VL/plft for given
			 * in VL/plft (row) and ingress_route vl increase (column)
			 *
			 * vl inc     | 0 (T1) | 1 (T2) | 2 (T1) | 3 (T2) | 4 (T1) | 5 (T2)
			 * -----------|        |        |        |        |        |
			 * in plft/vl |        |        |        |        |        |
			 * -----------+--------+--------+--------+--------+--------+--------
			 *     0      |   0    |   1    |   2    |   3    |   4    |   5
			 * -----------+--------+--------+--------+--------+--------+--------
			 *     1      |   2    |   1    |   4    |   3    |   6    |   5
			 * -----------+--------+--------+--------+--------+--------+--------
			 *     2      |   2    |   3    |   4    |   5    |   6    |   7
			 * -----------+--------+--------+--------+--------+--------+--------
			 *     3      |   4    |   3    |   6    |   5    |   8    |   7
			 * -----------+--------+--------+--------+--------+--------+--------
			 *     4      |   4    |   5    |   6    |   7    |   8    |   -
			 * -----------+--------+--------+--------+--------+--------+--------
			 *     5      |   6    |   5    |   8    |   7    |   -    |   -
			 * -----------+--------+--------+--------+--------+--------+--------
			 *
			 * From the table:
			 *
			 * For ingress_route_info.m_turn_type == T1:
			 * final_vl = (in_vl + 1) / 2 * 2 + vl_inc <= planes_number - 1
			 * max_in_vl is even -> in_vl + vl_inc <= planes_number - 1
			 *  -> max_in_vl <= planes_number - vl_inc - 1
			 * max_in_vl is even - round doun to even number
			 * max_in_vl <= (planes_number - vl_inc - 1) / 2 * 2
			 *
			 * For ingress_route_info.m_turn_type == T2:
			 * final_vl = (in_vl / 2) * 2 + vl_inc <= planes_number - 1
			 * max_in_vl is odd -> in_vl - 1 + vl_inc <= planes_number - 1
			 * in_vl <= planes_number - vl_inc
			 * max_in_vl is odd - convert down to odd number
			 * max_in_vl <= (planes_number - vl_inc + 1) / 2 * 2 - 1;
			 *
			 * The following teble provide max_plft for given vl_inc and max planes
			 * vl inc     | 0 (T1) | 1 (T2) | 2 (T1) | 3 (T2) | 4 (T1) | 5 (T2)
			 * -----------|        |        |        |        |        |
			 * max planes |        |        |        |        |        |
			 * -----------+--------+--------+--------+--------+--------+--------
			 *     4      |   2    |   3    |    0   |   1    |   -    |   -
			 * -----------+--------+--------+--------+--------+--------+--------
			 *     5      |   4    |   3    |    2   |   1    |   0    |   -
			 * -----------+--------+--------+--------+--------+--------+--------
			 *     6      |   4    |   5    |    2   |   3    |   0    |  1
			 * -----------+--------+--------+--------+--------+--------+--------
			 */

			if (ingress_route_info.m_vl_inc >= m_planes_number_) {
				continue; /* Round doun works only for positive number */
			}

			GroupData *p_group_data_array = group_data_array;
			if (remote_route_info.m_connection != NULL && /* Not the destination */
				(remote_route_info.m_connection->m_dim_idx == INVALID_DIM_INDEX ||
				 route_info.m_connection->m_dim_idx == INVALID_DIM_INDEX)) {
				p_group_data_array = invalid_dim_group_data_array;
			}

			int max_plft;
			if (ingress_route_info.m_turn_type == TURN_TYPE_2) {
				max_plft = (m_planes_number_ - ingress_route_info.m_vl_inc + 1) /
				    2 * 2 - 1;
			} else {
				max_plft = (m_planes_number_ - ingress_route_info.m_vl_inc - 1) /
				    2 * 2;
			}

			max_plft = std::min(max_plft, group_data_array_size - 1);
			for (int plft = 0; plft <= max_plft; ++plft) {
				p_group_data_array[plft].m_group_bitmask |=
					connection.m_ports_bitset;
				p_group_data_array[plft].m_group_size +=
					connection.m_ports.size();

				OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
					   "Add ports %s to group_data of plft_id: %u "
					   "total size: %u is_valid_dim: %u\n",
					   connection.m_ports_bitset.to_string().c_str(),
					   plft,
					   p_group_data_array[plft].m_group_size,
					   p_group_data_array == group_data_array);
			}
		}

		/*
		 *If no route through neighbors with valid coordinate
		 *use route through neighbors without valid coordinate
		 */
		if (invalid_dim_group_data_array[0].m_group_size) {

			for (int plft = 0; plft < group_data_array_size; ++plft) {
				if ((group_data_array[plft].m_group_size == 0) &&
					(invalid_dim_group_data_array[plft].m_group_size != 0)) {

					OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
						   "AR group on Switch GUID: 0x%" PRIx64 ", LID: %u: "
						   "to destination switch LID: %u plft_id:%u "
						   "uses neighbors without valid coordinate.\n",
						   sw_db_entry.m_general_sw_info.m_guid,
						   sw_db_entry.m_general_sw_info.m_lid,
						   sw_lid_num,
						   plft);

					group_data_array[plft].m_group_bitmask =
						invalid_dim_group_data_array[plft].m_group_bitmask;
					group_data_array[plft].m_group_size =
						invalid_dim_group_data_array[plft].m_group_size;
				}
			}
		}

		if (m_p_osm_log->level & OSM_LOG_DEBUG) {
			for (int plft = 0; plft < group_data_array_size; ++plft) {
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
					   "AR group on Switch GUID: 0x%" PRIx64 ", LID: %u "
					   "to destination switch LID: %u plft_id:%u "
					   "is %s size: %u\n",
					   sw_db_entry.m_general_sw_info.m_guid,
					   sw_db_entry.m_general_sw_info.m_lid,
					   sw_lid_num,
					   plft,
					   group_data_array[plft].m_group_bitmask.to_string().c_str(),
					   group_data_array[plft].m_group_size);
			}
		}

		/*
		 * Build kdor_group_data with discovered group_data for each plft
		 * group for plf == 0 is the largest group
		 */
		if (group_data_array[0].m_group_size > 1) {

			for (int plft = 0; plft < group_data_array_size; ++plft) {
				if (group_data_array[plft].m_group_size > 1) {
					if (m_p_osm_log->level & OSM_LOG_DEBUG) {
						OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
							   "Insert group bitmask: %s to "
							   "algorithm_data groups_map\n",
							   group_data_array[plft].m_group_bitmask.
							   to_string().c_str());
					}

					GroupBitmaskToDataMapInsertRes res =
						algorithm_data.m_groups_map.insert(
							make_pair(group_data_array[plft].m_group_bitmask,
								  group_data_array[plft]));
					kdor_group_data.group_data_array[plft] = &(res.first->second);

					if (res.second) { /* If new group */
						algorithm_data.m_groups.push_back(&res.first->second);
					}

				} else {
					kdor_group_data.group_data_array[plft] = NULL;
				}
			}

			/* Save discovered kdor_group_data for new destination */
			algorithm_data.AddLidToKdorGroupData(lid_num,
							     sw_lid_num,
							     kdor_group_data,
							     true,
							     m_p_osm_log);
		}
	}

	/*
	 * Step 2.2 -
	 *  try to assign group number for all groups.
	 */
	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG, "Step 2.2 AssignPortGroups \n");

	if (!algorithm_data.m_groups_map.empty()) {
		int rc = m_ar_mgr_.AssignPortGroups(sw_db_entry, algorithm_data);
		if (rc) {
			/* Sanity check, should not reach here */
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 9213: Failed to assign port groups.\n");
		}
	}

	/*
	 * Step 3 -
	 *  Update calculated ar_group_table:
	 *  At this point we have all assign groups in algorithm_data.m_assign_groups list
	 */
	m_ar_mgr_.SetPortGroups(sw_db_entry, algorithm_data);

	OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

void ArKdorAlgorithm::CalculateArPlfts(ARSWDataBaseEntry &sw_db_entry,
				       LidMapping *p_lid_mapping,
				       KdorAlgorithmData &algorithm_data,
				       LidToConnectionArray &lid_to_connections,
				       LidToPortArray &base_lid_to_port)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG, "Step 4 configure AR PLFTs \n");
	osm_advanced_routing_data_t *p_ar_data =
		&sw_db_entry.m_general_sw_info.m_p_osm_sw->ar_data;

	memset(lid_to_connections, 0, sizeof(lid_to_connections));
	BuildDstSwLidToConnection(sw_db_entry, lid_to_connections);

	/* save calculated port for base lid (in order to select same port to all vlids) */
	memset(base_lid_to_port, 0, sizeof(base_lid_to_port));

	uint8_t lid_state;
	uint16_t ar_group_id;
	for (unsigned dest_lid = 1;
		 dest_lid <= m_ar_mgr_.m_p_osm_subn->max_lid;
		 ++dest_lid) {

		u_int8_t ucast_lft_port = GetStaticUcastLftPort(sw_db_entry,
								p_lid_mapping,
								lid_to_connections,
								base_lid_to_port,
								dest_lid);

		if (ucast_lft_port == OSM_NO_PATH)	/* No route to this lid_num from this switch */
			continue;

		LidToKdorGroupDataMapConstIter ar_kdor_groups_iter =
			algorithm_data.m_lid_to_kdor_group_map.find(dest_lid);

		for (int plft_id = 0; plft_id < m_planes_number_; ++plft_id) {
			CalculateArPlft(sw_db_entry,
					algorithm_data,
					ucast_lft_port,
					dest_lid,
					plft_id,
					ar_kdor_groups_iter,
					lid_state,
					ar_group_id);

			if (m_p_osm_log->level & OSM_LOG_DEBUG) {
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
					   "LID %u plft id %u dest_lid: %u - "
					   "GroupNumber: %u LidState: %u Static Port %u\n",
					   sw_db_entry.m_general_sw_info.m_lid,
					   plft_id,
					   dest_lid,
					   ar_group_id, lid_state, ucast_lft_port);
			}

			osm_ar_set_required_plft_top(p_ar_data, plft_id,
						     m_ar_mgr_.m_p_osm_subn->max_lid);

		} /* End of plft_id loop */
	} /* End of loop destination lids */

	OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

void ArKdorAlgorithm::CalculateArPlft(ARSWDataBaseEntry &sw_db_entry,
				      KdorAlgorithmData &algorithm_data,
				      u_int8_t ucast_lft_port,
				      unsigned dest_lid,
				      int plft_id,
				      LidToKdorGroupDataMapConstIter &ar_kdor_groups_iter,
				      OUT uint8_t &lid_state,
				      OUT uint16_t &ar_group_id)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
		   "CalculateArPlft LID %u dest_lid:%u static_port: %u plft: %u\n",
		   sw_db_entry.m_general_sw_info.m_lid,
		   dest_lid,
		   ucast_lft_port,
		   plft_id);

	osm_advanced_routing_data_t *p_ar_data =
		&sw_db_entry.m_general_sw_info.m_p_osm_sw->ar_data;

	if (plft_id == (m_planes_number_ - 1) || /* last plane is static only */
		ar_kdor_groups_iter == algorithm_data.m_lid_to_kdor_group_map.end() ||
		ar_kdor_groups_iter->second.group_data_array[plft_id] == NULL) {

		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "No AR group - Set lid sate Static Group: 0\n");

		lid_state = (uint8_t)OSM_AR_IB_LID_STATE_STATIC;
		ar_group_id = 0;
		osm_ar_lft_set(p_ar_data, dest_lid, plft_id, ucast_lft_port,
			       lid_state, ar_group_id);

		OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
	}

	GroupData* p_group_data =
		ar_kdor_groups_iter->second.group_data_array[plft_id];

	if (p_group_data->m_group_table_number == ARMGR_MAX_GROUP_TABLES) {

		lid_state = (uint8_t)OSM_AR_IB_LID_STATE_STATIC;
		ar_group_id = 0;
		osm_ar_lft_set(p_ar_data, dest_lid, plft_id, ucast_lft_port,
			       lid_state, ar_group_id);

		if (m_p_osm_log->level & OSM_LOG_DEBUG) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "No AR group because group bitmask:%s was not mapped.\n",
				   p_group_data->m_group_bitmask.to_string().c_str());
		}

	} else {
		lid_state = (uint8_t)m_ar_mgr_.m_master_db.m_ar_mode;
		ar_group_id = p_group_data->m_group_number;
		/*
		 * Currently RN Not implemented
		 */
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG, "LID %u dest_lid:%u - "
			   "Set lid state %u Group:%u\n",
			   sw_db_entry.m_general_sw_info.m_lid, dest_lid,
			   lid_state,
			   ar_group_id);

		osm_ar_lft_set(p_ar_data, dest_lid, plft_id, ucast_lft_port,
			       lid_state, ar_group_id);
	}

	OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

uint8_t ArKdorAlgorithm::GetStaticUcastLftPort(ARSWDataBaseEntry &sw_db_entry,
					       LidMapping *p_lid_mapping,
					       SwitchConnection *dst_sw_lid_to_connection[],
					       uint8_t base_lid_to_port[],
					       uint16_t dest_lid)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);
	uint8_t port_number;
	SwitchConnection *connection;
	if (p_lid_mapping->m_lid_to_sw_lid_mapping[dest_lid] == 0) {
		OSM_AR_LOG_RETURN(m_p_osm_log, OSM_NO_PATH);
	}

	if (p_lid_mapping->m_lid_to_sw_lid_mapping[dest_lid] == dest_lid) {
		/* sw lid */
		connection = dst_sw_lid_to_connection[dest_lid];
		if (connection == NULL) {
			if (dest_lid == sw_db_entry.m_general_sw_info.m_lid) {
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
					   "GetStaticUcastLftPort for local switch lid: %u "
					   "return: %u\n",
					   dest_lid, 0);
				OSM_AR_LOG_RETURN(m_p_osm_log, 0);
			}
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 9214: Failed to find static connection to "
				   "destination lid: %u\n",
				   dest_lid);
			OSM_AR_LOG_RETURN(m_p_osm_log, OSM_NO_PATH);
		}
		/* No load balancing to switch lid */
		port_number = (*connection->m_next_port_for_routing)->port_num;
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Switch GUID: 0x%" PRIx64 ", LID: %u - "
			   "Use static port %u for destination switch lid: %u\n",
			   sw_db_entry.m_general_sw_info.m_guid,
			   sw_db_entry.m_general_sw_info.m_lid,
			   port_number, dest_lid);

		OSM_AR_LOG_RETURN(m_p_osm_log, port_number);
	}
	/* CA lid */
	uint16_t base_lid = p_lid_mapping->m_lid_to_base_lid_mapping[dest_lid];
	if (base_lid_to_port[base_lid] != 0) {
		port_number = base_lid_to_port[base_lid];

		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Switch GUID: 0x%" PRIx64 ", LID: %u - "
			   "Use static port %u for dest_lid: %u base_lid: %u\n",
			   sw_db_entry.m_general_sw_info.m_guid,
			   sw_db_entry.m_general_sw_info.m_lid,
			   port_number, dest_lid, base_lid);

		OSM_AR_LOG_RETURN(m_p_osm_log, port_number);
	}

	uint16_t sw_lid = p_lid_mapping->m_lid_to_sw_lid_mapping[dest_lid];
	connection = dst_sw_lid_to_connection[sw_lid];
	if (connection == NULL) {
		if (sw_lid == sw_db_entry.m_general_sw_info.m_lid) {
			port_number = OsmSwitchGetPortByLidUseNewLft(
			    sw_db_entry.m_general_sw_info.m_p_osm_sw, base_lid);

			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Switch GUID: 0x%" PRIx64 ", LID: %u - "
				   "Use static port %u for dest_lid: %u base_lid: %u"
				   "on local switch\n",
				   sw_db_entry.m_general_sw_info.m_guid,
				   sw_db_entry.m_general_sw_info.m_lid,
				   port_number, dest_lid, base_lid);

			OSM_AR_LOG_RETURN(m_p_osm_log, port_number);
		}
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 9215: Failed to find static connection to "
			   "switch lid:%u of destination lid: %u\n",
			   sw_lid, dest_lid);
		OSM_AR_LOG_RETURN(m_p_osm_log, OSM_NO_PATH);
	}
	port_number = (*connection->m_next_port_for_routing)->port_num;
	base_lid_to_port[base_lid] = port_number;
	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
		   "GetStaticUcastLftPort for base_lid: %u "
		   "of destination lid: %u set and return: %u\n",
		   base_lid, dest_lid, port_number);

	/* Increment next_port_for_routing */
	++connection->m_next_port_for_routing;
	if (connection->m_next_port_for_routing == connection->m_ports.end()) {
		connection->m_next_port_for_routing = connection->m_ports.begin();
	}
	OSM_AR_LOG_RETURN(m_p_osm_log, port_number);
}

void ArKdorAlgorithm::BuildDstSwLidToConnection(
	ARSWDataBaseEntry &sw_db_entry,
	SwitchConnection *dst_sw_lid_to_connection[])
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	SwitchConnection *connection;
	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_map_.begin();
	for (; sw_node_it != m_sw_map_.end();  ++sw_node_it) {
		if (&sw_node_it->second == &sw_db_entry) {
			continue;
		}
		ARSWDataBaseEntry &dst_sw_db_entry = sw_node_it->second;
		RouteInfo &route_info_to_dst =
			sw_db_entry.m_kdor_data->GetRouteInfo(dst_sw_db_entry.m_kdor_data);
		connection = route_info_to_dst.m_connection;
		if (connection == NULL) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 9216: -A- Failed to get RouteInfo from "
				   "Switch GUID: 0x%" PRIx64 ", LID: %u "
				   "to switch GUID: 0x%" PRIx64 ", LID: %u\n",
				   sw_db_entry.m_general_sw_info.m_guid,
				   sw_db_entry.m_general_sw_info.m_lid,
				   dst_sw_db_entry.m_general_sw_info.m_guid,
				   dst_sw_db_entry.m_general_sw_info.m_lid);
			continue;
		}
		dst_sw_lid_to_connection[dst_sw_db_entry.m_general_sw_info.m_lid] =
			connection;
		connection->m_next_port_for_routing = connection->m_ports.begin();
	}

	OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

void ArKdorAlgorithm::BuildKdorVl2VlPerOpVl()
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	/*
	 * For turn type 1 increase vl if vl%2 == 1
	 * For turn type 2 increase vl if vl%2 == 0
	 */
	m_turn_type_to_vl2vl_per_op_vls_[TURN_TYPE_0] = m_vl2vl_per_op_vls_;
	m_turn_type_to_vl2vl_per_op_vls_[TURN_TYPE_1] = m_turn_1_vl2vl_per_op_vls_;
	m_turn_type_to_vl2vl_per_op_vls_[TURN_TYPE_2] = m_turn_2_vl2vl_per_op_vls_;



	/* Currently support only one vl per plane, therefore: */
	uint8_t vls_per_stream = m_planes_number_;

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE, "BuildVl2Vl t1,t2 PerOpVl.\n");
	u_int8_t t1_vl2vl_array[IB_NUMBER_OF_SLS]; /* turn_type_1 */
	u_int8_t t2_vl2vl_array[IB_NUMBER_OF_SLS]; /* turn_type_2 */
	u_int8_t num_vls;
	u_int8_t max_vl;
	u_int8_t qos_max_vl;     /* max vl for the current qos */
	u_int8_t t1_out_vl;	     /* turn_type_1 */
	u_int8_t t2_out_vl;	     /* turn_type_2 */

	/*
	 * Turn type 1 is valid on even vls ((in_vl % 2) == 0)
	 * Turn type 2 is valid on odd vls ((in_vl % 2) != 0)
	 */

	memset(m_turn_1_vl2vl_per_op_vls_, 0, sizeof(m_turn_1_vl2vl_per_op_vls_));
	memset(m_turn_2_vl2vl_per_op_vls_, 0, sizeof(m_turn_2_vl2vl_per_op_vls_));

	for (uint8_t op_vls = 2; op_vls <= MAX_OP_VL_CODE; ++op_vls) {

		num_vls = ( 1 << (op_vls-1));
		max_vl = min((uint8_t)(num_vls - 1), (uint8_t)(IB_DROP_VL - 1));

		if (num_vls < vls_per_stream)
			continue; /* Not Supported */

		/* On invalid turn that my lead to loop we set to qos_max_vl */
		for (u_int8_t in_vl = 0; in_vl < IB_NUMBER_OF_SLS; ++in_vl) {
			/* vl < num_vls */
			u_int8_t vl = in_vl % num_vls;

			/*
			 * The qos groups number of given vl [1..last] is:
			 * (vl / vls_per_stream) + 1
			 *
			 * In each qos group there are vls_per_stream vls so for vl in specific
			 * qos group the max vl is:
			 */
			qos_max_vl = ((vl / vls_per_stream) + 1 ) * vls_per_stream - 1;
			qos_max_vl = min(qos_max_vl, max_vl);

			if ((in_vl % 2) == 0) {
				t1_out_vl = min(qos_max_vl, vl);
				t2_out_vl  = min(qos_max_vl, (uint8_t)(vl + 1));
			} else {
				t1_out_vl = min(qos_max_vl, (uint8_t)(vl + 1));
				t2_out_vl  = min(qos_max_vl, vl);
			}

			m_turn_1_vl2vl_per_op_vls_[op_vls][in_vl] = t1_out_vl;
			m_turn_2_vl2vl_per_op_vls_[op_vls][in_vl] = t2_out_vl;
		}
	} /* op_vls loop */

	if (OSM_LOG_DEBUG & m_p_osm_log->level) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Calculated kdor turns vl to vl mapping, vls_per_stream: %u\n",
			   vls_per_stream);

		for (uint8_t op_vls = 1; op_vls <= MAX_OP_VL_CODE; ++op_vls) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "op_vls: %u turn type 1 vl to vl mapping %s\n",
				   op_vls,
				   SLToVLMappingToStr(
					   m_turn_1_vl2vl_per_op_vls_[op_vls]).c_str());
		}
		for (uint8_t op_vls = 1; op_vls <= MAX_OP_VL_CODE; ++op_vls) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "op_vls: %u turn type 2 vl to vl mapping %s\n",
				   op_vls,
				   SLToVLMappingToStr(
					   m_turn_2_vl2vl_per_op_vls_[op_vls]).c_str());
		}
	}
	OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

void ArKdorAlgorithm::BuildKdorPlftMap()
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE, "BuildKdorPlftMap\n");
	ib_port_sl_to_plft_map_entry_t vl_to_plft_map;
	u_int8_t plft_id;

	memset(&vl_to_plft_map, 0, sizeof(vl_to_plft_map));

	plft_id = 0;
	for (u_int8_t in_vl = 0; in_vl < IB_NUMBER_OF_SLS; ++in_vl) {
		if ((in_vl % m_planes_number_) == 0) {
			plft_id = 0;
		}

		vl_to_plft_map.plft_port_sl[in_vl] = plft_id;
		plft_id++;
	}

	for (int i = 0; i < IB_PORT_SL_TO_PLFT_MAP_BLOCK_SIZE; ++i) {
		memcpy(&m_plft_map_.port[i], &vl_to_plft_map,
			   sizeof(vl_to_plft_map));
	}

	if (OSM_LOG_DEBUG & m_p_osm_log->level) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Calculated VL to plft mapping, planes_number: %u\n",
			   m_planes_number_);

		for (u_int8_t in_vl = 0; in_vl < IB_NUMBER_OF_SLS; ++in_vl) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "map VL: %u to plft id:%u\n",
				   in_vl, vl_to_plft_map.plft_port_sl[in_vl]);
		}
	}
	OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

void ArKdorAlgorithm::SetPlftMap()
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_map_.begin();
	for (; sw_node_it != m_sw_map_.end(); ++sw_node_it) {

		if (sw_node_it->second.in_temporary_error)
			continue;

		if (!IsFeatureActive(sw_node_it->second, m_algorithm_feature_)) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Switch GUID: 0x%" PRIx64 ", LID: %u - " \
				   "Algorithm (%u) not supported or not enabled, "
				   "Set PlftMap skipped.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid,
				   m_algorithm_feature_);

			/* clear AR configuration from sw - SM should configure this sw */
			sw_node_it->second.m_app_data = ARMGR_SW_NOT_UPDATED;
			continue;
		}

		SetPlftMap(sw_node_it->second);
	}

	OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

void ArKdorAlgorithm::SetPlftMap(ARSWDataBaseEntry &sw_db_entry)
{
	uint8_t max_block = sw_db_entry.m_general_sw_info.m_p_osm_sw->p_node->
	    node_info.num_ports / IB_PORT_SL_TO_PLFT_MAP_BLOCK_SIZE;

	for (uint8_t port_block = 0; port_block <= max_block; ++port_block) {
		osm_ar_port_sl_to_plft_set_block(sw_db_entry.m_general_sw_info.m_p_osm_sw,
						 port_block,
						 FALSE, /* calculated */
						 &m_plft_map_);
	}
}

bool ArKdorAlgorithm::IsValidCoordinate(ARSWDataBaseEntry &sw_db_entry)
{
	return (sw_db_entry.m_general_sw_info.m_p_osm_sw->coord !=
		OSM_SW_NO_COORD);
}

/*
 * KdorAlgorithmData
 */

void KdorAlgorithmData::AddLidToKdorGroupData(uint16_t lid_num,
					      uint16_t sw_lid_num,
					      KdorGroupData &kdor_group_data,
					      bool is_new_group,
					      osm_log_t *p_osm_log)
{
	OSM_AR_LOG(p_osm_log, OSM_LOG_DEBUG,
		   "Add lid_num:%u sw_lid_num: %u to KdorGroupData is_new_group: %u\n",
		   lid_num, sw_lid_num, is_new_group);

	if (is_new_group) {
		AddLidToKdorGroupData(sw_lid_num, kdor_group_data);
	}

	if (lid_num != sw_lid_num) {
		AddLidToKdorGroupData(lid_num, kdor_group_data);
	}
}

void KdorAlgorithmData::AddLidToKdorGroupData(uint16_t lid_num,
					      KdorGroupData &kdor_group_data)
{
	static int num_groups = MAX_KDOR_PLFT_NUMBER - 1;

	for (int plft = 0; plft < num_groups; ++plft) {
		/* might be NULL incase of group size <= 1 */
		if (kdor_group_data.group_data_array[plft] != NULL) {
			kdor_group_data.group_data_array[plft]->
			m_lids_list.push_back(lid_num);
		}
	}
	m_lid_to_kdor_group_map.insert(make_pair(lid_num, kdor_group_data));
}

/*
 * ArHcAlgorithm
 */

/*
 * Update connection::direction dimensions index and sign for given
 * connection on src switch
 * Valid connection direction is connection change of one and only one dimension.
 * If no change found (same connection for src and dst)
 * or more than one dimension changed, return non zero value
 */
int ArHcAlgorithm::SetDirection(osm_switch_t* src,
				SwitchConnection &connection)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	osm_switch_t* dst = connection.m_remote_switch;
	if (dst == NULL) {
		OSM_AR_LOG_RETURN(m_p_osm_log, 0);
	}

	if (src->coord == OSM_SW_NO_COORD ||
	    dst->coord == OSM_SW_NO_COORD) {
	    connection.m_dim_idx = INVALID_DIM_INDEX;
	    connection.m_dim_sign = (dim_sign_t)(1);

		if (m_p_osm_log->level & OSM_LOG_DEBUG) {
			u_int64_t src_node_guid = cl_ntoh64(osm_node_get_node_guid(src->p_node));
			u_int16_t src_node_lid = cl_ntoh16(osm_node_get_base_lid(src->p_node, 0));
			u_int64_t dst_node_guid = cl_ntoh64(osm_node_get_node_guid(src->p_node));
			u_int16_t dst_node_lid = cl_ntoh16(osm_node_get_base_lid(src->p_node, 0));

			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Invalid coordinate on set direction from "
				   "switch GUID: 0x%" PRIx64 ", LID: %u coord: 0x%x "
				   "to switch GUID: 0x%" PRIx64 ", LID: %u coord: 0x%x "
				   "dim_idx: %u  dim_sign: %d\n",
				   src_node_guid, src_node_lid, src->coord,
				   dst_node_guid, dst_node_lid, dst->coord,
				   connection.m_dim_idx, connection.m_dim_sign);
		}
		OSM_AR_LOG_RETURN(m_p_osm_log, 0);
	}

	bool diff_found = false;
	dim_index_t dim_idx = 0;
	uint16_t coord_diff = src->coord ^ dst->coord;

	while (coord_diff) {
		if (coord_diff & 0x1) {
			if (coord_diff == 0x1) {
				diff_found = true;
			}
			break;
		}
		coord_diff = coord_diff >> 1;
		dim_idx++;
	}

	if (!diff_found) {
		u_int64_t src_node_guid = cl_ntoh64(osm_node_get_node_guid(src->p_node));
		u_int16_t src_node_lid = cl_ntoh16(osm_node_get_base_lid(src->p_node, 0));
		u_int64_t dst_node_guid = cl_ntoh64(osm_node_get_node_guid(src->p_node));
		u_int16_t dst_node_lid = cl_ntoh16(osm_node_get_base_lid(src->p_node, 0));

		OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 9217: Switch GUID: 0x%" PRIx64 ", LID: %u coord: 0x%x"
			   "is not valid HC member of "
			   "switch GUID: 0x%" PRIx64 ", LID: %u coord: 0x%x\n",
			   src_node_guid, src_node_lid, src->coord,
			   dst_node_guid, dst_node_lid, dst->coord);

		OSM_AR_LOG_RETURN(m_p_osm_log, 1);
	}

	connection.m_dim_idx = dim_idx;
	if (src->coord >> dim_idx & 0x1) {
		connection.m_dim_sign = (dim_sign_t)(-1);
	} else {
		connection.m_dim_sign = (dim_sign_t)(1);
	}

	if (m_p_osm_log->level & OSM_LOG_DEBUG) {
		u_int64_t src_node_guid = cl_ntoh64(osm_node_get_node_guid(src->p_node));
		u_int16_t src_node_lid = cl_ntoh16(osm_node_get_base_lid(src->p_node, 0));
		u_int64_t dst_node_guid = cl_ntoh64(osm_node_get_node_guid(src->p_node));
		u_int16_t dst_node_lid = cl_ntoh16(osm_node_get_base_lid(src->p_node, 0));

		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Direction from switch GUID: 0x%" PRIx64 ", LID: %u coord: 0x%x "
			   "to switch GUID: 0x%" PRIx64 ", LID: %u coord: 0x%x "
			   "dim_idx: %u  dim_sign: %d\n",
			   src_node_guid, src_node_lid, src->coord,
			   dst_node_guid, dst_node_lid, dst->coord,
			   connection.m_dim_idx, connection.m_dim_sign);
	}

	OSM_AR_LOG_RETURN(m_p_osm_log, 0);
}

/*
 * CoordinateData
 */

/* Count the number of dimensions in given coordinate */
int CoordinateData::CountDimensions(uint64_t coordinate)
{
	int count = 0;

	while (coordinate) {
		if (coordinate & FIRST_DIM_BITS) {
			count++;
		}
		coordinate >>= BITS_PER_DIM;
	}

	return count;
}

/*
 * Count the number of dimensions in given coordinate
 * set first_dim to the first dimension in coordinate
 */
uint8_t CoordinateData::CountDimensions(uint64_t coordinate,
					uint8_t &first_dim)
{
	int count = 0;
	uint8_t dim = 0;
	first_dim = 0;

	while (coordinate) {
		dim++;
		if (coordinate & FIRST_DIM_BITS) {
			count++;
			if (!first_dim) {
				first_dim = dim;
			}
		}
		coordinate >>= BITS_PER_DIM;
	}

	return count;
}


/*
 * Returns dimension number if the given coordinate is one dimension,
 * otherwise returns 0.
 */
uint8_t CoordinateData::IsOneDimension(uint64_t coordinate)
{
	uint8_t dim = 0;

	while (coordinate) {
		OSM_AR_LOG(g_osm_log, OSM_LOG_DEBUG,
			   "IsOneDimension Coordinate: 0x%" PRIx64
			   " dim: %u (coordinate & FIRST_DIM_BITS):0x%" PRIx64
			   "\n",
			   coordinate, dim,(coordinate & FIRST_DIM_BITS));

		dim++;
		if (coordinate & FIRST_DIM_BITS) {
			break;
		}
		coordinate >>= BITS_PER_DIM;
	}

	coordinate >>= BITS_PER_DIM;

	if (coordinate) {
		return 0;
	}

	OSM_AR_LOG(g_osm_log, OSM_LOG_DEBUG,"IsOneDimension returned: %u\n",dim);
	return dim;
}

/*
 * Returns CoordinateIntersection of a line and coordinate not on the line.
 *
 * dim_coordinate: coordinate on the line.
 * dim: line direction.
 * coordinate: a coordinate not on the line.
 */
uint64_t CoordinateData::CoordinateIntersection(uint64_t dim_coordinate,
						uint8_t dim,
						uint64_t coordinate)
{
	uint64_t dim_mask = KDOR_DIM_MASK(dim);

	uint64_t res = (dim_coordinate & (~dim_mask)) |
		(coordinate & dim_mask);

	OSM_AR_LOG(g_osm_log, OSM_LOG_DEBUG,
		   "CoordinateIntersection of dim_coordinate: 0x%" PRIx64
		   " dim: %u and coordinate: 0x%" PRIx64
		   " return: 0x%" PRIx64 "\n",
		   dim_coordinate, dim, coordinate, res);

	return res;
}

void CoordinateData::SetCoordinate(uint64_t coordinate)
{
	m_coordinate = coordinate;

	if (m_coordinate_state != COORDINATE_STATE_SET) {
		sm_total_set_coordinates++;
		m_coordinate_state = OSM::COORDINATE_STATE_SET;
	}

	OSM_AR_LOG(g_osm_log, OSM_LOG_DEBUG,
		   "Set coordinate: 0x%" PRIx64 " to switch GUID: 0x%" PRIx64
		   ", LID: %u desc: %s\n",
		   coordinate,
		   m_db_entry.m_general_sw_info.m_guid,
		   m_db_entry.m_general_sw_info.m_lid,
		   m_db_entry.m_general_sw_info.m_p_osm_sw->p_node->node_desc.description);
}

void CoordinateData::DeleteCoordinate()
{
	if (m_coordinate_state == COORDINATE_STATE_SET) {
		OSM_AR_LOG(g_osm_log, OSM_LOG_DEBUG,
			   "Delete coordinate: 0x%" PRIx64 " from switch GUID: 0x%" PRIx64
			   ", LID: %u desc: %s\n",
			   m_coordinate,
			   m_db_entry.m_general_sw_info.m_guid,
			   m_db_entry.m_general_sw_info.m_lid,
			   m_db_entry.m_general_sw_info.m_p_osm_sw->p_node->node_desc.description);

		m_coordinate = 0;
		m_coordinate_state = OSM::COORDINATE_STATE_INIT;
		sm_total_set_coordinates--;
	}
}

/*
 * Add neighbor coordinate to the node CoordinateData. Set the node coordinate,
 * if there is enough neighbors information to do so.
 *
 * In order to set the coordinate three neighbors are needed, two of them not
 * on the same line.
 * Coordinates added to m_neighbors_coordinate are never on the same line,
 * therefore maximum two coordinate can be added to m_neighbors_coordinate.
 * On Adding new neighbor coordinate, setting the node coordinate is possible if:
 * 1) There are two neighbor coordinates in m_neighbors_coordinate
 * 2) There is one neighbor coordinate in m_neighbors_coordinate and
 *    m_neighbors_dim != 0 which means two coordinates on same line were added.
 */
int CoordinateData::AddNeighborCoordinate(uint64_t coordinate)
{
	OSM_AR_LOG(g_osm_log, OSM_LOG_DEBUG,
		   "AddNeighborCoordinate coordinate: 0x%" PRIx64
		   " to switch GUID: 0x%" PRIx64 ", LID: %u desc: %s\n",
		   coordinate,
		   m_db_entry.m_general_sw_info.m_guid,
		   m_db_entry.m_general_sw_info.m_lid,
		   m_db_entry.m_general_sw_info.m_p_osm_sw->p_node->node_desc.description);

	if (m_coordinate_state == COORDINATE_STATE_SET) {
		return 0;
	}

	if (m_neighbors_coordinate.empty()) {
		m_neighbors_coordinate.push_back(coordinate);
			OSM_AR_LOG(g_osm_log, OSM_LOG_DEBUG,
				   "Push coordinate 0x%" PRIx64
				   " to neighbors_coordinate of switch GUID: 0x%"
				   PRIx64 ", LID: %u \n",
				   coordinate,
				   m_db_entry.m_general_sw_info.m_guid,
				   m_db_entry.m_general_sw_info.m_lid);
		return 0;
	}

	if (m_neighbors_coordinate.size() == (size_t)1) {

		uint64_t dim_diff = coordinate ^ m_neighbors_coordinate.front();
		if (!dim_diff) {
			OSM_AR_LOG(g_osm_log, OSM_LOG_ERROR,
				   "ERR 9219: "
				   "-A- Two neighbors of switch GUID: 0x%" PRIx64
				   ", LID: %u have same coordinate: 0x%" PRIx64
				   "\n",
				   m_db_entry.m_general_sw_info.m_guid,
				   m_db_entry.m_general_sw_info.m_lid,
				   coordinate);
			return 1;
		}

		int dim = IsOneDimension(dim_diff);
		if (m_neighbors_dim == 0) {

			if (dim != 0) {
				/*
				 * Two neighbors and this node should define
				 * a plane, but this coordinate lies on same line
				 * together with the node and a neighbor so just
				 * set the line dimension.
				 */
				m_neighbors_dim = dim;
				return 0;
			}

			m_neighbors_coordinate.push_back(coordinate);
			OSM_AR_LOG(g_osm_log, OSM_LOG_DEBUG,
				   "Push coordinate 0x%" PRIx64
				   " to neighbors_coordinate of switch GUID: 0x%"
				   PRIx64 ", LID: %u \n",
				   coordinate,
				   m_db_entry.m_general_sw_info.m_guid,
				   m_db_entry.m_general_sw_info.m_lid);
			return 0;
		}
		/* m_neighbors_dim != 0 */
		if (dim == m_neighbors_dim) {
			return 0; /* additional neighbor on same m_neighbors_dim */
		}

		if (dim != 0) {
			OSM_LOG(g_osm_log, OSM_LOG_ERROR,
				"ERR 921A: failed to calculate coordinates for "
				"switch GUID: 0x%" PRIx64 ", LID: %u based on: "
				"neighbor coordinate 0x%" PRIx64 " on dim: %u "
				"and new neighbor coordinate 0x%" PRIx64
				" on dim: %u \n",
				m_db_entry.m_general_sw_info.m_guid,
				m_db_entry.m_general_sw_info.m_lid,
				m_neighbors_coordinate.front(),
				m_neighbors_dim,
				coordinate,
				dim);

			return 1;
		}

		/*
		 * Coordinate discovered by a neighbor not on a previous
		 * discovered line.
		 */
		OSM_AR_LOG(g_osm_log, OSM_LOG_DEBUG,
			   "AddNeighborCoordinate SetCoordinate intersection of "
			   "coordinate: 0x%" PRIx64 " on dim %u "
			   "and coordinate: 0x%" PRIx64 "\n",
			   m_neighbors_coordinate.front(),
			   m_neighbors_dim,
			   coordinate);
		SetCoordinate(CoordinateIntersection(m_neighbors_coordinate.front(),
						     m_neighbors_dim,
						     coordinate));
		return 0;
	}
	/* Intersect three coordinates */
	uint64_t dim_diff_1 = coordinate ^ m_neighbors_coordinate.front();
	if (!dim_diff_1) {
		OSM_AR_LOG(g_osm_log, OSM_LOG_ERROR,
			   "ERR 921B: "
			   "-A- Two neighbors of switch GUID: 0x%" PRIx64
			   ", LID: %u have same coordinate: 0x%" PRIx64 "\n",
			   m_db_entry.m_general_sw_info.m_guid,
			   m_db_entry.m_general_sw_info.m_lid,
			   coordinate);
		return 1;
	}
	uint64_t dim_diff_2 = coordinate ^ m_neighbors_coordinate.back();
	if (!dim_diff_2) {
		OSM_AR_LOG(g_osm_log, OSM_LOG_ERROR,
			   "ERR 921C: "
			   "-A- Two neighbors of switch GUID: 0x%" PRIx64
			   ", LID: %u have same coordinate: 0x%" PRIx64 "\n",
			   m_db_entry.m_general_sw_info.m_guid,
			   m_db_entry.m_general_sw_info.m_lid,
			   coordinate);
		return 1;
	}

	uint8_t first_dim_1;
	uint8_t first_dim_2;
	uint8_t dim_num_1 = CountDimensions(dim_diff_1, first_dim_1);
	uint8_t dim_num_2 = CountDimensions(dim_diff_2, first_dim_2);

	if ((dim_num_1 > 2) || (dim_num_2 > 2)) {
		OSM_LOG(g_osm_log, OSM_LOG_ERROR,
			"ERR 9220: failed to calculate coordinates for "
			"switch GUID: 0x%" PRIx64 ", LID: %u based on:"
			" 0x%" PRIx64 " ,"
			" 0x%" PRIx64 " ,"
			" 0x%" PRIx64 " \n",
			m_db_entry.m_general_sw_info.m_guid,
			m_db_entry.m_general_sw_info.m_lid,
			m_neighbors_coordinate.front(),
			m_neighbors_coordinate.back(),
			coordinate);
		return 1;
	}

	OSM_AR_LOG(g_osm_log, OSM_LOG_DEBUG,
		   "AddNeighborCoordinate SetCoordinate intersection of three "
		   "coordinates: 0x%" PRIx64 ", 0x%" PRIx64 ", 0x%" PRIx64 "\n",
		   m_neighbors_coordinate.front(),
		   m_neighbors_coordinate.back(),
		   coordinate);

	if (dim_num_1 == 1) {
		/*
		 * coordinate and m_neighbors_coordinate.front()
		 * are on the same line, on direction: first_dim_1
		 * The node coordinate is given by intersection with
		 * m_neighbors_coordinate.back() which is not on the same line.
		 */
		SetCoordinate(CoordinateIntersection(m_neighbors_coordinate.front(),
						     first_dim_1,
						     m_neighbors_coordinate.back()));
		return 0;
	}
	if (dim_num_2 == 1) {
		/*
		 * coordinate and m_neighbors_coordinate.back()
		 * are on the same line, on direction: first_dim_2
		 * The node coordinate is given by intersection with
		 * m_neighbors_coordinate.front() which is not on the same line.
		 */
		SetCoordinate(CoordinateIntersection(m_neighbors_coordinate.back(),
						     first_dim_2,
						     m_neighbors_coordinate.front()));
		return 0;
	}

	uint64_t dim_diff = dim_diff_1 & dim_diff_2;
	if (!IsOneDimension(dim_diff)) {
		OSM_LOG(g_osm_log, OSM_LOG_ERROR,
			"ERR 921D: -A- failed to calculate coordinates for "
			"switch GUID: 0x%" PRIx64 ", LID: %u based on:"
			" 0x%" PRIx64 " ,"
			" 0x%" PRIx64 " ,"
			" 0x%" PRIx64 " \n",
			m_db_entry.m_general_sw_info.m_guid,
			m_db_entry.m_general_sw_info.m_lid,
			m_neighbors_coordinate.front(),
			m_neighbors_coordinate.back(),
			coordinate);
		return 1;
	}

	SetCoordinate(coordinate ^ dim_diff);
	return 0;
}

void CoordinateData::RemoveNeighborCoordinate(uint64_t coordinate)
{
	list <uint64_t>::iterator iter = m_neighbors_coordinate.begin();
	for (; iter != m_neighbors_coordinate.end(); ++iter) {
		if (*iter == coordinate) {
			m_neighbors_coordinate.erase(iter);
			OSM_AR_LOG(g_osm_log, OSM_LOG_DEBUG,
				   "Erase coordinate 0x%" PRIx64
				   " from neighbors_coordinate of switch GUID: 0x%"
				   PRIx64 ", LID: %u \n",
				   coordinate,
				   m_db_entry.m_general_sw_info.m_guid,
				   m_db_entry.m_general_sw_info.m_lid);
			break;
		}
	}
}

/*
 * ArGhcAlgorithm
 */
int ArGhcAlgorithm::SetCoordinates()
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	OSM_LOG(g_osm_log, OSM_LOG_DEBUG,
		"Set Generalized Hypercube Coordinates\n");

	/*
	 * This routine assigns coordinates to each switch of a
	 * Generalized Hypercube topology.
	 * Coordinates calculation is performed in the followings
	 * steps:
	 * 1. Create graph of the switches of the topology (every vertex
	 *    represent switch, no parallel links, no representation to other
	 *    node types).
	 * 2. Select origin node and assign coordinates to it, and to its adjacent
	 *    vertices. This will set coordinates on all main axis.
	 * 3. Assign coordinates to the rest of the vertices.
	 */

	/* Init and set coordinates origin */
	CoordinatesInit();

	/*
	 * Set coordinates to all main axes nodes
	 * e.g: (X1, 0 , 0,...) , (0, X2 , 0,...) ...
	 * this method returns list of nodes on main planes
	 * main planes spreads between two main axes.
	 * e.g (X1, X2, 0, 0, ...) , (0, X2, x3, 0, ...), (X1, X2, 0, 0, ...) ...
	 */
	CoordinateDataList main_planes;
	if (CoordinatesSetMainAxes(main_planes)) {
		OSM_AR_LOG_RETURN(m_p_osm_log, 1);
	}

	if (CoordinatesSet(main_planes)) {
		OSM_AR_LOG_RETURN(m_p_osm_log, 1);
	}

	uint16_t prev_set_coordinates = 0;
	uint16_t switches_number = (uint16_t)m_sw_map_.size();
	/*
	 * Try to assign coordinates to switches that are not assigned yet, by
	 * expanding dimension in order to overcome missing switches on main axes
	 */
	while (switches_number > CoordinateData::sm_total_set_coordinates) {

		prev_set_coordinates = CoordinateData::sm_total_set_coordinates;
		if (ExpandDimension()) {
			OSM_AR_LOG_RETURN(m_p_osm_log, 1);
		}

		if (prev_set_coordinates == CoordinateData::sm_total_set_coordinates) {

			OSM_AR_LOG(g_osm_log, OSM_LOG_INFO,
				   "Failed to set all Generalized Hypercube "
				   "coordinates. set %u coordinates out of %u\n",
				   CoordinateData::sm_total_set_coordinates,
				   (uint16_t)m_sw_map_.size());
			break;
		}
	}
	OSM_AR_LOG_RETURN(m_p_osm_log, 0);
}


void ArGhcAlgorithm::CoordinatesInit()
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	OSM_LOG(m_p_osm_log, OSM_LOG_DEBUG,"CoordinatesInit\n");

	size_t max_neighbors = 0;
	m_origin_node_ = NULL;
	memset (m_max_dim_coord_, 0, sizeof(m_max_dim_coord_));
	CoordinateData::sm_total_set_coordinates = 0;

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_map_.begin();
	for (; sw_node_it != m_sw_map_.end(); ++sw_node_it) {

		KdorSwData *p_kdor_data = sw_node_it->second.m_kdor_data;
		CoordinateData &coordinate_data = p_kdor_data->GetCoordinateData();
		coordinate_data.Clear();

		ConnectionVec &connections = p_kdor_data->GetConnectionsVec();
		int connections_number = p_kdor_data->GetConnectionsNumber();
		for (int connection_idx = 0; connection_idx < connections_number;
		     ++connection_idx) {

			SwitchConnection &connection = connections[connection_idx];
			ARSWDataBaseEntry *remote_db_entry =
				connection.GetRemoteSwitchDbEntry();
			if (remote_db_entry == NULL) {
				continue; /* CAs connection */
			}

			GuidSetInsertRes res =
				coordinate_data.m_neighbors.insert(
					remote_db_entry->m_general_sw_info.m_guid);

			if (res.second) {
				OSM_LOG(m_p_osm_log, OSM_LOG_DEBUG,
					"Add CoordinateData of switch GUID: 0x%" PRIx64
					" to neighbors_data of "
					"switch GUID: 0x%" PRIx64 ", LID: %u \n",
					remote_db_entry->m_general_sw_info.m_guid,
					coordinate_data.m_db_entry.m_general_sw_info.m_guid,
					coordinate_data.m_db_entry.m_general_sw_info.m_lid);

				coordinate_data.m_neighbors_data.push_back(
					&remote_db_entry->m_kdor_data->
					GetCoordinateData());
			}
		}

		size_t neighbors_size = coordinate_data.m_neighbors.size();

		if ((neighbors_size > max_neighbors || m_origin_node_ == NULL) ||
		    ((neighbors_size == max_neighbors) &&
		     (sw_node_it->second.m_general_sw_info.m_guid <
		      m_origin_node_->m_general_sw_info.m_guid))) {

			max_neighbors = neighbors_size;
			m_origin_node_ = &(sw_node_it->second);

			OSM_LOG(g_osm_log, OSM_LOG_DEBUG,
				"Reset origin node to switch GUID: 0x%" PRIx64
				", LID: %u desc: %s number of neighbors: %u\n",
				m_origin_node_->m_general_sw_info.m_guid,
				m_origin_node_->m_general_sw_info.m_lid,
				m_origin_node_->m_general_sw_info.m_p_osm_sw->p_node->node_desc.description,
				(int)max_neighbors);
		}
	}
	OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

/*
 * Set coordinates to all main axes nodes
 * e.g: (X1, 0 , 0,...) , (0, X2 , 0,...) ...
 * this method returns list of nodes on main planes
 * e.g (X1, X2, 0, 0, ...) , (0, X2, x3, 0, ...), (X1, X2, 0, 0, ...) ...
 */
int ArGhcAlgorithm::CoordinatesSetMainAxes(CoordinateDataList &main_planes)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);
	int rc = 0;

	CoordinateData &coordinate_data =
		m_origin_node_->m_kdor_data->GetCoordinateData();

	coordinate_data.SetCoordinate(0);
	/* AddNeighborCoordinate 0 */
	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG, "Update origin neighbors\n");
	CoordinateDataList::const_iterator iter =
		coordinate_data.m_neighbors_data.begin();
	for (; iter != coordinate_data.m_neighbors_data.end(); ++iter) {
		(*iter)->AddNeighborCoordinate(0);
	}

	uint8_t dim = 0;
	iter = coordinate_data.m_neighbors_data.begin();
	for (; iter != coordinate_data.m_neighbors_data.end(); ++iter) {
		if ((*iter)->IsCoordinateSet()) {
			continue;
		}

		/*
		 * Advance to next unused dim
		 * we may return to previous dim in case of merge dimensions
		 */
		while ((++dim <= KDOR_MAX_DIM) && m_max_dim_coord_[dim]);

		if (dim > KDOR_MAX_DIM) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 921E: "
				   "Dimension number exceeded maximum of: %u\n",
				   KDOR_MAX_DIM);
			rc = 1;
			break;
		}
		rc = CoordinatesSetMainAxis(main_planes, dim, *(*iter));
		if (rc) {
			break;
		}
	}

	OSM_AR_LOG_RETURN(m_p_osm_log, rc);
}

/*
 * Set coordinates to nodes on a main axis on the given dimension dim
 * this method updates OUT parameter main_planes with nodes on main planes
 * coordinate_data: a node on the main axis
 */
int ArGhcAlgorithm::CoordinatesSetMainAxis(CoordinateDataList &main_planes,
					   uint8_t &dim,
					   CoordinateData &coordinate_data)
{
	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
		   "CoordinatesSetMainAxis dim: %u\n", dim);
	queue <CoordinateData *> axis_queue;
	coordinate_data.SetCoordinate(GetNextCoordinate(dim));
	axis_queue.push(&coordinate_data);
	return CoordinatesSetMainAxis(main_planes, axis_queue, dim);
}

int ArGhcAlgorithm::CoordinatesSetMainAxis(CoordinateDataList &main_planes,
					   queue <CoordinateData *> &axis_queue,
					   uint8_t &dim)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
		   "CoordinatesSetMainAxis dim: %u axis_queue.size: %u\n",
		   dim, (int)axis_queue.size());

	while (!axis_queue.empty()) {
		CoordinateData *p_curr_data = axis_queue.front();
		axis_queue.pop();

		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "CoordinatesSetMainAxis dim: %u "
				   "Handle neighbors of switch GUID: 0x%" PRIx64
				   ", LID: %u \n",
				   dim,
				   p_curr_data->m_db_entry.m_general_sw_info.m_guid,
				   p_curr_data->m_db_entry.m_general_sw_info.m_lid);

		CoordinateDataList::const_iterator iter =
			p_curr_data->m_neighbors_data.begin();
		for (; iter != p_curr_data->m_neighbors_data.end(); ++iter) {

			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "CoordinatesSetMainAxis dim: %u "
				   "Handle neighbor switch GUID: 0x%" PRIx64
				   ", LID: %u \n",
				   dim,
				   (*iter)->m_db_entry.m_general_sw_info.m_guid,
				   (*iter)->m_db_entry.m_general_sw_info.m_lid);

			if ((*iter)->IsCoordinateSet()) {
				continue;
			}

			if ((*iter)->m_neighbors_coordinate.size() == 2) {
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
					   "CoordinatesSetMainAxis dim: %u "
					   "Conflict found on switch GUID: 0x%" PRIx64
					   ", LID: %u. Merge is needed.\n",
					   dim,
					   (*iter)->m_db_entry.m_general_sw_info.m_guid,
					   (*iter)->m_db_entry.m_general_sw_info.m_lid);

				if (MergeMainAxes(axis_queue, *iter, dim)) {
					OSM_AR_LOG_RETURN(m_p_osm_log, 1);
				}
				/*
				 * p_curr_data and neighbor handled by
				 * MergeMainAxes. move to next in axis_queue
				 */
				break;
			}

			if (AddNeighborMainAxisCoordinate(
				dim, **iter, p_curr_data->m_coordinate)){
				OSM_AR_LOG_RETURN(m_p_osm_log, 1);
			}

			if ((*iter)->IsCoordinateSet()) {
				axis_queue.push(*iter);
				continue;
			}

			if ((*iter)->m_neighbors_coordinate.size() == 2) {
				/*
				 * A node that has two neighbors on main axes
				 * is on main plane
				 */
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
					   "Push switch GUID: 0x%" PRIx64 ", LID: %u "
					   "to main_planes\n",
					   (*iter)->m_db_entry.m_general_sw_info.m_guid,
					   (*iter)->m_db_entry.m_general_sw_info.m_lid);
				main_planes.push_back(*iter);
			}
		}
	}
	OSM_AR_LOG_RETURN(m_p_osm_log, 0);
}

int ArGhcAlgorithm::AddNeighborMainAxisCoordinate(
	uint8_t axis_dim,
	CoordinateData &coordinate_data,
	uint64_t neighbor_coordinate)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	if(coordinate_data.AddNeighborCoordinate(neighbor_coordinate)) {
		OSM_AR_LOG_RETURN(m_p_osm_log, 1);
	}

	if (coordinate_data.m_neighbors_dim == axis_dim) {
		if (m_max_dim_coord_[axis_dim] == FIRST_DIM_BITS) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 921F: "
				   "Dimension %u size exceeded maximum of: %u \n",
				   axis_dim, (int)FIRST_DIM_BITS);
			OSM_AR_LOG_RETURN(m_p_osm_log, 1);
		}
		coordinate_data.SetCoordinate(GetNextCoordinate(axis_dim));
	}
	OSM_AR_LOG_RETURN(m_p_osm_log, 0);
}

/*
 * Merge the given p_coordinate_data and all its neighbors with set coordinate
 * To a single main axis.
 * New discovered coordinate will be inserted to axis_queue
 */
int ArGhcAlgorithm::MergeMainAxes(queue <CoordinateData *> &axis_queue,
				  CoordinateData *p_coordinate_data,
				  uint8_t &dim)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	if (p_coordinate_data->m_neighbors_coordinate.size() != 2) {
		OSM_AR_LOG(g_osm_log, OSM_LOG_ERROR,
			   "ERR 9221: MergeMainAxes: switch GUID: 0x%" PRIx64
			   ", LID: %u has unexpected neighbors_coordinate number: %u\n",
			   p_coordinate_data->m_db_entry.m_general_sw_info.m_guid,
			   p_coordinate_data->m_db_entry.m_general_sw_info.m_lid,
			   (int)p_coordinate_data->m_neighbors_coordinate.size());
		OSM_AR_LOG_RETURN(m_p_osm_log, 1);
	}

	uint8_t from_dim, to_dim;
	/* 2 is the number of coordinates in m_neighbors_coordinate */
	uint8_t neighbor_dim[2] = {};

	list <uint64_t>::const_iterator iter =
		p_coordinate_data->m_neighbors_coordinate.begin();
	for (uint8_t dim_idx = 0;
	     iter != p_coordinate_data->m_neighbors_coordinate.end();
	     ++iter, ++dim_idx) {
		neighbor_dim[dim_idx] = CoordinateData::IsOneDimension(*iter);
		if (neighbor_dim[dim_idx] == 0) {
			OSM_AR_LOG(g_osm_log, OSM_LOG_ERROR,
				   "ERR 9222: "
				   "MergeMainAxes: switch GUID: 0x%" PRIx64
				   ", LID: %u has unexpected neighbor coordinate: 0x%"
				   PRIx64 "\n",
				   p_coordinate_data->m_db_entry.m_general_sw_info.m_guid,
				   p_coordinate_data->m_db_entry.m_general_sw_info.m_lid,
				   *iter);
			OSM_AR_LOG_RETURN(m_p_osm_log, 1);
		}
	}

	/* Set to_dim to be the minimal main axis neighbor dimension */
	if (neighbor_dim[0] < neighbor_dim[1]) {
		to_dim = neighbor_dim[0];
		from_dim = neighbor_dim[1];
	} else {
		to_dim = neighbor_dim[1];
		from_dim = neighbor_dim[0];
	}

	OSM_AR_LOG(g_osm_log, OSM_LOG_DEBUG,
		   "MergeMainAxes: switch GUID: 0x%" PRIx64
		   ", LID: %u dim: %u from_dim:%u to_dim:%u\n",
		   p_coordinate_data->m_db_entry.m_general_sw_info.m_guid,
		   p_coordinate_data->m_db_entry.m_general_sw_info.m_lid,
		   dim, from_dim, to_dim);

	uint8_t to_dim_mask = KDOR_DIM_MASK(to_dim);
	CoordinateDataList::const_iterator neighbors_iter =
		p_coordinate_data->m_neighbors_data.begin();
	for (; neighbors_iter != p_coordinate_data->m_neighbors_data.end();
	     ++neighbors_iter) {

		if ((*neighbors_iter)->IsCoordinateSet() &&
		    ((*neighbors_iter)->m_coordinate & ~to_dim_mask)) {
			if (MergeMainAxesToDim(axis_queue, *neighbors_iter,
					       to_dim)) {
				OSM_AR_LOG_RETURN(m_p_osm_log, 1);
			}
		}
	}

	OSM_AR_LOG(g_osm_log, OSM_LOG_DEBUG,
		   "MergeMainAxes to_dim:%u completed.\n",to_dim);

	m_max_dim_coord_[from_dim] = 0;
	/* in some cases dim may be different from from_dim */
	m_max_dim_coord_[dim] = 0;
	dim = to_dim; /* set coordinates added to axis_queue on main axis dim */

	OSM_AR_LOG_RETURN(m_p_osm_log, 0);
}

/*
 * Change main axis coordinate of the given p_coordinate_data, and all its
 * main axis neighbors, to a new coordinate on the given to_dim.
 * In addition update all updated nodes neighbors accordingly.
 * New discovered coordinate will be inserted to axis_queue
 */
int ArGhcAlgorithm::MergeMainAxesToDim(queue <CoordinateData *> &axis_queue,
				       CoordinateData *p_coordinate_data,
				       uint8_t to_dim)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	uint8_t to_dim_mask = KDOR_DIM_MASK(to_dim);
	queue <CoordinateData *> changed_coord_data_queue;
	changed_coord_data_queue.push(p_coordinate_data);

	while (!changed_coord_data_queue.empty()) {
		CoordinateData *p_curr_data = changed_coord_data_queue.front();
		changed_coord_data_queue.pop();

		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Merge main axis switch GUID: 0x%" PRIx64
			   ", LID: %u desc: %s to dim: %u\n",
			   p_curr_data->m_db_entry.m_general_sw_info.m_guid,
			   p_curr_data->m_db_entry.m_general_sw_info.m_lid,
			   p_curr_data->m_db_entry.m_general_sw_info.m_p_osm_sw->
				p_node->node_desc.description,
			   to_dim);

		if (m_max_dim_coord_[to_dim] == FIRST_DIM_BITS) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 9223: "
				   "Dimension %u size exceeded maximum of: %u\n",
				   to_dim, (int)FIRST_DIM_BITS);
			OSM_AR_LOG_RETURN(m_p_osm_log, 1);
		}
		uint64_t orig_coordinate = p_curr_data->m_coordinate;
		p_curr_data->DeleteCoordinate();
		p_curr_data->SetCoordinate(GetNextCoordinate(to_dim));

		CoordinateDataList::const_iterator iter =
			p_curr_data->m_neighbors_data.begin();
		for (; iter != p_curr_data->m_neighbors_data.end(); ++iter) {

			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "MergeMainAxes to dim: %u "
				   "Handle neighbor switch GUID: 0x%" PRIx64
				   ", LID: %u\n",
				   to_dim,
				   (*iter)->m_db_entry.m_general_sw_info.m_guid,
				   (*iter)->m_db_entry.m_general_sw_info.m_lid);

			if ((*iter)->IsCoordinateSet()) {

				/* Skip origin */
				if ((*iter)->m_coordinate == 0) {
					continue;
				}

				if ((*iter)->m_neighbors_coordinate.size() == 2) {
					(*iter)->DeleteCoordinate();
					/*
					 * In addition need to delete orig_coordinate
					 * from m_neighbors_coordinate of neighbors
					 * that are not on main axis
					 */
				} else if ((*iter)->m_coordinate & to_dim_mask) {
					/* We already update this main axis neighbor */
					continue;
				} else {
					OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
						   "Add switch GUID: 0x%" PRIx64
						   ", LID: %u coordinate: 0x%" PRIx64
						   " to Merge main axes queue to_dim:%u\n",
						   (*iter)->m_db_entry.m_general_sw_info.m_guid,
						   (*iter)->m_db_entry.m_general_sw_info.m_lid,
						   (*iter)->m_coordinate,
						   to_dim);

					changed_coord_data_queue.push(*iter);
					continue;
				}
			}

			/*
			 * Delete orig_coordinate from m_neighbors_coordinate
			 * of neighbors not on main axis
			 */
			(*iter)->RemoveNeighborCoordinate(orig_coordinate);

			if (AddNeighborMainAxisCoordinate(
				to_dim, **iter, p_curr_data->m_coordinate)){
				OSM_AR_LOG_RETURN(m_p_osm_log, 1);
			}

			if ((*iter)->IsCoordinateSet()) {
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
					   "MergeMainAxes to dim: %u "
					   "Add neighbor switch GUID: 0x%" PRIx64
					   ", LID: %u to axis_queue\n",
					   to_dim,
					   (*iter)->m_db_entry.m_general_sw_info.m_guid,
					   (*iter)->m_db_entry.m_general_sw_info.m_lid);

				/* New discovered coordinates will be handled by caller */
				axis_queue.push(*iter);
				continue;
			}
		}
	}
	OSM_AR_LOG_RETURN(m_p_osm_log, 0);
}

/* Set coordinate on all GHC topology given nodes on main planes */
int ArGhcAlgorithm::CoordinatesSet(CoordinateDataList &main_planes)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	/* Use list so we can easily clear the cueue if needed */
	list <CoordinateData *> new_coord_queue;
	CoordinateData *p_conflict_neighbor;

	bool conflict_discovered = false;
	for (int retry = 0; retry < KDOR_MAX_DIM; ++retry) {
		conflict_discovered = false;
		for (CoordinateDataList::const_iterator iter = main_planes.begin();
		     iter != main_planes.end(); ++iter) {

			if ((*iter)->m_neighbors_coordinate.size() != 2) {
				/* This might happen on main axis merge */
				continue;
			}
			(*iter)->SetCoordinate(
				(*iter)->m_neighbors_coordinate.front() ^
				(*iter)->m_neighbors_coordinate.back());

			p_conflict_neighbor = FindConflictingNeighbor(*iter);

			if (p_conflict_neighbor) {
				conflict_discovered = true;
				if (MergeMainPlaneCoordinate(
					main_planes, *iter, p_conflict_neighbor)) {
					OSM_AR_LOG_RETURN(m_p_osm_log, 1);
				}
				new_coord_queue.clear();
				break;
			}

			new_coord_queue.push_back(*iter);
		}

		if (!conflict_discovered) {
			break;
		}
	}

	if (conflict_discovered) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR, "ERR 9224: "
			   "CoordinatesSet failed to resolve conflicts \n");
		OSM_AR_LOG_RETURN(m_p_osm_log, 1);
	}

	if (SetNeighborsCoordinate(new_coord_queue)) {
		OSM_AR_LOG_RETURN(m_p_osm_log, 1);
	}

	OSM_AR_LOG_RETURN(m_p_osm_log, 0);
}

/* Update neighbors and set coordinates recursively */
int ArGhcAlgorithm::SetNeighborsCoordinate(list <CoordinateData *> &new_coord_queue)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);
	while (!new_coord_queue.empty()) {
		CoordinateData *p_curr_data = new_coord_queue.front();
		new_coord_queue.pop_front();

		/* AddNeighborCoordinate 0 */
		CoordinateDataList::const_iterator iter =
			p_curr_data->m_neighbors_data.begin();
		for (; iter != p_curr_data->m_neighbors_data.end(); ++iter) {
			if ((*iter)->IsCoordinateSet()) {
				continue;
			}

			if ((*iter)->AddNeighborCoordinate(p_curr_data->m_coordinate)) {
				OSM_AR_LOG_RETURN(m_p_osm_log, 1);
			}
			if ((*iter)->IsCoordinateSet()) {
				new_coord_queue.push_back(*iter);
			}
		}
	}
	OSM_AR_LOG_RETURN(m_p_osm_log, 0);
}

bool ArGhcAlgorithm::IsValidCoordinate(ARSWDataBaseEntry &sw_db_entry)
{
	return sw_db_entry.m_kdor_data->GetCoordinateData().IsCoordinateSet();
}

/*
 * In case a switch is missing on a main axis we have to expand the size
 * of the dimension of this main axis
 */
int ArGhcAlgorithm::ExpandDimension()
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	uint16_t rank = 0;
	/*Select the node with the highest calculated rank*/
	CoordinateData *p_coordinate_data = NULL;

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_map_.begin();
	for (; sw_node_it != m_sw_map_.end(); ++sw_node_it) {
		CoordinateData &current_coordinate_data =
			sw_node_it->second.m_kdor_data->GetCoordinateData();

		if (current_coordinate_data.IsCoordinateSet() ||
		    current_coordinate_data.m_neighbors_dim == 0) {
			continue;
		}

		/*
		 * Find the node with maximal neighbors with
		 * equal value of m_neighbors_dim
		 */
		uint16_t current_rank = 1;
		CoordinateDataList::const_iterator iter =
			current_coordinate_data.m_neighbors_data.begin();
		for (; iter != current_coordinate_data.m_neighbors_data.end(); ++iter) {
			if ((*iter)->IsCoordinateSet()) {
				continue;
			}

			if ((*iter)->m_neighbors_dim ==
			    current_coordinate_data.m_neighbors_dim) {
				current_rank++;
			}
		}

		if (current_rank > rank) {
			p_coordinate_data = &current_coordinate_data;
			rank = current_rank;
		}
	}

	if (p_coordinate_data) {

		uint8_t expanded_dim =
			p_coordinate_data->m_neighbors_dim;

		if (m_max_dim_coord_[expanded_dim] == FIRST_DIM_BITS) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 9225: "
				   "Failed to Expand dimension for "
				   "Switch GUID: 0x%" PRIx64 ", LID: %u "
				   "neighbor coord: 0x%" PRIx64
				   "Dimension %u size exceeded maximum of: %u \n",
				   p_coordinate_data->m_db_entry.m_general_sw_info.m_guid,
				   p_coordinate_data->m_db_entry.m_general_sw_info.m_lid,
				   p_coordinate_data->m_neighbors_coordinate.front(),
				   expanded_dim, (int)FIRST_DIM_BITS);
			OSM_AR_LOG_RETURN(m_p_osm_log, 1);
		}

		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Expand dimension %u\n", expanded_dim);

		uint64_t dim_coordinate = GetNextCoordinate(expanded_dim);
		p_coordinate_data->SetCoordinate(
			CoordinateData::CoordinateIntersection(
				p_coordinate_data->m_neighbors_coordinate.front(),
				expanded_dim,
				dim_coordinate));

		list <CoordinateData *> new_coord_queue;
		new_coord_queue.push_back(p_coordinate_data);

		if (SetNeighborsCoordinate(new_coord_queue)) {
			OSM_AR_LOG_RETURN(m_p_osm_log, 1);
		}
	}
	OSM_AR_LOG_RETURN(m_p_osm_log, 0);
}

/*
 * Validate node coordinate with respect to its neighbors
 * If conflict found, return the conflicting neighbor
 */
CoordinateData *ArGhcAlgorithm::FindConflictingNeighbor(CoordinateData *p_coordinate_data)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	uint8_t path_dim;
	CoordinateDataList::const_iterator iter =
		p_coordinate_data->m_neighbors_data.begin();
	for (; iter != p_coordinate_data->m_neighbors_data.end(); ++iter) {
		if (!(*iter)->IsCoordinateSet()) {
			continue;
		}

		uint8_t path_dim = CoordinateData::IsOneDimension(
			p_coordinate_data->m_coordinate ^  (*iter)->m_coordinate);

		if (!path_dim) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Discovered conflict between Switch GUID: 0x%"
				   PRIx64 ", LID: %u coord: 0x%"
				   PRIx64 " to switch GUID: 0x%" PRIx64
				   ", LID: %u coord: 0x%" PRIx64 "\n",
				   p_coordinate_data->m_db_entry.m_general_sw_info.m_guid,
				   p_coordinate_data->m_db_entry.m_general_sw_info.m_lid,
				   p_coordinate_data->m_coordinate,
				   (*iter)->m_db_entry.m_general_sw_info.m_guid,
				   (*iter)->m_db_entry.m_general_sw_info.m_lid,
				   (*iter)->m_coordinate);

			OSM_AR_LOG_RETURN(m_p_osm_log, *iter);
		}
	}
	OSM_AR_LOG_RETURN(m_p_osm_log, NULL);
}

int ArGhcAlgorithm::MergeMainPlaneCoordinate(CoordinateDataList &main_planes,
					     CoordinateData *p_coordinate_data,
					     CoordinateData *p_conflict_neighbor)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);
	uint64_t path_coordinate = p_coordinate_data->m_coordinate ^
		p_conflict_neighbor->m_coordinate;

	uint8_t dim[2] = {0,0};
	uint8_t idx = 0;

	while (path_coordinate) {
		OSM_AR_LOG(g_osm_log, OSM_LOG_DEBUG,
			   "MergeMainPlaneCoordinate Coordinate: 0x%" PRIx64
			   " dim: %u (coordinate & FIRST_DIM_BITS):0x%" PRIx64 "\n",
			   path_coordinate, dim[idx],(path_coordinate & FIRST_DIM_BITS));

		dim[idx]++;
		if (path_coordinate & FIRST_DIM_BITS) {
			if (idx == 0) {
				idx++;
				dim[idx] = dim[0];
			} else {
				break;
			}
		}
		path_coordinate >>= BITS_PER_DIM;
	}

	OSM_AR_LOG(g_osm_log, OSM_LOG_DEBUG,
		   "MergeMainPlaneCoordinate merge dimension %u to dimension %u\n",
		   dim[1], dim[0]);

	/* Find a node on main axis from dim as starting merge point */
	CoordinateData *p_merge_coordinate_data = NULL;
	CoordinateData *p_merge_neighbor = p_coordinate_data;

	uint64_t from_dim_mask = KDOR_DIM_MASK(dim[1]);
	if (p_conflict_neighbor->m_coordinate & from_dim_mask) {
		p_merge_neighbor = p_conflict_neighbor;
	}

	CoordinateDataList::const_iterator iter =
		p_merge_neighbor->m_neighbors_data.begin();
	for (; iter != p_merge_neighbor->m_neighbors_data.end(); ++iter) {
		if (!(*iter)->IsCoordinateSet()) {
			continue;
		}

		OSM_AR_LOG(g_osm_log, OSM_LOG_DEBUG,
			   "MergeMainPlaneCoordinate test neighbor Coordinate: 0x%"
			   PRIx64 "\n", (*iter)->m_coordinate);

		/* Not Main Axis Coordinate */
		if ((*iter)->m_neighbors_coordinate.size() == 2) {
			continue;
		}

		if ((*iter)->m_coordinate & from_dim_mask) {
			p_merge_coordinate_data = *iter;
			break;
		}
	}

	if (p_merge_coordinate_data == NULL) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR, "ERR 9226: "
			   "Merge main plane coordinate: "
			   "Failed to find merge staritig point near "
			   "switch GUID: 0x%" PRIx64 ", LID: %u coordinate: 0x%"
			   PRIx64 " on main axis dimension: %u \n",
			   p_merge_neighbor->m_db_entry.m_general_sw_info.m_guid,
			   p_merge_neighbor->m_db_entry.m_general_sw_info.m_lid,
			   p_merge_neighbor->m_coordinate,
			   dim[1]);

		OSM_AR_LOG_RETURN(m_p_osm_log, 1);
	}

	queue <CoordinateData *> axis_queue;
	if (MergeMainAxesToDim(axis_queue, p_merge_coordinate_data, dim[0])) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR, "ERR 9227: "
			   "MergeMainAxesToDim failed \n");
		OSM_AR_LOG_RETURN(m_p_osm_log, 1);
	}

	if (!axis_queue.empty() &&
	    CoordinatesSetMainAxis(main_planes, axis_queue, dim[0])) {
		OSM_AR_LOG_RETURN(m_p_osm_log, 1);
	}

	OSM_AR_LOG_RETURN(m_p_osm_log, 0);
}

/*
 * Update connection::direction dimensions index and sign for given
 * connection on src switch
 * Valid connection direction is when coordinate changes, between src and
 * remote nodes, by one and only one dimension.
 * If no change found (same coordinate for src and remote)
 * or more than one dimension changed, return non zero value
 */
int ArGhcAlgorithm::SetDirection(osm_switch_t* src,
				 SwitchConnection &connection)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	osm_switch_t* dst = connection.m_remote_switch;
	if (dst == NULL) {
		OSM_AR_LOG_RETURN(m_p_osm_log, 0);
	}

	CoordinateData &src_coord =
		((ARSWDataBaseEntry *)(src->priv))->m_kdor_data->GetCoordinateData();
	CoordinateData &dst_coord =
		((ARSWDataBaseEntry *)(dst->priv))->m_kdor_data->GetCoordinateData();


	if (!src_coord.IsCoordinateSet() || !dst_coord.IsCoordinateSet()) {
		connection.m_dim_idx = INVALID_DIM_INDEX;
		connection.m_dim_sign = (dim_sign_t)(1);

		if (m_p_osm_log->level & OSM_LOG_DEBUG) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Invalid coordinate on set direction from "
				   "switch GUID: 0x%" PRIx64 ", LID: %u "
				   "to switch GUID: 0x%" PRIx64 ", LID: %u "
				   "dim_idx: %u  dim_sign: %d\n",
				   src_coord.m_db_entry.m_general_sw_info.m_guid,
				   src_coord.m_db_entry.m_general_sw_info.m_lid,
				   dst_coord.m_db_entry.m_general_sw_info.m_guid,
				   dst_coord.m_db_entry.m_general_sw_info.m_lid,
				   connection.m_dim_idx, connection.m_dim_sign);
		}
		OSM_AR_LOG_RETURN(m_p_osm_log, 0);
	}

	/*
	 * Check if the coordinate change between src and remote is along one
	 * dimension and if yes (therefore valid), along which dimension.
	 */
	uint8_t dim = CoordinateData::IsOneDimension(
		src_coord.m_coordinate ^  dst_coord.m_coordinate);

	if (!dim) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 9228: Switch GUID: 0x%" PRIx64 ", LID: %u coord: 0x%"
			   PRIx64 " is not valid GHC member of "
			   "switch GUID: 0x%" PRIx64 ", LID: %u coord: 0x%"
			   PRIx64 "\n",
			   src_coord.m_db_entry.m_general_sw_info.m_guid,
			   src_coord.m_db_entry.m_general_sw_info.m_lid,
			   src_coord.m_coordinate,
			   dst_coord.m_db_entry.m_general_sw_info.m_guid,
			   dst_coord.m_db_entry.m_general_sw_info.m_lid,
			   dst_coord.m_coordinate);

		OSM_AR_LOG_RETURN(m_p_osm_log, 1);
	}

	connection.m_dim_idx = dim - 1;
	uint64_t dim_mask = KDOR_DIM_MASK(dim);

	if ((src_coord.m_coordinate & dim_mask) > (dst_coord.m_coordinate & dim_mask)) {
		connection.m_dim_sign = (dim_sign_t)(-1);
	} else {
		connection.m_dim_sign = (dim_sign_t)(1);
	}

	if (m_p_osm_log->level & OSM_LOG_DEBUG) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Direction from switch GUID: 0x%" PRIx64
			   ", LID: %u coord: 0x%" PRIx64
			   " to switch GUID: 0x%" PRIx64
			   ", LID: %u coord: 0x%" PRIx64
			   " (src_coord.m_coordinate & dim_mask): 0x%" PRIx64
			   " (dst_coord.m_coordinate & dim_mask): 0x%" PRIx64
			   " dim_mask: 0x%" PRIx64
			   " dim_idx: %u"
			   " dim_sign: %d\n",
			   src_coord.m_db_entry.m_general_sw_info.m_guid,
			   src_coord.m_db_entry.m_general_sw_info.m_lid,
			   src_coord.m_coordinate,
			   dst_coord.m_db_entry.m_general_sw_info.m_guid,
			   dst_coord.m_db_entry.m_general_sw_info.m_lid,
			   dst_coord.m_coordinate,
			   (src_coord.m_coordinate & dim_mask),
			   (dst_coord.m_coordinate & dim_mask),
			   dim_mask,
			   connection.m_dim_idx, connection.m_dim_sign);
	}

	OSM_AR_LOG_RETURN(m_p_osm_log, 0);
}
