/*
 * Copyright (c) 2021-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * Copyright (c) 2006-2009 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2009 HNR Consulting. All rights reserved.
 * Copyright (c) 2012 Lawrence Livermore National Lab.  All rights reserved.
 * Copyright (c) 2014-2020 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    OSM Congestion Control configuration implementation
 *
 * Author:
 *    Albert Chu, LLNL
 *    Alex Minchiu, Mellanox
 *    Daniel Klein, Mellanox
 *    Or Nechemia, Mellanox
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>

#include <iba/ib_types.h>
#include <complib/cl_debug.h>
#include <complib/cl_math.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_CONGESTION_CONTROL_C
#include <opensm/osm_subnet.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_log.h>
#include <opensm/osm_subnet.h>
#include <opensm/osm_congestion_control.h>
#include <opensm/osm_unhealthy_ports.h>
#include <opensm/osm_key_mgr.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_inform.h>
#include <dirent.h>
#include <errno.h>

#define CONGESTION_CONTROL_INITIAL_TID_VALUE 0x7A93
#define OSM_CC_DEFAULT_PRINT_SUPPRESS_TIMEOUT 5000000
#define CC_ATTR_MOD_ALL_PORTS CL_HTON32(0x000000ff)

extern int parse_congestion_control_policy(struct osm_opensm *osm_ptr);
extern int parse_ppcc_algo_file(struct osm_opensm *osm_ptr, char *algo_file_path, char *file_name);
extern int osm_print_num_received(IN uint32_t num_received);
extern uint64_t osm_trap_get_key(IN uint16_t lid, IN uint8_t port_num,
				 IN osm_madw_t * p_ntci);
static ib_api_status_t congestion_control_send_set_keyinfo(osm_subn_t * p_subn,
							   osm_port_t * p_port);

int cc_compare_profiles(IN const void * p_profiles_name_1, IN const void * p_profiles_name_2)
{
	return strcmp((const char *)p_profiles_name_1, (const char *)p_profiles_name_2);
}

int ppcc_compare_algo_version(IN const void * p_algo_version_1, IN const void * p_algo_version_2)
{
	return memcmp((const ib_ppcc_algo_info_t *)p_algo_version_1, (const ib_ppcc_algo_info_t *)p_algo_version_2, sizeof(ib_ppcc_algo_info_t));
}

void osm_cc_mgr_construct(IN osm_cc_mgr_t * p_mgr)
{
	cl_fmap_init(&p_mgr->cc_profiles_map, cc_compare_profiles);
	cl_qlist_init(&p_mgr->sl_profiles_conf);
	cl_qlist_init(&p_mgr->vl_profiles_conf);
	cl_qlist_init(&p_mgr->sw_conf);
	cl_qlist_init(&p_mgr->ca_conf);
	cl_fmap_init(&p_mgr->ca_algo_conf, ppcc_algo_compare_strings);
	cl_fmap_init(&p_mgr->algo_version_to_id_map, ppcc_compare_algo_version);
}

void osm_cc_destroy_ppcc_ca_algo(osm_ppcc_ca_algo_t **pp_ppcc_ca_algo)
{
	osm_ppcc_ca_algo_t *p_ppcc_ca_algo = *pp_ppcc_ca_algo;
	int i;

	if (!p_ppcc_ca_algo)
		return;

	for (i = 0; i < PPCC_ALGO_MAX_PARAMS; i++) {
		if (!p_ppcc_ca_algo->param_list[i])
			break;

		free(p_ppcc_ca_algo->param_list[i]);
	}

	free(*pp_ppcc_ca_algo);
	*pp_ppcc_ca_algo = NULL;
}

void osm_cc_mgr_destroy(IN osm_cc_mgr_t * p_mgr)
{
	cl_fmap_item_t *p_item, *p_next_item;
	cl_map_item_t *p_map_item, *p_next_map_item;
	cl_fmap_item_t *p_algo_item, *p_next_algo_item;
	osm_cc_sl_conf_t *p_sl_conf = NULL;
	osm_cc_vl_conf_t *p_vl_conf = NULL;
	osm_cc_ca_conf_t *p_ca_conf = NULL;
	osm_cc_sw_conf_t *p_sw_conf = NULL;
	osm_cc_ca_algo_conf_t *p_ca_algo_conf = NULL;
	osm_ppcc_ca_algos_import_conf_t *p_ca_algos_import_conf = NULL;
	osm_ppcc_ca_algo_t *p_ppcc_ca_algo = NULL;

	p_next_item = cl_fmap_head(&p_mgr->cc_profiles_map);
	while (p_next_item != cl_fmap_end(&p_mgr->cc_profiles_map)) {
		p_item = p_next_item;
		p_next_item = cl_fmap_next(p_item);
		cl_fmap_remove_item(&p_mgr->cc_profiles_map, p_item);
		free((osm_cc_profiles_conf_t *) p_item);
	}

	while (cl_qlist_count(&p_mgr->sl_profiles_conf) > 0) {
		p_sl_conf =
		    (osm_cc_sl_conf_t *) cl_qlist_remove_head(&p_mgr->sl_profiles_conf);
		free(p_sl_conf);
	}

	while (cl_qlist_count(&p_mgr->vl_profiles_conf) > 0) {
		p_vl_conf =
		    (osm_cc_vl_conf_t *) cl_qlist_remove_head(&p_mgr->vl_profiles_conf);
		free(p_vl_conf);
	}

	while (cl_qlist_count(&p_mgr->sw_conf) > 0) {
		p_sw_conf =
		    (osm_cc_sw_conf_t *) cl_qlist_remove_head(&p_mgr->sw_conf);
		free(p_sw_conf);
	}

	while (cl_qlist_count(&p_mgr->ca_conf) > 0) {
		p_ca_conf =
		    (osm_cc_ca_conf_t *) cl_qlist_remove_head(&p_mgr->ca_conf);
		free(p_ca_conf);
	}

	p_next_algo_item = cl_fmap_head(&p_mgr->ca_algo_conf);
	while (p_next_algo_item != cl_fmap_end(&p_mgr->ca_algo_conf)) {
		p_algo_item = p_next_algo_item;
		p_next_algo_item = cl_fmap_next(p_algo_item);
		cl_fmap_remove_item(&p_mgr->ca_algo_conf, p_algo_item);

		p_ca_algo_conf = (osm_cc_ca_algo_conf_t *) p_algo_item;
		p_next_item = cl_fmap_head(&p_ca_algo_conf->param_name_to_index);
		while (p_next_item != cl_fmap_end(&p_ca_algo_conf->param_name_to_index)) {
			p_item = p_next_item;
			p_next_item = cl_fmap_next(p_item);
			cl_fmap_remove_item(&p_ca_algo_conf->param_name_to_index, p_item);
			free((osm_cc_algo_param_index_t *) p_item);
		}

		free(p_ca_algo_conf);
	}

	p_next_item = cl_fmap_head(&p_mgr->algo_version_to_id_map);
	while (p_next_item != cl_fmap_end(&p_mgr->algo_version_to_id_map)) {
		p_item = p_next_item;
		p_next_item = cl_fmap_next(p_item);
		cl_fmap_remove_item(&p_mgr->algo_version_to_id_map, p_item);
		free((osm_ppcc_algo_id_map_item_t *) p_item);
	}

	if (p_mgr->ca_algo_import_conf) {
		p_ca_algos_import_conf = p_mgr->ca_algo_import_conf;
		p_next_map_item = cl_qmap_head(&p_mgr->ca_algo_import_conf->ca_algos);
		while (p_next_map_item != cl_qmap_end(&p_mgr->ca_algo_import_conf->ca_algos)) {
			p_map_item = p_next_map_item;
			p_next_map_item = cl_qmap_next(p_map_item);
			cl_qmap_remove_item(&p_mgr->ca_algo_import_conf->ca_algos, p_map_item);
			p_ppcc_ca_algo = (osm_ppcc_ca_algo_t *) p_map_item;
			osm_cc_destroy_ppcc_ca_algo(&p_ppcc_ca_algo);
		}
		if (p_ca_algos_import_conf->p_current_ca_algo)
			osm_cc_destroy_ppcc_ca_algo(&p_ca_algos_import_conf->p_current_ca_algo);
		free(p_ca_algos_import_conf);
		p_mgr->ca_algo_import_conf = NULL;
	}
}

static void cc_mark_resend_algo_config_set(osm_port_t *p_port)
{
	int i;

	for (i = 0; i < MAX_PPCC_ALGO_SLOTS; i++)
		p_port->p_cc_data->resend_algo_config_set[i] = TRUE;
}

static void cc_mark_resend_param_config(osm_port_t *p_port)
{
	int i;

	for (i = 0; i < MAX_PPCC_ALGO_SLOTS; i++)
		p_port->p_cc_data->resend_param_config[i] = TRUE;
}

/* Clear CA NP data and mark for resend */
static void cc_clear_ca_subn_np_params(osm_port_t *p_port)
{
	p_port->p_cc_data->resend_mask |= OSM_CC_RESEND_CA_NP_PARAMETERS;
	memset(&p_port->p_cc_data->subn_mads.hca_np_params, 0,
	       sizeof(p_port->p_cc_data->subn_mads.hca_np_params));
}

/* Clear CA RP data and mark for resend */
static void cc_clear_ca_subn_rp_params(osm_port_t *p_port)
{
	p_port->p_cc_data->resend_mask |= OSM_CC_RESEND_CA_RP_PARAMETERS;
	memset(&p_port->p_cc_data->subn_mads.hca_rp_params, 0,
	       sizeof(p_port->p_cc_data->subn_mads.hca_rp_params));
}

/* Clear PPCC CA algo data and mark for resend */
static void cc_clear_ca_subn_algo_config_set(osm_port_t *p_port)
{
	cc_mark_resend_algo_config_set(p_port);
	memset(&p_port->p_cc_data->subn_mads.algo_config, 0,
	       sizeof(p_port->p_cc_data->subn_mads.algo_config));
}

/* Clear PPCC CA algo data and mark for resend */
static void cc_clear_ca_subn_algo_params_config(osm_port_t *p_port)
{
	p_port->p_cc_data->resend_mask |= OSM_CC_RESEND_PPCC_CA_ALGO_CONFIG_PARAMS;
	cc_mark_resend_param_config(p_port);
	memset(&p_port->p_cc_data->subn_mads.config_param, 0,
	       sizeof(p_port->p_cc_data->subn_mads.config_param));
}

static void cc_clear_ca_subn_data(osm_port_t *p_port)
{
	/*
	 * Mark CA for re-query capabilities and resend NP and NR
	 * configurations.
	 */
	p_port->p_cc_data->resend_mask |= OSM_CC_RESEND_ALL;
	p_port->p_cc_data->resend_mask |= OSM_CC_RESEND_CC_ENH_INFO;
	p_port->p_cc_data->resend_mask |= OSM_CC_HCA_GENERAL_SETTINGS;
	p_port->p_cc_data->resend_mask |= OSM_CC_RESEND_CA_NP_PARAMETERS;
	p_port->p_cc_data->resend_mask |= OSM_CC_RESEND_CA_RP_PARAMETERS;
	p_port->p_cc_data->resend_mask |= OSM_CC_RESEND_PPCC_CA_ALGO_CONFIG_GET;
	cc_mark_resend_algo_config_set(p_port);
	cc_mark_resend_param_config(p_port);

	/* Clear data */
	memset(&p_port->p_cc_data->subn_mads, 0, sizeof(p_port->p_cc_data->subn_mads));
}

static void cc_clear_sw_subn_ports_data(osm_switch_t *p_switch)
{
	uint8_t port;
	osm_node_t *p_node = p_switch->p_node;

	/* Mark switch to resend SL mapping configurations */
	p_switch->cc_data.resend_mask |= OSM_CC_RESEND_SW_SL_MAPPING;

	/* Clear data from ports */
	for (port = 0; port < p_switch->num_ports; port++) {
		osm_physp_t *p_physp = osm_node_get_physp_ptr(p_node, port);
		memset(&p_physp->cc_data.subn_mads, 0,
		       sizeof(p_physp->cc_data.subn_mads));
	}
}

static void cc_clear_sw_calc_data(osm_switch_t *p_switch)
{
	uint8_t port;
	osm_node_t *p_node = p_switch->p_node;

	memset(&p_switch->cc_data.calc_set_sw_general_settings, 0,
	       sizeof(p_switch->cc_data.calc_set_sw_general_settings));

	/* Clear data from ports */
	for (port = 0; port < p_switch->num_ports; port++) {
		osm_physp_t *p_physp = osm_node_get_physp_ptr(p_node, port);
		if (!p_physp)
			continue;

		memset(&p_physp->cc_data.calc_mads, 0,
		       sizeof(p_physp->cc_data.calc_mads));
		p_physp->cc_data.vl_calculated = FALSE;
		p_physp->cc_data.sl_calculated = FALSE;
	}
}

static void cc_clear_calc_switches(struct osm_opensm *p_osm)
{
	cl_map_item_t *p_item;

	for (p_item = cl_qmap_head(&p_osm->subn.sw_guid_tbl);
	     p_item != cl_qmap_end(&p_osm->subn.sw_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		osm_switch_t *p_switch = (osm_switch_t *) p_item;

		cc_clear_sw_calc_data(p_switch);
		p_switch->cc_data.calculated = FALSE;
	}
}

static void cc_clear_calc_hcas(struct osm_opensm *p_osm)
{
	cl_map_item_t *p_item;

	for (p_item = cl_qmap_head(&p_osm->subn.port_guid_tbl);
	     p_item != cl_qmap_end(&p_osm->subn.port_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		osm_port_t *p_port = (osm_port_t *) p_item;
		osm_node_t *p_node = p_port->p_node;

		if (osm_node_get_type(p_node) != IB_NODE_TYPE_CA)
			continue;

		memset(&p_port->p_cc_data->calc_mads, 0,
		       sizeof(p_port->p_cc_data->calc_mads));
		p_port->p_cc_data->calculated = FALSE;
	}
}

static void cc_clear_calc_data(struct osm_opensm *p_osm)
{
	cc_clear_calc_switches(p_osm);
	cc_clear_calc_hcas(p_osm);
}

static void cc_clear_sw_subn_data(osm_switch_t *p_switch)
{
	/* Mark switch for re-query capabilities and resend configurations */
	p_switch->cc_data.resend_mask |= OSM_CC_RESEND_ALL;
	p_switch->cc_data.resend_mask |= OSM_CC_RESEND_CC_ENH_INFO;
	p_switch->cc_data.resend_mask |= OSM_CC_RESEND_GET_SW_GENERAL_SETTINGS;
	p_switch->cc_data.resend_mask |= OSM_CC_RESEND_SET_SW_GENERAL_SETTINGS;
	p_switch->cc_data.resend_mask |= OSM_CC_RESEND_SW_SL_MAPPING;

	cc_clear_sw_subn_ports_data(p_switch);
}

static boolean_t cc_check_ca_specific_mad_need_update(const osm_port_t *p_port, uint16_t mask)
{
	return p_port->p_cc_data->resend_mask & mask;
}

/* Check if CA is marked for full update */
static boolean_t cc_check_ca_need_update(const osm_port_t *p_port, uint16_t mask)
{
	uint16_t cc_key_bits =
	    p_port->p_cc_data->resend_mask & OSM_CC_KEY_UPDATE_BITS;
	return p_port->p_cc_data->resend_mask & (OSM_CC_RESEND_ALL | mask | cc_key_bits);
}

static boolean_t cc_check_sw_specific_mad_need_update(const osm_switch_t *p_switch, uint16_t mask)
{
	return p_switch->cc_data.resend_mask & mask;
}

/* Check if switch is marked for full update */
static boolean_t cc_check_sw_need_update(const osm_switch_t *p_switch, uint16_t mask)
{
	return p_switch->cc_data.resend_mask & (OSM_CC_RESEND_ALL | mask);
}

static boolean_t cc_check_ca_ppcc_need_update(const osm_port_t *p_port,
					      osm_cc_update_bits_t update_bits, uint8_t algo_slot)
{
	boolean_t need_update = FALSE;

	switch(update_bits) {
	case OSM_CC_RESEND_PPCC_CA_ALGO_CONFIG_SET:
		need_update = p_port->p_cc_data->resend_algo_config_set[algo_slot];
		break;
	case OSM_CC_RESEND_PPCC_CA_ALGO_CONFIG_PARAMS:
		need_update = p_port->p_cc_data->resend_param_config[algo_slot];
		break;
	default:
		break;
	}

	return (need_update || cc_check_ca_need_update(p_port, update_bits));
}

/* Disable CA need update flags */
static inline void cc_disable_ca_need_update(osm_port_t *p_port, uint16_t mask)
{
	uint16_t cc_key_bits =
	    p_port->p_cc_data->resend_mask & OSM_CC_KEY_UPDATE_BITS;
	p_port->p_cc_data->resend_mask &= ~mask;
	p_port->p_cc_data->resend_mask |= cc_key_bits;
}

static inline void cc_disable_ca_ppcc_need_update(osm_port_t *p_port,
						  osm_cc_update_bits_t update_bits, uint8_t algo_slot)
{
	switch(update_bits) {
	case OSM_CC_RESEND_PPCC_CA_ALGO_CONFIG_SET:
		p_port->p_cc_data->resend_algo_config_set[algo_slot] = FALSE;
		break;
	case OSM_CC_RESEND_PPCC_CA_ALGO_CONFIG_PARAMS:
		p_port->p_cc_data->resend_param_config[algo_slot] = FALSE;
		break;
	default:
		break;
	}

	cc_disable_ca_need_update(p_port, update_bits);
}

/* Disable switch need update flags */
static inline void cc_disable_sw_need_update(osm_switch_t *p_switch, uint16_t mask)
{
	p_switch->cc_data.resend_mask &= ~mask;
}

/*
 * Mark devices that needs to be updated and reset information
 * of configuration that is set on the subnet.
 */
static void cc_mark_need_update(struct osm_opensm * p_osm)
{
	cl_map_item_t *p_next;

	p_next = cl_qmap_head(&p_osm->subn.port_guid_tbl);
	while (p_next != cl_qmap_end(&p_osm->subn.port_guid_tbl)) {
		osm_port_t *p_port = (osm_port_t *)p_next;
		osm_node_t *p_node =  p_port->p_node;
		osm_switch_t *p_switch = p_node->sw;
		p_next = cl_qmap_next(p_next);

		/* Check if device is marked as needs update */
		if (!p_osm->subn.need_update &&
		    !osm_port_get_physp(p_port)->need_update &&
		    !(p_switch && p_switch->need_update)) {
			continue;
		}

		if (p_switch) {
			cc_clear_sw_subn_data(p_switch);
		} else if (osm_node_get_type(p_port->p_node) == IB_NODE_TYPE_CA) {
			cc_clear_ca_subn_data(p_port);
		}
	}
}

osm_cc_group_type_enum osm_cc_get_group_type(IN const char *str)
{
	if (!str || !strcasecmp(str, "null"))
		return OSM_CC_GROUP_NOT_DEFINED;
	else if (!strcasecmp(str, "ALL_SWITCHES"))
		return OSM_CC_GROUP_ALL_SWITCHES;
	else if (!strcasecmp(str, "ALL_CAS"))
		return OSM_CC_GROUP_ALL_CAS;
	else if (!strcasecmp(str, "SWITCH_PORTS"))
		return OSM_CC_GROUP_SWITCH_PORTS;
	else
		return OSM_CC_GROUP_PORT_GROUP;
}

static
void print_profiles_conf(struct osm_opensm * p_osm, osm_cc_profiles_conf_t *p_profiles_conf)
{
	int i;
	osm_cc_profile_t *p_profile;

	for(i = 0; i < 3; i++) {
		p_profile = &p_profiles_conf->profiles[i];
		OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
			"profile_name:%s (mode:%u), "
			"index:%u percent:%u, min:%lu, max:%lu\n",
			p_profiles_conf->name, p_profiles_conf->mode, i,
			p_profile->percent, p_profile->min, p_profile->max);
	}
}

static
void print_profiles_map(struct osm_opensm * p_osm, cl_fmap_t *p_profiles)
{
	cl_fmap_item_t *p_item;

	for (p_item = cl_fmap_head(p_profiles);
	     p_item != cl_fmap_end(p_profiles);
	     p_item = cl_fmap_next(p_item)) {
		print_profiles_conf(p_osm, (osm_cc_profiles_conf_t *)p_item);
	}
}

static
void print_sl_conf(struct osm_opensm * p_osm, cl_qlist_t *p_sl_conf_list)
{
	cl_list_item_t *p_item;
	osm_cc_sl_conf_t *p_sl_conf;

	p_item = cl_qlist_head(p_sl_conf_list);
	while (p_item != cl_qlist_end(p_sl_conf_list)) {
		p_sl_conf = (osm_cc_sl_conf_t *)p_item;
		OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
			"CC group %d SL to profile: "
			"%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d\n",
			p_sl_conf->group_type,
			p_sl_conf->sl2profile_num[0], p_sl_conf->sl2profile_num[1],
			p_sl_conf->sl2profile_num[2], p_sl_conf->sl2profile_num[3],
			p_sl_conf->sl2profile_num[4], p_sl_conf->sl2profile_num[5],
			p_sl_conf->sl2profile_num[6], p_sl_conf->sl2profile_num[7],
			p_sl_conf->sl2profile_num[8], p_sl_conf->sl2profile_num[9],
			p_sl_conf->sl2profile_num[10], p_sl_conf->sl2profile_num[11],
			p_sl_conf->sl2profile_num[12], p_sl_conf->sl2profile_num[13],
			p_sl_conf->sl2profile_num[14], p_sl_conf->sl2profile_num[15]);
		p_item = cl_qlist_next(p_item);
	}
}

static
void print_vl_conf(struct osm_opensm * p_osm, cl_qlist_t *p_vl_conf_list)
{
	cl_list_item_t *p_item;
	osm_cc_vl_conf_t *p_vl_conf;
	osm_cc_profiles_conf_t **p_vl_to_profiles;

	p_item = cl_qlist_head(p_vl_conf_list);
	while (p_item != cl_qlist_end(p_vl_conf_list)) {
		p_vl_conf = (osm_cc_vl_conf_t *)p_item;
		p_vl_to_profiles = p_vl_conf->vl_to_profiles;
		OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
			"CC group %d VL to profile: "
			"%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s\n",
			p_vl_conf->group_type,
			p_vl_to_profiles[0]->name, p_vl_to_profiles[1]->name,
			p_vl_to_profiles[2]->name, p_vl_to_profiles[3]->name,
			p_vl_to_profiles[4]->name, p_vl_to_profiles[5]->name,
			p_vl_to_profiles[6]->name, p_vl_to_profiles[7]->name,
			p_vl_to_profiles[8]->name, p_vl_to_profiles[9]->name,
			p_vl_to_profiles[10]->name, p_vl_to_profiles[11]->name,
			p_vl_to_profiles[12]->name, p_vl_to_profiles[13]->name,
			p_vl_to_profiles[14]->name);
		p_item = cl_qlist_next(p_item);
	}
}

static
void print_ca_algo_conf(struct osm_opensm * p_osm, cl_fmap_t *p_ca_algo_conf_map)
{
	cl_fmap_item_t *p_item;
	osm_cc_ca_algo_conf_t *p_ca_algo_conf;
	int i;

	for (p_item = cl_fmap_head(p_ca_algo_conf_map);
	     p_item != cl_fmap_end(p_ca_algo_conf_map);
	     p_item = cl_fmap_next(p_item)) {
		p_ca_algo_conf = (osm_cc_ca_algo_conf_t *)p_item;
		OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
			"CC CA algorithm name: %s, file_name: %s, version: %u.%u.%u\n",
			p_ca_algo_conf->name, p_ca_algo_conf->file_name,
			cl_ntoh16(p_ca_algo_conf->algo_version.algo_id),
				  p_ca_algo_conf->algo_version.algo_major_version,
			p_ca_algo_conf->algo_version.algo_minor_version);

		for (i = 0; i < p_ca_algo_conf->param_num; i++) {
			osm_cc_algo_param_t *p_algo_param = &p_ca_algo_conf->algo_param_list[i];
			OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
				"Parameter num: %u, name: %s, value: %u, min_value: %u, max_value: %u\n",
				i, p_algo_param->param_name, p_algo_param->value,
				p_algo_param->min_value, p_algo_param->max_value);
		}
	}
}

static
void print_ca_algo_import_conf(struct osm_opensm * p_osm,
			       osm_ppcc_ca_algos_import_conf_t *p_ca_algos_import_conf)
{
	osm_ppcc_dictionary_entry_t *entry;
	osm_ppcc_ca_algo_t *p_algo;
	cl_map_item_t *p_map_item;
	int i;

	if (!p_ca_algos_import_conf)
		return;

	OSM_LOG(&p_osm->log, OSM_LOG_DEBUG, "algo_num: %d\n",
		cl_qmap_count(&p_ca_algos_import_conf->ca_algos));
	for (p_map_item = cl_qmap_head(&p_ca_algos_import_conf->ca_algos);
	     p_map_item != cl_qmap_end(&p_ca_algos_import_conf->ca_algos);
	     p_map_item = cl_qmap_next(p_map_item)) {
		p_algo = (osm_ppcc_ca_algo_t *) p_map_item;
		OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
			"algo_id: %d, file_name: %s\n",
			p_algo->algo_id, p_algo->file_name);
		for (i = 0; i < PPCC_ALGO_MAX_PARAMS; i++) {
			entry = p_algo->param_list[i];
			if (!entry)
				break;

			OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
				"param_name: %s, value: %d\n",
				entry->str, entry->value);
		}
	}
}

static
void print_ca_conf(struct osm_opensm * p_osm, cl_qlist_t *p_ca_conf_list)
{
	cl_list_item_t *p_item;
	osm_cc_ca_conf_t *p_ca_conf;
	osm_ppcc_ca_conf_t *p_ppcc_ca_conf;
	osm_cc_hca_np_params_conf_t *p_hca_np_params;

	for (p_item = cl_qlist_head(p_ca_conf_list);
	     p_item != cl_qlist_end(p_ca_conf_list);
	     p_item = cl_qlist_next(p_item)) {
		p_ca_conf = (osm_cc_ca_conf_t *)p_item;

		if (p_ca_conf->cc_type == OSM_CC_V0_CC)
			continue;

		p_ppcc_ca_conf = &p_ca_conf->ppcc_ca_conf;
		p_hca_np_params = &p_ca_conf->hca_np_params;

		OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
			"group_type: %d, enable_notify: %d, enable_react: %d, "
			"min_time_between_cnps: %d, "
			"cnp_sl_mode: %d, cnp_sl: %d, "
			"rtt_sl_mode: %d, rtt_resp_sl: %d\n",
			p_ca_conf->group_type,
			p_ca_conf->enable_notify, p_ca_conf->enable_react,
			p_hca_np_params->min_time_between_cnps,
			p_hca_np_params->cnp_sl_mode,
			p_hca_np_params->cnp_sl,
			p_hca_np_params->rtt_sl_mode,
			p_hca_np_params->rtt_resp_sl);

		OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
			"algo_sl_map: %d, %d, %d, %d, %d, %d, %d, %d, "
			"%d, %d, %d, %d, %d, %d, %d, %d\n",
			p_ppcc_ca_conf->algo_sl_map[0], p_ppcc_ca_conf->algo_sl_map[1],
			p_ppcc_ca_conf->algo_sl_map[2], p_ppcc_ca_conf->algo_sl_map[3],
			p_ppcc_ca_conf->algo_sl_map[4], p_ppcc_ca_conf->algo_sl_map[5],
			p_ppcc_ca_conf->algo_sl_map[6], p_ppcc_ca_conf->algo_sl_map[7],
			p_ppcc_ca_conf->algo_sl_map[8], p_ppcc_ca_conf->algo_sl_map[9],
			p_ppcc_ca_conf->algo_sl_map[10], p_ppcc_ca_conf->algo_sl_map[11],
			p_ppcc_ca_conf->algo_sl_map[12], p_ppcc_ca_conf->algo_sl_map[13],
			p_ppcc_ca_conf->algo_sl_map[14], p_ppcc_ca_conf->algo_sl_map[15]);
	}
}

static void cc_mad_post(osm_congestion_control_t *p_cc, osm_madw_t *p_madw,
			osm_physp_t *p_physp, osm_port_t *p_port,
			ib_net16_t attr_id, ib_net32_t attr_mod,
			uint8_t method_id)
{
	ib_cc_mad_t *p_cc_mad;
	uint8_t port;
	ib_net16_t dest_lid;
	osm_node_t *p_node = p_port->p_node;

	OSM_LOG_ENTER(p_cc->log);

	port = osm_physp_get_port_num(p_physp);

	p_cc_mad = osm_madw_get_cc_mad_ptr(p_madw);

	p_cc_mad->header.base_ver = 1;
	p_cc_mad->header.mgmt_class = IB_MCLASS_CC;
	p_cc_mad->header.class_ver = 2;
	p_cc_mad->header.method = method_id;
	p_cc_mad->header.status = 0;
	p_cc_mad->header.class_spec = 0;
	p_cc_mad->header.trans_id =
		cl_hton64((uint64_t) cl_atomic_inc(&p_cc->trans_id) &
			  (uint64_t) (0xFFFFFFFF));
	if (p_cc_mad->header.trans_id == 0)
		p_cc_mad->header.trans_id =
			cl_hton64((uint64_t) cl_atomic_inc(&p_cc->trans_id) &
				  (uint64_t) (0xFFFFFFFF));
	p_cc_mad->header.attr_id = attr_id;
	p_cc_mad->header.additional_status = 0;
	p_cc_mad->header.attr_mod = attr_mod;

	dest_lid = osm_node_get_base_lid(p_node, port);
	if (p_cc->subn->opt.cc_key_enable)
		p_cc_mad->cc_key = p_port->cc_key;

	memset(p_cc_mad->log_data, '\0', IB_CC_LOG_DATA_SIZE);

	p_madw->mad_addr.dest_lid = dest_lid;
	p_madw->mad_addr.addr_type.gsi.remote_qp = IB_QP1;
	p_madw->mad_addr.addr_type.gsi.remote_qkey = IB_QP1_WELL_KNOWN_Q_KEY;
	p_madw->resp_expected = TRUE;
	p_madw->fail_msg = CL_DISP_MSGID_NONE;

	p_madw->context.cc_context.node_guid = osm_node_get_node_guid(p_node);
	p_madw->context.cc_context.port_guid = osm_physp_get_port_guid(p_physp);
	p_madw->context.cc_context.port = port;
	p_madw->context.cc_context.mad_method = method_id;
	p_madw->context.cc_context.attr_mod = attr_mod;

	p_madw->dest_guid = p_madw->context.cc_context.node_guid;

	cl_spinlock_acquire(&p_cc->mad_queue_lock);
	cl_atomic_inc(&p_cc->outstanding_mads);
	cl_qlist_insert_tail(&p_cc->mad_queue, &p_madw->list_item);
	cl_spinlock_release(&p_cc->mad_queue_lock);

	cl_event_signal(&p_cc->cc_poller_wakeup);

	OSM_LOG_EXIT(p_cc->log);
}


static ib_api_status_t
cc_send_hca_general_settings(osm_sm_t * p_sm,
			     osm_node_t *p_node,
			     osm_port_t *p_port)
{
	osm_congestion_control_t *p_cc = &p_sm->p_subn->p_osm->cc;
	osm_madw_t *p_madw = NULL;
	ib_cc_mad_t *p_cc_mad = NULL;
	ib_cc_hca_general_settings_t *p_ca_general_setting = NULL;
	osm_physp_t *p_physp = p_port->p_physp;

	OSM_LOG_ENTER(p_sm->p_log);

	if (!cc_check_ca_need_update(p_port, OSM_CC_HCA_GENERAL_SETTINGS) &&
	    !memcmp(&p_port->p_cc_data->calc_mads.hca_general_settings,
		    &p_port->p_cc_data->subn_mads.hca_general_settings,
		    sizeof(p_port->p_cc_data->calc_mads.hca_general_settings)))
		return IB_SUCCESS;

	p_madw = osm_mad_pool_get(p_cc->mad_pool, p_cc->bind_handle[p_sm->p_subn->sm_port_idx],
				  MAD_BLOCK_SIZE, NULL);
	if (p_madw == NULL) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR C110: "
			"failed to allocate mad\n");
		return IB_INSUFFICIENT_MEMORY;
	}

	p_cc_mad = osm_madw_get_cc_mad_ptr(p_madw);

	p_ca_general_setting = ib_cc_mad_get_mgt_data_ptr(p_cc_mad);

	memcpy(p_ca_general_setting,
	       &p_port->p_cc_data->calc_mads.hca_general_settings,
	       sizeof(p_port->p_cc_data->calc_mads.hca_general_settings));

	cc_mad_post(p_cc, p_madw, p_physp, p_port,
		    IB_MAD_ATTR_CC_HCA_GENERAL_SETTINGS, 0, IB_MAD_METHOD_SET);

	OSM_LOG_EXIT(p_sm->p_log);

	return IB_SUCCESS;
}

static ib_api_status_t cc_send_hca_rp_params(osm_sm_t * p_sm,
					     osm_node_t *p_node,
					     osm_port_t *p_port)
{
	osm_congestion_control_t *p_cc = &p_sm->p_subn->p_osm->cc;
	osm_madw_t *p_madw = NULL;
	ib_cc_mad_t *p_cc_mad = NULL;
	ib_cc_hca_rp_parameters_t *p_hca_rp_params = NULL;
	osm_physp_t *p_physp = p_port->p_physp;

	OSM_LOG_ENTER(p_sm->p_log);

	if (!cc_check_ca_need_update(p_port, OSM_CC_RESEND_CA_RP_PARAMETERS) &&
	    !memcmp(&p_port->p_cc_data->calc_mads.hca_rp_params,
		    &p_port->p_cc_data->subn_mads.hca_rp_params,
		    sizeof(p_port->p_cc_data->subn_mads.hca_rp_params)))
		return IB_SUCCESS;

	p_madw = osm_mad_pool_get(p_cc->mad_pool, p_cc->bind_handle[p_sm->p_subn->sm_port_idx],
				  MAD_BLOCK_SIZE, NULL);
	if (p_madw == NULL) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR C110: "
			"failed to allocate mad\n");
		return IB_INSUFFICIENT_MEMORY;
	}

	p_cc_mad = osm_madw_get_cc_mad_ptr(p_madw);
	p_hca_rp_params = ib_cc_mad_get_mgt_data_ptr(p_cc_mad);

	memcpy(p_hca_rp_params,
	       &p_port->p_cc_data->calc_mads.hca_rp_params,
	       sizeof(p_port->p_cc_data->calc_mads.hca_rp_params));

	cc_mad_post(p_cc, p_madw, p_physp, p_port,
		    IB_MAD_ATTR_CC_HCA_RP_PARAMETERS, 0, IB_MAD_METHOD_SET);

	OSM_LOG_EXIT(p_sm->p_log);

	return IB_SUCCESS;
}

static ib_api_status_t cc_send_hca_np_params(osm_sm_t * p_sm,
					     osm_node_t *p_node,
					     osm_port_t *p_port)
{
	osm_congestion_control_t *p_cc = &p_sm->p_subn->p_osm->cc;
	osm_madw_t *p_madw = NULL;
	ib_cc_mad_t *p_cc_mad = NULL;
	ib_cc_hca_np_parameters_t *p_hca_np_params = NULL;
	osm_physp_t *p_physp = p_port->p_physp;

	OSM_LOG_ENTER(p_sm->p_log);

	if (!cc_check_ca_need_update(p_port, OSM_CC_RESEND_CA_NP_PARAMETERS) &&
	    !memcmp(&p_port->p_cc_data->calc_mads.hca_np_params,
		    &p_port->p_cc_data->subn_mads.hca_np_params,
		    sizeof(p_port->p_cc_data->subn_mads.hca_np_params)))
		return IB_SUCCESS;

	p_madw = osm_mad_pool_get(p_cc->mad_pool, p_cc->bind_handle[p_sm->p_subn->sm_port_idx],
				  MAD_BLOCK_SIZE, NULL);
	if (p_madw == NULL) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR C110: "
			"failed to allocate mad\n");
		return IB_INSUFFICIENT_MEMORY;
	}

	p_cc_mad = osm_madw_get_cc_mad_ptr(p_madw);
	p_hca_np_params = ib_cc_mad_get_mgt_data_ptr(p_cc_mad);

	memcpy(p_hca_np_params,
	       &p_port->p_cc_data->calc_mads.hca_np_params,
	       sizeof(p_port->p_cc_data->calc_mads.hca_np_params));

	cc_mad_post(p_cc, p_madw, p_physp, p_port,
		    IB_MAD_ATTR_CC_HCA_NP_PARAMETERS, 0, IB_MAD_METHOD_SET);

	OSM_LOG_EXIT(p_sm->p_log);

	return IB_SUCCESS;
}

static inline boolean_t cc_is_port_configurable(osm_sm_t * p_sm,
						osm_port_t * p_port)
{
	return (p_port->disc_through == OSM_DISC_NORMAL &&
		p_port->is_reachable &&
		(!lookup_isolated_port(p_sm, p_port->p_node, p_port->p_physp->port_num) ||
		 osm_node_get_type(p_port->p_node) == IB_NODE_TYPE_SWITCH) &&
		(!p_sm->p_subn->opt.cc_key_enable ||
		 p_port->p_cc_data->key_configured));
}

static inline boolean_t cc_memcmp_sw_set_general_settings_calc_subn(IN const osm_switch_t * p_switch)
{
	const ib_cc_sw_general_settings_t *p_calc_set_sw_general_settings;
	const ib_cc_sw_general_settings_t *p_subn_set_sw_general_settings;

	p_calc_set_sw_general_settings = &p_switch->cc_data.calc_set_sw_general_settings;
	p_subn_set_sw_general_settings = &p_switch->cc_data.subn_set_sw_general_settings;

	return ((p_calc_set_sw_general_settings->enable != p_subn_set_sw_general_settings->enable) ||
		(p_calc_set_sw_general_settings->aqs_weight != p_subn_set_sw_general_settings->aqs_weight) ||
		(p_calc_set_sw_general_settings->aqs_time != p_subn_set_sw_general_settings->aqs_time));
}

static ib_api_status_t cc_send_get_sw_general_settings(osm_sm_t * p_sm,
						       osm_switch_t *p_switch)
{
	osm_congestion_control_t *p_cc = &p_sm->p_subn->p_osm->cc;
	osm_physp_t *p_physp;
	osm_madw_t *p_madw = NULL;
	osm_node_t *p_node = p_switch->p_node;
	osm_port_t *p_port = NULL;

	OSM_LOG_ENTER(p_sm->p_log);

	p_physp = osm_node_get_physp_ptr(p_node, 0);
	p_port = osm_get_port_by_guid(p_sm->p_subn, p_physp->port_guid);
	if (!p_port) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR C12E: "
			"Port GUID 0x%" PRIx64 " not in table\n",
			cl_ntoh64(p_physp->port_guid));
		return IB_INVALID_PARAMETER;
	}

	if (!cc_is_port_configurable(p_sm, p_port))
		return IB_SUCCESS;

	if (!cc_check_sw_need_update(p_switch, OSM_CC_RESEND_GET_SW_GENERAL_SETTINGS))
		return IB_SUCCESS;

	p_madw = osm_mad_pool_get(p_cc->mad_pool, p_cc->bind_handle[p_sm->p_subn->sm_port_idx],
				  MAD_BLOCK_SIZE, NULL);
	if (p_madw == NULL) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR C12D: failed to allocate mad\n");
		return IB_INSUFFICIENT_MEMORY;
	}

	cc_mad_post(p_cc, p_madw, p_physp, p_port,
		    IB_MAD_ATTR_CC_SW_GENERAL_SETTINGS, 0, IB_MAD_METHOD_GET);

	OSM_LOG_EXIT(p_sm->p_log);

	return IB_SUCCESS;
}

static ib_api_status_t cc_send_cc_enh_info(osm_sm_t * p_sm,
					   osm_port_t *p_port)
{
	osm_congestion_control_t *p_cc = &p_sm->p_subn->p_osm->cc;
	osm_madw_t *p_madw = NULL;
	osm_node_t *p_node = p_port->p_node;
	osm_physp_t *p_physp;

	OSM_LOG_ENTER(p_sm->p_log);

	if (!cc_is_port_configurable(p_sm, p_port))
		return IB_SUCCESS;

	if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH) {
		if (!cc_check_sw_need_update(p_node->sw, OSM_CC_RESEND_CC_ENH_INFO))
			return IB_SUCCESS;
		p_physp = osm_node_get_physp_ptr(p_node, 0);
	} else {
		if (!cc_check_ca_need_update(p_port, OSM_CC_RESEND_CC_ENH_INFO))
			return IB_SUCCESS;
		p_physp = p_port->p_physp;
	}

	p_madw = osm_mad_pool_get(p_cc->mad_pool, p_cc->bind_handle[p_sm->p_subn->sm_port_idx],
				  MAD_BLOCK_SIZE, NULL);
	if (p_madw == NULL) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR C110: "
			"failed to allocate mad\n");
		return IB_INSUFFICIENT_MEMORY;
	}

	cc_mad_post(p_cc, p_madw, p_physp, p_port,
		    IB_MAD_ATTR_CC_ENH_INFO, 0, IB_MAD_METHOD_GET);

	OSM_LOG_EXIT(p_sm->p_log);

	return IB_SUCCESS;
}

static ib_api_status_t cc_send_ppcc_ca_algo_config(osm_sm_t *p_sm,
						   osm_port_t *p_port,
						   uint8_t algo_slot,
						   uint8_t method_id)
{
	osm_congestion_control_t *p_cc = &p_sm->p_subn->p_osm->cc;
	ib_ppcc_hca_algo_config_t *p_ca_ppcc_ca_algo_config = NULL;
	osm_madw_t *p_madw = NULL;
	ib_cc_mad_t *p_cc_mad = NULL;
	ib_net32_t attr_mod;

	OSM_LOG_ENTER(p_sm->p_log);

	if (method_id == IB_MAD_METHOD_GET) {
		if (!cc_check_ca_need_update(p_port, OSM_CC_RESEND_PPCC_CA_ALGO_CONFIG_GET))
			return IB_SUCCESS;
	} else if (method_id == IB_MAD_METHOD_SET) {
		if (!cc_check_ca_ppcc_need_update(p_port, OSM_CC_RESEND_PPCC_CA_ALGO_CONFIG_SET, algo_slot) &&
		    !memcmp(&p_port->p_cc_data->calc_mads.algo_config[algo_slot],
			    &p_port->p_cc_data->subn_mads.algo_config[algo_slot],
			    sizeof(p_port->p_cc_data->calc_mads.algo_config[algo_slot])))
			return IB_SUCCESS;
	}

	attr_mod = cl_hton32(((uint32_t) algo_slot) << 24);
	p_madw = osm_mad_pool_get(p_cc->mad_pool, p_cc->bind_handle[p_sm->p_subn->sm_port_idx],
				  MAD_BLOCK_SIZE, NULL);
	if (p_madw == NULL) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR C110: "
			"failed to allocate mad\n");
		return IB_INSUFFICIENT_MEMORY;
	}

	p_cc_mad = osm_madw_get_cc_mad_ptr(p_madw);
	p_ca_ppcc_ca_algo_config = ib_cc_mad_get_mgt_data_ptr(p_cc_mad);
	if (method_id == IB_MAD_METHOD_SET)
		memcpy(p_ca_ppcc_ca_algo_config, &p_port->p_cc_data->calc_mads.algo_config[algo_slot],
		       sizeof(p_port->p_cc_data->calc_mads.algo_config[algo_slot]));
	else
		p_ca_ppcc_ca_algo_config->encap_type = 2;


	cc_mad_post(p_cc, p_madw, p_port->p_physp, p_port,
		    IB_MAD_ATTR_PPCC_HCA_ALGO_CONFIG, attr_mod, method_id);

	OSM_LOG_EXIT(p_sm->p_log);

	return IB_SUCCESS;
}

static ib_api_status_t cc_send_ppcc_ca_algo_disable(osm_sm_t *p_sm, osm_port_t *p_port,
						    uint8_t algo_slot)
{
	osm_congestion_control_t *p_cc = &p_sm->p_subn->p_osm->cc;
	ib_ppcc_hca_algo_config_t *p_ca_ppcc_ca_algo_config = NULL;
	osm_madw_t *p_madw = NULL;
	ib_cc_mad_t *p_cc_mad = NULL;
	ib_net32_t attr_mod;

	OSM_LOG_ENTER(p_sm->p_log);

	attr_mod = cl_hton32(((uint32_t) algo_slot) << 24);
	p_madw = osm_mad_pool_get(p_cc->mad_pool, p_cc->bind_handle[p_sm->p_subn->sm_port_idx],
				  MAD_BLOCK_SIZE, NULL);
	if (p_madw == NULL) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR C110: "
			"failed to allocate mad\n");
		return IB_INSUFFICIENT_MEMORY;
	}

	p_cc_mad = osm_madw_get_cc_mad_ptr(p_madw);
	p_ca_ppcc_ca_algo_config = ib_cc_mad_get_mgt_data_ptr(p_cc_mad);
	memset(p_ca_ppcc_ca_algo_config, 0, sizeof(ib_ppcc_hca_algo_config_t));

	cc_mad_post(p_cc, p_madw, p_port->p_physp, p_port,
		    IB_MAD_ATTR_PPCC_HCA_ALGO_CONFIG, attr_mod, IB_MAD_METHOD_SET);

	OSM_LOG_EXIT(p_sm->p_log);

	return IB_SUCCESS;
}

static ib_api_status_t cc_send_ppcc_ca_algo_params(osm_sm_t *p_sm,
						   osm_port_t *p_port,
						   uint8_t algo_slot)
{
	osm_congestion_control_t *p_cc = &p_sm->p_subn->p_osm->cc;
	ib_ppcc_hca_config_param_t *p_ppcc_hca_algo_params_config = NULL;
	osm_madw_t *p_madw = NULL;
	ib_cc_mad_t *p_cc_mad = NULL;
	ib_net32_t attr_mod;

	OSM_LOG_ENTER(p_sm->p_log);

	if (!cc_check_ca_ppcc_need_update(p_port, OSM_CC_RESEND_PPCC_CA_ALGO_CONFIG_PARAMS, algo_slot) &&
	    !memcmp(&p_port->p_cc_data->calc_mads.config_param[algo_slot],
		    &p_port->p_cc_data->subn_mads.config_param[algo_slot],
		    sizeof(p_port->p_cc_data->calc_mads.config_param[algo_slot])))
		return IB_SUCCESS;

	attr_mod = cl_hton32(((uint32_t) algo_slot) << 24);
	p_madw = osm_mad_pool_get(p_cc->mad_pool, p_cc->bind_handle[p_sm->p_subn->sm_port_idx],
				  MAD_BLOCK_SIZE, NULL);
	if (p_madw == NULL) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR C110: "
			"failed to allocate mad\n");
		return IB_INSUFFICIENT_MEMORY;
	}

	p_cc_mad = osm_madw_get_cc_mad_ptr(p_madw);
	p_ppcc_hca_algo_params_config = ib_cc_mad_get_mgt_data_ptr(p_cc_mad);
	memcpy(p_ppcc_hca_algo_params_config, &p_port->p_cc_data->calc_mads.config_param[algo_slot],
	       sizeof(p_port->p_cc_data->calc_mads.config_param[algo_slot]));

	cc_mad_post(p_cc, p_madw, p_port->p_physp, p_port,
		    IB_MAD_ATTR_PPCC_HCA_ALGO_PARAMS_CONFIG, attr_mod, IB_MAD_METHOD_SET);

	OSM_LOG_EXIT(p_sm->p_log);

	return IB_SUCCESS;
}

static ib_api_status_t
cc_send_port_profile_settings(osm_sm_t * p_sm,
			      osm_node_t *p_node,
			      osm_physp_t *p_physp,
			      ib_net32_t attr_mod)
{
	osm_congestion_control_t *p_cc = &p_sm->p_subn->p_osm->cc;
	uint8_t force_update;
	osm_madw_t *p_madw = NULL;
	ib_cc_mad_t *p_cc_mad = NULL;
	osm_physp_t *p_physp0 = NULL;
	ib_cc_port_profile_settings_t *p_port_profile_setting = NULL;
	boolean_t set_all_vls = TRUE;
	uint8_t vl, max_vl_num, max_op_vls;
	osm_port_t *p_port;

	OSM_LOG_ENTER(p_sm->p_log);

	OSM_ASSERT(p_node->sw);

	/*
	 * Avoid sending CongestionPortProfileSettings if
	 * SET CongestionSwitchGeneralSettings isn't configured yet,
	 * because PortProfile is calculated according to GeneralSettings.
	 */
	if (cc_check_sw_specific_mad_need_update(p_node->sw, OSM_CC_RESEND_SET_SW_GENERAL_SETTINGS) ||
	    cc_memcmp_sw_set_general_settings_calc_subn(p_node->sw))
		return IB_SUCCESS;

	max_op_vls = osm_get_max_op_vls_by_node(p_sm->p_subn, p_node);
	if (attr_mod != CC_ATTR_MOD_ALL_PORTS)
		max_vl_num = 1U << (ib_port_info_get_op_vls(&p_physp->port_info) - 1);
	else
		max_vl_num = 1U << (max_op_vls - 1);

	if (max_vl_num == IB_MAX_NUM_VLS)
		max_vl_num = IB_DROP_VL;

	p_physp0 = osm_node_get_physp_ptr(p_node, 0);
	p_port = osm_get_port_by_guid(p_sm->p_subn, p_physp->port_guid);
	if (!p_port) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR C122: "
			"Port GUID 0x%" PRIx64 " not in table\n",
			cl_ntoh64(p_physp->port_guid));
		return IB_INVALID_PARAMETER;
	}

	force_update = cc_check_sw_need_update(p_node->sw, 0);

	/*
	 * HACK: check if can set all VLs profiles using single MAD, by
	 * comparing all VLs profiles to VL0 profile.
	 */
	for (vl = 1; vl < max_vl_num; vl++) {
		if (memcmp(&p_physp->cc_data.calc_mads.vl_to_profiles[0],
			   &p_physp->cc_data.calc_mads.vl_to_profiles[vl],
			   sizeof(p_physp->cc_data.calc_mads.vl_to_profiles[vl]))) {
			set_all_vls = FALSE;
			break;
		}
	}

	for (vl = 0; vl < max_vl_num; vl++) {
		if (!force_update &&
		    !memcmp(&p_physp->cc_data.calc_mads.vl_to_profiles[vl],
			    &p_physp->cc_data.subn_mads.vl_to_profiles[vl],
			    sizeof(p_physp->cc_data.calc_mads.vl_to_profiles[vl])))
			continue;

		p_madw = osm_mad_pool_get(p_cc->mad_pool, p_cc->bind_handle[p_sm->p_subn->sm_port_idx],
				  MAD_BLOCK_SIZE, NULL);
		if (p_madw == NULL) {
			OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR C110: "
				"failed to allocate mad\n");
			return IB_INSUFFICIENT_MEMORY;
		}

		p_cc_mad = osm_madw_get_cc_mad_ptr(p_madw);

		p_port_profile_setting = ib_cc_mad_get_mgt_data_ptr(p_cc_mad);

		memcpy(p_port_profile_setting,
		       &p_physp->cc_data.calc_mads.vl_to_profiles[vl],
		       sizeof(p_physp->cc_data.calc_mads.vl_to_profiles[vl]));

		/* Set VL mask of outgoing MAD */
		if (set_all_vls)
			p_port_profile_setting->vl_mask = cl_hton16((1 << max_vl_num) - 1);
		else
			p_port_profile_setting->vl_mask = cl_hton16(1 << vl);

		cc_mad_post(p_cc, p_madw, p_physp0, p_port,
			    IB_MAD_ATTR_CC_PORT_SW_PROFILE_SETTINGS,
			    attr_mod, IB_MAD_METHOD_SET);

		if (set_all_vls)
			/*
			 * We sent single set request for all vls, no need to
			 * continue
			 */
			break;
	}

	OSM_LOG_EXIT(p_sm->p_log);

	return IB_SUCCESS;
}

static ib_api_status_t
cc_send_sl_mapping_settings(osm_sm_t * p_sm,
			    osm_node_t *p_node,
			    osm_physp_t *p_physp,
			    ib_net32_t attr_mod)
{
	osm_congestion_control_t *p_cc = &p_sm->p_subn->p_osm->cc;
	osm_madw_t *p_madw = NULL;
	ib_cc_mad_t *p_cc_mad = NULL;
	osm_physp_t *p_physp0 = NULL;
	ib_cc_sl_mapping_settings_t *p_cc_sl_setting = NULL;
	osm_port_t *p_port;

	OSM_LOG_ENTER(p_sm->p_log);
	OSM_ASSERT(p_node->sw);

	p_physp0 = osm_node_get_physp_ptr(p_node, 0);
	p_port = osm_get_port_by_guid(p_sm->p_subn, p_physp->port_guid);
	if (!p_port) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR C121: "
			"Port GUID 0x%" PRIx64 " not in table\n",
			cl_ntoh64(p_physp->port_guid));
		return IB_INVALID_PARAMETER;
	}

	if (!cc_check_sw_need_update(p_node->sw, OSM_CC_RESEND_SW_SL_MAPPING) &&
	    !memcmp(&p_physp->cc_data.calc_mads.sl_mapping,
		    &p_physp->cc_data.subn_mads.sl_mapping,
		    sizeof(p_physp->cc_data.calc_mads.sl_mapping)))
		return IB_SUCCESS;

	p_madw = osm_mad_pool_get(p_cc->mad_pool, p_cc->bind_handle[p_sm->p_subn->sm_port_idx],
				  MAD_BLOCK_SIZE, NULL);
	if (p_madw == NULL) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR C110: "
			"failed to allocate mad\n");
		return IB_INSUFFICIENT_MEMORY;
	}

	p_cc_mad = osm_madw_get_cc_mad_ptr(p_madw);

	p_cc_sl_setting = ib_cc_mad_get_mgt_data_ptr(p_cc_mad);

	memcpy(p_cc_sl_setting,
	       &p_physp->cc_data.calc_mads.sl_mapping,
	       sizeof(p_physp->cc_data.calc_mads.sl_mapping));

	cc_mad_post(p_cc, p_madw, p_physp0, p_port,
		    IB_MAD_ATTR_CC_SL_MAPPING_SETTINGS,
		    attr_mod, IB_MAD_METHOD_SET);

	OSM_LOG_EXIT(p_sm->p_log);

	return IB_SUCCESS;
}

static ib_api_status_t cc_send_set_sw_general_settings(osm_sm_t * p_sm,
						       osm_switch_t *p_switch)
{
	osm_congestion_control_t *p_cc = &p_sm->p_subn->p_osm->cc;
	osm_physp_t *p_physp;
	osm_madw_t *p_madw = NULL;
	ib_cc_mad_t *p_cc_mad = NULL;
	ib_cc_sw_general_settings_t *p_set_sw_general_settings = NULL;
	osm_port_t *p_port = NULL;
	osm_node_t *p_node = p_switch->p_node;

	OSM_LOG_ENTER(p_sm->p_log);

	p_physp = osm_node_get_physp_ptr(p_node, 0);
	p_port = osm_get_port_by_guid(p_sm->p_subn, p_physp->port_guid);
	if (!p_port) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR C10C: "
			"Port GUID 0x%" PRIx64 " not in table\n",
			cl_ntoh64(p_physp->port_guid));
		return IB_INVALID_PARAMETER;
	}

	if (!cc_is_port_configurable(p_sm, p_port))
		return IB_SUCCESS;

	/* Send MAD if forced update or non-read only fields were changed */
	if (!cc_check_sw_need_update(p_switch, OSM_CC_RESEND_SET_SW_GENERAL_SETTINGS) &&
	    !cc_memcmp_sw_set_general_settings_calc_subn(p_switch))
		return IB_SUCCESS;

	p_madw = osm_mad_pool_get(p_cc->mad_pool, p_cc->bind_handle[p_sm->p_subn->sm_port_idx],
				  MAD_BLOCK_SIZE, NULL);
	if (p_madw == NULL) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR C101: "
			"failed to allocate mad\n");
		return IB_INSUFFICIENT_MEMORY;
	}

	p_cc_mad = osm_madw_get_cc_mad_ptr(p_madw);

	p_set_sw_general_settings = ib_cc_mad_get_mgt_data_ptr(p_cc_mad);

	memcpy(p_set_sw_general_settings,
	       &p_switch->cc_data.calc_set_sw_general_settings,
	       sizeof(p_switch->cc_data.calc_set_sw_general_settings));

	cc_mad_post(p_cc, p_madw, p_physp, p_port,
		    IB_MAD_ATTR_CC_SW_GENERAL_SETTINGS, 0, IB_MAD_METHOD_SET);

	OSM_LOG_EXIT(p_sm->p_log);

	return IB_SUCCESS;
}

static ib_api_status_t cc_send_to_ext_ports(osm_sm_t * p_sm, osm_node_t *p_node,
					    ib_api_status_t (send)(osm_sm_t *, osm_node_t *,
								   osm_physp_t *, ib_net32_t attr_mod))
{
	int ret = 0;
	osm_physp_t *p_ext_port_physp;
	uint8_t	ext_port_num;
	ib_api_status_t status;
	uint8_t num_ports = osm_node_get_num_physp(p_node);

	for (ext_port_num = 1; ext_port_num < num_ports; ext_port_num++) {
		p_ext_port_physp = osm_node_get_physp_ptr(p_node, ext_port_num);
		if (!p_ext_port_physp->p_remote_physp)
			continue;

		/* skip special ports */
		if (osm_phsyp_get_special_port_type(p_ext_port_physp->p_remote_physp))
			continue;

		status = send(p_sm, p_node, p_ext_port_physp,
			      cl_hton32((uint32_t)ext_port_num));
		if (status != IB_SUCCESS)
			ret = -1;
	}

	return ret;
}

static ib_api_status_t configure_profiles(struct osm_opensm *p_osm)
{
	int ret = 0;
	cl_map_item_t *p_item;
	ib_api_status_t status;
	osm_physp_t *p_physp;
	osm_port_t *p_port = NULL;

	for (p_item = cl_qmap_head(&p_osm->subn.sw_guid_tbl);
	     p_item != cl_qmap_end(&p_osm->subn.sw_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		osm_switch_t *p_switch = (osm_switch_t *) p_item;
		osm_node_t *p_node = p_switch->p_node;

		p_physp = osm_node_get_physp_ptr(p_node, 0);
		p_port = osm_get_port_by_guid(p_osm->sm.p_subn, p_physp->port_guid);
		if (!p_port) {
			OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR C10D: "
				"Port GUID 0x%" PRIx64 " not in table\n",
				cl_ntoh64(p_physp->port_guid));
			continue;
		}

		if (!ib_cc_get_enh_info_ver0(&p_node->cc_enh_info))
			continue;

		/* Skip device when congestion control is disabled on it */
		if (!ib_cc_sw_get_enable_cc(&p_switch->cc_data.calc_set_sw_general_settings)) {
			/* Mark SW for future update */
			cc_clear_sw_subn_ports_data(p_switch);
			continue;
		}

		if (!cc_is_port_configurable(&p_osm->sm, p_port))
			continue;

		/* Skip if switch didn't respond to CongestionSwitchGeneralSettings GET request first */
		if (cc_check_sw_specific_mad_need_update(p_switch, OSM_CC_RESEND_GET_SW_GENERAL_SETTINGS))
			continue;

		if (p_switch->cc_data.is_vl_conf_for_all_ports)
			status = cc_send_port_profile_settings(&p_osm->sm, p_node, p_physp,
							       CC_ATTR_MOD_ALL_PORTS);
		else
			status = cc_send_to_ext_ports(&p_osm->sm, p_node,
						      cc_send_port_profile_settings);

		if (status != IB_SUCCESS)
			ret = -1;

		if (p_switch->cc_data.is_sl_conf_for_all_ports)
			status = cc_send_sl_mapping_settings(&p_osm->sm, p_node, p_physp,
							     CC_ATTR_MOD_ALL_PORTS);
		else
			status = cc_send_to_ext_ports(&p_osm->sm, p_node,
						      cc_send_sl_mapping_settings);

		if (status != IB_SUCCESS)
			ret = -1;
	}

	return ret;
}

static ib_api_status_t configure_switches(struct osm_opensm *p_osm)
{
	int ret = 0;
	cl_map_item_t *p_next;

	p_next = cl_qmap_head(&p_osm->subn.sw_guid_tbl);
	while (p_next != cl_qmap_end(&p_osm->subn.sw_guid_tbl)) {
		osm_switch_t *p_switch = (osm_switch_t *) p_next;
		ib_api_status_t status;

		p_next = cl_qmap_next(p_next);

		if (!ib_cc_get_enh_info_ver0(&p_switch->p_node->cc_enh_info))
			continue;

		/* Skip if switch didn't respond to GET request first */
		if (cc_check_sw_specific_mad_need_update(p_switch, OSM_CC_RESEND_GET_SW_GENERAL_SETTINGS))
			continue;

		status = cc_send_set_sw_general_settings(&p_osm->sm, p_switch);
		if (status != IB_SUCCESS)
			ret = -1;
	}

	return ret;
}

static ib_api_status_t configure_hcas(struct osm_opensm *p_osm)
{
	int ret = 0;
	cl_map_item_t *p_next;
	ib_cc_hca_general_settings_t *p_cc_hca_gs;
	boolean_t ver0_supported, ver1_supported;
	int algo_slot;

	p_next = cl_qmap_head(&p_osm->subn.port_guid_tbl);
	while (p_next != cl_qmap_end(&p_osm->subn.port_guid_tbl)) {
		osm_port_t *p_port = (osm_port_t *) p_next;
		osm_node_t *p_node = p_port->p_node;
		ib_api_status_t status;

		p_next = cl_qmap_next(p_next);

		if (osm_node_get_type(p_node) != IB_NODE_TYPE_CA)
			continue;

		ver0_supported = ib_cc_get_enh_info_ver0(&p_node->cc_enh_info);
		ver1_supported = (ib_cc_get_enh_info_ver1(&p_node->cc_enh_info) &&
				  ib_cc_enh_info_is_cap_supported(&p_node->cc_enh_info,
								  IB_CC_CAPABILITY_MASK_ALGO_CONFIG_SUPPORTED_BIT));
		if (!ver0_supported && !ver1_supported)
			continue;

		if (!cc_is_port_configurable(&p_osm->sm, p_port))
			continue;

		if (cc_check_ca_specific_mad_need_update(p_port, OSM_CC_HCA_GENERAL_SETTINGS) ||
		    memcmp(&p_port->p_cc_data->calc_mads.hca_general_settings,
			   &p_port->p_cc_data->subn_mads.hca_general_settings,
			   sizeof(p_port->p_cc_data->calc_mads.hca_general_settings)))
			continue;

		p_cc_hca_gs = &p_port->p_cc_data->calc_mads.hca_general_settings;

		/* Send RP configuration only in ver0 CC */
		if (!ver1_supported) {
			/* Skip RP configuration if device not configured for reaction */
			if (ib_cc_hca_get_en_react(p_cc_hca_gs)) {
				status = cc_send_hca_rp_params(&p_osm->sm, p_node, p_port);
				if (status != IB_SUCCESS)
					ret = -1;
			} else {
				/* Mark RP for future update */
				cc_clear_ca_subn_rp_params(p_port);
			}
		}

		if (ver1_supported) {
			if (ib_cc_hca_get_en_notify(p_cc_hca_gs) &&
			    ib_cc_hca_get_en_react(p_cc_hca_gs)) {
				for (algo_slot = 0; algo_slot < p_port->p_cc_data->algo_info_array_size; algo_slot++) {
					if (p_port->p_cc_data->subn_algo_info[algo_slot].algo_id == 0 ||
					    !ib_ppcc_get_hca_algo_en(&p_port->p_cc_data->calc_mads.algo_config[algo_slot])) {
						cc_disable_ca_ppcc_need_update(p_port, OSM_CC_RESEND_PPCC_CA_ALGO_CONFIG_PARAMS, algo_slot);
						continue;
					}
 
					status = cc_send_ppcc_ca_algo_params(&p_osm->sm, p_port, algo_slot);
					if (status != IB_SUCCESS)
						return status;
				}
			} else {
				cc_clear_ca_subn_algo_params_config(p_port);
			}
		}
		/* Skip NP configuration if device not configured for notifications */
		if (ib_cc_hca_get_en_notify(p_cc_hca_gs)) {
			status = cc_send_hca_np_params(&p_osm->sm, p_node, p_port);
			if (status != IB_SUCCESS)
				ret = -1;
		} else {
			/* Mark NP for future update */
			cc_clear_ca_subn_np_params(p_port);
		}

	}

	return ret;
}

static inline boolean_t cc_profiles_conf_is_percent_mode(const osm_cc_profiles_conf_t *p_profiles_conf)
{
	return p_profiles_conf->mode == 1;
}

static uint64_t cc_sw_profile_min_legal(osm_switch_t *p_switch)
{
	uint64_t step_size_kb, step_size_bytes;
	uint8_t aqs_weight;
	boolean_t granularity_supported;

	step_size_kb = ib_cc_sw_get_cap_cc_profile_step_size(&p_switch->cc_data.subn_get_sw_general_settings);
	step_size_bytes = step_size_kb * BYTES_PER_KB;
	granularity_supported = ib_cc_enh_info_is_cap_supported(&p_switch->p_node->cc_enh_info,
								IB_CC_CAPABILITY_MASK_GRANULARITY_SUPPORTED_BIT);

	/* Take aqs_weight from calc because the SET wasn't sent yet */
	aqs_weight = ib_cc_sw_get_aqs_weight(&p_switch->cc_data.calc_set_sw_general_settings);

	/* Return KB/Bytes according to granularity support */
	return (1 << aqs_weight) * (granularity_supported ? step_size_kb : step_size_bytes);
}

static uint64_t cc_sw_profile_max_legal(osm_switch_t *p_switch)
{
	ib_cc_sw_general_settings_t *p_get_sw_general_settings;
	uint64_t step_size_kb, total_buffer_size_bytes;
	boolean_t granularity_supported;

	p_get_sw_general_settings = &p_switch->cc_data.subn_get_sw_general_settings;
	step_size_kb = ib_cc_sw_get_cap_cc_profile_step_size(p_get_sw_general_settings);
	total_buffer_size_bytes = cl_ntoh32(ib_cc_sw_get_cap_total_buffer_size(p_get_sw_general_settings));
	granularity_supported = ib_cc_enh_info_is_cap_supported(&p_switch->p_node->cc_enh_info,
								IB_CC_CAPABILITY_MASK_GRANULARITY_SUPPORTED_BIT);

	/* Return KB/Bytes according to granularity support */
	if (granularity_supported) {
		uint64_t total_buffer_size_kb = total_buffer_size_bytes / BYTES_PER_KB;
		return ROUNDDOWN(MIN(total_buffer_size_kb, IB_CC_PORT_PROFILE_SETTINGS_MAXIMAL_MAX_VALUE),
				 step_size_kb);
	} else {
		uint64_t step_size_bytes = step_size_kb * BYTES_PER_KB;
		return ROUNDDOWN(MIN(total_buffer_size_bytes, IB_CC_PORT_PROFILE_SETTINGS_MAXIMAL_MAX_VALUE),
				 step_size_bytes);
	}
}

static void calculate_sw_vl_to_profiles(struct osm_opensm *p_osm,
					ib_cc_port_profile_settings_t *vl_to_profiles,
					osm_cc_vl_conf_t *p_vl_conf,
					osm_switch_t *p_switch)
{
	osm_node_t *p_node = p_switch->p_node;
	uint64_t step_size_bytes, min_cap, max_cap;
	int vl, profile;
	boolean_t granularity_supported;

	step_size_bytes = ib_cc_sw_get_cap_cc_profile_step_size(&p_switch->cc_data.subn_get_sw_general_settings) * BYTES_PER_KB;
	min_cap = cc_sw_profile_min_legal(p_switch);
	max_cap = cc_sw_profile_max_legal(p_switch);
	granularity_supported = ib_cc_enh_info_is_cap_supported(&p_node->cc_enh_info,
								IB_CC_CAPABILITY_MASK_GRANULARITY_SUPPORTED_BIT);
	for (vl = 0; vl < IB_DROP_VL; vl++) {
		osm_cc_profiles_conf_t *p_profiles_conf = p_vl_conf->vl_to_profiles[vl];
		uint64_t min, max, min_bytes, max_bytes;
		boolean_t is_percent_mode;

		is_percent_mode = cc_profiles_conf_is_percent_mode(p_profiles_conf);

		/* No granularity on percent mode */
		if (granularity_supported && !is_percent_mode) {
			ib_cc_port_profile_set_granularity(&vl_to_profiles[vl], 1);
		} else {
			ib_cc_port_profile_set_granularity(&vl_to_profiles[vl], 0);
		}
		ib_cc_port_profile_set_mode(&vl_to_profiles[vl], p_profiles_conf->mode);

		for (profile = 0; profile < OSM_CC_MAX_PROFILE_NUM; profile++) {
			min_bytes = p_profiles_conf->profiles[profile].min;
			max_bytes = p_profiles_conf->profiles[profile].max;

			if (is_percent_mode) {
				/* min and max are percentages and don't need further calculations */
				ib_cc_port_profiles_set_profile(&vl_to_profiles[vl],
								profile,
								min_bytes,
								max_bytes,
								p_profiles_conf->profiles[profile].percent);
				continue;
			}

			/* min and max need to be a multiple of step_size */
			min_bytes = ROUNDNEAREST(min_bytes, step_size_bytes);
			max_bytes = ROUNDNEAREST(max_bytes, step_size_bytes);

			/* If switch supports granularity, use KBs */
			if (granularity_supported) {
				min = min_bytes / BYTES_PER_KB;
				max = max_bytes / BYTES_PER_KB;
			} else {
				min = min_bytes;
				max = max_bytes;
			}

			/*
			 * Check that: min_cap <= min <= max <= max_cap.
			 * Note that min_cap and max_cap are already in
			 * KB/Bytes according to granularity support.
			 */
			if (min < min_cap) {
				OSM_LOG(&p_osm->log, OSM_LOG_INFO, "Switch 0x%016" PRIx64
					" can't accept profile min smaller than %" PRIu64 " %s."
					" changing VL %d profile %d min to this value.\n",
					cl_ntoh64(osm_node_get_node_guid(p_node)), min_cap,
					granularity_supported ? "KBs" : "Bytes", vl, profile);
				min = min_cap;
				max = MAX(max, min);
			}
			if (max > max_cap) {
				OSM_LOG(&p_osm->log, OSM_LOG_INFO, "Switch 0x%016" PRIx64
					" can't accept profile max larger than %" PRIu64 " %s."
					" changing VL %d profile %d max to this value.\n",
					cl_ntoh64(osm_node_get_node_guid(p_node)), max_cap,
					granularity_supported ? "KBs" : "Bytes", vl, profile);
				max = max_cap;
				min = MIN(min, max);
			}

			ib_cc_port_profiles_set_profile(&vl_to_profiles[vl], profile, min, max,
							p_profiles_conf->profiles[profile].percent);
		}
	}
}

static void calculate_vl_profiles_mad_data(struct osm_opensm *p_osm,
					   cl_fmap_t *cc_profiles_map,
					   osm_cc_vl_conf_t *p_vl_conf)
{
	cl_map_item_t *p_item;
	ib_cc_port_profile_settings_t vl_to_profiles[IB_MAX_NUM_VLS];
	osm_physp_t *p_physp;
	uint8_t num_ports, port_num;
	boolean_t all_switch_ports = FALSE;

	if (p_vl_conf->group_type == OSM_CC_GROUP_SWITCH_PORTS)
		all_switch_ports = TRUE;

	for (p_item = cl_qmap_head(&p_osm->subn.sw_guid_tbl);
	     p_item != cl_qmap_end(&p_osm->subn.sw_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		osm_switch_t *p_switch = (osm_switch_t *) p_item;
		osm_node_t *p_node = p_switch->p_node;

		if (!ib_cc_get_enh_info_ver0(&p_node->cc_enh_info))
			continue;

		/* Skip if switch didn't respond to CongestionSwitchGeneralSettings GET request first */
		if (cc_check_sw_specific_mad_need_update(p_switch, OSM_CC_RESEND_GET_SW_GENERAL_SETTINGS))
			continue;

		memset(vl_to_profiles, 0, sizeof(ib_cc_port_profile_settings_t) * IB_MAX_NUM_VLS);
		calculate_sw_vl_to_profiles(p_osm, vl_to_profiles, p_vl_conf, p_switch);

		num_ports = osm_node_get_num_physp(p_node);
		p_switch->cc_data.is_vl_conf_for_all_ports = all_switch_ports;

		/* If the VL configuration is the same for all the ports save it on port 0 */
		port_num = p_switch->cc_data.is_vl_conf_for_all_ports ? 0 : 1;
		for ( ; port_num < num_ports; port_num++) {
			p_physp = osm_node_get_physp_ptr(p_node, port_num);
			if (p_physp->cc_data.vl_calculated)
				continue;

			memcpy(p_physp->cc_data.calc_mads.vl_to_profiles,
			       vl_to_profiles, sizeof(ib_cc_port_profile_settings_t) * IB_MAX_NUM_VLS);
			p_physp->cc_data.vl_calculated = TRUE;
		}
	}
}

static void calculate_sl_profiles_mad_data(struct osm_opensm *p_osm,
					   osm_cc_sl_conf_t *p_sl_conf)
{
	cl_map_item_t *p_item;
	osm_physp_t *p_physp;
	uint8_t num_ports, port_num, sl_num;
	ib_cc_sl_mapping_settings_t cc_sl_mapping;
	boolean_t all_switch_ports = FALSE;

	memset(&cc_sl_mapping, 0, sizeof(ib_cc_sl_mapping_settings_t));

	if (p_sl_conf->group_type == OSM_CC_GROUP_SWITCH_PORTS)
		all_switch_ports = TRUE;

	for (sl_num = 0; sl_num < IB_NUMBER_OF_SLS; sl_num++) {
		ib_cc_sl_set_profile_num(&cc_sl_mapping, sl_num,
					 p_sl_conf->sl2profile_num[sl_num]);
	}

	for (p_item = cl_qmap_head(&p_osm->subn.sw_guid_tbl);
	     p_item != cl_qmap_end(&p_osm->subn.sw_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		osm_switch_t *p_switch = (osm_switch_t *) p_item;
		osm_node_t *p_node = p_switch->p_node;

		num_ports = osm_node_get_num_physp(p_node);
		p_switch->cc_data.is_sl_conf_for_all_ports = all_switch_ports;

		/* If the SL configuration is the same for all the ports save it on port 0 */
		port_num = p_switch->cc_data.is_sl_conf_for_all_ports ? 0 : 1;
		for ( ; port_num < num_ports; port_num++) {
			p_physp = osm_node_get_physp_ptr(p_node, port_num);
			if (p_physp->cc_data.sl_calculated)
				continue;

			memcpy(&p_physp->cc_data.calc_mads.sl_mapping,
			       &cc_sl_mapping, sizeof(cc_sl_mapping));

			p_physp->cc_data.sl_calculated = TRUE;
		}
	}
}

static void calculate_sw_mad_data(struct osm_opensm *p_osm,
				  osm_cc_sw_conf_t *p_sw_conf)
{
	cl_map_item_t *p_item;

	for (p_item = cl_qmap_head(&p_osm->subn.sw_guid_tbl);
	     p_item != cl_qmap_end(&p_osm->subn.sw_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		osm_switch_t *p_switch = (osm_switch_t *) p_item;
		uint64_t step_size_kb, step_size_bytes, step_size;
		uint8_t aqs_weight = p_sw_conf->aqs_weight;
		boolean_t granularity_supported;

		if (p_osm->subn.opt.mlnx_congestion_control == OSM_CAPABILITY_DISABLED) {
			ib_cc_sw_set_enable_cc(&p_switch->cc_data.calc_set_sw_general_settings, 0);
			continue;
		}

		if (p_switch->cc_data.calculated)
			continue;

		/* Skip if switch didn't respond to GET request */
		if (cc_check_sw_specific_mad_need_update(p_switch, OSM_CC_RESEND_GET_SW_GENERAL_SETTINGS))
			continue;

		granularity_supported = ib_cc_enh_info_is_cap_supported(&p_switch->p_node->cc_enh_info,
									IB_CC_CAPABILITY_MASK_GRANULARITY_SUPPORTED_BIT);
		step_size_kb = p_switch->cc_data.subn_get_sw_general_settings.cap_cc_profile_step_size;
		step_size_bytes = step_size_kb * BYTES_PER_KB;

		ib_cc_sw_set_enable_cc(&p_switch->cc_data.calc_set_sw_general_settings,
				       p_sw_conf->enable);

		/*
		 * aqs_weight needs to be set low enough so that the following holds:
		 * step_size(KB/Bytes) * 2^aqs_weight <= IB_CC_PORT_PROFILE_SETTINGS_MAXIMAL_MAX_VALUE,
		 * where the units of step_size depend on whether granularity is supported.
		 * Otherwise there's no possible configuration for port profiles.
		 */
		step_size = granularity_supported ? step_size_kb : step_size_bytes;
		while (step_size * (1 << aqs_weight) > IB_CC_PORT_PROFILE_SETTINGS_MAXIMAL_MAX_VALUE) {
			aqs_weight--;
		}

		p_switch->cc_data.calc_set_sw_general_settings.aqs_weight = aqs_weight;
		p_switch->cc_data.calc_set_sw_general_settings.aqs_time = p_sw_conf->aqs_time;
		p_switch->cc_data.calculated = TRUE;
	}
}

static void calculate_hca_rp_params_mad_data(OUT ib_cc_hca_rp_parameters_t *p_hca_rp_params,
					     IN osm_cc_ca_conf_t *p_ca_conf)
{
	ib_cc_hca_rp_params_set_tgt_rate(p_hca_rp_params,
					 p_ca_conf->hca_rp_params.enable_clamp_tgt_rate);
	ib_cc_hca_rp_params_set_tgt_rate_at(p_hca_rp_params,
					    p_ca_conf->hca_rp_params.enable_clamp_tgt_rate_after_time_inc);
	p_hca_rp_params->rpg_time_reset = cl_hton32(p_ca_conf->hca_rp_params.rpg_time_reset);
	p_hca_rp_params->rpg_byte_reset = cl_hton32(p_ca_conf->hca_rp_params.rpg_byte_reset);
	p_hca_rp_params->rpg_threshold = p_ca_conf->hca_rp_params.rpg_threshold;
	p_hca_rp_params->rpg_max_rate = cl_hton32(p_ca_conf->hca_rp_params.rpg_max_rate);
	p_hca_rp_params->rpg_ai_rate = cl_hton16(p_ca_conf->hca_rp_params.rpg_ai_rate);
	p_hca_rp_params->rpg_hai_rate = cl_hton16(p_ca_conf->hca_rp_params.rpg_hai_rate);
	p_hca_rp_params->rpg_gd = p_ca_conf->hca_rp_params.rpg_gd;
	p_hca_rp_params->rpg_min_dec_fac = p_ca_conf->hca_rp_params.rpg_min_dec_fac;
	p_hca_rp_params->rpg_min_rate = cl_hton32(p_ca_conf->hca_rp_params.rpg_min_rate);
	p_hca_rp_params->rate_to_set_on_first_cnp = cl_hton32(p_ca_conf->hca_rp_params.rate_to_set_on_first_cnp);
	p_hca_rp_params->dce_tcp_g = cl_hton16(p_ca_conf->hca_rp_params.dce_tcp_g);
	p_hca_rp_params->dce_tcp_rtt = cl_hton32(p_ca_conf->hca_rp_params.dce_tcp_rtt);
	p_hca_rp_params->rate_reduce_monitor_period = cl_hton32(p_ca_conf->hca_rp_params.rate_reduce_monitor_period);
	p_hca_rp_params->initial_alpha_value = cl_hton16(p_ca_conf->hca_rp_params.initial_alpha_value);
}

static void calculate_ca_mad_data(struct osm_opensm *p_osm,
				  osm_cc_ca_conf_t *p_ca_conf)
{
	osm_ppcc_algo_data_t algo_id_data_map[IB_NUMBER_OF_SLS];
	cl_map_item_t *p_next;
	int sl_num, algo_id;

	if (p_ca_conf->cc_type == OSM_CC_V1_PPCC) {
		memset(algo_id_data_map, 0, sizeof(algo_id_data_map));
		for (sl_num = 0; sl_num < IB_NUMBER_OF_SLS; sl_num++) {
			algo_id = p_ca_conf->ppcc_ca_conf.algo_sl_map[sl_num];
			if (algo_id)
				algo_id_data_map[algo_id - 1].sl_mask |= cl_hton16(1 << sl_num);
		}
	}

	p_next = cl_qmap_head(&p_ca_conf->port_group->ports);
	while (p_next != cl_qmap_end(&p_ca_conf->port_group->ports)) {
		osm_port_t *p_port = ((osm_port_group_member_t*)p_next)->port;
		osm_node_t *p_node = p_port->p_node;

		p_next = cl_qmap_next(p_next);

		if (osm_node_get_type(p_node) != IB_NODE_TYPE_CA)
			continue;

		if (p_osm->subn.opt.mlnx_congestion_control == OSM_CAPABILITY_DISABLED) {
			ib_cc_hca_set_en_notify(&p_port->p_cc_data->calc_mads.hca_general_settings, 0);
			ib_cc_hca_set_en_react(&p_port->p_cc_data->calc_mads.hca_general_settings, 0);
			continue;
		}

		if (p_port->p_cc_data->calculated)
			continue;

		ib_cc_hca_set_en_notify(&p_port->p_cc_data->calc_mads.hca_general_settings,
					p_ca_conf->enable_notify);
		ib_cc_hca_set_en_react(&p_port->p_cc_data->calc_mads.hca_general_settings,
				       p_ca_conf->enable_react);

		p_port->p_cc_data->calc_mads.hca_np_params.min_time_between_cnps = cl_hton16(p_ca_conf->hca_np_params.min_time_between_cnps);
		ib_cc_hca_set_cnp_sl_mode(&p_port->p_cc_data->calc_mads.hca_np_params, p_ca_conf->hca_np_params.cnp_sl_mode);
		p_port->p_cc_data->calc_mads.hca_np_params.cnp_sl = p_ca_conf->hca_np_params.cnp_sl;
		ib_cc_hca_set_rtt_sl_mode(&p_port->p_cc_data->calc_mads.hca_np_params, p_ca_conf->hca_np_params.rtt_sl_mode);
		p_port->p_cc_data->calc_mads.hca_np_params.rtt_resp_sl = p_ca_conf->hca_np_params.rtt_resp_sl;

		if (p_ca_conf->cc_type == OSM_CC_V0_CC) {
			calculate_hca_rp_params_mad_data(&p_port->p_cc_data->calc_mads.hca_rp_params, p_ca_conf);
		} else {
			memcpy(p_port->p_cc_data->algo_id_data_map, algo_id_data_map, sizeof(algo_id_data_map));
		}

		p_port->p_cc_data->calculated = TRUE;
	}
}

static void calculate_algo_mads_data(IN struct osm_opensm *p_osm)
{
	cl_fmap_t *p_algo_version_to_id_map = &p_osm->sm.cc_mgr.algo_version_to_id_map;
	osm_ppcc_algo_id_map_item_t *algo_id_map_item;
	cl_fmap_item_t *p_fmap_item;
	cl_map_item_t *p_item;
	int algo_slot;

	for (p_item = cl_qmap_head(&p_osm->subn.port_guid_tbl);
	     p_item != cl_qmap_end(&p_osm->subn.port_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		osm_port_t *p_port = (osm_port_t *) p_item;
		osm_node_t *p_node = p_port->p_node;
		ib_net16_t sl_mask;
		ib_ppcc_algo_info_t *p_subn_algo_info = p_port->p_cc_data->subn_algo_info;

		if (osm_node_get_type(p_node) != IB_NODE_TYPE_CA)
			continue;

		for (algo_slot = 0; algo_slot < p_port->p_cc_data->algo_info_array_size; algo_slot++) {
			if (p_subn_algo_info[algo_slot].algo_id == 0)
				continue;

			p_fmap_item = cl_fmap_get(p_algo_version_to_id_map,
						  &p_subn_algo_info[algo_slot]);
			if (p_fmap_item == cl_fmap_end(p_algo_version_to_id_map)) {
				OSM_LOG(&p_osm->log, OSM_LOG_VERBOSE,
					"Algorithm with algo_info: algo_id: %u, "
					"algo_major_version: %u, algo_minor_version: %u "
					"is not configured in the CC configuration file. Disable the algorithm\n",
					cl_ntoh16(p_subn_algo_info[algo_slot].algo_id), p_subn_algo_info[algo_slot].algo_major_version,
					p_subn_algo_info[algo_slot].algo_minor_version);

				/* Disable algorithm on the device */
				ib_ppcc_set_hca_algo_en(&p_port->p_cc_data->calc_mads.algo_config[algo_slot], 0);
				ib_ppcc_set_hca_trace_en(&p_port->p_cc_data->calc_mads.algo_config[algo_slot], 0);
				ib_ppcc_set_hca_counter_en(&p_port->p_cc_data->calc_mads.algo_config[algo_slot], 0);

				continue;
			}

			algo_id_map_item = (osm_ppcc_algo_id_map_item_t *) p_fmap_item;

			sl_mask = p_port->p_cc_data->algo_id_data_map[algo_id_map_item->algo_id - 1].sl_mask;

			p_port->p_cc_data->calc_mads.algo_config[algo_slot].sl_bit_mask = sl_mask;
			ib_ppcc_set_hca_algo_en(&p_port->p_cc_data->calc_mads.algo_config[algo_slot], 1);
			ib_ppcc_set_hca_trace_en(&p_port->p_cc_data->calc_mads.algo_config[algo_slot], 1);
			ib_ppcc_set_hca_counter_en(&p_port->p_cc_data->calc_mads.algo_config[algo_slot], 1);

			memcpy(&p_port->p_cc_data->calc_mads.config_param[algo_slot],
			       &p_osm->sm.cc_mgr.algo_id_params_mad_map[algo_id_map_item->algo_id],
			       sizeof(p_port->p_cc_data->calc_mads.config_param[algo_slot]));
		}
	}
}

static void cc_calculate_ca_algo_mad_data(IN struct osm_opensm *p_osm,
					  IN osm_ppcc_ca_algos_import_conf_t *p_ca_algos_import_conf)
{
	ib_ppcc_hca_config_param_t *p_algo_id_params_mad_map = p_osm->sm.cc_mgr.algo_id_params_mad_map;
	osm_ppcc_dictionary_entry_t *entry;
	cl_fmap_item_t *p_algo_item, *p_item;
	cl_map_item_t *p_map_item;
	osm_cc_algo_param_index_t *p_map_algo_param_index;
	osm_cc_ca_algo_conf_t *p_ca_algo_conf;
	osm_ppcc_ca_algo_t *p_algo_import_conf;
	ib_ppcc_hca_config_param_t *p_ppcc_hca_config_param_mad;
	int i, j;

	memset(p_algo_id_params_mad_map, 0, sizeof(p_osm->sm.cc_mgr.algo_id_params_mad_map));
	for (p_map_item = cl_qmap_head(&p_ca_algos_import_conf->ca_algos);
	     p_map_item != cl_qmap_end(&p_ca_algos_import_conf->ca_algos);
	     p_map_item = cl_qmap_next(p_map_item)) {
		p_algo_import_conf = (osm_ppcc_ca_algo_t *) p_map_item;
		p_ppcc_hca_config_param_mad = &p_algo_id_params_mad_map[p_algo_import_conf->algo_id];

		p_algo_item = cl_fmap_get(&p_osm->sm.cc_mgr.ca_algo_conf, p_algo_import_conf->file_name);
		p_ca_algo_conf = (osm_cc_ca_algo_conf_t *) p_algo_item;

		memset(p_ppcc_hca_config_param_mad, 0, sizeof(ib_ppcc_hca_config_param_t));
		p_ppcc_hca_config_param_mad->encap_type = 2;
		p_ppcc_hca_config_param_mad->encap_len = p_ca_algo_conf->param_num * 4;

		/* Set default algo parameters values */
		for (i = 0; i < p_ca_algo_conf->param_num; i++)
			p_ppcc_hca_config_param_mad->encapsulation[i] = cl_ntoh32(p_ca_algo_conf->algo_param_list[i].value);

		for (j = 0; j < PPCC_ALGO_MAX_PARAMS; j++) {
			entry = p_algo_import_conf->param_list[j];
			if (!entry)
				break;

			p_item = cl_fmap_get(&p_ca_algo_conf->param_name_to_index, entry->str);
			p_map_algo_param_index = (osm_cc_algo_param_index_t *) p_item;
			p_ppcc_hca_config_param_mad->encapsulation[p_map_algo_param_index->param_index] = cl_ntoh32(entry->value);
		}
	}
}


static void calculate_mad_data(struct osm_opensm *p_osm, osm_cc_mgr_t *p_mgr)
{
	cl_list_item_t *p_item;

	if (p_mgr->ca_algo_import_conf)
		cc_calculate_ca_algo_mad_data(p_osm, p_mgr->ca_algo_import_conf);

	p_item = cl_qlist_head(&p_mgr->ca_conf);
	while (p_item != cl_qlist_end(&p_mgr->ca_conf)) {
		calculate_ca_mad_data(p_osm, (osm_cc_ca_conf_t *)p_item);
		p_item = cl_qlist_next(p_item);
	}

	p_item = cl_qlist_head(&p_mgr->sw_conf);
	while (p_item != cl_qlist_end(&p_mgr->sw_conf)) {
		calculate_sw_mad_data(p_osm, (osm_cc_sw_conf_t *)p_item);
		p_item = cl_qlist_next(p_item);
	}

	p_item = cl_qlist_head(&p_mgr->vl_profiles_conf);
	while (p_item != cl_qlist_end(&p_mgr->vl_profiles_conf)) {
		calculate_vl_profiles_mad_data(p_osm, &p_mgr->cc_profiles_map,
					       (osm_cc_vl_conf_t *)p_item);
		p_item = cl_qlist_next(p_item);
	}

	p_item = cl_qlist_head(&p_mgr->sl_profiles_conf);
	while (p_item != cl_qlist_end(&p_mgr->sl_profiles_conf)) {

		calculate_sl_profiles_mad_data(p_osm, (osm_cc_sl_conf_t *)p_item);

		p_item = cl_qlist_next(p_item);
	}
}

static boolean_t congestion_control_is_cc_class_supported(osm_port_t * p_port)
{
	return ib_general_info_is_cap_enabled(&p_port->p_node->general_info,
					      IB_GENERAL_INFO_IS_CC_SUPPORTED_BIT);
}


static ib_api_status_t get_sw_general_settings(struct osm_opensm *p_osm)
{
	ib_api_status_t status;
	cl_map_item_t *p_item;

	for (p_item = cl_qmap_head(&p_osm->subn.sw_guid_tbl);
	     p_item != cl_qmap_end(&p_osm->subn.sw_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		osm_switch_t *p_switch = (osm_switch_t *) p_item;

		if (!ib_cc_get_enh_info_ver0(&p_switch->p_node->cc_enh_info))
			continue;

		status = cc_send_get_sw_general_settings(&p_osm->sm, p_switch);
		if (status != IB_SUCCESS)
			return status;
	}

	return IB_SUCCESS;
}

static int get_enh_cc_info(struct osm_opensm *p_osm)
{
	ib_api_status_t status;
	cl_map_item_t *p_next;

	p_next = cl_qmap_head(&p_osm->subn.port_guid_tbl);
	while (p_next != cl_qmap_end(&p_osm->subn.port_guid_tbl)) {
		osm_port_t *p_port = (osm_port_t *) p_next;
		osm_node_t *p_node = p_port->p_node;
		p_next = cl_qmap_next(p_next);

		if (!(congestion_control_is_cc_class_supported(p_port))) {
			/* Mark that device does not support MLNX congestion control */
			memset(&p_node->cc_enh_info, 0, sizeof(ib_cc_enh_info_t));
			continue;
		}

		status = cc_send_cc_enh_info(&p_osm->sm, p_port);
		if (status != IB_SUCCESS)
			return -1;
	}

	return IB_SUCCESS;
}

static int cc_configure_ppcc_algo_config(struct osm_opensm *p_osm, uint8_t method_id)
{
	ib_cc_hca_general_settings_t *p_cc_hca_gs;
	ib_api_status_t status;
	cl_map_item_t *p_item;
	boolean_t ver1_supported;
	int algo_slot;

	for (p_item = cl_qmap_head(&p_osm->subn.port_guid_tbl);
	     p_item != cl_qmap_end(&p_osm->subn.port_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		osm_port_t *p_port = (osm_port_t *) p_item;
		osm_node_t *p_node = p_port->p_node;
		boolean_t disable_algos = FALSE;

		if (osm_node_get_type(p_node) != IB_NODE_TYPE_CA)
			continue;

		ver1_supported = ib_cc_get_enh_info_ver1(&p_node->cc_enh_info) &&
			ib_cc_enh_info_is_cap_supported(&p_node->cc_enh_info,
							IB_CC_CAPABILITY_MASK_ALGO_CONFIG_SUPPORTED_BIT);
		if (!ver1_supported)
			continue;

		if (!cc_is_port_configurable(&p_osm->sm, p_port))
			continue;

		if (cc_check_ca_specific_mad_need_update(p_port, OSM_CC_HCA_GENERAL_SETTINGS) ||
		    memcmp(&p_port->p_cc_data->calc_mads.hca_general_settings,
			   &p_port->p_cc_data->subn_mads.hca_general_settings,
			   sizeof(p_port->p_cc_data->calc_mads.hca_general_settings)))
			continue;

		p_cc_hca_gs = &p_port->p_cc_data->calc_mads.hca_general_settings;
		if (method_id == IB_MAD_METHOD_GET) {
			if (ib_cc_hca_get_en_notify(p_cc_hca_gs) &&
			    ib_cc_hca_get_en_react(p_cc_hca_gs)) {
				status = cc_send_ppcc_ca_algo_config(&p_osm->sm, p_port, 0, method_id);
				if (status != IB_SUCCESS)
					return status;
			}
		} else {
			if (ib_cc_hca_get_en_notify(p_cc_hca_gs) &&
			    ib_cc_hca_get_en_react(p_cc_hca_gs)) {
				if (cc_check_ca_specific_mad_need_update(p_port, OSM_CC_RESEND_CA_NP_PARAMETERS) ||
				    memcmp(&p_port->p_cc_data->calc_mads.hca_np_params,
					   &p_port->p_cc_data->subn_mads.hca_np_params,
					   sizeof(p_port->p_cc_data->subn_mads.hca_np_params)))
					disable_algos = TRUE;

				for (algo_slot = 0; algo_slot < p_port->p_cc_data->algo_info_array_size; algo_slot++) {
					if (p_port->p_cc_data->subn_algo_info[algo_slot].algo_id == 0)
						continue;

					if (disable_algos || p_port->p_cc_data->resend_param_config[algo_slot] ||
					    memcmp(&p_port->p_cc_data->calc_mads.config_param[algo_slot],
						   &p_port->p_cc_data->subn_mads.config_param[algo_slot],
						   sizeof(p_port->p_cc_data->calc_mads.config_param[algo_slot]))) {
						cc_send_ppcc_ca_algo_disable(&p_osm->sm, p_port, algo_slot);
						continue;
					}

					status = cc_send_ppcc_ca_algo_config(&p_osm->sm, p_port, algo_slot, method_id);
					if (status != IB_SUCCESS)
						return status;
				}
			} else {
				cc_clear_ca_subn_algo_config_set(p_port);
			}
		}
	}

	return IB_SUCCESS;
}

static int set_hca_general_settings(struct osm_opensm *p_osm)
{
	ib_api_status_t status;
	cl_map_item_t *p_item;
	boolean_t ver0_supported, ver1_supported;
	int ret = 0;

	for (p_item = cl_qmap_head(&p_osm->subn.port_guid_tbl);
	     p_item != cl_qmap_end(&p_osm->subn.port_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		osm_port_t *p_port = (osm_port_t *) p_item;
		osm_node_t *p_node = p_port->p_node;

		if (osm_node_get_type(p_node) != IB_NODE_TYPE_CA)
			continue;

		ver0_supported = ib_cc_get_enh_info_ver0(&p_node->cc_enh_info);
		ver1_supported = (ib_cc_get_enh_info_ver1(&p_node->cc_enh_info) &&
				  ib_cc_enh_info_is_cap_supported(&p_node->cc_enh_info,
								  IB_CC_CAPABILITY_MASK_ALGO_CONFIG_SUPPORTED_BIT));
		if (!ver0_supported && !ver1_supported)
			continue;

		if (!cc_is_port_configurable(&p_osm->sm, p_port))
			continue;

		status = cc_send_hca_general_settings(&p_osm->sm, p_node, p_port);
		if (status != IB_SUCCESS)
			ret = -1;
	}

	return ret;
}

static void cc_unmark_devices(struct osm_opensm *p_osm)
{
	cl_map_item_t *p_item;
	osm_port_t *p_port;
	osm_node_t *p_node;
	osm_switch_t *p_switch;
	unsigned int cc_sw_counter = 0;
	unsigned int cc_ca_counter = 0;

	for (p_item = cl_qmap_head(&p_osm->subn.port_guid_tbl);
	     p_item != cl_qmap_end(&p_osm->subn.port_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_port = (osm_port_t *)p_item;
		p_node = p_port->p_node;

		if (p_node->sw) {
			if (cc_is_port_configurable(&p_osm->sm, p_port))
				cc_disable_sw_need_update(p_node->sw, OSM_CC_RESEND_ALL);
			p_switch = p_node->sw;
			if (p_switch->cc_data.subn_set_sw_general_settings.enable)
				cc_sw_counter++;
			continue;
		}

		if (osm_node_get_type(p_node) == IB_NODE_TYPE_CA) {
			if (cc_is_port_configurable(&p_osm->sm, p_port))
				cc_disable_ca_need_update(p_port, OSM_CC_RESEND_ALL);
			if (p_port->p_cc_data->subn_mads.hca_general_settings.enable_mask & 0xC0)
				cc_ca_counter++;
			continue;
		}
	}

	OSM_LOG(&p_osm->log, OSM_LOG_VERBOSE,
		"congestion control configured on %u switches %u CAs\n",
		cc_sw_counter, cc_ca_counter);
}

static int cc_configure_subnet(struct osm_opensm *p_osm)
{
	int ret = 0;

	cl_plock_acquire(&p_osm->lock);

	if (get_enh_cc_info(p_osm)) {
		ret = -1;
		goto Exit;
	}

	cl_plock_release(&p_osm->lock);
	osm_congestion_control_wait_pending_transactions(p_osm);
	cl_plock_acquire(&p_osm->lock);

	if (get_sw_general_settings(p_osm)) {
		ret = -1;
		goto Exit;
	}

	cl_plock_release(&p_osm->lock);
	osm_congestion_control_wait_pending_transactions(p_osm);
	cl_plock_acquire(&p_osm->lock);

	cc_clear_calc_data(p_osm);
	calculate_mad_data(p_osm, &p_osm->sm.cc_mgr);

	/*
	 * Configuration order on switches should be:
	 * 1. CongestionSwitchGeneralSetting
	 * 2. CongestionPortProfileSetting
	 * 3. CongestionSLMappingSetting?
	 */

	if (configure_switches(p_osm))
		ret = -1;

	cl_plock_release(&p_osm->lock);
	osm_congestion_control_wait_pending_transactions(p_osm);
	cl_plock_acquire(&p_osm->lock);

	if (configure_profiles(p_osm))
		ret = -1;

	if (set_hca_general_settings(p_osm))
		ret = -1;

	cl_plock_release(&p_osm->lock);
	osm_congestion_control_wait_pending_transactions(p_osm);
	cl_plock_acquire(&p_osm->lock);

	if (cc_configure_ppcc_algo_config(p_osm, IB_MAD_METHOD_GET))
		ret = -1;

	cl_plock_release(&p_osm->lock);
	osm_congestion_control_wait_pending_transactions(p_osm);
	cl_plock_acquire(&p_osm->lock);

	calculate_algo_mads_data(p_osm);
	if (configure_hcas(p_osm))
		ret = -1;

	cl_plock_release(&p_osm->lock);
	osm_congestion_control_wait_pending_transactions(p_osm);
	cl_plock_acquire(&p_osm->lock);

	/* Enable PPCC algorithms after setting the params */
	if (cc_configure_ppcc_algo_config(p_osm, IB_MAD_METHOD_SET))
		ret = -1;

Exit:
	cl_plock_release(&p_osm->lock);

	osm_congestion_control_wait_pending_transactions(p_osm);

	return ret;
}

static ib_api_status_t congestion_control_send_req_keyinfo(osm_sm_t * p_sm,
							   osm_port_t * p_port)
{
	osm_opensm_t *p_osm = p_sm->p_subn->p_osm;
	osm_madw_t *p_madw = NULL;
	osm_node_t *p_node = p_port->p_node;
	osm_physp_t *p_physp;

	OSM_LOG_ENTER(p_sm->p_log);

	if (p_port->disc_through != OSM_DISC_NORMAL)
		return IB_SUCCESS;

	if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH)
		p_physp = osm_node_get_physp_ptr(p_node, 0);
	else
		p_physp = p_port->p_physp;

	p_madw = osm_mad_pool_get(&p_osm->mad_pool,
				  p_osm->cc.bind_handle[p_sm->p_subn->sm_port_idx],
				  MAD_BLOCK_SIZE, NULL);

	if (p_madw == NULL) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR C129: "
			"failed to allocate mad\n");
		return IB_INSUFFICIENT_MEMORY;
	}

	cl_plock_acquire(&p_osm->lock);
	cc_mad_post(&p_osm->cc, p_madw, p_physp, p_port,
		    IB_MAD_ATTR_CONG_KEY_INFO, 0, IB_MAD_METHOD_GET);

	cl_plock_release(&p_osm->lock);

	OSM_LOG_EXIT(p_sm->p_log);

	return IB_SUCCESS;
}

static void congestion_control_setup_key_configuration(struct osm_opensm * p_osm)
{
	osm_port_t *p_port;
	cl_map_item_t *p_item;

	OSM_LOG_ENTER(&p_osm->log);

	if (p_osm->subn.opt.cc_key_enable == OSM_CAPABILITY_IGNORED)
		goto Exit;

	for (p_item = cl_qmap_head(&p_osm->subn.port_guid_tbl);
	     p_item != cl_qmap_end(&p_osm->subn.port_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_port = (osm_port_t *)p_item;

		/*
		 * Set cached key at port key. Hence, key configuration MADs
		 * are send with cached key, if there is any.
		 */
		if (!(p_port->p_cc_data->resend_mask & OSM_CC_RESEND_GET_KEY_INFO))
			osm_key_mgr_get_cached_key(p_osm, p_port,IB_MCLASS_CC,
						   OSM_KEY_TYPE_MANAGER);

		osm_key_mgr_generate_key(p_osm, p_port, IB_MCLASS_CC,
					 OSM_KEY_TYPE_MANAGER);

		if (p_port->is_new || p_port->p_physp->need_update ||
		    (p_port->p_node->sw && p_port->p_node->sw->need_update) ||
		    p_osm->subn.first_time_master_sweep ||
		    p_osm->subn.force_reroute || p_osm->subn.coming_out_of_standby) {
			p_port->p_cc_data->resend_mask |= OSM_CC_RESEND_GET_KEY_INFO;
			p_port->p_cc_data->resend_mask |= OSM_CC_RESEND_CC_KEY_SUPPORT;
		}

		if (p_osm->cc.recalculate_keys)
			p_port->p_cc_data->resend_mask |= OSM_CC_RESEND_CC_KEY;
	}

Exit:
	OSM_LOG_EXIT(&p_osm->log);
}

static int ppcc_parse_algo_files(struct osm_opensm *p_osm)
{
	DIR *algo_dir;
	struct dirent *de;
	char file_path[PATH_MAX];
	int ret = 0;
	errno = 0;

	algo_dir = opendir(p_osm->subn.opt.ppcc_algo_dir);
	if (!algo_dir) {
		OSM_LOG(&p_osm->log, OSM_LOG_VERBOSE,
			"Failed to open ppcc_algo directory '%s' - %s\n",
			p_osm->subn.opt.ppcc_algo_dir, strerror(errno));
		return 0;
	}

	while ((de = readdir(algo_dir))) {
		/* Ignore all hidden files */
		if (de->d_name[0] == '.')
			continue;

		snprintf(file_path, sizeof(file_path), "%s/%s",
			 p_osm->subn.opt.ppcc_algo_dir, de->d_name);
		if (parse_ppcc_algo_file(p_osm, file_path, de->d_name)) {
			ret = 1;
			goto Exit;
		}
	}
	
Exit:
	closedir(algo_dir);
	return ret;	
}


int osm_congestion_control_setup(struct osm_opensm *p_osm, boolean_t reload_configuration)
{
	osm_cc_mgr_t *p_mgr = &p_osm->sm.cc_mgr;
	int ret = 0;

	OSM_LOG_ENTER(&p_osm->log);

	if (reload_configuration)
		osm_cc_mgr_destroy(p_mgr);

	cc_mark_need_update(p_osm);

	/*
	 * Do nothing unless the most recent routing attempt was successful.
	 */
	if (!p_osm->subn.routing_engine_used)
		goto exit;

	if (p_osm->subn.opt.cc_key_enable) {
		congestion_control_setup_key_configuration(p_osm);
		if (osm_congestion_control_configure_subnet_keys(p_osm)) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR C126: "
				"CC keys configuration has failed\n");
			ret = 1;
			goto exit;
		}
	}

	if (p_osm->subn.opt.mlnx_congestion_control == OSM_CAPABILITY_ENABLED &&
	    reload_configuration) {
		if (ppcc_parse_algo_files(p_osm)) {
			ret = 1;
			goto exit;
		}
		if (parse_congestion_control_policy(p_osm)) {
			ret = 1;
			goto exit;
		}

		if (OSM_LOG_IS_ACTIVE_V2(&p_osm->log, OSM_LOG_DEBUG)) {
			print_profiles_map(p_osm, &p_mgr->cc_profiles_map);
			print_sl_conf(p_osm, &p_mgr->sl_profiles_conf);
			print_vl_conf(p_osm, &p_mgr->vl_profiles_conf);
			print_ca_algo_conf(p_osm, &p_mgr->ca_algo_conf);
			print_ca_algo_import_conf(p_osm, p_mgr->ca_algo_import_conf);
			print_ca_conf(p_osm, &p_mgr->ca_conf);
		}
	}


	ret = cc_configure_subnet(p_osm);

	cc_unmark_devices(p_osm);
exit:
	if (ret)
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR C100: Errors during congestion control configuration\n");

	OSM_LOG_EXIT(&p_osm->log);

	return ret;
}

int osm_congestion_control_wait_pending_transactions(struct osm_opensm *p_osm)
{
	osm_congestion_control_t *cc = &p_osm->sm.p_subn->p_osm->cc;


	while (1) {
		unsigned count = cc->outstanding_mads;
		if (!count || osm_exit_flag)
			break;
		cl_event_wait_on(&cc->outstanding_mads_done_event,
				 EVENT_NO_TIMEOUT,
				 TRUE);
	}

	return osm_exit_flag;
}

static inline void decrement_outstanding_mads(osm_congestion_control_t *p_cc)
{
	uint32_t outstanding;

	outstanding = cl_atomic_dec(&p_cc->outstanding_mads);
	if (!outstanding)
		cl_event_signal(&p_cc->outstanding_mads_done_event);

	cl_atomic_dec(&p_cc->outstanding_mads_on_wire);
	cl_event_signal(&p_cc->sig_mads_on_wire_continue);
}

const char *cc_get_cc_attr_str(IN ib_net16_t attr)
{
	switch (attr) {
	case IB_MAD_ATTR_CC_ENH_INFO:
		return "EnhancedCongestionInfo";
	case IB_MAD_ATTR_CC_SW_GENERAL_SETTINGS:
		return "CongestionSwitchGeneralSettings";
	case IB_MAD_ATTR_CC_PORT_SW_PROFILE_SETTINGS:
		return "CongestionPortProfileSettings";
	case IB_MAD_ATTR_CC_SL_MAPPING_SETTINGS:
		return "CongestionSLMappingSettings";
	case IB_MAD_ATTR_CC_HCA_GENERAL_SETTINGS:
		return "CongestionHCAGeneralSettings";
	case IB_MAD_ATTR_CC_HCA_RP_PARAMETERS:
		return "CongestionHCARPParameters";
	case IB_MAD_ATTR_CC_HCA_NP_PARAMETERS:
		return "CongestionHCANPParameters";
	case IB_MAD_ATTR_PPCC_HCA_ALGO_CONFIG:
		return "CongestionHCAAlgoConfig";
	case IB_MAD_ATTR_PPCC_HCA_ALGO_PARAMS_CONFIG:
		return "CongestionHCAParamConfig";
	case IB_MAD_ATTR_CONG_KEY_INFO:
		return "CongestionKeyInfo";
	case IB_MAD_ATTR_CLASS_PORT_INFO:
		return "ClassPortInfo";
	case IB_MAD_ATTR_NOTICE:
		return "Notice";
	}

	return "UNKNOWN";
}

void dump_cc_mad(ib_cc_mad_t *p_cc_mad, uint32_t buf_size,
		 osm_log_t *p_log, ib_net64_t dest_guid)
{
		char buf[BUF_SIZE];
		buf[0] = '\0';
		unsigned n;
		int i;
		if (!p_cc_mad)
			return;

		n = sprintf(buf,
			    "CC Mad dump:\n"
			    "\t\t\t\tbase_ver................0x%X\n"
			    "\t\t\t\tmgmt_class..............0x%X\n"
			    "\t\t\t\tclass_ver...............0x%X\n"
			    "\t\t\t\tmethod..................0x%X (%s)\n",
			    p_cc_mad->header.base_ver, p_cc_mad->header.mgmt_class,
			    p_cc_mad->header.class_ver, p_cc_mad->header.method,
			    ib_get_sm_method_str(p_cc_mad->header.method));

		n += snprintf(buf + n, buf_size - n,
			      "\t\t\t\tstatus..................0x%X\n",
			      cl_ntoh16(p_cc_mad->header.status));

		n += snprintf(buf + n, buf_size - n,
			      "\t\t\t\ttrans_id................0x%" PRIx64 "\n"
			      "\t\t\t\tattr_id.................0x%X (%s)\n"
			      "\t\t\t\tadditional_status.......0x%X\n"
			      "\t\t\t\tattr_mod................0x%X\n"
			      "\t\t\t\tcc_key..................0x%016" PRIx64 "\n"
			      "\t\t\t\tdest_guid...............0x%016" PRIx64 "\n",
			      cl_ntoh64(p_cc_mad->header.trans_id),
			      cl_ntoh16(p_cc_mad->header.attr_id),
			      cc_get_cc_attr_str(p_cc_mad->header.attr_id),
			      cl_ntoh16(p_cc_mad->header.additional_status),
			      cl_ntoh32(p_cc_mad->header.attr_mod),
			      cl_ntoh64(p_cc_mad->cc_key),
			      cl_ntoh64(dest_guid));

		for (i = 0; i < 64; i += 16) {
			n += snprintf(buf + n, buf_size - n,
				      "\n\t\t\t\t%02X %02X %02X %02X "
				      "%02X %02X %02X %02X"
				      "   %02X %02X %02X %02X %02X %02X %02X %02X\n",
				      p_cc_mad->mgt_data[i],
				      p_cc_mad->mgt_data[i + 1],
				      p_cc_mad->mgt_data[i + 2],
				      p_cc_mad->mgt_data[i + 3],
				      p_cc_mad->mgt_data[i + 4],
				      p_cc_mad->mgt_data[i + 5],
				      p_cc_mad->mgt_data[i + 6],
				      p_cc_mad->mgt_data[i + 7],
				      p_cc_mad->mgt_data[i + 8],
				      p_cc_mad->mgt_data[i + 9],
				      p_cc_mad->mgt_data[i + 10],
				      p_cc_mad->mgt_data[i + 11],
				      p_cc_mad->mgt_data[i + 12],
				      p_cc_mad->mgt_data[i + 13],
				      p_cc_mad->mgt_data[i + 14],
				      p_cc_mad->mgt_data[i + 15]);
		}

	osm_log(p_log, OSM_LOG_FRAMES, "%s\n", buf);
}

static void cc_rcv_mad_port_profile_settings(osm_node_t *p_node, uint8_t port_num,
					     ib_cc_port_profile_settings_t *p_cc_port_profiles)
{
	osm_physp_t *p_physp;
	osm_switch_t *p_sw;
	uint16_t vl_mask;
	uint8_t port, port_end, vl;

	OSM_ASSERT(p_node && p_node->sw);
	p_sw = p_node->sw;
	OSM_ASSERT(port_num == 0xff || (port_num && port_num < p_sw->num_ports));

	vl_mask = cl_ntoh16(p_cc_port_profiles->vl_mask);

	/*
	 * Store MAD without VL mask in order to all comparing profiles
	 * configuration of different VLs.
	 */
	p_cc_port_profiles->vl_mask = 0;

	if (port_num == 0xff) {
		port = 0;
		port_end = p_sw->num_ports;
	} else {
		port = port_num;
		port_end = port_num + 1;
	}

	for (; port < port_end; port++) {
		p_physp = osm_node_get_physp_ptr(p_node, port);
		for (vl = 0; vl < IB_MAX_NUM_VLS; vl++) {
			if ((vl_mask & (1 << vl)) == 0)
				continue;
			/*
			 * HACK: In Profile settings MAD the switch can change the buffers size.
			 * In this case the received MAD is not equal to the one OpenSM sent.
			 * To prevent the sending of the MAD every HEAVY SWEEP
			 * save the data that was sent and not the subn data.
			 */
			memcpy(&p_physp->cc_data.subn_mads.vl_to_profiles[vl],
			       &p_physp->cc_data.calc_mads.vl_to_profiles[vl],
			       sizeof(ib_cc_port_profile_settings_t));
		}
	}
}

static void cc_rcv_mad_sl_settings(osm_node_t *p_node, uint8_t port_num,
				   ib_cc_sl_mapping_settings_t *p_cc_sl_setting)
{
	osm_physp_t *p_physp;
	osm_switch_t *p_sw;
	uint8_t port, port_end, sl;

	OSM_ASSERT(p_node && p_node->sw);
	p_sw = p_node->sw;
	OSM_ASSERT(port_num == 0xff || (port_num && port_num < p_sw->num_ports));

	if (port_num == 0xff) {
		port = 0;
		port_end = p_sw->num_ports;
	} else {
		port = port_num;
		port_end = port_num + 1;
	}

	for (; port < port_end; port++) {
		p_physp = osm_node_get_physp_ptr(p_node, port);
		for (sl = 0; sl < IB_NUMBER_OF_SLS; sl++) {
			memcpy(&p_physp->cc_data.subn_mads.sl_mapping,
			       p_cc_sl_setting,
			       sizeof(ib_cc_sl_mapping_settings_t));
		}
	}

	/* HACK - we set the mask once per switch, instead of per port */
	cc_disable_sw_need_update(p_sw, OSM_CC_RESEND_SW_SL_MAPPING);
}

static ib_api_status_t congestion_control_send_set_cpi(osm_subn_t * p_subn,
						       osm_port_t * p_port)
{
	osm_opensm_t *p_osm = p_subn->p_osm;
	osm_congestion_control_t *p_cc = &p_subn->p_osm->cc;
	const ib_cc_mad_t *p_cc_mad;
	ib_class_port_info_t *cpi_data;
	osm_madw_t *p_madw;
	ib_gid_t gid;
	ib_api_status_t status = IB_SUCCESS;

	OSM_LOG_ENTER(&p_osm->log);

	p_madw = osm_mad_pool_get(p_cc->mad_pool, p_cc->bind_handle[p_subn->sm_port_idx],
				  MAD_BLOCK_SIZE, NULL);
	if (!p_madw) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR C113: "
			"Unable to allocate MAD\n");
		status = IB_INSUFFICIENT_MEMORY;
		goto Exit;
	}
	p_cc_mad = osm_madw_get_cc_mad_ptr(p_madw);
	cpi_data = (ib_class_port_info_t *)ib_cc_mad_get_mgt_data_ptr(p_cc_mad);
	memcpy(cpi_data, &p_port->p_cc_data->cpi, sizeof(ib_class_port_info_t));

	gid.unicast.prefix = 0;
	gid.unicast.interface_id = 0;
	cpi_data->trap_gid = gid;
	cpi_data->trap_tc_sl_fl = 0;
	cpi_data->trap_lid = p_subn->sm_base_lid;
	cpi_data->trap_pkey = cl_hton16(0xFFFF);
	cpi_data->trap_hop_qp = cl_hton32(1);
	cpi_data->trap_qkey = IB_QP1_WELL_KNOWN_Q_KEY;

	if (!memcmp(cpi_data, &p_port->p_cc_data->cpi, sizeof(ib_class_port_info_t))) {
		p_port->p_cc_data->resend_mask &= ~OSM_CC_RESEND_TRAP_LID;
		osm_mad_pool_put(p_cc->mad_pool, p_madw);
		goto Exit;
	}

	cl_plock_acquire(&p_osm->lock);
	cc_mad_post(p_cc, p_madw, p_port->p_physp, p_port,
		    IB_MAD_ATTR_CLASS_PORT_INFO, 0, IB_MAD_METHOD_SET);
	cl_plock_release(&p_osm->lock);

Exit:
	OSM_LOG_EXIT(&p_osm->log);
	return status;
}

static void congestion_control_rcv_sw_general_settings(osm_congestion_control_t * p_cc,
						       osm_madw_t * p_madw,
						       osm_port_t * p_port)
{
	osm_opensm_t *p_osm = p_cc->osm;
	const ib_cc_mad_t *p_cc_mad;
	ib_cc_sw_general_settings_t *p_sw_general_settings;
	const osm_cc_context_t *p_cc_context;

	OSM_LOG_ENTER(&p_osm->log);

	p_cc_mad = (ib_cc_mad_t *)osm_madw_get_mad_ptr(p_madw);
	p_sw_general_settings = ib_cc_mad_get_mgt_data_ptr(p_cc_mad);
	p_cc_context = (osm_cc_context_t *)&p_madw->context;

	if (p_cc_context->mad_method == IB_MAD_METHOD_GET) {
		memcpy(&p_port->p_node->sw->cc_data.subn_get_sw_general_settings,
		       p_sw_general_settings,
		       sizeof(ib_cc_sw_general_settings_t));
		cc_disable_sw_need_update(p_port->p_node->sw, OSM_CC_RESEND_GET_SW_GENERAL_SETTINGS);
	} else {
		memcpy(&p_port->p_node->sw->cc_data.subn_set_sw_general_settings,
		       p_sw_general_settings,
		       sizeof(ib_cc_sw_general_settings_t));
		cc_disable_sw_need_update(p_port->p_node->sw, OSM_CC_RESEND_SET_SW_GENERAL_SETTINGS);
	}

	OSM_LOG_EXIT(&p_osm->log);
}

static void congestion_control_rcv_cpi(osm_congestion_control_t * p_cc,
				       osm_madw_t * p_madw,
				       osm_port_t * p_port)
{
	osm_opensm_t *p_osm = p_cc->osm;
	const ib_cc_mad_t *p_cc_mad;
	ib_class_port_info_t *data;
	osm_cc_context_t *p_cc_context;

	OSM_LOG_ENTER(&p_osm->log);

	p_cc_mad = (ib_cc_mad_t *)osm_madw_get_mad_ptr(p_madw);
	data = ib_cc_mad_get_mgt_data_ptr(p_cc_mad);
	p_cc_context = (osm_cc_context_t *)&p_madw->context;

	if (p_cc_mad->header.status != 0) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR C12A: "
			"Received class port info GET_RESP with status != success\n");
		goto Exit;
	}

	if (p_cc_context->mad_method == IB_MAD_METHOD_GET) {
		p_port->p_cc_data->resend_mask &= ~OSM_CC_RESEND_CC_KEY_SUPPORT;

		p_port->p_cc_data->resend_mask |= OSM_CC_RESEND_TRAP_LID;
		congestion_control_send_set_cpi(p_cc->subn, p_port);

	} else if (p_cc_context->mad_method == IB_MAD_METHOD_SET) {
		p_port->p_cc_data->resend_mask &= ~OSM_CC_RESEND_TRAP_LID;
		if (!p_port->p_cc_data->key_configured)
			p_port->p_cc_data->resend_mask |= OSM_CC_RESEND_CC_KEY;
	}
	memcpy(&p_port->p_cc_data->cpi, (ib_class_port_info_t *)data,
	       sizeof(ib_class_port_info_t));

Exit:
	OSM_LOG_EXIT(&p_osm->log);
}

static void congestion_control_rcv_keyinfo(osm_congestion_control_t * p_cc,
					   osm_madw_t * p_madw,
					   osm_port_t * p_port)
{
	osm_opensm_t *p_osm = p_cc->osm;
	osm_subn_opt_t *p_opt = &p_osm->subn.opt;
	const ib_mad_t *p_mad;
	const ib_cc_mad_t *p_cc_mad;
	ib_cong_key_info_t *p_mad_info_data;
	osm_cc_context_t *p_cc_context;

	OSM_LOG_ENTER(&p_osm->log);

	OSM_ASSERT(p_madw);
	p_mad = osm_madw_get_mad_ptr(p_madw);
	p_cc_context = (osm_cc_context_t *)&p_madw->context;

	if (p_mad->status != 0) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR C10E: "
			"Received KeyInfo GET_RESP with status (0x%x)\n",
			cl_ntoh16(p_mad->status));
		goto Exit;
	}

	if (p_cc_context->mad_method == IB_MAD_METHOD_GET) {
		p_cc_mad = osm_madw_get_cc_mad_ptr(p_madw);
		p_mad_info_data = ib_cc_mad_get_mgt_data_ptr(p_cc_mad);

		/*
		 * If response holds a key - it means key is no longer protected
		 * and this is the port's key.
		 * If response's key is 0, there are 2 options -
		 * 	1. protect_bit is 1, means key is protected and SM holds the right key.
		 * 	2. otherwise, key is 0.
		 */

		if (p_mad_info_data->protect_bit) {
			/* Key parameters were not changed */
			if (p_mad_info_data->lease_period == cl_hton16(p_opt->cc_key_lease_period) &&
			    p_mad_info_data->protect_bit ==
			    cl_hton16((uint16_t)p_opt->cc_key_protect_bit << 15)) {
				/*
				 * Send CPI Get in order to check if need to update
				 * values of trap_lid etc.
				 */
				p_port->p_cc_data->resend_mask |= OSM_CC_RESEND_CC_KEY_SUPPORT;
			}

		} else {
			/* Updating the port with current key */
			p_port->cc_key = p_mad_info_data->cc_key;
		}

		p_port->p_cc_data->resend_mask &= ~OSM_CC_RESEND_GET_KEY_INFO;

	} else if (p_cc_context->mad_method == IB_MAD_METHOD_SET &&
	    (p_port->p_cc_data->resend_mask & OSM_CC_RESEND_CC_KEY)) {
		p_port->p_cc_data->resend_mask &= ~OSM_CC_RESEND_CC_KEY;

		/* Done key configuration */
		p_port->p_cc_data->key_configured = TRUE;
		p_cc->num_configured++;
	}

Exit:
	OSM_LOG_EXIT(&p_osm->log);
}

static void cc_statistic_rcv_counter_increase(ib_net16_t attr_id, osm_subn_t *p_subn)
{
	/* CC class is currently sent only on the main port */
	uint64_t *stats_counters = p_subn->p_osm->stats_counters[0];

	switch (attr_id) {
	case IB_MAD_ATTR_CLASS_PORT_INFO:
		stats_counters[CC_CLASS_PORT_INFO_RCV_COUNT]++;
		break;
	case IB_MAD_ATTR_CONG_KEY_INFO:
		stats_counters[CC_KEY_INFO_RCV_COUNT]++;
		break;
	case IB_MAD_ATTR_CC_ENH_INFO:
		stats_counters[CC_ENH_INFO_RCV_COUNT]++;
		break;
	case IB_MAD_ATTR_CC_HCA_GENERAL_SETTINGS:
		stats_counters[CC_HCA_GENERAL_SETTINGS_RCV_COUNT]++;
		break;
	case IB_MAD_ATTR_CC_PORT_SW_PROFILE_SETTINGS:
		stats_counters[CC_PORT_SW_PROFILE_SETTINGS_RCV_COUNT]++;
		break;
	case IB_MAD_ATTR_CC_SL_MAPPING_SETTINGS:
		stats_counters[CC_SL_MAPPING_SETTINGS_RCV_COUNT]++;
		break;
	case IB_MAD_ATTR_CC_SW_GENERAL_SETTINGS:
		stats_counters[CC_SW_GENERAL_SETTINGS_RCV_COUNT]++;
		break;
	case IB_MAD_ATTR_CC_HCA_RP_PARAMETERS:
		stats_counters[CC_HCA_RP_PARAMETERS_RCV_COUNT]++;
		break;
	case IB_MAD_ATTR_CC_HCA_NP_PARAMETERS:
		stats_counters[CC_HCA_NP_PARAMETERS_RCV_COUNT]++;
		break;
	case IB_MAD_ATTR_PPCC_HCA_ALGO_CONFIG:
		stats_counters[PPCC_HCA_ALGO_CONFIG_RCV_COUNT]++;
		break;
	case IB_MAD_ATTR_PPCC_HCA_ALGO_PARAMS_CONFIG:
		stats_counters[PPCC_HCA_ALGO_PARAMS_CONFIG_RCV_COUNT]++;
		break;
	default:
		break;
	}
}

static void cc_statistic_counter_increase(ib_net16_t attr_id, osm_subn_t *p_subn)
{
	/* CC class is currently sent only on the main port */
	uint64_t *stats_counters = p_subn->p_osm->stats_counters[0];

	switch (attr_id) {
	case IB_MAD_ATTR_CLASS_PORT_INFO:
		stats_counters[CC_CLASS_PORT_INFO_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_CONG_KEY_INFO:
		stats_counters[CC_KEY_INFO_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_CC_ENH_INFO:
		stats_counters[CC_ENH_INFO_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_CC_HCA_GENERAL_SETTINGS:
		stats_counters[CC_HCA_GENERAL_SETTINGS_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_CC_PORT_SW_PROFILE_SETTINGS:
		stats_counters[CC_PORT_SW_PROFILE_SETTINGS_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_CC_SL_MAPPING_SETTINGS:
		stats_counters[CC_SL_MAPPING_SETTINGS_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_CC_SW_GENERAL_SETTINGS:
		stats_counters[CC_SW_GENERAL_SETTINGS_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_CC_HCA_RP_PARAMETERS:
		stats_counters[CC_HCA_RP_PARAMETERS_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_CC_HCA_NP_PARAMETERS:
		stats_counters[CC_HCA_NP_PARAMETERS_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_PPCC_HCA_ALGO_CONFIG:
		stats_counters[PPCC_HCA_ALGO_CONFIG_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_PPCC_HCA_ALGO_PARAMS_CONFIG:
		stats_counters[PPCC_HCA_ALGO_PARAMS_CONFIG_SEND_COUNT]++;
		break;
	default:
		break;
	}
}

static void cc_rcv_mad(void *context, void *data)
{
	osm_congestion_control_t *p_cc = context;
	osm_opensm_t *p_osm = p_cc->osm;
	osm_madw_t *p_madw = data;
	ib_cc_mad_t *p_cc_mad;
	osm_madw_context_t *p_mad_context = &p_madw->context;
	ib_mad_t *p_mad = osm_madw_get_mad_ptr(p_madw);
	ib_net64_t node_guid = p_mad_context->cc_context.node_guid;
	ib_net64_t port_guid = p_mad_context->cc_context.port_guid;
	uint8_t port = p_mad_context->cc_context.port;
	ib_net32_t attr_mod = p_mad_context->cc_context.attr_mod;
	uint8_t algo_slot;
	osm_port_t *p_port;

	OSM_LOG_ENTER(p_cc->log);

	OSM_LOG(p_cc->log, OSM_LOG_VERBOSE,
		"Processing received MAD status 0x%x for "
		"attr ID 0x%x mod 0x%x node 0x%" PRIx64 " port %u\n",
		cl_ntoh16(p_mad->status), cl_ntoh16(p_mad->attr_id),
		cl_ntoh32(p_mad_context->cc_context.attr_mod),
		cl_ntoh64(node_guid), port);

	p_cc_mad = osm_madw_get_cc_mad_ptr(p_madw);

	cc_statistic_rcv_counter_increase(p_cc_mad->header.attr_id, p_cc->subn);

	cl_plock_acquire(&p_osm->lock);

	p_port = osm_get_port_by_guid(p_cc->subn, port_guid);
	if (!p_port) {
		OSM_LOG(p_cc->log, OSM_LOG_ERROR, "ERR C109: "
			"Port GUID 0x%" PRIx64 " not in table\n",
			cl_ntoh64(port_guid));
		cl_plock_release(&p_osm->lock);
		goto Exit;
	}

	p_port->cc_timeout_count = 0;

	if (p_cc_mad->header.status) {
		if (p_cc_mad->header.status & IB_MAD_STATUS_UNSUP_CLASS_VER
		    || p_cc_mad->header.status & IB_MAD_STATUS_UNSUP_METHOD
		    || p_cc_mad->header.status & IB_MAD_STATUS_UNSUP_METHOD_ATTR)
			p_port->cc_unavailable_flag = TRUE;
		OSM_LOG(p_cc->log, OSM_LOG_ERROR, "ERR C10A:"
			" Received invalid status from port GUID 0x%" PRIx64 ","
			" status:0x%x, attr_id:0x%x\n",
			cl_ntoh64(port_guid),
			cl_ntoh16(p_cc_mad->header.status),
			cl_ntoh16(p_cc_mad->header.attr_id));
		cl_plock_release(&p_osm->lock);
		goto Exit;
	} else {
		p_port->cc_unavailable_flag = FALSE;
	}

	if (p_cc_mad->header.attr_id == IB_MAD_ATTR_CC_ENH_INFO) {
		ib_cc_enh_info_t *p_cc_enh_info = ib_cc_mad_get_mgt_data_ptr(p_cc_mad);
		osm_node_t *p_node = p_port->p_node;

		memcpy(&p_port->p_node->cc_enh_info, p_cc_enh_info, sizeof(ib_cc_enh_info_t));
		if (osm_node_get_type(p_port->p_node) == IB_NODE_TYPE_CA) {
			cc_disable_ca_need_update(p_port, OSM_CC_RESEND_CC_ENH_INFO);
		} else {
			cc_disable_sw_need_update(p_node->sw, OSM_CC_RESEND_CC_ENH_INFO);
		}
	} else if (p_cc_mad->header.attr_id == IB_MAD_ATTR_CC_HCA_GENERAL_SETTINGS) {
		ib_cc_hca_general_settings_t *p_ca_general_setting = ib_cc_mad_get_mgt_data_ptr(p_cc_mad);
		memcpy(&p_port->p_cc_data->subn_mads.hca_general_settings, p_ca_general_setting,
		       sizeof(ib_cc_hca_general_settings_t));
		cc_disable_ca_need_update(p_port, OSM_CC_HCA_GENERAL_SETTINGS);
	} else if (p_cc_mad->header.attr_id == IB_MAD_ATTR_CC_PORT_SW_PROFILE_SETTINGS) {
		ib_cc_port_profile_settings_t *p_cc_port_profiles = ib_cc_mad_get_mgt_data_ptr(p_cc_mad);
		ib_net32_t attr_mod = p_mad_context->cc_context.attr_mod;
		uint8_t port_num = cl_ntoh32(attr_mod) & 0xff;
		cc_rcv_mad_port_profile_settings(p_port->p_node, port_num, p_cc_port_profiles);
	} else if (p_cc_mad->header.attr_id == IB_MAD_ATTR_CC_SL_MAPPING_SETTINGS) {
		ib_cc_sl_mapping_settings_t *p_cc_sl_setting = ib_cc_mad_get_mgt_data_ptr(p_cc_mad);
		ib_net32_t attr_mod = p_mad_context->cc_context.attr_mod;
		uint8_t port_num = cl_ntoh32(attr_mod) & 0xff;
		cc_rcv_mad_sl_settings(p_port->p_node, port_num, p_cc_sl_setting);
	} else if (p_cc_mad->header.attr_id == IB_MAD_ATTR_CC_SW_GENERAL_SETTINGS) {
		congestion_control_rcv_sw_general_settings(p_cc, p_madw, p_port);
	} else if (p_cc_mad->header.attr_id == IB_MAD_ATTR_CC_HCA_RP_PARAMETERS) {
		ib_cc_hca_rp_parameters_t *p_hca_rp_params = ib_cc_mad_get_mgt_data_ptr(p_cc_mad);
		memcpy(&p_port->p_cc_data->subn_mads.hca_rp_params, p_hca_rp_params,
		       sizeof(p_port->p_cc_data->subn_mads.hca_rp_params));
		cc_disable_ca_need_update(p_port, OSM_CC_RESEND_CA_RP_PARAMETERS);
	} else if (p_cc_mad->header.attr_id == IB_MAD_ATTR_CC_HCA_NP_PARAMETERS) {
		ib_cc_hca_rp_parameters_t *p_hca_np_params = ib_cc_mad_get_mgt_data_ptr(p_cc_mad);
		memcpy(&p_port->p_cc_data->subn_mads.hca_np_params, p_hca_np_params,
		       sizeof(p_port->p_cc_data->subn_mads.hca_np_params));
		cc_disable_ca_need_update(p_port, OSM_CC_RESEND_CA_NP_PARAMETERS);
	} else if (p_mad->attr_id == IB_MAD_ATTR_CLASS_PORT_INFO){
		congestion_control_rcv_cpi(p_cc, p_madw, p_port);
	} else if (p_mad->attr_id == IB_MAD_ATTR_CONG_KEY_INFO) {
		congestion_control_rcv_keyinfo(p_cc, p_madw, p_port);
	} else if (p_mad->attr_id == IB_MAD_ATTR_PPCC_HCA_ALGO_CONFIG) {
		ib_ppcc_hca_algo_config_t *p_ppcc_hca_algo_config = ib_cc_mad_get_mgt_data_ptr(p_cc_mad);

		if (p_mad_context->cc_context.mad_method == IB_MAD_METHOD_GET) {
			cc_disable_ca_need_update(p_port, OSM_CC_RESEND_PPCC_CA_ALGO_CONFIG_GET);
			if (p_ppcc_hca_algo_config->encap_type == 2) {
				memcpy(&p_port->p_cc_data->subn_algo_info,
				       &p_ppcc_hca_algo_config->encapsulation.encap_algo_info.algo_info,
				       sizeof(p_port->p_cc_data->subn_algo_info));
				p_port->p_cc_data->algo_info_array_size = p_ppcc_hca_algo_config->encap_len / 4;
			}
		} else if (p_mad_context->cc_context.mad_method == IB_MAD_METHOD_SET) {
			algo_slot = (cl_ntoh32(attr_mod) >> 24) & 0x0f;
			memcpy(&p_port->p_cc_data->subn_mads.algo_config[algo_slot], p_ppcc_hca_algo_config,
			       sizeof(ib_ppcc_hca_algo_config_t));
			cc_disable_ca_ppcc_need_update(p_port, OSM_CC_RESEND_PPCC_CA_ALGO_CONFIG_SET, algo_slot);
		}
	} else if (p_mad->attr_id == IB_MAD_ATTR_PPCC_HCA_ALGO_PARAMS_CONFIG) {
		ib_ppcc_hca_config_param_t *p_ppcc_hca_config_param = ib_cc_mad_get_mgt_data_ptr(p_cc_mad);
		algo_slot = (cl_ntoh32(attr_mod) >> 24) & 0x0f;

		memcpy(&p_port->p_cc_data->subn_mads.config_param[algo_slot], p_ppcc_hca_config_param,
		       sizeof(ib_ppcc_hca_config_param_t));
		cc_disable_ca_ppcc_need_update(p_port, OSM_CC_RESEND_PPCC_CA_ALGO_CONFIG_PARAMS, algo_slot);
	} else {
		OSM_LOG(p_cc->log, OSM_LOG_ERROR, "ERR C10A: "
			"Unexpected MAD attribute ID %u received\n",
			cl_ntoh16(p_cc_mad->header.attr_id));
	}

	cl_plock_release(&p_osm->lock);

Exit:
	decrement_outstanding_mads(p_cc);
	osm_mad_pool_put(p_cc->mad_pool, p_madw);
	OSM_LOG_EXIT(p_cc->log);
}

static void cc_poller_send(osm_congestion_control_t *p_cc,
			   osm_madw_t *p_madw)
{
	osm_subn_opt_t *p_opt = &p_cc->subn->opt;
	cl_status_t sts;
	ib_api_status_t status;
	osm_madw_context_t mad_context = p_madw->context;
	ib_net16_t attr_id = p_madw->p_mad->attr_id;
	boolean_t resp_expected = p_madw->resp_expected;

	status = osm_vendor_send(p_cc->bind_handle[p_cc->subn->sm_port_idx], p_madw, p_madw->resp_expected);
	if (status == IB_SUCCESS) {
		cc_statistic_counter_increase(attr_id, p_cc->subn);
		if (resp_expected)
			cl_atomic_inc(&p_cc->outstanding_mads_on_wire);
		while (p_cc->outstanding_mads_on_wire >
		       (int32_t)p_opt->cc_max_outstanding_mads) {
wait:
			sts = cl_event_wait_on(&p_cc->sig_mads_on_wire_continue,
					       EVENT_NO_TIMEOUT, TRUE);
			if (sts != CL_SUCCESS)
				goto wait;
		}
	} else
		OSM_LOG(p_cc->log, OSM_LOG_ERROR, "ERR C104: "
			"send failed to node 0x%" PRIx64 "port %u\n",
			cl_ntoh64(mad_context.cc_context.node_guid),
			mad_context.cc_context.port);
}



static void cc_poller(void *p_ptr)
{
	osm_congestion_control_t *p_cc = p_ptr;
	osm_madw_t *p_madw;
	ib_cc_mad_t *p_cc_mad;

	OSM_LOG_ENTER(p_cc->log);

	if (p_cc->thread_state == OSM_THREAD_STATE_NONE)
		p_cc->thread_state = OSM_THREAD_STATE_RUN;

	while (p_cc->thread_state == OSM_THREAD_STATE_RUN) {
		cl_spinlock_acquire(&p_cc->mad_queue_lock);

		p_madw = (osm_madw_t *) cl_qlist_remove_head(&p_cc->mad_queue);
		p_cc_mad = osm_madw_get_cc_mad_ptr(p_madw);

		cl_spinlock_release(&p_cc->mad_queue_lock);
		if (p_madw != (osm_madw_t *) cl_qlist_end(&p_cc->mad_queue)) {
			if (OSM_LOG_IS_ACTIVE_V2(p_cc->log, OSM_LOG_FRAMES))
				dump_cc_mad(p_cc_mad, BUF_SIZE, p_cc->log,
					    p_madw->dest_guid);
			cc_poller_send(p_cc, p_madw);
		} else {
			cl_event_wait_on(&p_cc->cc_poller_wakeup,
					 EVENT_NO_TIMEOUT, TRUE);
		}
	}

	OSM_LOG_EXIT(p_cc->log);
}

static ib_api_status_t congestion_control_send_repress(IN osm_congestion_control_t * p_cc,
						       IN const osm_madw_t * p_req_madw)
{
	osm_madw_t *resp_madw;
	ib_api_status_t status = IB_SUCCESS;
	osm_opensm_t *p_osm = p_cc->osm;
	ib_mad_t *p_resp_mad;
	ib_cc_mad_t *p_cc_mad;
	ib_net16_t slid;
	osm_port_t *p_port;

	OSM_LOG_ENTER(&p_osm->log);

	OSM_ASSERT(p_req_madw);

	/* do nothing if we are exiting ... */
	if (osm_exit_flag)
		goto Exit;

	slid = p_req_madw->mad_addr.dest_lid;
	p_port = osm_get_port_by_lid(&p_osm->subn, slid);
	if (!p_port) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR C12B: "
			"Could not find port of source lid %u\n",
			cl_ntoh16(slid));
		status = IB_INVALID_LID;
		goto Exit;
	}

	resp_madw = osm_mad_pool_get(&p_osm->mad_pool,
				     osm_madw_get_bind_handle(p_req_madw),
				     MAD_BLOCK_SIZE, NULL);
	if (resp_madw == NULL) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR C116: Unable to acquire MAD\n");
		status = IB_INSUFFICIENT_RESOURCES;
		goto Exit;
	}

	/*
	   Copy the request mad to the response mad, then just
	   update the necessary fields.
	 */
	memcpy((void *)resp_madw->p_mad, (void *)p_req_madw->p_mad, sizeof(ib_mad_t));
	p_resp_mad = (ib_mad_t *)resp_madw->p_mad;

	resp_madw->mad_addr.dest_lid = slid;
	resp_madw->mad_addr.addr_type.gsi.remote_qp = IB_QP1;
	resp_madw->mad_addr.addr_type.gsi.remote_qkey =
		IB_QP1_WELL_KNOWN_Q_KEY;

	p_resp_mad->method = IB_MAD_METHOD_TRAP_REPRESS;
	resp_madw->status = 0;
	resp_madw->resp_expected = FALSE;
	resp_madw->fail_msg = CL_DISP_MSGID_NONE;

	p_cc_mad = osm_madw_get_cc_mad_ptr(resp_madw);
	p_cc_mad->cc_key = p_port->cc_key;

	/* post trap repress to cc poller */
	cl_spinlock_acquire(&p_cc->mad_queue_lock);
	cl_qlist_insert_tail(&p_cc->mad_queue, &resp_madw->list_item);
	cl_spinlock_release(&p_cc->mad_queue_lock);

	OSM_LOG(&p_osm->log, OSM_LOG_VERBOSE, "Responding with trap repress "
		"to trap from LID %u\n", cl_ntoh16(slid));

	cl_event_signal(&p_cc->cc_poller_wakeup);

Exit:
	OSM_LOG_EXIT(&p_osm->log);
	return status;
}

static int congestion_control_create_cc_notice_1600(osm_sm_t * sm,
						    ib_mad_notice_attr_t * p_notice,
						    ib_mad_notice_attr_t * p_ntci_1600)
{
	ib_net16_t slid;
	osm_port_t *p_port;

	slid = p_notice->data_details.ntc_cckey_0.slid;
	p_port = osm_get_port_by_lid(sm->p_subn, slid);
	if (!p_port) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR C118: "
			"Could not find source physical port\n");
		return 1;
	}
	/* Copy attributes, not include data_details and issuer gid*/
	memcpy((void *)p_ntci_1600,(void *)p_notice , 10);

	p_ntci_1600->generic_type &= 0x7F;
	p_ntci_1600->data_details.vendor_data_details.v_data_details.ntc_1600.slid = slid;
	p_ntci_1600->data_details.vendor_data_details.v_data_details.ntc_1600.attribute_id =
		p_notice->data_details.ntc_cckey_0.attribute_id;
	p_ntci_1600->data_details.vendor_data_details.v_data_details.ntc_1600.method =
		p_notice->data_details.ntc_cckey_0.method;
	p_ntci_1600->data_details.vendor_data_details.v_data_details.ntc_1600.mclass = IB_MCLASS_CC;
	p_ntci_1600->data_details.vendor_data_details.v_data_details.ntc_1600.sgid =
		p_notice->data_details.ntc_cckey_0.source_gid;
	p_ntci_1600->data_details.vendor_data_details.v_data_details.ntc_1600.physp_sguid =
		p_port->guid;
	p_ntci_1600->data_details.vendor_data_details.v_data_details.ntc_1600.key =
		p_notice->data_details.ntc_cckey_0.cc_key;
	p_ntci_1600->data_details.vendor_data_details.trap_num = cl_ntoh16(1600);
	p_ntci_1600->issuer_gid = p_notice->issuer_gid;

	return 0;
}

static void congestion_control_trap_process_req(void * context, void * data)
{
	osm_congestion_control_t *p_cc = context;
	osm_opensm_t *p_osm = p_cc->osm;
	osm_sm_t *sm = &p_osm->sm;
	osm_madw_t *p_madw = data;
	ib_cc_mad_t *p_cc_mad;
	uint8_t payload[sizeof(ib_mad_notice_attr_t)];
	ib_mad_notice_attr_t *p_ntci = (ib_mad_notice_attr_t *)payload;
	ib_mad_notice_attr_t p_ntci_1600;
	osm_madw_t tmp_madw; /* we need a copy to last after repress */
	ib_api_status_t status;
	uint64_t trap_key, tid;
	uint32_t num_received;
	uint8_t method;
	ib_net16_t source_lid;
	osm_port_t * source_port;

	OSM_LOG_ENTER(&p_osm->log);

	OSM_ASSERT(p_madw);
	OSM_ASSERT(p_madw->p_mad->mgmt_class == IB_MCLASS_CC);

	memset(payload, 0, sizeof(payload));
	memset(&tmp_madw, 0, sizeof(tmp_madw));

	p_cc_mad = (ib_cc_mad_t *)p_madw->p_mad;
	method = p_madw->p_mad->method;

	if (method != IB_MAD_METHOD_TRAP) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR C11F: "
			"Unsupported method 0x%X\n", method);
		goto Exit;
	}

	memcpy(payload, p_cc_mad->mgt_data, sizeof(payload));
	p_ntci = (ib_mad_notice_attr_t *)payload;
	if (p_ntci->g_or_v.generic.trap_num !=
		CC_TRAP_NUM_KEY_VIOLATION) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR C120: "
			"Unsupported trap number 0x%X\n",
			cl_ntoh16(p_ntci->g_or_v.generic.trap_num));
		goto Exit;
	}

	memcpy(&tmp_madw, p_madw, sizeof(tmp_madw));

	osm_dump_key_violation(&p_osm->log, p_ntci, FILE_ID, OSM_LOG_VERBOSE,
			       IB_MCLASS_CC);

	source_lid = p_madw->mad_addr.dest_lid;
	source_port = osm_get_port_by_lid(sm->p_subn, source_lid);
	if (!source_port) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR C11D: "
			"Trap source port was not found by lid %u\n",
			cl_ntoh16(source_lid));
		goto Exit;
	}

	/*
	 * If SM does not hold the correct key, sending trap repress and
	 * Set KeyInfo will just generate more key violations.
	 */
	if (!source_port->p_cc_data->key_configured) {
		OSM_LOG(&p_osm->log, OSM_LOG_VERBOSE, "Ignores trap,"
			" key is not configured yet for port 0x%"PRIx64"\n",
			cl_ntoh64(source_port->guid));
		goto Exit;
	}

	status = congestion_control_send_repress(p_cc, p_madw);
	if (status != IB_SUCCESS) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR C11C: "
			"Failed sending repress (%s) to port lid %u\n",
			ib_get_err_str(status), cl_ntoh16(source_lid));
		goto Exit;
	}

	status = congestion_control_send_set_keyinfo(sm->p_subn, source_port);
	if (status != IB_SUCCESS) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR C12C:"
			" Failed sending SET KeyInfo (%s) to port lid %u,"
			" due to trap receive\n",
			ib_get_err_str(status), cl_ntoh16(source_lid));
		goto Exit;
	}

	/* try to find it in the aging tracker */
	trap_key = osm_trap_get_key(source_lid,
				    source_port->p_physp->port_num,
				    p_madw);
	num_received = cl_event_wheel_num_regs(&sm->trap_aging_tracker,
					       trap_key);
	if (num_received >= 10 &&
	    osm_print_num_received(num_received)) {
		OSM_LOG(&p_osm->log, OSM_LOG_VERBOSE,
			"Received trap %u times consecutively\n",
			num_received);
	}

	/* restart the aging anyway */
	cl_event_wheel_reg(&sm->trap_aging_tracker, trap_key,
			   cl_get_time_stamp() + OSM_CC_DEFAULT_PRINT_SUPPRESS_TIMEOUT,
			   NULL, NULL);

	/* If was already registered do nothing more */
	if (num_received >= 10)
		goto Exit;

	tid = cl_ntoh64(p_madw->p_mad->trans_id);
	if (congestion_control_create_cc_notice_1600(sm, p_ntci, &p_ntci_1600))
		goto Exit;

	cl_plock_acquire(&p_osm->lock);
	status = osm_report_notice(sm->p_log, sm->p_subn, &p_ntci_1600,
				   (uint64_t)cl_atomic_inc(&sm->p_subn->tid) &
				   (uint64_t)(0xFFFFFFFF),
				   tid);
	cl_plock_release(&p_osm->lock);
	if (status != IB_SUCCESS) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR C123: "
			"Error sending trap reports (%s)\n",
			ib_get_err_str(status));
		goto Exit;
	}
Exit:
	OSM_LOG_EXIT(&p_osm->log);
}

ib_api_status_t osm_congestion_control_init(osm_congestion_control_t * p_cc,
					    struct osm_opensm *p_osm,
					    const osm_subn_opt_t * p_opt)
{
	ib_api_status_t status = IB_SUCCESS;

	OSM_LOG_ENTER(&p_osm->log);

	memset(p_cc, 0, sizeof(*p_cc));

	p_cc->osm = p_osm;
	p_cc->subn = &p_osm->subn;
	p_cc->sm = &p_osm->sm;
	p_cc->log = &p_osm->log;
	p_cc->mad_pool = &p_osm->mad_pool;
	p_cc->trans_id = CONGESTION_CONTROL_INITIAL_TID_VALUE;
	memcpy(p_cc->vendor, p_osm->p_gmp_vendor, sizeof(p_cc->vendor));

	p_cc->cc_disp_h = cl_disp_register(&p_osm->gmp_disp, OSM_MSG_MAD_CC,
					   cc_rcv_mad, p_cc);
	if (p_cc->cc_disp_h == CL_DISP_INVALID_HANDLE)
		goto Exit;

	p_cc->cc_trap_disp_h = cl_disp_register(&p_osm->gmp_trap_disp,
						OSM_MSG_MAD_CC_NOTICE,
						congestion_control_trap_process_req,
						p_cc);
	if (p_cc->cc_trap_disp_h == CL_DISP_INVALID_HANDLE){
		OSM_LOG(p_cc->log, OSM_LOG_ERROR, "ERR C111:"
			"could not register dispatcher handler\n");
		return IB_INSUFFICIENT_RESOURCES;
	}

	cl_qlist_init(&p_cc->mad_queue);

	status = cl_spinlock_init(&p_cc->mad_queue_lock);
	if (status != IB_SUCCESS)
		goto Exit;

	cl_event_construct(&p_cc->cc_poller_wakeup);
	status = cl_event_init(&p_cc->cc_poller_wakeup, FALSE);
	if (status != IB_SUCCESS)
		goto Exit;

	cl_event_construct(&p_cc->outstanding_mads_done_event);
	status = cl_event_init(&p_cc->outstanding_mads_done_event, FALSE);
	if (status != IB_SUCCESS)
		goto Exit;

	cl_event_construct(&p_cc->sig_mads_on_wire_continue);
	status = cl_event_init(&p_cc->sig_mads_on_wire_continue, FALSE);
	if (status != IB_SUCCESS)
		goto Exit;

	p_cc->thread_state = OSM_THREAD_STATE_NONE;

	status = cl_thread_init(&p_cc->cc_poller, cc_poller, p_cc,
				"cc poller");
	if (status != IB_SUCCESS)
		goto Exit;

	p_cc->recalculate_keys = TRUE;
	p_cc->num_configured = 0;

	status = IB_SUCCESS;
Exit:
	OSM_LOG_EXIT(p_cc->log);
	return status;
}

static void cc_mad_recv_callback(osm_madw_t * p_madw, void *bind_context,
				 osm_madw_t * p_req_madw)
{
	osm_congestion_control_t *p_cc = bind_context;
	ib_net64_t dest_guid;

	OSM_LOG_ENTER(p_cc->log);

	OSM_ASSERT(p_madw);

	dest_guid = (p_madw->p_mad->method == IB_MAD_METHOD_GET_RESP) ?
		    p_req_madw->dest_guid : p_madw->dest_guid;

	if (OSM_LOG_IS_ACTIVE_V2(p_cc->log, OSM_LOG_FRAMES))
		dump_cc_mad((ib_cc_mad_t *)p_madw->p_mad, BUF_SIZE, p_cc->log,
			    dest_guid);


	/* Do not decrement outstanding mads here, do it in the dispatcher */

	switch (p_madw->p_mad->method) {
	case IB_MAD_METHOD_GET_RESP:
		OSM_ASSERT(p_req_madw != NULL);
		osm_madw_copy_context(p_madw, p_req_madw);
		osm_mad_pool_put(p_cc->mad_pool, p_req_madw);

		if (cl_disp_post(p_cc->cc_disp_h, OSM_MSG_MAD_CC,
				 p_madw, NULL, NULL) != CL_SUCCESS) {
			OSM_LOG(p_cc->log, OSM_LOG_ERROR, "ERR C105: "
				"Congestion Control Dispatcher post failed\n");
			osm_mad_pool_put(p_cc->mad_pool, p_madw);
		}
		break;
	case IB_MAD_METHOD_TRAP:
		OSM_ASSERT(p_req_madw == NULL);
		if (cl_disp_post(p_cc->cc_trap_disp_h, OSM_MSG_MAD_CC_NOTICE,
				 p_madw, NULL, NULL) != CL_SUCCESS) {
			OSM_LOG(p_cc->log, OSM_LOG_ERROR, "ERR C127: "
				"CC MAD post to management class trap "
				"dispatcher has failed\n");
			osm_mad_pool_put(p_cc->mad_pool, p_madw);
		}
		break;
	default:
		OSM_LOG(p_cc->log, OSM_LOG_ERROR, "ERR C128: "
			"Method (%u) not supported by Congestion control "
			"management class\n", p_madw->p_mad->method);
		osm_mad_pool_put(p_cc->mad_pool, p_madw);
		if (p_req_madw)
			osm_mad_pool_put(p_cc->mad_pool, p_req_madw);
	}
	OSM_LOG_EXIT(p_cc->log);
}

static void cc_mad_send_err_callback(void *bind_context,
				     osm_madw_t * p_madw)
{
	osm_congestion_control_t *p_cc = bind_context;
	osm_madw_context_t *p_madw_context = &p_madw->context;
	osm_opensm_t *p_osm = p_cc->osm;
	uint64_t node_guid = p_madw_context->cc_context.node_guid;
	uint64_t port_guid = p_madw_context->cc_context.port_guid;
	uint8_t port = p_madw_context->cc_context.port;
	osm_port_t *p_port;
	int log_flag = 1;

	OSM_LOG_ENTER(p_cc->log);

	cl_plock_acquire(&p_osm->lock);

	if (OSM_LOG_IS_ACTIVE_V2(p_cc->log, OSM_LOG_FRAMES))
		dump_cc_mad(osm_madw_get_cc_mad_ptr(p_madw), BUF_SIZE,
			    p_cc->log, p_madw->dest_guid);

	p_port = osm_get_port_by_guid(p_cc->subn, port_guid);
	if (!p_port) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR C10B: "
			"Port GUID 0x%" PRIx64 " not in table\n",
			cl_ntoh64(port_guid));
		cl_plock_release(&p_osm->lock);
		goto Exit;
	}

	/* If timed out before, don't bothering logging again
	 * we assume no CC support
	 */
	if (p_madw->status == IB_TIMEOUT
	    && p_port->cc_timeout_count)
		log_flag = 0;

	if (log_flag)
		OSM_LOG(p_cc->log, OSM_LOG_ERROR, "ERR C106: MAD Error (%s): "
			"attr id = 0x%x LID %u GUID 0x%016" PRIx64 " port %u "
			"TID 0x%" PRIx64 "\n",
			ib_get_err_str(p_madw->status),
			cl_ntoh16(p_madw->p_mad->attr_id),
			cl_ntoh16(p_madw->mad_addr.dest_lid),
			cl_ntoh64(node_guid),
			port,
			cl_ntoh64(p_madw->p_mad->trans_id));

	if (p_madw->status == IB_TIMEOUT) {
		/* CC class is currently sent only on the main port */
		p_osm->stats_counters[0][CC_MAD_TIMEOUTS_COUNT]++;
		p_port->cc_timeout_count++;
		if (p_port->cc_timeout_count > OSM_CC_TIMEOUT_COUNT_THRESHOLD
		    && !p_port->cc_unavailable_flag) {
			p_port->cc_unavailable_flag = TRUE;
			p_port->cc_timeout_count = 0;
		}
	} else {
		OSM_LOG(p_cc->log, OSM_LOG_ERROR,
			"ERR C10F: Subnet initialization error\n");
		p_cc->subn->subnet_initialization_error = TRUE;
	}

	cl_plock_release(&p_osm->lock);

	decrement_outstanding_mads(p_cc);
Exit:
	osm_mad_pool_put(p_cc->mad_pool, p_madw);
	OSM_LOG_EXIT(&p_osm->log);
}

ib_api_status_t osm_congestion_control_bind(osm_congestion_control_t * p_cc,
					    ib_net64_t port_guid)
{
	osm_bind_info_t bind_info;
	ib_api_status_t status = IB_SUCCESS;
	uint8_t port_index = p_cc->subn->sm_port_idx;

	OSM_LOG_ENTER(p_cc->log);

	bind_info.port_guid = port_guid;
	bind_info.mad_class = IB_MCLASS_CC;
	bind_info.class_version = 2;
	bind_info.is_responder = FALSE;
	bind_info.is_report_processor = FALSE;
	bind_info.is_trap_processor = TRUE;
	bind_info.recv_q_size = OSM_SM_DEFAULT_QP1_RCV_SIZE;
	bind_info.send_q_size = OSM_SM_DEFAULT_QP1_SEND_SIZE;
	bind_info.timeout = p_cc->subn->opt.transaction_timeout;
	bind_info.retries = p_cc->subn->opt.transaction_retries;

	OSM_LOG(p_cc->log, OSM_LOG_VERBOSE,
		"Binding to port GUID 0x%" PRIx64 ", port_index: %u\n",
		cl_ntoh64(port_guid), port_index);

	if (p_cc->bind_handle[port_index] == OSM_BIND_INVALID_HANDLE)
		p_cc->bind_handle[port_index] =
			osm_vendor_bind(p_cc->vendor[port_index], &bind_info,
					p_cc->mad_pool,
					cc_mad_recv_callback,
					cc_mad_send_err_callback, p_cc);
	else
		p_cc->bind_handle[port_index] =
			osm_vendor_rebind(p_cc->vendor[port_index], &bind_info,
					  p_cc->mad_pool,
					  cc_mad_recv_callback,
					  cc_mad_send_err_callback,
					  p_cc, p_cc->bind_handle[port_index]);

	if (p_cc->bind_handle[port_index] == OSM_BIND_INVALID_HANDLE) {
		status = IB_ERROR;
		OSM_LOG(p_cc->log, OSM_LOG_ERROR,
			"ERR C107: Vendor specific bind failed (%s)\n",
			ib_get_err_str(status));
		goto Exit;
	}

Exit:
	OSM_LOG_EXIT(p_cc->log);
	return status;
}

void osm_congestion_control_shutdown(osm_congestion_control_t * p_cc)
{
	uint8_t port_index;

	OSM_LOG_ENTER(p_cc->log);
	for (port_index = 0; port_index < OSM_MAX_BINDING_PORTS; port_index++) {
		if (p_cc->bind_handle[port_index] != CL_DISP_INVALID_HANDLE)
			osm_vendor_unbind(p_cc->bind_handle[port_index]);
	}
	cl_disp_unregister(p_cc->cc_disp_h);
	cl_event_signal(&p_cc->outstanding_mads_done_event);

	OSM_LOG_EXIT(p_cc->log);
}

void osm_congestion_control_destroy(osm_congestion_control_t * p_cc)
{
	osm_madw_t *p_madw;

	OSM_LOG_ENTER(p_cc->log);

	p_cc->thread_state = OSM_THREAD_STATE_EXIT;

	cl_event_signal(&p_cc->sig_mads_on_wire_continue);
	cl_event_signal(&p_cc->cc_poller_wakeup);

	cl_thread_destroy(&p_cc->cc_poller);

	cl_spinlock_acquire(&p_cc->mad_queue_lock);

	while (!cl_is_qlist_empty(&p_cc->mad_queue)) {
		p_madw = (osm_madw_t *) cl_qlist_remove_head(&p_cc->mad_queue);
		osm_mad_pool_put(p_cc->mad_pool, p_madw);
	}

	cl_spinlock_release(&p_cc->mad_queue_lock);

	cl_spinlock_destroy(&p_cc->mad_queue_lock);

	cl_event_destroy(&p_cc->cc_poller_wakeup);
	cl_event_destroy(&p_cc->outstanding_mads_done_event);
	cl_event_destroy(&p_cc->sig_mads_on_wire_continue);

	OSM_LOG_EXIT(p_cc->log);
}

void osm_congestion_control_invalidate(osm_congestion_control_t * p_cc)
{
	osm_opensm_t *p_osm = p_cc->osm;
	cl_map_item_t *p_item;
	osm_port_t *p_port;
	osm_node_t *p_node;
	osm_switch_t *p_switch;

	for (p_item = cl_qmap_head(&p_osm->subn.port_guid_tbl);
	     p_item != cl_qmap_end(&p_osm->subn.port_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_port = (osm_port_t *)p_item;
		p_node = p_port->p_node;

		if (osm_node_get_type(p_node) == IB_NODE_TYPE_CA) {
			cc_clear_ca_subn_data(p_port);
			continue;
		}

		if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH) {
			p_switch = p_node->sw;
			cc_clear_sw_subn_data(p_switch);
			continue;
		}
	}
}

static ib_api_status_t congestion_control_send_req_cpi(osm_sm_t * p_sm,
						       osm_port_t * p_port)
{
	osm_opensm_t *p_osm = p_sm->p_subn->p_osm;
	osm_madw_t *p_madw = NULL;
	osm_node_t *p_node = p_port->p_node;
	osm_physp_t *p_physp;

	OSM_LOG_ENTER(p_sm->p_log);

	if (p_port->disc_through != OSM_DISC_NORMAL)
		return IB_SUCCESS;

	if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH)
		p_physp = osm_node_get_physp_ptr(p_node, 0);
	else
		p_physp = p_port->p_physp;

	p_madw = osm_mad_pool_get(&p_osm->mad_pool,
				  p_osm->cc.bind_handle[p_sm->p_subn->sm_port_idx],
				  MAD_BLOCK_SIZE, NULL);

	if (p_madw == NULL) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR C112: "
			"failed to allocate mad\n");
		return IB_INSUFFICIENT_MEMORY;
	}

	cl_plock_acquire(&p_osm->lock);
	cc_mad_post(&p_osm->cc, p_madw, p_physp, p_port,
		    IB_MAD_ATTR_CLASS_PORT_INFO, 0, IB_MAD_METHOD_GET);

	cl_plock_release(&p_osm->lock);

	OSM_LOG_EXIT(p_sm->p_log);

	return IB_SUCCESS;
}

static void congestion_control_send_cpi(osm_sm_t * p_sm,
				        osm_port_t * p_port)
{
	ib_api_status_t status;
	boolean_t send_req_cpi = TRUE;

	if ((!p_port->p_cc_data->key_configured &&
	    (p_port->p_cc_data->resend_mask & OSM_CC_RESEND_CC_KEY_SUPPORT)) ||
	    p_sm->p_subn->coming_out_of_standby ||
	    p_sm->p_subn->first_time_master_sweep)
		goto req_cpi;

	if (p_port->is_new || p_port->p_physp->need_update ||
	    (p_port->p_node->sw && p_port->p_node->sw->need_update))
		goto req_cpi;

	/* continue key configuration */
	if (p_port->p_cc_data->resend_mask & OSM_CC_RESEND_CC_KEY_SUPPORT) {
		goto req_cpi;
	} else if (p_port->p_cc_data->resend_mask & OSM_CC_RESEND_TRAP_LID) {
		status = congestion_control_send_set_cpi(p_sm->p_subn, p_port);
		if (status != IB_SUCCESS) {
			OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR C115: "
			"failed to update port (guid 0x%016" PRIx64 ") "
			"with CC classportinfo\n",
			cl_ntoh64(p_port->guid));
		}
	}

	send_req_cpi = FALSE;
req_cpi:
	if (send_req_cpi &&
	    !(p_port->p_cc_data->resend_mask & OSM_CC_RESEND_GET_KEY_INFO)) {
		p_port->p_cc_data->resend_mask |= OSM_CC_RESEND_CC_KEY_SUPPORT;
		congestion_control_send_req_cpi(p_sm, p_port);
	}
}

static ib_api_status_t congestion_control_send_set_keyinfo(osm_subn_t * p_subn,
							   osm_port_t * p_port)
{
	osm_opensm_t *p_osm = p_subn->p_osm;
	osm_congestion_control_t *p_cc = &p_osm->cc;
	osm_madw_t *p_madw = NULL;
	ib_cc_mad_t *p_cc_mad = NULL;
	ib_cong_key_info_t p_cc_key_info_data, *p_mad_info_data = NULL;
	osm_subn_opt_t *p_opt = &p_subn->opt;
	ib_api_status_t status = IB_SUCCESS;

	OSM_LOG_ENTER(&p_osm->log);

	p_madw = osm_mad_pool_get(p_cc->mad_pool, p_cc->bind_handle[p_subn->sm_port_idx],
				  MAD_BLOCK_SIZE, NULL);
	if (!p_madw) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR C119: "
			"failed to allocate mad\n");
		status = IB_INSUFFICIENT_MEMORY;
		goto Exit;
	}

	p_cc_mad = osm_madw_get_cc_mad_ptr(p_madw);
	p_mad_info_data = ib_cc_mad_get_mgt_data_ptr(p_cc_mad);

	p_cc_key_info_data.cc_key = osm_key_mgr_generate_key(p_osm, p_port, IB_MCLASS_CC,
							     OSM_KEY_TYPE_MANAGER);
	p_cc_key_info_data.protect_bit = cl_hton16((uint16_t)p_opt->cc_key_protect_bit << 15);
	p_cc_key_info_data.lease_period = cl_hton16(p_opt->cc_key_lease_period);
	p_cc_key_info_data.violations = 0;

	memcpy(p_mad_info_data, &p_cc_key_info_data, sizeof(ib_cong_key_info_t));
	cl_plock_acquire(&p_osm->lock);
	cc_mad_post(p_cc, p_madw, p_port->p_physp, p_port,
		    IB_MAD_ATTR_CONG_KEY_INFO, 0, IB_MAD_METHOD_SET);
	cl_plock_release(&p_osm->lock);

	/* Updating the port structure after sending the keyInfo mad
	 * with previous key */
	p_port->cc_key = p_cc_key_info_data.cc_key;
	osm_key_mgr_set_cached_key(p_osm, p_port, IB_MCLASS_CC, OSM_KEY_TYPE_MANAGER);

Exit:
	OSM_LOG_EXIT(&p_osm->log);
	return status;
}

static void congestion_control_configure_keyinfo(osm_opensm_t * p_osm)
{
	osm_port_t *p_port;
	cl_map_item_t *p_item;
	ib_api_status_t status;

	OSM_LOG_ENTER(&p_osm->log);

	if (p_osm->subn.opt.cc_key_enable == OSM_CAPABILITY_IGNORED)
		return;

	/* Send KeyInfo to all ports which cckey is yet to be configured
	 * and support the feature */
	for (p_item = cl_qmap_head(&p_osm->subn.port_guid_tbl);
	     p_item != cl_qmap_end(&p_osm->subn.port_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_port = (osm_port_t *)p_item;

		if (!p_port->is_reachable || p_port->cc_unavailable_flag ||
		    (osm_node_get_type(p_port->p_node) == IB_NODE_TYPE_CA &&
		     osm_physp_get_port_state(p_port->p_physp) != IB_LINK_ACTIVE))
			continue;

		if (p_port->p_cc_data->resend_mask &
		    (OSM_CC_RESEND_CC_KEY_SUPPORT | OSM_CC_RESEND_TRAP_LID | OSM_CC_RESEND_GET_KEY_INFO))
			continue;

		if (p_port->p_cc_data->key_configured &&
		    !(p_osm->cc.recalculate_keys || p_osm->cc.cc_key_param_update)) {
			p_osm->cc.num_configured++;
			continue;
		}

		if (!congestion_control_is_cc_class_supported(p_port))
			continue;

		if ((p_port->p_cc_data->resend_mask & OSM_CC_RESEND_CC_KEY) ||
		    p_osm->cc.cc_key_param_update) {
			p_port->p_cc_data->resend_mask |= OSM_CC_RESEND_CC_KEY;
			status = congestion_control_send_set_keyinfo(&p_osm->subn,
								     p_port);
			if (status != IB_SUCCESS) {
				OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR C125: "
				"failed to update port (guid 0x%016" PRIx64 ") "
				"by CC keyinfo\n",
				cl_ntoh64(p_port->guid));
			}
		}
	}
	OSM_LOG_ENTER(&p_osm->log);
}

static void congestion_control_configure_cpi(osm_opensm_t * p_osm)
{
	osm_port_t *p_port;
	cl_map_item_t *p_item;

	for (p_item = cl_qmap_head(&p_osm->subn.port_guid_tbl);
	     p_item != cl_qmap_end(&p_osm->subn.port_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_port = (osm_port_t *)p_item;
		if (!p_port->is_reachable ||
		    (osm_node_get_type(p_port->p_node) == IB_NODE_TYPE_CA &&
		     osm_physp_get_port_state(p_port->p_physp) != IB_LINK_ACTIVE))
			continue;
		if (congestion_control_is_cc_class_supported(p_port) &&
		    !p_port->cc_unavailable_flag &&
		    (p_port->p_cc_data->resend_mask &
		     (OSM_CC_RESEND_CC_KEY_SUPPORT | OSM_CC_RESEND_TRAP_LID)))
			congestion_control_send_cpi(&p_osm->sm, p_port);
	}
}

static void congestion_control_req_keys(struct osm_opensm * p_osm)
{
	osm_port_t *p_port;
	cl_map_item_t *p_item;

	/* Send KeyInfo Get to all relevant ports in order to get current CC keys */
	for (p_item = cl_qmap_head(&p_osm->subn.port_guid_tbl);
	     p_item != cl_qmap_end(&p_osm->subn.port_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_port = (osm_port_t *)p_item;
		if (!p_port->is_reachable ||
		    (osm_node_get_type(p_port->p_node) == IB_NODE_TYPE_CA &&
		     osm_physp_get_port_state(p_port->p_physp) != IB_LINK_ACTIVE) ||
		    !congestion_control_is_cc_class_supported(p_port) ||
		    !(p_port->p_cc_data->resend_mask & OSM_CC_RESEND_GET_KEY_INFO)) {
			continue;
		}

		if (p_port->p_cc_data->key_configured &&
		    !(p_port->p_cc_data->resend_mask & OSM_CC_RESEND_GET_KEY_INFO))
			continue;

		if (congestion_control_send_req_keyinfo(&p_osm->sm, p_port)) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR C124: "
			"failed to update port (guid 0x%016" PRIx64 ") "
			"by CC keyinfo\n",
			cl_ntoh64(p_port->guid));
		}
	}
}

int osm_congestion_control_configure_subnet_keys(struct osm_opensm * p_osm)
{
	int ret = 0;

	OSM_LOG_ENTER(&p_osm->log);

	p_osm->cc.num_configured = 0;

	congestion_control_req_keys(p_osm);
	if (osm_congestion_control_wait_pending_transactions(p_osm)) {
		ret = 1;
		goto Exit;
	}

	congestion_control_configure_cpi(p_osm);
	if (osm_congestion_control_wait_pending_transactions(p_osm)) {
		ret = 1;
		goto Exit;
	}

	congestion_control_configure_keyinfo(p_osm);
	if (osm_congestion_control_wait_pending_transactions(p_osm)) {
		ret = 1;
		goto Exit;
	}

	p_osm->cc.cc_key_param_update = FALSE;
	p_osm->cc.recalculate_keys = FALSE;

	if (p_osm->subn.opt.cc_key_enable == OSM_CAPABILITY_ENABLED)
		OSM_LOG(&p_osm->log, OSM_LOG_VERBOSE,
			"%u out of %d ports were successfully configured for CC key\n",
			p_osm->cc.num_configured, cl_qmap_count(&p_osm->subn.port_guid_tbl));

Exit:
	OSM_LOG_EXIT(&p_osm->log);
	return ret;
}
