/*
 * Copyright (c) 2021-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * Copyright (c) 2004-2009 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2006 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <complib/cl_debug.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_DB_PACK_C
#include <opensm/osm_db_pack.h>
#include <opensm/osm_log.h>
#include <opensm/osm_helper.h>
#include <sys/time.h>

#define LIST_INIT_MIN_ITEMS 10
#define PORT_PAIR_MAX_LENGTH 10
#define GUID_PAIR_MAX_LENGTH 40
#define KEYS_MAX_LENGTH 40

static inline void pack_guid(uint64_t guid, char *p_guid_str)
{
	sprintf(p_guid_str, "0x%016" PRIx64, guid);
}

static inline uint64_t unpack_guid(char *p_guid_str)
{
	return strtoull(p_guid_str, NULL, 0);
}

static inline uint8_t unpack_port(char *p_num)
{
	unsigned long tmp_port = strtoul(p_num, NULL, 0);
	OSM_ASSERT(tmp_port < 0x100);
	return (uint8_t)tmp_port;
}

static inline void pack_lids(uint16_t min_lid, uint16_t max_lid, char *lid_str)
{
	sprintf(lid_str, "0x%04x 0x%04x", min_lid, max_lid);
}

static inline int unpack_lids(IN char *p_lid_str, OUT uint16_t * p_min_lid,
			      OUT uint16_t * p_max_lid)
{
	unsigned long tmp;
	char *p_next;
	char *p_num;
	char lids_str[24];

	strncpy(lids_str, p_lid_str, 23);
	lids_str[23] = '\0';
	p_num = strtok_r(lids_str, " \t", &p_next);
	if (!p_num)
		return 1;
	tmp = strtoul(p_num, NULL, 0);
	if (tmp >= 0xC000)
		return 1;

	*p_min_lid = (uint16_t) tmp;

	p_num = strtok_r(NULL, " \t", &p_next);
	if (!p_num)
		return 1;
	tmp = strtoul(p_num, NULL, 0);
	if (tmp >= 0xC000)
		return 1;

	*p_max_lid = (uint16_t) tmp;

	return 0;
}

static inline void pack_flid(uint16_t flid, char *flid_str)
{
	sprintf(flid_str, "0x%04x", flid);
}

static inline int unpack_flid(IN char *p_flid_str, OUT uint16_t * p_flid)
{
	unsigned long tmp;

	tmp = strtoul(p_flid_str, NULL, 0);
	if (tmp > IB_LID_UCAST_END_HO)
		return 1;

	*p_flid = (uint16_t) tmp;

	return 0;
}

static inline long int unpack_flid_key(IN char *p_flid_str)
{
	return strtol(p_flid_str, NULL, 16);
}

static inline void pack_neighbor(uint64_t guid, uint8_t portnum, char *p_str)
{
	sprintf(p_str, "0x%016" PRIx64 ":%u", guid, portnum);
}

static inline int unpack_neighbor(char *p_str, uint64_t *guid,
				  uint8_t *portnum)
{
	char tmp_str[24];
	char *p_num, *p_next;
	unsigned long tmp_port;

	strncpy(tmp_str, p_str, 23);
	tmp_str[23] = '\0';
	p_num = strtok_r(tmp_str, ":", &p_next);
	if (!p_num)
		return 1;
	if (guid)
		*guid = strtoull(p_num, NULL, 0);

	p_num = strtok_r(NULL, ":", &p_next);
	if (!p_num)
		return 1;
	if (portnum) {
		tmp_port = strtoul(p_num, NULL, 0);
		OSM_ASSERT(tmp_port < 0x100);
		*portnum = (uint8_t) tmp_port;
	}

	return 0;
}

static inline void pack_group_id(uint16_t group_ids[3], char *group_id_str)
{
	sprintf(group_id_str, "0x%04x,0x%04x,0x%04x", group_ids[0], group_ids[1], group_ids[2]);
}

static inline int unpack_group_id(IN char *p_group_id_str,
				  OUT uint16_t group_ids[3])
{
	unsigned tmp0, tmp1, tmp2;
	int n;

	tmp0 = tmp1 = tmp2 = 0xFFFF;

	n = sscanf(p_group_id_str, "0x%x,0x%x,0x%x", &tmp0, &tmp1, &tmp2);
	if (n < 1)
		return 1;

	if (tmp0 > 0xFFFF || tmp1 > 0xFFFF || tmp2 > 0xFFFF)
		return 1;

	group_ids[0] = tmp0;
	group_ids[1] = tmp1;
	group_ids[2] = tmp2;

	return 0;
}

static inline void pack_flid_group_id(uint16_t group_id, char *group_id_str)
{
	sprintf(group_id_str, "0x%04x", group_id);
}

static inline int unpack_flid_group_id(IN char *p_group_id_str,	OUT uint16_t *p_group_id)
{
	unsigned long tmp;

	tmp = strtoul(p_group_id_str, NULL, 0);
	if (tmp >= 0xFFFF)
		return 1;

	*p_group_id = (uint16_t) tmp;

	return 0;
}

/*
 * Split a string into 2 tokens using a given delimiter. If the delimiter exists
 * more than once, then the splitting is done by the 1st occurrence. If the
 * delimiter is not found then 1 is returned, and the the result tokens are
 * undefined.
 * The caller should pass a mutable buffer at least as long as the input str.
 */
static
inline int split_str_to_2(char *p_str, char *buf, int buf_size,
			  char* delimiter, char **p_token1, char **p_token2)
{
	char *p_next;

	strncpy(buf, p_str, buf_size - 1);
	*p_token1 = strtok_r(buf, delimiter, &p_next);
	if (!*p_token1)
		return 1;

	*p_token2 = strtok_r(NULL, "", &p_next);
	if (!*p_token2)
		return 1;
	return 0;
}

static inline void pack_guid_pair(uint64_t guid1, uint64_t guid2, char *p_str)
{
	sprintf(p_str, "0x%016" PRIx64 "-0x%016" PRIx64,
		cl_hton64(guid1), cl_hton64(guid2));
}

static inline void pack_time_ports_pair(char* time, char* ports, char *p_str)
{
	sprintf(p_str, "%s-%s", time, ports);
}

static inline void pack_port_pair(uint8_t port1, uint8_t port2, char *p_str)
{
	sprintf(p_str, "%u-%u", port1, port2);
}

static inline
int unpack_guid_pair(char *p_str, uint64_t *guid1, uint64_t *guid2)
{
	char buf[GUID_PAIR_MAX_LENGTH], *p_token1, *p_token2;

	if (split_str_to_2(p_str, buf, sizeof(buf), "-",
			   &p_token1, &p_token2) == 1)
		return 1;

	*guid1 = cl_ntoh64(unpack_guid(p_token1));
	*guid2 = cl_ntoh64(unpack_guid(p_token2));
	return 0;
}

static inline
int unpack_port_pair(char *p_str, uint8_t *port1, uint8_t *port2)
{
	char buf[PORT_PAIR_MAX_LENGTH], *p_token1, *p_token2;

	if (split_str_to_2(p_str, buf, sizeof(buf), "-",
			   &p_token1, &p_token2) == 1)
		return 1;

	*port1 = unpack_port(p_token1);
	*port2 = unpack_port(p_token2);
	return 0;
}

static inline
int unpack_time_ports_pair(char *p_str, uint8_t *port1, uint8_t *port2, char *time)
{
	char buf[PORT_PAIR_MAX_LENGTH + MAX_DATE_STR_LEN + 1], *p_token1, *p_token2;

	if (split_str_to_2(p_str, buf, sizeof(buf), "-",
			   &p_token1, &p_token2) == 1)
		return 1;

	strncpy(time, p_token1, MAX_DATE_STR_LEN);
	if(unpack_port_pair(p_token2, port1, port2))
		return 1;
	return 0;
}

static inline void pack_key(uint64_t key, char *p_keys_str)
{
	sprintf(p_keys_str, "0x%016" PRIx64, key);
}

static inline uint64_t unpack_key(char *p_key_str)
{
	return strtoull(p_key_str, NULL, 0);
}


/******************************************************************************/
/*                                guid2lid                                    */
/******************************************************************************/

int osm_db_guids(IN osm_db_domain_t * p_domain,
		 OUT cl_qlist_t * p_guid_list)
{
	char *p_key;
	cl_list_t keys;
	osm_db_guid_elem_t *p_guid_elem;

	cl_list_construct(&keys);
	cl_list_init(&keys, LIST_INIT_MIN_ITEMS);

	if (osm_db_keys(p_domain, &keys))
		return 1;

	while ((p_key = cl_list_remove_head(&keys)) != NULL) {
		p_guid_elem =
		    (osm_db_guid_elem_t *)calloc(1, sizeof(osm_db_guid_elem_t));
		OSM_ASSERT(p_guid_elem != NULL);

		p_guid_elem->guid = unpack_guid(p_key);
		cl_qlist_insert_head(p_guid_list, &p_guid_elem->item);
	}

	cl_list_destroy(&keys);
	return 0;
}

int osm_db_guid2lid_get(IN osm_db_domain_t * p_g2l, IN uint64_t guid,
			OUT uint16_t * p_min_lid, OUT uint16_t * p_max_lid)
{
	char guid_str[20];
	char *p_lid_str;
	uint16_t min_lid, max_lid;

	pack_guid(guid, guid_str);
	p_lid_str = osm_db_lookup(p_g2l, guid_str);
	if (!p_lid_str)
		return 1;
	if (unpack_lids(p_lid_str, &min_lid, &max_lid))
		return 1;

	if (p_min_lid)
		*p_min_lid = min_lid;
	if (p_max_lid)
		*p_max_lid = max_lid;

	return 0;
}

int osm_db_guid2lid_set(IN osm_db_domain_t * p_g2l, IN uint64_t guid,
			IN uint16_t min_lid, IN uint16_t max_lid)
{
	char guid_str[20];
	char lid_str[16];

	pack_guid(guid, guid_str);
	pack_lids(min_lid, max_lid, lid_str);

	return osm_db_update(p_g2l, guid_str, lid_str);
}

int osm_db_guid2lid_delete(IN osm_db_domain_t * p_g2l, IN uint64_t guid)
{
	char guid_str[20];
	pack_guid(guid, guid_str);
	return osm_db_delete(p_g2l, guid_str);
}

/******************************************************************************/
/*                                guid2flid                                   */
/******************************************************************************/

int osm_db_guid2flid_get(IN osm_db_domain_t * p_g2fl, IN uint64_t guid,
			 OUT uint16_t * p_flid)
{
	char guid_str[20];
	char *p_flid_str;
	uint16_t flid;

	pack_guid(guid, guid_str);
	p_flid_str = osm_db_lookup(p_g2fl, guid_str);
	if (!p_flid_str)
		return 1;
	if (unpack_flid(p_flid_str, &flid))
		return 1;

	if (p_flid)
		*p_flid = flid;

	return 0;
}

int osm_db_guid2flid_set(IN osm_db_domain_t * p_g2fl, IN uint64_t guid, IN uint16_t flid)
{
	char guid_str[20];
	char flid_str[16];

	pack_guid(guid, guid_str);
	pack_flid(flid, flid_str);

	return osm_db_update(p_g2fl, guid_str, flid_str);
}

int osm_db_guid2flid_delete(IN osm_db_domain_t * p_g2fl, IN uint64_t guid)
{
	char guid_str[20];
	pack_guid(guid, guid_str);
	return osm_db_delete(p_g2fl, guid_str);
}

/******************************************************************************/
/*                                neighbor                                    */
/******************************************************************************/

int osm_db_neighbor_guids(IN osm_db_domain_t * p_neighbor,
			  OUT cl_qlist_t * p_neighbor_list)
{
	char *p_key;
	cl_list_t keys;
	osm_db_neighbor_elem_t *p_neighbor_elem;

	cl_list_construct(&keys);
	cl_list_init(&keys, LIST_INIT_MIN_ITEMS);

	if (osm_db_keys(p_neighbor, &keys))
		return 1;

	while ((p_key = cl_list_remove_head(&keys)) != NULL) {
		p_neighbor_elem =
		    (osm_db_neighbor_elem_t *)calloc(1, sizeof(osm_db_neighbor_elem_t));
		OSM_ASSERT(p_neighbor_elem != NULL);

		unpack_neighbor(p_key, &p_neighbor_elem->guid,
				&p_neighbor_elem->portnum);
		cl_qlist_insert_head(p_neighbor_list, &p_neighbor_elem->item);
	}

	cl_list_destroy(&keys);
	return 0;
}

int osm_db_neighbor_get(IN osm_db_domain_t * p_neighbor, IN uint64_t guid1,
			IN uint8_t portnum1, OUT uint64_t * p_guid2,
			OUT uint8_t * p_portnum2)
{
	char neighbor_str[24];
	char *p_other_str;
	uint64_t temp_guid;
	uint8_t temp_portnum;

	pack_neighbor(guid1, portnum1, neighbor_str);
	p_other_str = osm_db_lookup(p_neighbor, neighbor_str);
	if (!p_other_str)
		return 1;
	if (unpack_neighbor(p_other_str, &temp_guid, &temp_portnum))
		return 1;

	if (p_guid2)
		*p_guid2 = temp_guid;
	if (p_portnum2)
		*p_portnum2 = temp_portnum;

	return 0;
}

int osm_db_neighbor_set(IN osm_db_domain_t * p_neighbor, IN uint64_t guid1,
			IN uint8_t portnum1, IN uint64_t guid2,
			IN uint8_t portnum2)
{
	char n1_str[24], n2_str[24];

	pack_neighbor(guid1, portnum1, n1_str);
	pack_neighbor(guid2, portnum2, n2_str);

	return osm_db_update(p_neighbor, n1_str, n2_str);
}

int osm_db_neighbor_delete(IN osm_db_domain_t * p_neighbor, IN uint64_t guid,
			   IN uint8_t portnum)
{
	char n_str[24];

	pack_neighbor(guid, portnum, n_str);
	return osm_db_delete(p_neighbor, n_str);
}

/******************************************************************************/
/*                                  an2an                                     */
/******************************************************************************/

int osm_db_an2an_get_all(IN osm_db_domain_t * p_domain,
			 OUT cl_qlist_t * p_elements)
{
	char *p_key, *p_value;
	cl_list_t keys;
	osm_db_an2an_elem_t *p_elem;

	cl_list_construct(&keys);
	cl_list_init(&keys, LIST_INIT_MIN_ITEMS);

	if (osm_db_keys(p_domain, &keys))
		return 1;

	while ((p_key = cl_list_remove_head(&keys)) != NULL) {
		p_elem =
		    (osm_db_an2an_elem_t *)calloc(1, sizeof(osm_db_an2an_elem_t));
		OSM_ASSERT(p_elem != NULL);
		p_value = osm_db_lookup(p_domain, p_key);

		if (unpack_guid_pair(p_key, &p_elem->guid1, &p_elem->guid2) ||
		    unpack_time_ports_pair(p_value, &p_elem->portnum1,
					   &p_elem->portnum2, p_elem->timestamp_str))
			free(p_elem);
		else
			cl_qlist_insert_head(p_elements, &p_elem->item);
	}

	cl_list_destroy(&keys);
	return 0;
}

int osm_db_an2an_get(IN osm_db_domain_t * p_domain,
		     IN uint64_t sw_guid1, IN uint64_t sw_guid2,
		     OUT uint8_t * port1, OUT uint8_t * port2, OUT char *time)
{
	char p_key[40];
	char *p_value;

	pack_guid_pair(sw_guid1, sw_guid2, p_key);
	p_value = osm_db_lookup(p_domain, p_key);
	if (!p_value)
		return 1;
	return unpack_time_ports_pair(p_value, port1, port2, time);
}

static inline void update_timestamp(char *time)
{
	struct timeval tv;
	gettimeofday(&tv, NULL);
	sprint_time_str(&tv, time, MAX_DATE_STR_LEN, 1);
}

int osm_db_an2an_set(IN osm_db_domain_t * p_domain,
		     IN uint64_t sw_guid1, IN uint64_t sw_guid2,
		     IN uint8_t port1, IN uint8_t port2)
{
	char p_key[GUID_PAIR_MAX_LENGTH], p_ports[PORT_PAIR_MAX_LENGTH];
	char p_time[MAX_DATE_STR_LEN] = { 0 };
	char p_value[MAX_DATE_STR_LEN + 1 + PORT_PAIR_MAX_LENGTH];

	pack_guid_pair(sw_guid1, sw_guid2, p_key);
	pack_port_pair(port1, port2, p_ports);
	update_timestamp(p_time);

	pack_time_ports_pair(p_time, p_ports, p_value);
	return osm_db_update(p_domain, p_key, p_value);
}

int osm_db_an2an_delete(IN osm_db_domain_t * p_domain,
			IN uint64_t sw_guid1, IN uint64_t sw_guid2)
{
	char p_key[40];

	pack_guid_pair(sw_guid1, sw_guid2, p_key);
	return osm_db_delete(p_domain, p_key);
}

/******************************************************************************/
/*                                  guid2key                                  */
/******************************************************************************/
int osm_db_guid2key_get(IN osm_db_domain_t * p_g2key, IN uint64_t guid,
			OUT uint64_t * key)
{
	char guid_str[20];
	char *p_key_str;

	pack_guid(guid, guid_str);
	p_key_str = osm_db_lookup(p_g2key, guid_str);
	if (!p_key_str)
		return 1;

	if (key)
		*key = unpack_key(p_key_str);

	return 0;
}

int osm_db_guid2key_set(IN osm_db_domain_t * p_g2key, IN uint64_t guid,
			 IN uint64_t key)
{
	char guid_str[20];
	char key_str[20];

	pack_guid(guid, guid_str);
	pack_key(key, key_str);

	return osm_db_update(p_g2key, guid_str, key_str);
}

int osm_db_guid2key_delete(IN osm_db_domain_t * p_g2key, IN uint64_t guid)
{
	char guid_str[20];

	pack_guid(guid, guid_str);
	return osm_db_delete(p_g2key, guid_str);
}

/******************************************************************************/
/*                                  guid2groupid                             */
/******************************************************************************/

int osm_db_guid2groupid_get(IN osm_db_domain_t * p_g2groupid, IN uint64_t guid,
			    OUT uint16_t group_ids[3])
{
	char guid_str[20];
	char *p_group_id_str;

	pack_guid(guid, guid_str);
	p_group_id_str = osm_db_lookup(p_g2groupid, guid_str);
	if (!p_group_id_str)
		return 1;

	if (unpack_group_id(p_group_id_str, group_ids))
		return 1;

	return 0;
}

int osm_db_guid2groupid_set(IN osm_db_domain_t * p_g2groupid, IN uint64_t guid,
			    IN uint16_t group_ids[3])
{
	char guid_str[20];
	char group_id_str[32];

	pack_guid(guid, guid_str);
	pack_group_id(group_ids, group_id_str);

	return osm_db_update(p_g2groupid, guid_str, group_id_str);
}

int osm_db_guid2groupid_delete(IN osm_db_domain_t * p_g2groupid, IN uint64_t guid)
{
	char guid_str[20];
	pack_guid(guid, guid_str);

	return osm_db_delete(p_g2groupid, guid_str);
}

/******************************************************************************/
/*                                  flid2groupid                             */
/******************************************************************************/
int osm_db_flid2groupid_flids(IN osm_db_domain_t * p_flid2groupid,
			      OUT cl_qlist_t * p_flid_list)
{
	char *p_key;
	cl_list_t keys;
	osm_db_flid_elem_t *p_flid_elem;

	cl_list_construct(&keys);
	cl_list_init(&keys, LIST_INIT_MIN_ITEMS);

	if (osm_db_keys(p_flid2groupid, &keys))
		return 1;

	while ((p_key = cl_list_remove_head(&keys)) != NULL) {
		p_flid_elem =
		    (osm_db_flid_elem_t *)calloc(1, sizeof(osm_db_flid_elem_t));
		OSM_ASSERT(p_flid_elem != NULL);

		p_flid_elem->flid = (uint16_t)unpack_flid_key(p_key);
		cl_qlist_insert_head(p_flid_list, &p_flid_elem->item);
	}

	cl_list_destroy(&keys);
	return 0;
}

int osm_db_flid2groupid_get(IN osm_db_domain_t * p_flid2groupid, IN uint16_t flid,
			    OUT uint16_t * p_group_id)
{
	char flid_str[16];
	char *p_group_id_str;
	uint16_t group_id;

	pack_flid(flid, flid_str);
	p_group_id_str = osm_db_lookup(p_flid2groupid, flid_str);
	if (!p_group_id_str)
		return 1;

	if (unpack_flid_group_id(p_group_id_str, &group_id))
		return 1;

	if (p_group_id)
		*p_group_id = group_id;

	return 0;
}

int osm_db_flid2groupid_set(IN osm_db_domain_t * p_flid2groupid, IN uint16_t flid,
			    IN uint16_t group_id)
{
	char flid_str[16];
	char group_id_str[16];

	pack_flid(flid, flid_str);
	pack_flid_group_id(group_id, group_id_str);

	return osm_db_update(p_flid2groupid, flid_str, group_id_str);
}

int osm_db_flid2groupid_delete(IN osm_db_domain_t * p_flid2groupid, IN uint16_t flid)
{
	char flid_str[16];

	pack_flid(flid, flid_str);
	return osm_db_delete(p_flid2groupid, flid_str);
}

/******************************************************************************/
/*                                  guid2plane                                 */
/******************************************************************************/

static inline void pack_plane(uint8_t plane, char * plane_str)
{
	sprintf(plane_str, "%u", plane);
}

static inline int unpack_plane(IN char * p_plane_str,
			       OUT uint8_t * plane)
{
	unsigned tmp;
	int n;

	n = sscanf(p_plane_str, "%u", &tmp);
	if (n < 1)
		return 1;

	if (tmp > 0xFF)
		return 1;

	*plane = tmp;

	return 0;
}

int osm_db_guid2plane_get(IN osm_db_domain_t * p_g2plane, IN uint64_t guid,
			  OUT uint8_t * plane)
{
	char guid_str[20];
	char *p_plane_str;

	pack_guid(guid, guid_str);
	p_plane_str = osm_db_lookup(p_g2plane, guid_str);
	if (!p_plane_str)
		return 1;

	if (unpack_plane(p_plane_str, plane))
		return 1;

	return 0;
}

int osm_db_guid2plane_set(IN osm_db_domain_t * p_g2plane, IN uint64_t guid,
			  IN uint8_t plane)
{
	char guid_str[20];
	char plane_str[32];

	pack_guid(guid, guid_str);
	pack_plane(plane, plane_str);

	return osm_db_update(p_g2plane, guid_str, plane_str);
}

int osm_db_guid2plane_delete(IN osm_db_domain_t * p_g2plane, IN uint64_t guid)
{
	char guid_str[20];
	pack_guid(guid, guid_str);

	return osm_db_delete(p_g2plane, guid_str);
}
