/*
 * Copyright (c) 2021-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * Copyright (c) 2009 Sun Microsystems, Inc. All rights reserved.
 * Copyright (c) 2004-2009 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2011 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Various OpenSM dumpers
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <iba/ib_types.h>
#include <complib/cl_qmap.h>
#include <complib/cl_debug.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_DUMP_C
#include <opensm/osm_opensm.h>
#include <opensm/osm_db_pack.h>
#include <opensm/osm_log.h>
#include <opensm/osm_node.h>
#include <opensm/osm_switch.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_msgdef.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_vport.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_router.h>
#include <opensm/osm_version.h>
#include <opensm/osm_unhealthy_ports.h>
#include <opensm/osm_an2an.h>
#include <opensm/osm_gpu.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>

#define LID_MATRIX_DUMP_FILE "opensm-lid-matrix"
#define LFT_DUMP_FILE "opensm-lfts"
#define FDBS_DUMP_FILE "opensm.fdbs"
#define MCFDBS_DUMP_FILE "opensm.mcfdbs"
#define SL2VL_DUMP_FILE "opensm-sl2vl.dump"
#define SUBNET_LST_DUMP_FILE "opensm-subnet"
#define VIRT_DUMP_FILE "opensm-virtualization.dump"
#define ROUTERS_DUMP_FILE "opensm-routers.dump"
#define SMDB_DUMP_FILE "opensm-smdb"
#define AR_DUMP_FILE "opensm.far"
#define PERFLOG_DUMP_FILE "opensm-perflog"
#define PERFLOG_DELAY_PREFIX 	"SM_PERFLOG_DELAY"

/*
 * SMDB_FILE_VERSION FORMAT: <N1.N2>
 * When removing field/table, increase N1, and reset N2.
 * When Adding new field/table, increase N2.
 */
#define SMDB_FILE_VERSION "1.4"
#define SMDB_INDEX_TABLE_NAME "INDEX_TABLE"

#define MAX_NAME_LEN 128
#define MAX_STR_LEN 64
#define MAX_DATE_STR_LEN 32
#define MAX_TOPOLOGY_INFO_LEN 32
static uint64_t virt_dumped_port_counter = 0;

typedef struct dump_ctx {
	osm_opensm_t *p_osm;
	boolean_t rch_dump;
} dump_ctx_t;


typedef unsigned int(*dump_smdb_tbl_pfn_t)(FILE *, void *);

typedef struct smdb_tbl_record {
	const char *name;
	dump_smdb_tbl_pfn_t dump_tbl_pfn;
	boolean_t sub_topology_table;
	boolean_t (*display_section) (struct osm_subn *p_subn);
	long int offset;
	long int size;
	long int line;
	long int rows;
} smdb_tbl_record_t;

typedef struct smdb_tbl_context {
	osm_subn_t *p_subn;
	smdb_tbl_record_t *p_records;
	boolean_t dump_all;
} smdb_tbl_context_t;

static inline char *get_physp_speed_act_str(osm_physp_t *p_physp)
{
	osm_node_t *p_node = p_physp->p_node;
	char *link_speed_act_str;
	ib_net32_t capability_mask;
	ib_net16_t capability_mask2;
	uint8_t link_speed_act;

	if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH) {
		capability_mask =
		    osm_node_get_physp_ptr(p_node, 0)->port_info.capability_mask;
		capability_mask2 =
		    osm_node_get_physp_ptr(p_node, 0)->port_info.capability_mask2;
	} else {
		capability_mask = p_physp->port_info.capability_mask;
		capability_mask2 = p_physp->port_info.capability_mask2;
	}

	link_speed_act = ib_port_info_get_link_speed_active(&p_physp->port_info);

	if (link_speed_act == IB_LINK_SPEED_ACTIVE_2_5)
		link_speed_act_str = "2.5";
	else if (link_speed_act == IB_LINK_SPEED_ACTIVE_5)
		link_speed_act_str = "5";
	else if (link_speed_act == IB_LINK_SPEED_ACTIVE_10)
		link_speed_act_str = "10";
	else
		link_speed_act_str = "??";

	if (p_physp->ext_port_info.link_speed_active & FDR10_BIT)
		link_speed_act_str = "FDR10";

	if (capability_mask & IB_PORT_CAP_HAS_EXT_SPEEDS) {
		link_speed_act =
			ib_port_info_get_link_speed_ext_active(&p_physp->port_info);
		if (link_speed_act == IB_LINK_SPEED_EXT_ACTIVE_14)
			link_speed_act_str = "14";
		else if (link_speed_act == IB_LINK_SPEED_EXT_ACTIVE_25)
			link_speed_act_str = "25";
		else if (link_speed_act == IB_LINK_SPEED_EXT_ACTIVE_50)
			link_speed_act_str = "50";
		else if (link_speed_act == IB_LINK_SPEED_EXT_ACTIVE_100)
			link_speed_act_str = "100";
		else if (link_speed_act != IB_LINK_SPEED_EXT_ACTIVE_NONE)
			link_speed_act_str = "??";
	}

	if ((capability_mask & IB_PORT_CAP_HAS_CAP_MASK2) &&
	    (capability_mask2 & IB_PORT_CAP2_IS_EXT_SPEEDS2_SUPPORTED)) {
		link_speed_act = ib_port_info_get_link_speed_ext_active2(&p_physp->port_info);
		if (link_speed_act == IB_LINK_SPEED_EXT2_ACTIVE_200)
			link_speed_act_str = "200";
		else if (link_speed_act != IB_LINK_SPEED_EXT2_ACTIVE_NONE)
			link_speed_act_str = "??";
	}

	return link_speed_act_str;
}

static inline char *get_physp_link_width_act_str(osm_physp_t *p_physp,
						 ib_net16_t cap_mask2)
{
	char *link_width_act_str;

	if (p_physp->port_info.link_width_active == 1)
		link_width_act_str = "1x";
	else if (p_physp->port_info.link_width_active == 2)
		link_width_act_str = "4x";
	else if (p_physp->port_info.link_width_active == 4)
		link_width_act_str = "8x";
	else if (p_physp->port_info.link_width_active == 8)
		link_width_act_str = "12x";
	else link_width_act_str = "??";

	if (cap_mask2 & IB_PORT_CAP2_IS_LINK_WIDTH_2X_SUPPORTED) {
		if (p_physp->port_info.link_width_active == 16)
			link_width_act_str = "2x";
	}

	return link_width_act_str;
}

static inline char *get_physp_port_state_str(osm_physp_t *p_physp)
{
	uint8_t port_state = ib_port_info_get_port_state(&p_physp->port_info);

	return (port_state == IB_LINK_ACTIVE ? "ACT" :
		port_state == IB_LINK_ARMED ? "ARM" :
		port_state == IB_LINK_INIT ? "INI" : "DWN");
}

static inline void get_date_str(char *str, int len, boolean_t full)
{
	struct timeval tv;

	*str = '\0';
	gettimeofday(&tv, NULL);
	sprint_time_str(&tv, str, len, full);
}

static inline char *get_disc_state_str(disc_state_t disc_state)
{
	switch (disc_state) {
		case OSM_NOT_DISC:
			return "NONE";
		case OSM_DISC_IGNORED:
			return "IGNORED";
		case OSM_DISC_NORMAL:
			return "NORMAL";
		default:
			return "N/A";
	}
}

static void dump_ucast_path_distribution(cl_map_item_t * item, FILE * file,
					 void *cxt)
{
	osm_node_t *p_node;
	osm_node_t *p_remote_node;
	uint8_t i;
	uint8_t num_ports;
	uint32_t num_paths;
	ib_net64_t remote_guid_ho;
	osm_switch_t *p_sw = (osm_switch_t *) item;

	p_node = p_sw->p_node;
	num_ports = p_sw->num_ports;

	fprintf(file, "dump_ucast_path_distribution: Switch 0x%" PRIx64 "\n"
		"Port : Path Count Through Port",
		cl_ntoh64(osm_node_get_node_guid(p_node)));

	for (i = 0; i < num_ports; i++) {
		num_paths = osm_switch_path_count_get(p_sw, i);
		fprintf(file, "\n %03u : %u", i, num_paths);
		if (i == 0) {
			fprintf(file, " (switch management port)");
			continue;
		}

		p_remote_node = osm_node_get_remote_node(p_node, i, NULL);
		if (p_remote_node == NULL)
			continue;

		remote_guid_ho =
		    cl_ntoh64(osm_node_get_node_guid(p_remote_node));

		switch (osm_node_get_type(p_remote_node)) {
		case IB_NODE_TYPE_SWITCH:
			fprintf(file, " (link to switch");
			break;
		case IB_NODE_TYPE_ROUTER:
			fprintf(file, " (link to router");
			break;
		case IB_NODE_TYPE_CA:
			fprintf(file, " (link to CA");
			break;
		default:
			fprintf(file, " (link to unknown node type");
			break;
		}

		fprintf(file, " 0x%" PRIx64 ")", remote_guid_ho);
	}

	fprintf(file, "\n");
}

static void dump_ucast_routes(cl_map_item_t * item, FILE * file, void *dump_cxt)
{
	const osm_node_t *p_node;
	osm_port_t *p_port;
	uint8_t port_num;
	uint8_t num_hops;
	uint8_t best_hops;
	uint8_t best_port;
	uint16_t max_lid_ho;
	uint16_t lid_ho, base_lid;
	boolean_t direct_route_exists = FALSE;
	boolean_t dor;
	osm_switch_t *p_sw = (osm_switch_t *) item;
	ib_net64_t sw_guid;
	ib_net64_t remote_guid;
	osm_opensm_t *p_osm = ((dump_ctx_t*)dump_cxt)->p_osm;
	osm_lid_mgr_t *p_lid_mgr = &p_osm->sm.lid_mgr;
	boolean_t is_flid;

	p_node = p_sw->p_node;

	max_lid_ho = p_sw->max_lid_ho;

	fprintf(file, "dump_ucast_routes: "
		"Switch 0x%016" PRIx64 "\nLID    : Port : Hops : Optimal\n",
		cl_ntoh64(osm_node_get_node_guid(p_node)));

	dor = osm_routing_engine_is_dor(p_osm->subn.routing_engine_used);

	for (lid_ho = 1; lid_ho <= max_lid_ho; lid_ho++) {
		fprintf(file, "0x%04X : ", lid_ho);
		is_flid = osm_lid_mgr_is_flid(p_lid_mgr, lid_ho);

		p_port = osm_get_port_by_lid_ho(&p_osm->subn, lid_ho);
		if (!p_port && !is_flid) {
			fprintf(file, "UNREACHABLE\n");
			continue;
		}

		port_num = osm_switch_get_port_by_lid(p_sw, lid_ho,
						      OSM_NEW_LFT);

		if (port_num == OSM_NO_PATH) {
			/*
			   This may occur if there are 'holes' in the existing
			   LID assignments.  Running SM with --reassign_lids
			   will reassign and compress the LID range.  The
			   subnet should work fine either way.
			 */
			fprintf(file, "UNREACHABLE\n");
			continue;
		}

		if (is_flid) {
			fprintf(file, "%03u  : HOPS UNKNOWN\n", port_num);
			continue;
		}
		/*
		   Switches can lie about which port routes a given
		   lid due to a recent reconfiguration of the subnet.
		   Therefore, ensure that the hop count is better than
		   OSM_NO_PATH.
		 */
		if (p_port->p_node->sw) {
			/* Target LID is switch.
			   Get its base lid and check hop count for this base LID only. */
			base_lid = osm_node_get_base_lid(p_port->p_node, 0);
			base_lid = cl_ntoh16(base_lid);
			num_hops =
			    osm_switch_get_hop_count(p_sw, base_lid, port_num);
		} else {
			/* Target LID is not switch (CA or router).
			   Check if we have route to this target from current switch. */
			num_hops =
			    osm_switch_get_hop_count(p_sw, lid_ho, port_num);
			if (num_hops != OSM_NO_PATH) {
				direct_route_exists = TRUE;
				base_lid = lid_ho;
			} else {
				osm_physp_t *p_physp = p_port->p_physp;

				if (!p_physp || !p_physp->p_remote_physp ||
				    !p_physp->p_remote_physp->p_node->sw)
					num_hops = OSM_NO_PATH;
				else {
					base_lid =
					    osm_node_get_base_lid(p_physp->
								  p_remote_physp->
								  p_node, 0);
					base_lid = cl_ntoh16(base_lid);
					sw_guid = osm_node_get_node_guid(p_sw->p_node);
					remote_guid = osm_node_get_node_guid(p_physp->
								p_remote_physp->p_node);
					num_hops =
					    sw_guid == remote_guid ? 0 :
					    osm_switch_get_hop_count(p_sw,
								     base_lid,
								     port_num);
				}
			}
		}

		if (num_hops == OSM_NO_PATH) {
			fprintf(file, "%03u  : HOPS UNKNOWN\n", port_num);
			continue;
		}

		best_hops = osm_switch_get_least_hops(p_sw, base_lid);
		if (!p_port->p_node->sw && !direct_route_exists) {
			best_hops++;
			num_hops++;
		}

		fprintf(file, "%03u  : %02u   : ", port_num, num_hops);

		if (best_hops == num_hops)
			fprintf(file, "yes");
		else {
			/* No LMC Optimization */
			best_port = osm_switch_recommend_path(p_sw, p_port,
							      lid_ho, 1, TRUE,
							      FALSE, dor,
							      p_osm->subn.opt.port_shifting,
							      NULL, OSM_NEW_LFT,
							      p_osm->subn.opt.port_profile_switch_nodes);
			fprintf(file, "No %u hop path possible via port %u!",
				best_hops, best_port);
		}

		fprintf(file, "\n");
	}
}

static void dump_mcast_routes(cl_map_item_t * item, FILE * file, void *cxt)
{
	osm_switch_t *p_sw = (osm_switch_t *) item;
	osm_mcast_tbl_t *p_tbl;
	int16_t mlid_ho = 0;
	int16_t mlid_start_ho;
	uint8_t position = 0;
	int16_t block_num = 0;
	boolean_t first_mlid;
	boolean_t first_port;
	const osm_node_t *p_node;
	uint16_t i, j;
	uint16_t mask_entry;
	char sw_hdr[256];
	char mlid_hdr[32];

	p_node = p_sw->p_node;

	p_tbl = osm_switch_get_mcast_tbl_ptr(p_sw);

	sprintf(sw_hdr, "\nSwitch 0x%016" PRIx64 "\nLID    : Out Port(s)\n",
		cl_ntoh64(osm_node_get_node_guid(p_node)));
	first_mlid = TRUE;
	while (block_num <= p_tbl->max_block_in_use) {
		mlid_start_ho = (uint16_t) (block_num * IB_MCAST_BLOCK_SIZE);
		for (i = 0; i < IB_MCAST_BLOCK_SIZE; i++) {
			mlid_ho = mlid_start_ho + i;
			position = 0;
			first_port = TRUE;
			sprintf(mlid_hdr, "0x%04X :",
				mlid_ho + IB_LID_MCAST_START_HO);
			while (position <= p_tbl->max_position) {
				mask_entry =
				    cl_ntoh16((*p_tbl->
					       p_mask_tbl)[block_num][i + IB_MCAST_BLOCK_SIZE * position]);
				if (mask_entry == 0) {
					position++;
					continue;
				}
				for (j = 0; j < 16; j++) {
					if ((1 << j) & mask_entry) {
						if (first_mlid) {
							fprintf(file, "%s",
								sw_hdr);
							first_mlid = FALSE;
						}
						if (first_port) {
							fprintf(file, "%s",
								mlid_hdr);
							first_port = FALSE;
						}
						fprintf(file, " 0x%03X ",
							j + (position * 16));
					}
				}
				position++;
			}
			if (first_port == FALSE)
				fprintf(file, "\n");
		}
		block_num++;
	}
}

static void dump_lid_matrix(cl_map_item_t * item, FILE * file, void *cxt)
{
	osm_switch_t *p_sw = (osm_switch_t *) item;
	osm_opensm_t *p_osm = cxt;
	osm_node_t *p_node = p_sw->p_node;
	unsigned max_lid = p_sw->max_lid_ho;
	unsigned max_port = p_sw->num_ports;
	uint16_t lid;
	uint8_t port;

	fprintf(file, "Switch: guid 0x%016" PRIx64 "\n",
		cl_ntoh64(osm_node_get_node_guid(p_node)));
	for (lid = 1; lid <= max_lid; lid++) {
		osm_port_t *p_port;
		if (osm_switch_get_least_hops(p_sw, lid) == OSM_NO_PATH)
			continue;
		fprintf(file, "0x%04x:", lid);
		for (port = 0; port < max_port; port++)
			fprintf(file, " %02x",
				osm_switch_get_hop_count(p_sw, lid, port));
		p_port = osm_get_port_by_lid_ho(&p_osm->subn, lid);
		if (p_port)
			fprintf(file, " # portguid 0x%016" PRIx64,
				cl_ntoh64(osm_port_get_guid(p_port)));
		fprintf(file, "\n");
	}
}

static void dump_ucast_lfts(cl_map_item_t * item, FILE * file, void *dump_cxt)
{
	osm_switch_t *p_sw = (osm_switch_t *) item;
	osm_opensm_t *p_osm = ((dump_ctx_t*)dump_cxt)->p_osm;
	osm_node_t *p_node = p_sw->p_node;
	unsigned max_lid = p_sw->max_lid_ho;
	unsigned max_port = p_sw->num_ports;
	uint16_t lid;
	uint16_t dumped_lids;
	uint8_t port;

	dumped_lids = 0;
	fprintf(file, "Unicast lids [0-%u] of switch Lid %u guid 0x%016"
		PRIx64 " (\'%s\'):\n",
		max_lid, cl_ntoh16(osm_node_get_base_lid(p_node, 0)),
		cl_ntoh64(osm_node_get_node_guid(p_node)), p_node->print_desc);
	for (lid = 0; lid <= max_lid; lid++) {
		osm_port_t *p_port;
		port = osm_switch_get_port_by_lid(p_sw, lid, OSM_NEW_LFT);

		if (port >= max_port)
			continue;
		dumped_lids++;
		fprintf(file, "0x%04x %03u # ", lid, port);

		p_port = osm_get_port_by_lid_ho(&p_osm->subn, lid);
		if (p_port) {
			p_node = p_port->p_node;
			fprintf(file, "%s portguid 0x%016" PRIx64 ": \'%s\'",
				ib_get_node_type_str(osm_node_get_type(p_node)),
				cl_ntoh64(osm_port_get_guid(p_port)),
				p_node->print_desc);
		} else
			fprintf(file, "unknown node and type");
		fprintf(file, "\n");
	}
	fprintf(file, "%u lids dumped\n", max_lid);
}

static void dump_topology_node(cl_map_item_t * item, FILE * file, void *cxt)
{
	osm_node_t *p_node = (osm_node_t *) item;
	uint32_t cPort;
	osm_node_t *p_nbnode;
	osm_physp_t *p_physp, *p_default_physp, *p_rphysp;
	char *link_speed_act_str;
	char *link_width_act_str;
	char *port_state_str;
	ib_net16_t cap_mask2;

	if (!p_node->node_info.num_ports)
		return;

	for (cPort = 1; cPort < osm_node_get_num_physp(p_node); cPort++) {

		p_physp = osm_node_get_physp_ptr(p_node, cPort);
		if (!p_physp)
			continue;

		p_rphysp = p_physp->p_remote_physp;
		if (!p_rphysp)
			continue;

		OSM_ASSERT(cPort == p_physp->port_num);

		if (p_node->node_info.node_type == IB_NODE_TYPE_SWITCH)
			p_default_physp = osm_node_get_physp_ptr(p_node, 0);
		else
			p_default_physp = p_physp;
		cap_mask2 = p_default_physp->port_info.capability_mask2;

		fprintf(file, "{ %s%s Ports:%02X SystemGUID:%016" PRIx64
			" NodeGUID:%016" PRIx64 " PortGUID:%016" PRIx64
			" VenID:%06X DevID:%04X Rev:%08X {%s} LID:%04X PN:%02X } ",
			p_node->node_info.node_type == IB_NODE_TYPE_SWITCH ?
			"SW" : p_node->node_info.node_type ==
			IB_NODE_TYPE_CA ? "CA" : p_node->node_info.node_type ==
			IB_NODE_TYPE_ROUTER ? "Rt" : "**",
			p_default_physp->port_info.base_lid ==
			p_default_physp->port_info.
			master_sm_base_lid ? "-SM" : "",
			p_node->node_info.num_ports,
			cl_ntoh64(p_node->node_info.sys_guid),
			cl_ntoh64(p_node->node_info.node_guid),
			cl_ntoh64(p_physp->port_guid),
			cl_ntoh32(ib_node_info_get_vendor_id
				  (&p_node->node_info)),
			cl_ntoh16(p_node->node_info.device_id),
			cl_ntoh32(p_node->node_info.revision),
			p_node->print_desc,
			cl_ntoh16(p_default_physp->port_info.base_lid), cPort);

		p_nbnode = p_rphysp->p_node;

		if (p_nbnode->node_info.node_type == IB_NODE_TYPE_SWITCH)
			p_default_physp = osm_node_get_physp_ptr(p_nbnode, 0);
		else
			p_default_physp = p_rphysp;

		fprintf(file, "{ %s%s Ports:%02X SystemGUID:%016" PRIx64
			" NodeGUID:%016" PRIx64 " PortGUID:%016" PRIx64
			" VenID:%08X DevID:%04X Rev:%08X {%s} LID:%04X PN:%02X } ",
			p_nbnode->node_info.node_type == IB_NODE_TYPE_SWITCH ?
			"SW" : p_nbnode->node_info.node_type ==
			IB_NODE_TYPE_CA ? "CA" :
			p_nbnode->node_info.node_type == IB_NODE_TYPE_ROUTER ?
			"Rt" : "**",
			p_default_physp->port_info.base_lid ==
			p_default_physp->port_info.
			master_sm_base_lid ? "-SM" : "",
			p_nbnode->node_info.num_ports,
			cl_ntoh64(p_nbnode->node_info.sys_guid),
			cl_ntoh64(p_nbnode->node_info.node_guid),
			cl_ntoh64(p_rphysp->port_guid),
			cl_ntoh32(ib_node_info_get_vendor_id
				  (&p_nbnode->node_info)),
			cl_ntoh32(p_nbnode->node_info.device_id),
			cl_ntoh32(p_nbnode->node_info.revision),
			p_nbnode->print_desc,
			cl_ntoh16(p_default_physp->port_info.base_lid),
			p_rphysp->port_num);

		port_state_str = get_physp_port_state_str(p_physp);
		link_width_act_str = get_physp_link_width_act_str(p_physp, cap_mask2);
		link_speed_act_str = get_physp_speed_act_str(p_physp);

		fprintf(file, "PHY=%s LOG=%s SPD=%s\n",
			link_width_act_str,
			port_state_str,
			link_speed_act_str);
	}
}

static void dump_sl2vl_tbl(cl_map_item_t * item, FILE * file, void *cxt)
{
	osm_port_t *p_port = (osm_port_t *) item;
	osm_node_t *p_node = p_port->p_node;
	uint32_t in_port, out_port,
		 num_ports = p_node->node_info.num_ports;
	ib_net16_t base_lid = osm_port_get_base_lid(p_port);
	osm_physp_t *p_physp;
	ib_slvl_table_t *p_tbl;
	int i, n;
	char buf[1024];
	char * header_line =	"#in out : 0  1  2  3  4  5  6  7  8  9  10 11 12 13 14 15";
	char * separator_line = "#--------------------------------------------------------";

	if (!num_ports)
		return;

	fprintf(file, "%s 0x%016" PRIx64 ", base LID %d, "
		"\"%s\"\n%s\n%s\n",
		ib_get_node_type_str(p_node->node_info.node_type),
		cl_ntoh64(p_port->guid), cl_ntoh16(base_lid),
		p_node->print_desc, header_line, separator_line);

	if (p_node->node_info.node_type == IB_NODE_TYPE_SWITCH) {
		for (out_port = 0; out_port <= num_ports; out_port++){
			p_physp = osm_node_get_physp_ptr(p_node, out_port);

			/* no need to print SL2VL table for port that is down */
			if (!p_physp || !p_physp->p_remote_physp)
				continue;

			for (in_port = 0; in_port <= num_ports; in_port++) {
				p_tbl = osm_physp_get_slvl_tbl(p_physp, in_port);
				for (i = 0, n = 0; i < 16; i++)
					n += sprintf(buf + n, " %-2d",
						ib_slvl_table_get(p_tbl, i));
				fprintf(file, "%-3d %-3d :%s\n",
					in_port, out_port, buf);
			}
		}
	} else {
		p_physp = p_port->p_physp;
		p_tbl = osm_physp_get_slvl_tbl(p_physp, 0);
		for (i = 0, n = 0; i < 16; i++)
			n += sprintf(buf + n, " %-2d",
					ib_slvl_table_get(p_tbl, i));
		fprintf(file, "%-3d %-3d :%s\n", 0, 0, buf);
	}

	fprintf(file, "%s\n\n", separator_line);
}

static void dump_ar_routes(cl_map_item_t * item, FILE * file, void *dump_cxt)
{
	osm_switch_t *p_sw = (osm_switch_t *) item;
	osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;
	uint8_t plft_idx, num_plfts, port, state, i;
	uint16_t plft_top;
	uint16_t lid_ho, group_id, sl_mask;
	const osm_node_t *p_node;

	if (p_sw->ar_configured != OSM_SW_AR_CONF_CONFIGURED)
		return;

	p_node = p_sw->p_node;
	sl_mask = p_ar_data->req_ar_info.enabled_and_cap_fields & IB_AR_INFO_CAP_MASK_BY_SL_EN ?
		  cl_ntoh16(p_ar_data->req_ar_info.enable_by_sl_mask) : OSM_ALL_SL_VL_AR_ENABLED;
	fprintf(file,
		"Switch 0x%016" PRIx64 " ar_en: %u fr_en: %u rn_en: %u whbf_en: %u pfrn_en: %u\n"
		"#switch-name=%s\n"
		"ar_en_sl: (0x%x) ",
		cl_ntoh64(osm_node_get_node_guid(p_node)),
		osm_ar_info_is_ar_enabled(&p_ar_data->req_ar_info) ? 1 : 0,
		osm_ar_info_is_fr_enabled(&p_ar_data->req_ar_info) ? 1 : 0,
		osm_ar_info_is_rn_xmit_enabled(&p_ar_data->req_ar_info)? 1 : 0,
		osm_ar_info_is_whbf_enabled(&p_ar_data->req_ar_info) ? 1 : 0,
		osm_ar_info_is_pfrn_enabled(&p_ar_data->req_ar_info)? 1 : 0,
		p_sw->p_node->print_desc,
		sl_mask);

	sl_mask = osm_ar_info_is_hbf_enabled(&p_ar_data->req_ar_info) ?
		  cl_ntoh16(p_ar_data->req_ar_info.enable_by_sl_mask_hbf) : 0;
	fprintf(file, "hbf_sl: (0x%x)\n", sl_mask);

	/* when HBF is enabled */
	if (osm_ar_info_is_hbf_enabled(&p_ar_data->req_ar_info)) {
		char seed_value_str[20] = {'R', 'e', 's', 'e', 'r', 'v', 'e', 'd', '\0'};
		uint8_t seed_type = ib_hbf_config_get_seed_type(&p_ar_data->req_hbf_config);
		if (seed_type == OSM_HBF_SEED_TYPE_CONFIGURABLE) {
			sprintf(seed_value_str, "0x%08x",
				cl_ntoh32(p_ar_data->req_hbf_config.seed));
		}
		fprintf(file, "#HBF configuration: hash_type=%s seed_type=%s seed_value=%s "
			"fields_enable=0x%016lx\n",
			ib_hbf_config_get_hbf_type(&p_ar_data->req_hbf_config) ==
			    OSM_HBF_HASH_TYPE_CRC ? "CRC" : "XOR",
			seed_type == OSM_HBF_SEED_TYPE_RANDOM ? "Random" : "Config",
			seed_value_str, cl_ntoh64(p_ar_data->req_hbf_config.fields_enable));
	} else {
		fprintf(file, "#HBF configuration: N/A\n");
	}
	fprintf(file, "\nByTransportDisable: (0x%x). ",
		p_ar_data->req_ar_info.by_trns_disable);
	for (i = 0; i < IB_AR_INFO_TRANSPORT_NUM; i++) {
		if (p_ar_data->req_ar_info.by_trns_disable & (1 << i))
			fprintf(file, "%s, ", ib_ar_info_get_transport_name((1 << i)));
	}
	fprintf(file, "\n\n");

	osm_ar_print_subgroups(file, p_sw);

	fprintf(file, "\nLFT Definition:\n");
	/* print PLFTs or one AR LFT as well */
	num_plfts = p_ar_data->num_plfts ? p_ar_data->num_plfts : 1;
	for (plft_idx = 0; plft_idx < num_plfts; plft_idx++) {
		plft_top = p_ar_data->num_plfts ?
				osm_ar_get_plft_top(p_ar_data, plft_idx) :
				p_sw->max_lid_ho;

		fprintf(file, "\nPLFT_NUM: %u, top 0x%x\n\n", plft_idx, plft_top);
		fprintf(file,
			"LID       Static Port    Lid State      Group\n"
			"-----------------------------------------------\n");
		for (lid_ho = 1; lid_ho <= plft_top; lid_ho++) {
			/* print calculated AR LFT tables */
			osm_ar_lft_entry_get(p_sw, plft_idx, lid_ho,
					     &port, &state, &group_id);
			/* don't print unassigned lids */
			if (port == OSM_NO_PATH)
				continue;
			else
				fprintf(file, "0x%04x    %-15u%-15s%-u\n",
					lid_ho,
					port,
					osm_ar_get_port_state_str(state),
					group_id);
		}
	}
	fprintf(file, "\n");
}

static void print_node_report(cl_map_item_t * item, FILE * file, void *cxt)
{
	osm_node_t *p_node = (osm_node_t *) item;
	osm_opensm_t *osm = cxt;
	const osm_physp_t *p_physp, *p_remote_physp;
	const ib_port_info_t *p_pi;
	uint8_t port_num;
	uint32_t num_ports;
	uint8_t node_type;

	node_type = osm_node_get_type(p_node);

	num_ports = osm_node_get_num_physp(p_node);
	port_num = node_type == IB_NODE_TYPE_SWITCH ? 0 : 1;
	for (; port_num < num_ports; port_num++) {
		p_physp = osm_node_get_physp_ptr(p_node, port_num);
		if (!p_physp)
			continue;

		fprintf(file, "%-11s : %s : %02X :",
			osm_get_manufacturer_str(p_node),
			osm_get_node_type_str_fixed_width(node_type), port_num);

		p_pi = &p_physp->port_info;

		/*
		 * Port state is not defined for base switch port 0
		 */
		if (port_num == 0 &&
		    ib_switch_info_is_enhanced_port0(&p_node->sw->switch_info) == FALSE)
			fprintf(file, "     :");
		else
			fprintf(file, " %s :",
				osm_get_port_state_str_fixed_width
				(ib_port_info_get_port_state(p_pi)));

		/*
		 * LID values are only meaningful in select cases.
		 */
		if (ib_port_info_get_port_state(p_pi) != IB_LINK_DOWN
		    && ((node_type == IB_NODE_TYPE_SWITCH && port_num == 0)
			|| node_type != IB_NODE_TYPE_SWITCH))
			fprintf(file, " %04X :  %01X  :",
				cl_ntoh16(p_pi->base_lid),
				ib_port_info_get_lmc(p_pi));
		else
			fprintf(file, "      :     :");

		if (port_num == 0 &&
		    ib_switch_info_is_enhanced_port0(&p_node->sw->switch_info) == FALSE)
			fprintf(file, "      :     :      ");
		else
			fprintf(file, " %s : %s : %s ",
				osm_get_mtu_str
				(ib_port_info_get_neighbor_mtu(p_pi)),
				osm_get_lwa_str(p_pi->link_width_active),
				osm_get_lsa_str
				(ib_port_info_get_link_speed_active(p_pi),
				 ib_port_info_get_link_speed_ext_active(p_pi),
				 ib_port_info_get_link_speed_ext_active2(p_pi),
				 ib_port_info_get_port_state(p_pi),
				 p_physp->ext_port_info.link_speed_active));

		if (osm_physp_get_port_guid(p_physp) == osm->subn.sm_port_guid)
			fprintf(file, "* %016" PRIx64 " *",
				cl_ntoh64(osm_physp_get_port_guid(p_physp)));
		else
			fprintf(file, ": %016" PRIx64 " :",
				cl_ntoh64(osm_physp_get_port_guid(p_physp)));

		if (port_num
		    && (ib_port_info_get_port_state(p_pi) != IB_LINK_DOWN)) {
			p_remote_physp = osm_physp_get_remote(p_physp);
			if (p_remote_physp)
				fprintf(file, " %016" PRIx64 " (%02X)",
					cl_ntoh64(osm_physp_get_port_guid
						  (p_remote_physp)),
					osm_physp_get_port_num(p_remote_physp));
			else
				fprintf(file, " UNKNOWN");
		}

		fprintf(file, "\n");
	}

	fprintf(file, "------------------------------------------------------"
		"------------------------------------------------\n");
}

struct dump_context {
	osm_opensm_t *p_osm;
	FILE *file;
	void (*func) (cl_map_item_t *, FILE *, void *);
	void *cxt;
};

struct dump_qmap_context {
	cl_qmap_t *map;
	void (*func) (cl_map_item_t *, FILE *, void *);
	void *cxt;
};

static void dump_item(cl_map_item_t * item, void *cxt)
{
	((struct dump_context *)cxt)->func(item,
					   ((struct dump_context *)cxt)->file,
					   ((struct dump_context *)cxt)->cxt);
}

static void dump_qmap(FILE * file, cl_qmap_t * map,
		      void (*func) (cl_map_item_t *, FILE *, void *), void *cxt)
{
	struct dump_context dump_context;

	dump_context.file = file;
	dump_context.func = func;
	dump_context.cxt = cxt;

	cl_qmap_apply_func(map, dump_item, &dump_context);
}

static void dump_qmap_to_file(FILE *file, void *cxt)
{
    dump_qmap(file,
              ((struct dump_qmap_context *)cxt)->map,
              ((struct dump_qmap_context *)cxt)->func,
	      ((struct dump_qmap_context *)cxt)->cxt);
}

void osm_dump_qmap_to_file(osm_opensm_t * p_osm, const char *file_path,
			   cl_qmap_t * map,
			   void (*func) (cl_map_item_t *, FILE *, void *),
			   void *cxt)
{
	struct dump_qmap_context context;

	context.map = map;
	context.func = func;
	context.cxt = cxt;

	osm_dump_func_to_file(p_osm, file_path, dump_qmap_to_file,
			      (void *)&context);
}

static void dump_ar_func_to_file(osm_opensm_t * p_osm, const char *file_path,
				 void (*func) (FILE *, void *),
				 void *cxt)
{
	char path_tmp[1032];
	FILE *file;
	int status = 0;

	snprintf(path_tmp, sizeof(path_tmp), "%s.tmp", file_path);

	file = fopen(path_tmp, "w");
	if (!file) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"cannot create file \'%s\': %s\n",
			path_tmp, strerror(errno));
		return;
	}

	fprintf(file, "File version: 4\n");

	func(file, cxt);

	fclose(file);

	status = rename(path_tmp, file_path);
	if (status) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"Failed to rename file:%s (err:%s)\n",
			path_tmp, strerror(errno));
	}
}

static void dump_ar_qmap_to_file(osm_opensm_t * p_osm, const char *file_path,
				 cl_qmap_t * map,
				 void (*func) (cl_map_item_t *, FILE *, void *),
				 void *cxt)
{
	struct dump_qmap_context context;

	context.map = map;
	context.func = func;
	context.cxt = cxt;

	dump_ar_func_to_file(p_osm, file_path, dump_qmap_to_file, (void *)&context);
}

void osm_dump_func_to_file(osm_opensm_t * p_osm, const char *file_path,
			   void (*func) (FILE *, void *),
			   void *cxt)
{
	char path_tmp[1032];
	FILE *file;
	int status = 0;

	snprintf(path_tmp, sizeof(path_tmp), "%s.tmp", file_path);

	file = fopen(path_tmp, "w");
	if (!file) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"cannot create file \'%s\': %s\n",
			path_tmp, strerror(errno));
		return;
	}
	func(file, cxt);

	fclose(file);

	status = rename(path_tmp, file_path);
	if (status) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"Failed to rename file:%s (err:%s)\n",
			path_tmp, strerror(errno));
	}
}


static void print_report(osm_opensm_t * osm, FILE * file)
{
	fprintf(file, "\n==================================================="
		"====================================================\n"
		"Vendor      : Ty : #  : Sta : LID  : LMC : MTU  : LWA :"
		" LSA  : Port GUID        : Neighbor Port (Port #)\n");
	dump_qmap(stdout, &osm->subn.node_guid_tbl, print_node_report, osm);
}

void osm_dump_mcast_routes(osm_opensm_t * osm)
{
	char path[1024];
	if (OSM_LOG_IS_ACTIVE_V2(&osm->log, OSM_LOG_ROUTING)) {
		/* multicast routes */
		snprintf(path, sizeof(path), "%s/%s", osm->subn.opt.dump_files_dir, MCFDBS_DUMP_FILE);
		osm_dump_qmap_to_file(osm, path,
				      &osm->subn.sw_guid_tbl,
				      dump_mcast_routes, osm);
	}
}

void osm_dump_router_db(osm_opensm_t * osm)
{
	cl_qmap_t *p_rtr_guid_tbl = &osm->sm.p_subn->rtr_guid_tbl;
	cl_fmap_t *p_prefix_rtr_table = &osm->sm.router_mgr.prefix_rtr_table;
	char path[1024];
	FILE *file;
	cl_map_item_t *p_map_item;
	osm_router_t  *p_router;
	osm_router_tbl_data_t *p_rtr_data;
	cl_map_iterator_t map_iter1;
	cl_fmap_t *p_router_subnets;
	cl_fmap_item_t *p_item;
	osm_router_prefix_rtrs_map_t *p_subnet_routers;
	osm_subnet_prefix_pkey_t *p_prefix_pkey;

	snprintf(path, sizeof(path), "%s/%s",
		 osm->subn.opt.dump_files_dir, ROUTERS_DUMP_FILE);
	file = fopen(path, "w");

	if (!file) {
		OSM_LOG(&osm->log, OSM_LOG_ERROR,
			"cannot create file \'%s\': %s\n",
			path, strerror(errno));
		return;
	}

	fprintf(file, "Routers count %u\n", cl_qmap_count(p_rtr_guid_tbl));
	for (p_map_item = cl_qmap_head(p_rtr_guid_tbl);
	     p_map_item != cl_qmap_end(p_rtr_guid_tbl);
	     p_map_item = cl_qmap_next(p_map_item)) {
		p_router = (osm_router_t *) p_map_item;
		p_router_subnets = &p_router->subnet_prefix_tbl;
		fprintf(file, "\tRouter 0x%016" PRIx64 ":\n",
			cl_ntoh64(p_router->p_port->guid));
		for (p_rtr_data = (osm_router_tbl_data_t *)cl_fmap_head(
							     p_router_subnets);
		     p_rtr_data != (osm_router_tbl_data_t *)cl_fmap_end(
							     p_router_subnets);
		     p_rtr_data = (osm_router_tbl_data_t *)cl_fmap_next(
						    &p_rtr_data->map_item)) {
			if (p_rtr_data->is_adjacent)
				fprintf(file, "subnet prefix: 0x%016" PRIx64
					", pkey: 0x%04" PRIx16 ", sm lid: 0x%04"
					PRIx16 "\n",
					cl_ntoh64(p_rtr_data->subnet_prefix),
					cl_ntoh16(p_rtr_data->pkey),
					cl_ntoh16(p_rtr_data->master_sm_lid));
			else
				fprintf(file, "subnet prefix: 0x%016" PRIx64
					", pkey: 0x%04" PRIx16 "\n",
					cl_ntoh64(p_rtr_data->subnet_prefix),
					cl_ntoh16(p_rtr_data->pkey));
		}
		fprintf(file, "\n");
	}
	fprintf(file, "<subnet prefix, pkey> -> routers, size %zu\n",
		cl_fmap_count(p_prefix_rtr_table));
	for (p_item = cl_fmap_head(p_prefix_rtr_table);
	     p_item != cl_fmap_end(p_prefix_rtr_table);
	     p_item = cl_fmap_next(p_item)) {
		p_subnet_routers = (osm_router_prefix_rtrs_map_t *)p_item;
		p_prefix_pkey = (osm_subnet_prefix_pkey_t *)cl_fmap_key(p_item);
		fprintf(file, "\tSubnet 0x%016" PRIx64 " pkey 0x%04" PRIx16
			": ",
			cl_ntoh64(p_prefix_pkey->subnet_prefix),
			cl_ntoh16(p_prefix_pkey->pkey));
		for (map_iter1 = cl_map_head(&p_subnet_routers->rtrs);
		     map_iter1 != cl_map_end(&p_subnet_routers->rtrs);
		     map_iter1 = cl_map_next(map_iter1)) {
			fprintf(file, " 0x%016" PRIx64" |",
				cl_ntoh64(cl_map_key(map_iter1)));
		}
		fprintf(file, "\n");
	}
	fclose(file);
}

void osm_dump_lfts(osm_opensm_t * osm, osm_subn_t *p_subn, char* prefix)
{
	dump_ctx_t dump_ctx;
	char lfts_name[MAX_NAME_LEN];
	char path[1024];

	memset(lfts_name, 0, MAX_NAME_LEN);
	snprintf(lfts_name, MAX_NAME_LEN, "%s%s.dump", LFT_DUMP_FILE, prefix);

	if (OSM_LOG_IS_ACTIVE_V2(&osm->log, OSM_LOG_ROUTING)) {
		dump_ctx.p_osm = osm;
		snprintf(path, sizeof(path), "%s/%s", osm->subn.opt.dump_files_dir, lfts_name);
		osm_dump_qmap_to_file(osm, path, &p_subn->sw_guid_tbl,
				      dump_ucast_lfts, &dump_ctx);
	}
}

void osm_dump_subn_all(osm_opensm_t * osm, osm_subn_t *p_subn, char* prefix,
		       unsigned dump_flags)
{
	dump_ctx_t dump_ctx;
	char lid_matrix_name[MAX_NAME_LEN];
	char fdbs_name[MAX_NAME_LEN];
	char lst_name[MAX_NAME_LEN];
	char ar_name[MAX_NAME_LEN];
	char path[1024];

	if ((dump_flags & OSM_DUMP_FLAG_SMDB) && osm->subn.opt.sm_db_dump)
		osm_dump_sm_db(osm);

	if (dump_flags & OSM_DUMP_FLAG_ROUTING &&
	    OSM_LOG_IS_ACTIVE_V2(&osm->log, OSM_LOG_ROUTING)) {
		memset(lid_matrix_name, 0, MAX_NAME_LEN);
		memset(fdbs_name, 0, MAX_NAME_LEN);

		snprintf(lid_matrix_name, MAX_NAME_LEN, "%s%s.dump",
			 LID_MATRIX_DUMP_FILE, prefix);
		snprintf(fdbs_name, MAX_NAME_LEN, "%s%s", FDBS_DUMP_FILE, prefix);

		dump_ctx.p_osm = osm;
		dump_ctx.rch_dump = (strlen(prefix) > 0);

		/* unicast routes */
		snprintf(path, sizeof(path), "%s/%s", osm->subn.opt.dump_files_dir, lid_matrix_name);
		osm_dump_qmap_to_file(osm, path, &p_subn->sw_guid_tbl,
				      dump_lid_matrix, osm);
		osm_dump_lfts(osm, p_subn, prefix);
		if (OSM_LOG_IS_ACTIVE_V2(&osm->log, OSM_LOG_DEBUG) &&
		    !dump_ctx.rch_dump)
			dump_qmap(stdout, &p_subn->sw_guid_tbl,
				  dump_ucast_path_distribution, osm);

		/* An attempt to get osm_switch_recommend_path to report the
		   same routes that a sweep would assign. */
		if (p_subn->opt.scatter_ports)
			srandom(p_subn->opt.scatter_ports);

		snprintf(path, sizeof(path), "%s/%s", osm->subn.opt.dump_files_dir, fdbs_name);
		osm_dump_qmap_to_file(osm, path,
				      &p_subn->sw_guid_tbl,
				      dump_ucast_routes, &dump_ctx);
		if (!dump_ctx.rch_dump) {
			/* multicast routes */
			snprintf(path, sizeof(path), "%s/%s",
				 osm->subn.opt.dump_files_dir, MCFDBS_DUMP_FILE);
			osm_dump_qmap_to_file(osm, path,
					      &p_subn->sw_guid_tbl,
					      dump_mcast_routes, osm);
			/* SL2VL tables */
			if (p_subn->opt.qos ||
			    (p_subn->routing_engine_used &&
			     p_subn->routing_engine_used->update_sl2vl)) {
				snprintf(path, sizeof(path), "%s/%s",
					 osm->subn.opt.dump_files_dir, SL2VL_DUMP_FILE);
				osm_dump_qmap_to_file(osm, path,
						      &p_subn->port_guid_tbl,
						      dump_sl2vl_tbl, osm);
			}
		}
		if (osm->subn.opt.dump_ar) {
			memset(ar_name, 0, MAX_NAME_LEN);
			snprintf(ar_name, MAX_NAME_LEN, "%s%s", AR_DUMP_FILE, prefix);
			snprintf(path, sizeof(path), "%s/%s",
				 osm->subn.opt.dump_files_dir, ar_name);
			dump_ar_qmap_to_file(osm, path,
					     &p_subn->sw_guid_tbl,
					     dump_ar_routes, &dump_ctx);
		}
	}

	if (dump_flags & OSM_DUMP_FLAG_SUBNET_LST) {
		memset(lst_name, 0, MAX_NAME_LEN);
		snprintf(lst_name, MAX_NAME_LEN, "%s%s.lst", SUBNET_LST_DUMP_FILE,
			 prefix);

		if (osm->subn.opt.enable_subnet_lst) {
			if (osm->subn.opt.subnet_lst_file)
				snprintf(path, sizeof(path), "%s", osm->subn.opt.subnet_lst_file);
			else
				snprintf(path, sizeof(path), "%s/%s",
					 osm->subn.opt.dump_files_dir, lst_name);
			osm_dump_qmap_to_file(osm, path,
					      &p_subn->node_guid_tbl, dump_topology_node,
					      osm);
		}

		if (OSM_LOG_IS_ACTIVE_V2(&osm->log, OSM_LOG_VERBOSE))
			print_report(osm, stdout);
	}

	if (dump_flags & OSM_DUMP_FLAG_VIRTUALIZATION)
		osm_dump_virtualization(osm);

	if (dump_flags & OSM_DUMP_FLAG_ROUTERS)
		osm_dump_router_db(osm);
}

void static print_stat_header(osm_opensm_t * p_osm, char * header, int sm_port_idx) {
	if (!p_osm->statistic_dump_file[sm_port_idx])
		return;

	p_osm->statistic_file_size[sm_port_idx] +=
			fprintf(p_osm->statistic_dump_file[sm_port_idx], "%s\n", header);
}

void static osm_dump_statistic_per_port(osm_opensm_t * p_osm, int sm_port_idx,
					boolean_t *p_is_main_file_trancated)
{
	char path[1024];
	char old_path[1024];
	char date_str[MAX_STR_LEN];
	FILE *file = p_osm->statistic_dump_file[sm_port_idx];
	int i, port_idx;
	struct stat file_stat;
	uint64_t counter = 0;
	boolean_t old_stats_file_exists = FALSE;
	char *tmp_buf = NULL;
	boolean_t is_main_file = sm_port_idx == OSM_MAX_BINDING_PORTS ? TRUE : FALSE;
	char header[] = "TIME_STAMP, HEAVY_SWEEP_COUNT, LIGHT_SWEEP_COUNT, "
			"SM_MAD_TIMEOUTS_COUNT, SA_MAD_TIMEOUTS_COUNT, "
			"CC_MAD_TIMEOUTS_COUNT, VS_MAD_TIMEOUTS_COUNT, N2N_MAD_TIMEOUTS_COUNT, "
			"NODE_DESC_RCV_COUNT, NODE_DESC_SEND_COUNT, "
			"NODE_INFO_RCV_COUNT, NODE_INFO_SEND_COUNT, SWITCH_INFO_RCV_COUNT, "
			"SWITCH_INFO_SEND_COUNT, GUID_INFO_RCV_COUNT, GUID_INFO_SEND_COUNT, "
			"PORT_INFO_RCV_COUNT, PORT_INFO_SEND_COUNT, SPST_RCV_COUNT, SPST_SEND_COUNT, "
			"P_KEY_TBL_RCV_COUNT, P_KEY_TBL_SEND_COUNT, "
			"SL_VL_MAP_RCV_COUNT, SL_VL_MAP_SEND_COUNT, "
			"VL_ARBIT_TBL_RCV_COUNT, VL_ARBIT_TBL_SEND_COUNT, "
			"GENERAL_INFO_RCV_COUNT, GENERAL_INFO_SEND_COUNT, "
			"LFT_RCV_COUNT, LFT_SEND_COUNT, AR_INFO_RCV_COUNT, AR_INFO_SEND_COUNT, "
			"EXTENDED_NODE_INFO_RCV_COUNT, EXTENDED_NODE_INFO_SEND_COUNT, "
			"AR_GROUP_TABLE_RCV_COUNT, AR_GROUP_TABLE_SEND_COUNT, "
			"AR_GROUP_TABLE_RCV_COPY_COUNT, AR_GROUP_TABLE_COPY_SEND_COUNT, "
			"AR_LFT_RCV_COUNT, AR_LFT_SEND_COUNT, PLFT_INFO_RCV_COUNT, PLFT_INFO_SEND_COUNT, "
			"PLFT_DEF_RCV_COUNT, PLFT_DEF_SEND_COUNT, PORT_SL_TO_PLFT_MAP_RCV_COUNT, "
			"PORT_SL_TO_PLFT_MAP_SEND_COUNT, PLFT_MAP_RCV_COUNT, PLFT_MAP_SEND_COUNT, "
			"RN_SUB_GROUP_DIRECTION_TABLE_RCV_COUNT, RN_SUB_GROUP_DIRECTION_TABLE_SEND_COUNT, "
			"RN_GEN_STRING_TABLE_RCV_COUNT, RN_GEN_STRING_TABLE_SEND_COUNT, "
			"RN_RCV_STRING_TABLE_RCV_COUNT, RN_RCV_STRING_TABLE_SEND_COUNT, "
			"RN_XMIT_PORT_MASK_RCV_COUNT, RN_XMIT_PORT_MASK_SEND_COUNT, "
			"RN_GEN_BY_SUB_GROUP_PRIORITY_RCV_COUNT, RN_GEN_BY_SUB_GROUP_PRIORITY_SEND_COUNT, "
			"MFT_RCV_COUNT, MFT_SEND_COUNT, "
			"HIERARCHY_INFO_RCV_COUNT, HIERARCHY_INFO_SEND_COUNT, SM_INFO_RCV_COUNT, "
			"SM_INFO_SEND_COUNT, MEPI_RCV_COUNT, MEPI_SEND_COUNT, "
			"ROUTER_INFO_RCV_COUNT, ROUTER_INFO_RCV_COUNT, "
			"RTR_ADJ_SITE_LOCAL_TBL_RCV_COUNT, RTR_ADJ_SITE_LOCAL_TBL_SEND_COUNT, "
			"RTR_NEXT_HOP_TBL_RCV_COUNT, RTR_NEXT_HOP_TBL_SEND_COUNT, "
			"RTR_ADJ_ROUTER_LID_INFO_RCV_COUNT, RTR_ADJ_ROUTER_LID_INFO_SEND_COUNT, "
			"RTR_ROUTER_LID_TBL_RCV_COUNT, RTR_ROUTER_LID_TBL_SEND_COUNT, "
			"QOS_CONFIG_SL_RCV_COUNT, QOS_CONFIG_SL_SEND_COUNT, "
			"VIRTUALIZATION_INFO_RCV_COUNT, VIRTUALIZATION_INFO_SEND_COUNT, "
			"VPORT_PORT_STATE_TBL_RCV_COUNT, VPORT_PORT_STATE_TBL_SEND_COUNT, "
			"VPORT_NODE_INFO_RCV_COUNT, VPORT_NODE_INFO_SEND_COUNT, "
			"VPORT_PORT_INFO_RCV_COUNT, VPORT_PORT_INFO_SEND_COUNT, "
			"VPORT_PKEY_TABLE_RCV_COUNT, VPORT_PKEY_TABLE_SEND_COUNT, "
			"VPORT_NODE_DESCRIPTION_RCV_COUNT, VPORT_NODE_DESCRIPTION_SEND_COUNT, "
			"VPORT_GUID_INFO_RCV_COUNT, VPORT_GUID_INFO_SEND_COUNT, "
			"VS_CLASS_PORT_INFO_RCV_COUNT, VS_CLASS_PORT_INFO_SEND_COUNT, "
			"VS_KEY_INFO_RCV_COUNT, VS_KEY_INFO_SEND_COUNT, "
			"CC_KEY_INFO_RCV_COUNT, CC_KEY_INFO_SEND_COUNT, CC_CLASS_PORT_INFO_RCV_COUNT, "
			"CC_CLASS_PORT_INFO_SEND_COUNT, CC_ENH_INFO_RCV_COUNT, "
			"CC_ENH_INFO_SEND_COUNT, CC_HCA_GENERAL_SETTINGS_RCV_COUNT, "
			"CC_HCA_GENERAL_SETTINGS_SEND_COUNT, CC_PORT_SW_PROFILE_SETTINGS_RCV_COUNT, "
			"CC_PORT_SW_PROFILE_SETTINGS_SEND_COUNT, CC_SL_MAPPING_SETTINGS_RCV_COUNT, "
			"CC_SL_MAPPING_SETTINGS_SEND_COUNT, CC_SW_GENERAL_SETTINGS_RCV_COUNT, "
			"CC_SW_GENERAL_SETTINGS_SEND_COUNT, CC_HCA_RP_PARAMETERS_RCV_COUNT, "
			"CC_HCA_RP_PARAMETERS_SEND_COUNT, CC_HCA_NP_PARAMETERS_RCV_COUNT, "
			"CC_HCA_NP_PARAMETERS_SEND_COUNT, PPCC_HCA_ALGO_CONFIG_RCV_COUNT, "
			"PPCC_HCA_ALGO_CONFIG_SEND_COUNT, PPCC_HCA_ALGO_PARAMS_CONFIG_RCV_COUNT, "
			"PPCC_HCA_ALGO_PARAMS_CONFIG_SEND_COUNT, N2N_CLASS_PORT_INFO_RCV_COUNT, "
			"N2N_CLASS_PORT_INFO_SEND_COUNT, N2N_KEY_INFO_RCV_COUNT, "
			"N2N_KEY_INFO_SEND_COUNT, N2N_NEIGHBORS_INFO_RCV_COUNT, "
			"N2N_NEIGHBORS_INFO_SEND_COUNT, NOTICE_128_RCV_COUNT, "
			"NOTICE_128_SEND_COUNT, NOTICE_130_RCV_COUNT, NOTICE_130_SEND_COUNT, "
			"NOTICE_131_RCV_COUNT, NOTICE_131_SEND_COUNT, NOTICE_144_RCV_COUNT, "
			"NOTICE_144_SEND_COUNT, NOTICE_145_RCV_COUNT, NOTICE_145_SEND_COUNT, "
			"NOTICE_256_RCV_COUNT, NOTICE_256_SEND_COUNT, NOTICE_257_RCV_COUNT, "
			"NOTICE_257_SEND_COUNT, NOTICE_258_RCV_COUNT, NOTICE_258_SEND_COUNT, "
			"NOTICE_259_RCV_COUNT, NOTICE_259_SEND_COUNT, NOTICE_1144_RCV_COUNT, "
			"NOTICE_1144_SEND_COUNT, NOTICE_1146_RCV_COUNT, NOTICE_1146_SEND_COUNT, "
			"NODE_RECORD_RCV, NODE_RECORD_SEND, PORT_INFO_RECORD_RCV, PORT_INFO_RECORD_SEND, "
			"INFORM_INFO_RECORD_RCV, INFORM_INFO_RECORD_SEND, "
			"SERVICE_RECORD_RCV, SERVICE_RECORD_SEND, PATH_RECORD_RCV, PATH_RECORD_SEND, "
			"MC_MEMBER_RECORD_GET_RCV, MC_MEMBER_RECORD_SET_RCV, MC_MEMBER_RECORD_SEND, "
			"GUID_INFO_RECORD_RCV, GUID_INFO_RECORD_SEND, "
			"UNTRUSTED_MAD_DROP_COUNT, SA_ETM_EXCEEDING_MAX_SUBSCRIPTIONS_COUNT, "
			"SL_VL_MAP_TBL_RECORD_RCV, SL_VL_MAP_TBL_RECORD_SEND, SWITCH_INFO_RECORD_RCV, "
			"SWITCH_INFO_RECORD_SEND, LINEAR_FORWARD_TBL_RECORD_RCV, "
			"LINEAR_FORWARD_TBL_RECORD_SEND, MULTICAST_FORWARD_TBL_RECORD_RCV, "
			"MULTICAST_FORWARD_TBL_RECORD_SEND, VL_ARBI_TBL_RECORD_RCV, VL_ARBI_TBL_RECORD_SEND, "
			"SM_INFO_RECORD_RCV, SM_INFO_RECORD_SEND, P_KEY_TBL_RECORD_RCV, "
			"P_KEY_TBL_RECORD_SEND, LINK_RECORD_RCV, LINK_RECORD_SEND, "
			"MULTI_PATH_RECORD_RCV, MULTI_PATH_RECORD_SEND, HBF_CONFIG_RCV_COUNT, HBF_CONFIG_SEND_COUNT,"
			"WHBF_CONFIG_RCV_COUNT, WHBF_CONFIG_SEND_COUNT, "
			"PFRN_CONFIG_RCV_COUNT, PFRN_CONFIG_SEND_COUNT, "
			"PROFILES_CONFIG_RCV_COUNT, PROFILES_CONFIG_SEND_COUNT, "
			"CREDIT_WATCHDOG_CONFIG_RCV_COUNT, CREDIT_WATCHDOG_CONFIG_SEND_COUNT, "
			"BER_CONFIG_RCV_COUNT, BER_CONFIG_SEND_COUNT, "
			"ALID_INFO_RCV_COUNT, ALID_INFO_SEND_COUNT, "
			"RAIL_FILTER_CONFIG_RCV_COUNT, RAIL_FILTER_CONFIG_SEND_COUNT, "
			"NOTICE_1601_SEND_COUNT";

	if(!file){
		if (is_main_file)
			snprintf(path, sizeof(path), "%s/%s.dump",
				 p_osm->subn.opt.dump_files_dir,
				 OSM_DEFAULT_STATISTICS_DUMP_FILE);
		else
			snprintf(path, sizeof(path), "%s/%s.%d.dump",
				 p_osm->subn.opt.dump_files_dir,
				 OSM_DEFAULT_STATISTICS_DUMP_FILE, sm_port_idx);

		file = fopen(path, "a+");
		if (!file) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
				"cannot create statistics file \'%s\': %s\n",
				path, strerror(errno));
			goto Exit;
		}
		tmp_buf = calloc(1, sizeof(header) + 2);
		if (!tmp_buf) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
				"ERR 0602: Couldn't allocate buffer to read first"
				"line in stats counters file\n");
			fclose(file);
			goto Exit;
		}

		/*
		 * To support backward compatibility,
		 * in case the previous header is different from the current.
		 * Make a backup of the old file and create
		 * a new file with the new statistics.
		 */
		if (fgets(tmp_buf, sizeof(header) + 1, file)) {
			if (memcmp(header, tmp_buf, sizeof(header) - 1) ||
			    (tmp_buf[sizeof(header) - 1] != '\n')) {
				fclose(file);

				if (is_main_file)
					snprintf(old_path, sizeof(old_path),
						 "%s/%s.dump.old",
						 p_osm->subn.opt.dump_files_dir,
						 OSM_DEFAULT_STATISTICS_DUMP_FILE);
				else
					snprintf(old_path, sizeof(old_path),
						 "%s/%s.%d.dump.old",
						 p_osm->subn.opt.dump_files_dir,
						 OSM_DEFAULT_STATISTICS_DUMP_FILE,
						 sm_port_idx);

				/* Check if old statistics file exists */
				file = fopen(old_path, "r");
				if (file) {
					old_stats_file_exists = TRUE;
					fclose(file);
				}

				if (rename(path, old_path)) {
					OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
						"ERR 0603: Failed when tried to rename \'%s\' to \'%s\': %s\n",
						path, old_path, strerror(errno));
					goto Exit;
				}

				file = fopen(path, "a");
				if (!file) {
					OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
						"ERR 0604: Cannot create statistics file \'%s\': %s\n",
						path, strerror(errno));
					goto Exit;
				}

				OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
					"Statistics file was updated due to change in the columns. "
					"Old statistics were saved in the file %s."
					"%s\n", old_path, old_stats_file_exists ?
					" Previous old statistics file was overridden" : "");
			}
		}

		if (fstat(fileno(file), &file_stat)){
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
				"fstat failed, cannot get dump file statistics,"
				" %s\n", strerror(errno));
			fclose(file);
			goto Exit;
		}
		p_osm->statistic_dump_file[sm_port_idx] = file;
		/* convert size in MB to bytes */
		p_osm->statistic_max_size = p_osm->subn.opt.osm_stats_dump_limit << 20;
		p_osm->statistic_file_size[sm_port_idx] = file_stat.st_size;
		if (p_osm->statistic_file_size[sm_port_idx] == 0)
			print_stat_header(p_osm, header, sm_port_idx);
	}

	if (p_osm->statistic_max_size &&
	    ((is_main_file && (p_osm->statistic_file_size[sm_port_idx] > p_osm->statistic_max_size)) ||
	     *p_is_main_file_trancated)) {
		if (is_main_file) {
			OSM_LOG(&p_osm->log, OSM_LOG_VERBOSE, "Truncating main statistic dump file\n");
			*p_is_main_file_trancated = TRUE;
		} else {
			OSM_LOG(&p_osm->log, OSM_LOG_VERBOSE,
				"Truncating statistic dump file for SM port index: %d\n", sm_port_idx);
		}

		osm_truncate_file(p_osm->statistic_dump_file[sm_port_idx], &p_osm->statistic_file_size[sm_port_idx]);
		print_stat_header(p_osm, header, sm_port_idx);
	}

	get_date_str(date_str, MAX_STR_LEN, FALSE);
	p_osm->statistic_file_size[sm_port_idx] += fprintf(file,"%s, ", date_str);

	if (is_main_file) {
		for (i = 0; i < OSM_STATISTIC_COUNTERS_SIZE; i++) {
			counter = 0;

			for (port_idx = 0; port_idx < OSM_MAX_BINDING_PORTS; port_idx++)
				counter += p_osm->stats_counters[port_idx][i];

			p_osm->statistic_file_size[sm_port_idx] +=
				fprintf(file, "%" PRIu64 "%s", counter,
					(i == OSM_STATISTIC_COUNTERS_SIZE -1 ? "":","));
		}
	} else if (p_osm->subn.opt.osm_stats_dump_per_sm_port) {
		for (i = 0; i < OSM_STATISTIC_COUNTERS_SIZE; i++) {
			p_osm->statistic_file_size[sm_port_idx] +=
				fprintf(file, "%" PRIu64 "%s", p_osm->stats_counters[sm_port_idx][i],
					(i == OSM_STATISTIC_COUNTERS_SIZE - 1 ? "":","));
		}
	}

	p_osm->statistic_file_size[sm_port_idx] += fprintf(file,"\n");
	fflush(file);

Exit:
	if (tmp_buf)
		free(tmp_buf);
}

void osm_dump_statistic(osm_opensm_t * p_osm) {
	int port_idx;
	boolean_t is_main_file_trancated = FALSE;

	osm_dump_statistic_per_port(p_osm, OSM_MAX_BINDING_PORTS, &is_main_file_trancated);
	if (p_osm->subn.opt.osm_stats_dump_per_sm_port)
		for (port_idx = 0; port_idx < p_osm->subn.opt.num_guids; port_idx++)
			osm_dump_statistic_per_port(p_osm, port_idx, &is_main_file_trancated);
}

void osm_sm_dump_stat_callback(IN void *context)
{
	osm_sm_t *sm = context;
	OSM_LOG_ENTER(sm->p_log);

	if(sm->p_subn->opt.osm_stats_interval){
		CL_PLOCK_ACQUIRE(sm->p_lock);
		osm_sm_signal(sm, OSM_SIGNAL_DUMP_STATISTIC);
		cl_timer_start(&sm->dump_statistics_timer,
			       sm->p_subn->opt.osm_stats_interval);
		CL_PLOCK_RELEASE(sm->p_lock);
	}

	OSM_LOG_EXIT(sm->p_log);
}

void osm_dump_log_perf(IN osm_opensm_t *p_osm, osm_perflog_event_t *p_perflog_event,
		       osm_perflog_event_state_enum event_state)
{
	char end_date_str[MAX_STR_LEN];
	char start_date_str[MAX_STR_LEN];
	char path[256];
	char backup_path[256];
	uint64_t duration_ms;
	struct timeval tv_end_time;
	unsigned long current_len, diff;
	FILE *file = p_osm->perflog_mgr.perflog_dump_file;
	unsigned long report_size;
	int result;

	OSM_LOG_ENTER(&p_osm->log);

	if (event_state == OSM_PERFLOG_EVENT_START) {
		current_len = strlen(p_osm->perflog_mgr.current_str);
		if (p_perflog_event->subtask_num == 1) {
			snprintf(p_osm->perflog_mgr.current_str + current_len, p_osm->perflog_mgr.perflog_max_size - current_len,
				 "%*s\"subtasks\": [\n", p_perflog_event->indent - OSM_PERFLOG_KEY_INDENT, "");
			current_len = strlen(p_osm->perflog_mgr.current_str);
		}
		snprintf(p_osm->perflog_mgr.current_str + current_len, p_osm->perflog_mgr.perflog_max_size - current_len,
			 "%s  %*s{\n"
			 "%*s\"name\": \"%s\",\n",
			 p_perflog_event->subtask_num > 1 ? ",\n" : "",
			 p_perflog_event->indent - OSM_PERFLOG_KEY_INDENT, "",
			 p_perflog_event->indent, "", p_perflog_event->event_name);
	} else {
		gettimeofday(&tv_end_time, NULL);
		duration_ms = ((uint64_t) (tv_end_time.tv_sec - p_perflog_event->start_time.tv_sec) * 1000000) +
			      ((uint64_t) (tv_end_time.tv_usec - p_perflog_event->start_time.tv_usec));
		sprint_time_str(&p_perflog_event->start_time, start_date_str, MAX_STR_LEN, TRUE);
		sprint_time_str(&tv_end_time, end_date_str, MAX_STR_LEN, TRUE);
		current_len = strlen(p_osm->perflog_mgr.current_str);
		if (p_perflog_event->subtasks) {
			snprintf(p_osm->perflog_mgr.current_str + current_len,
				 p_osm->perflog_mgr.perflog_max_size - current_len,
				 "\n%*s],\n", p_perflog_event->indent, "");
			current_len = strlen(p_osm->perflog_mgr.current_str);
		}
		snprintf(p_osm->perflog_mgr.current_str + current_len, p_osm->perflog_mgr.perflog_max_size - current_len,
			 "%*s\"start_time\": \"%s\",\n"
			 "%*s\"end_time\": \"%s\",\n"
			 "%*s\"duration\": %lu\n"
			 "  %*s}",
			 p_perflog_event->indent, "", start_date_str,
			 p_perflog_event->indent, "", end_date_str,
			 p_perflog_event->indent, "", duration_ms,
			 p_perflog_event->indent - OSM_PERFLOG_KEY_INDENT, "");
		if (cl_is_qlist_empty(&p_osm->subn.perflog_list)) {
			if (!file) {
				snprintf(path, sizeof(path), "%s/%s.json",
					 p_osm->subn.opt.dump_files_dir, PERFLOG_DUMP_FILE);

				file = fopen(path, "r");
				if (file) {
					fclose(file);
					snprintf(backup_path, sizeof(backup_path), "%s/%s.backup.json",
						 p_osm->subn.opt.dump_files_dir, PERFLOG_DUMP_FILE);
					if (rename(path, backup_path)) {
						OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
							"ERR 0642: Failed when tried to rename \'%s\' to \'%s\': %s\n",
							path, backup_path, strerror(errno));
						goto Exit;
					}
				}
				file = fopen(path, "w");
				if (!file) {
					OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
						"ERR 0635: Cannot create performance log file \'%s\': %s\n",
						path, strerror(errno));
					goto Exit;
				}
				p_osm->perflog_mgr.perflog_dump_file = file;
				fprintf(file, "[\n");
				fprintf(file, "]\n");
			}
			report_size = strlen(p_osm->perflog_mgr.current_str) + 1;
			diff = p_osm->perflog_mgr.perflog_max_size - report_size;
			if (p_osm->perflog_mgr.perflog_file_size > diff) {
				fclose(file);

				snprintf(path, sizeof(path), "%s/%s.json",
					 p_osm->subn.opt.dump_files_dir, PERFLOG_DUMP_FILE);
				snprintf(backup_path, sizeof(backup_path), "%s/%s.backup.json",
					 p_osm->subn.opt.dump_files_dir, PERFLOG_DUMP_FILE);

				if (rename(path, backup_path)) {
					OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
						"ERR 0638: Failed when tried to rename \'%s\' to \'%s\': %s\n",
						path, backup_path, strerror(errno));
					goto Exit;
				}

				file = fopen(path, "w");
				if (!file) {
					OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
						"ERR 0639: Cannot create performance log file \'%s\': %s\n",
						path, strerror(errno));
					goto Exit;
				}
				p_osm->perflog_mgr.perflog_dump_file = file;
				p_osm->perflog_mgr.perflog_file_size = 0;
				p_osm->perflog_mgr.reports_num = 0;
				fprintf(file, "[\n");
				fprintf(file, "]\n");
			}

			p_osm->perflog_mgr.perflog_file_size += report_size;
			p_osm->perflog_mgr.reports_num++;
			if (p_osm->perflog_mgr.reports_num > 1) {
				result = fseek(file, -3, SEEK_END);
				if (result) {
					OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
						"ERR 0640: fseek failed with return value (%d), cannot "
						" write performance report\n", result);
					goto Exit;
				}
				fprintf(file, ",\n");
			} else {
				result = fseek(file, -2, SEEK_END);
				if (result) {
					OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
						"ERR 0641: fseek failed with return value (%d), cannot "
						" write performance report\n", result);
					goto Exit;
				}
			}

			fprintf(file, "%s\n]\n", p_osm->perflog_mgr.current_str);
			*p_osm->perflog_mgr.current_str = '\0';
			fflush(file);
		}
	}

Exit:
	OSM_LOG_EXIT(&p_osm->log);
}

static void perflog_pop_sub_events(osm_opensm_t *p_osm, char *event_name)
{
	osm_subn_t *p_subn = &p_osm->subn;
	boolean_t is_start_event_exists = FALSE;
	osm_perflog_event_t *p_perflog_event;
	cl_list_item_t *p_item = NULL;

	for (p_item = cl_qlist_head(&p_subn->perflog_list);
	     p_item != cl_qlist_end(&p_subn->perflog_list);
	     p_item = cl_qlist_next(p_item)) {
		p_perflog_event = (osm_perflog_event_t *) p_item;
		if (!strcmp(p_perflog_event->event_name, event_name)) {
			is_start_event_exists = TRUE;
			break;
		}
	}

	if (is_start_event_exists) {
		/* Pop all the sub events of the given event including the event */
		p_perflog_event = (osm_perflog_event_t *) cl_qlist_remove_tail(&p_subn->perflog_list);
		while (strcmp(p_perflog_event->event_name, event_name)) {
			osm_dump_log_perf(p_osm, p_perflog_event, OSM_PERFLOG_EVENT_END);
			free(p_perflog_event);
			p_perflog_event = (osm_perflog_event_t *) cl_qlist_remove_tail(&p_subn->perflog_list);
		}

		osm_dump_log_perf(p_osm, p_perflog_event, OSM_PERFLOG_EVENT_END);
		free(p_perflog_event);
	}
}

void osm_perflog(osm_opensm_t *p_osm, const char * base_name, const char *suffix,
		 osm_perflog_event_state_enum event_state)
{
	osm_subn_t *p_subn = &p_osm->subn;
	osm_perflog_event_t *p_perflog_event_current = NULL;
	osm_perflog_event_t *p_perflog_event;
	cl_list_item_t *p_list_item = NULL;
	char env_var_name[sizeof(PERFLOG_DELAY_PREFIX) + OSM_PERFLOG_EVENT_NAME_LEN + 1];
	char name[OSM_PERFLOG_EVENT_NAME_LEN];
	char *seconds_str = NULL;
	long seconds = 0;

	if ((base_name == NULL) || (suffix == NULL))
		return;

	if (cl_is_qlist_empty(&p_subn->perflog_list) && event_state == OSM_PERFLOG_EVENT_END)
		return;

	if (strcmp(suffix, ""))
		snprintf(name, sizeof(name), "%s_%s", base_name, suffix);
	else
		snprintf(name, sizeof(name), "%s", base_name);

	if (event_state == OSM_PERFLOG_EVENT_START) {
		snprintf(env_var_name, sizeof(env_var_name), "%s_%s", PERFLOG_DELAY_PREFIX, name);
		seconds_str = getenv(env_var_name);
		if (seconds_str != NULL) {
			seconds = strtol(seconds_str, NULL, 0);
			/* Check that number of seconds can be used by sleep */
			if (0 < seconds && seconds < UINT32_MAX) {
				OSM_LOG(&p_subn->p_osm->log, OSM_LOG_INFO,
					"'%s': Start sleeping for: %ld\n", name, seconds);
				sleep(seconds);
			}
		}
	}
	p_list_item = cl_qlist_tail(&p_subn->perflog_list);
	if (!cl_is_qlist_empty(&p_subn->perflog_list))
		p_perflog_event_current = (osm_perflog_event_t *) p_list_item;
	if (p_perflog_event_current && !strcmp(p_perflog_event_current->event_name, name)) {
		/* Same task can't start before the first one ends */
		OSM_ASSERT(event_state != OSM_PERFLOG_EVENT_START);
		if (event_state == OSM_PERFLOG_EVENT_START)
			return;

		if (event_state == OSM_PERFLOG_EVENT_END) {
			p_perflog_event_current = (osm_perflog_event_t *) cl_qlist_remove_tail(&p_subn->perflog_list);
			osm_dump_log_perf(p_osm, p_perflog_event_current, event_state);
			free(p_perflog_event_current);
		}
	} else {
		/* Cant end a task different from the current one */
		OSM_ASSERT(event_state != OSM_PERFLOG_EVENT_END);

		if (event_state == OSM_PERFLOG_EVENT_START) {
			p_perflog_event = (osm_perflog_event_t *)calloc(1, sizeof(osm_perflog_event_t));
			if (!p_perflog_event) {
				OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
					"ERR 0634: Not enough memory to add perflog data\n");
				return;
			}

			if (p_perflog_event_current) {
				p_perflog_event_current->subtasks++;
				p_perflog_event->subtask_num = p_perflog_event_current->subtasks;
			}

			strncpy(p_perflog_event->event_name, name, OSM_PERFLOG_EVENT_NAME_LEN - 1);
			gettimeofday(&p_perflog_event->start_time, NULL);
			p_perflog_event->indent = (cl_qlist_count(&p_subn->perflog_list) + 1) * OSM_PERFLOG_KEY_INDENT;
			cl_qlist_insert_tail(&p_subn->perflog_list, &p_perflog_event->list_item);

			osm_dump_log_perf(p_osm, p_perflog_event, event_state);
		} else if (event_state == OSM_PERFLOG_EVENT_END) {
			perflog_pop_sub_events(p_osm, name);
		}
	}
}

static unsigned int dump_smdb_sm_tbl(FILE *file, void *context)
{
	osm_subn_t *p_subn = (osm_subn_t *)context;
	char host_str[MAX_STR_LEN];
	pid_t pid = getpid();

	fprintf(file,
		"Version, PID, HostName, PortGUID, SubnetPrefix, LID, StartTime, RoutingEngine\n");
	if (gethostname(host_str, MAX_STR_LEN))
		host_str[0] = '\0';

	fprintf(file, "%s, "			/* Version */
		      "%d, "			/* PID */
		      "%s, "			/* HostName */
		      "0x%016" PRIx64 ", "	/* PortGUID */
		      "0x%016" PRIx64 ", "	/* SubnetPrefix */
		      "%d, "			/* LID */
		      "%s, "			/* StartTime */
		      "%s\n",			/* RoutingEngine */
		      OSM_VERSION, pid, host_str,
		      cl_ntoh64(p_subn->p_osm->subn.sm_port_guid),
		      cl_ntoh64(p_subn->opt.subnet_prefix),
		      cl_ntoh16(p_subn->p_osm->subn.sm_base_lid),
		      p_subn->p_osm->start_time_str,
		      (p_subn->routing_engine_used ?
		       p_subn->routing_engine_used->get_name(p_subn->routing_engine_used) :
		       "N/A"));

	/* single row table */
	return 1;
}

static unsigned int dump_smdb_sms_tbl(FILE *file, void *context)
{
	osm_subn_t *p_subn = (osm_subn_t *)context;
	osm_port_t *p_sm_port;
	cl_map_item_t *p_map_item;
	unsigned int rows = 0;
	osm_remote_sm_t *p_remote_sm;

	fprintf(file,
		"PortGUID, LID, Priority, State\n");

	fprintf(file,
		"0x%016" PRIx64 ", "	/* PortGUID */
		"%5d, "			/* LID */
		"%2d, "			/* Priority */
		"%2d\n",			/* State */
		cl_ntoh64(p_subn->p_osm->subn.sm_port_guid),
		cl_ntoh16(p_subn->p_osm->subn.sm_base_lid),
		p_subn->p_osm->subn.active_sm_priority,
		p_subn->p_osm->subn.sm_state);
	rows++;

	for (p_map_item = cl_qmap_head(&p_subn->sm_guid_tbl);
	     p_map_item != cl_qmap_end(&p_subn->sm_guid_tbl);
	     p_map_item = cl_qmap_next(p_map_item)) {
		p_remote_sm = (osm_remote_sm_t *)p_map_item;
		p_sm_port = osm_get_port_by_guid(p_subn, p_remote_sm->smi.guid);
		fprintf(file,
			"0x%016" PRIx64 ", "	/* PortGUID */
			"%5d, "			/* LID */
			"%2d, "			/* Priority */
			"%2d\n",			/* State */
			cl_ntoh64(p_remote_sm->smi.guid),
			p_sm_port ? cl_ntoh16(osm_port_get_base_lid(p_sm_port)) : 0,
			ib_sminfo_get_priority(&p_remote_sm->smi),
			ib_sminfo_get_state(&p_remote_sm->smi));
		rows++;
	}

	return rows;
}

static unsigned int dump_smdb_sm_ports_tbl(FILE *file, void *context)
{
	osm_subn_t *p_subn = (osm_subn_t *)context;
	unsigned int rows = 0;
	int i;

	fprintf(file,
		"PortGUID, Status\n");

	for (i = 0; i < p_subn->opt.num_guids; i++) {
		fprintf(file,
			"0x%016" PRIx64 ", "	/* PortGUID */
			"%9s\n",		/* Status */
			cl_ntoh64(p_subn->opt.guid[i]),
			p_subn->active_sm_guid[i] ? "VALID" : "INVALID");
		rows++;
	}

	return rows;
}

static unsigned int dump_smdb_nodes_tbl(FILE *file, void *context)
{
	osm_subn_t *p_subn = (osm_subn_t *)context;
	cl_map_item_t *p_map_item;
	osm_node_t *p_node;
	unsigned int rows = 0;

	fprintf(file,
		"SystemImageGUID, NodeGUID, NodeType, ExtNodeType, NumPorts, "
		"VendorID, DeviceID, NodeDesc\n");

	for (p_map_item = cl_qmap_head(&p_subn->node_guid_tbl);
	     p_map_item != cl_qmap_end(&p_subn->node_guid_tbl);
	     p_map_item = cl_qmap_next(p_map_item)) {
		p_node = (osm_node_t *)p_map_item;
		fprintf(file,
			"0x%016" PRIx64 ", "	/* SystemImageGUID */
			"0x%016" PRIx64 ", "	/* NodeGUID */
			"%d, "			/* NodeType */
			"%d, "			/* ExtendedNodeType */
			"%3d, "			/* NumPorts */
			"0x%08x, "		/* VendorID */
			"0x%04x, "		/* DeviceID */
			"\"%s\"\n",		/* NodeDesc */
			cl_ntoh64(p_node->node_info.sys_guid),
			cl_ntoh64(osm_node_get_node_guid(p_node)),
			osm_node_get_type(p_node),
			osm_node_get_extended_node_type(p_node),
			p_node->node_info.num_ports,
			cl_ntoh32(ib_node_info_get_vendor_id(&p_node->node_info)),
			cl_ntoh16(p_node->node_info.device_id),
			p_node->print_desc);
		rows++;
	}

	return rows;
}

static unsigned int dump_smdb_switches_tbl(FILE *file, void *context)
{
	osm_subn_t *p_subn = (osm_subn_t *)context;
	cl_map_item_t *p_map_item;
	osm_node_t *p_node;
	osm_switch_t *p_sw;
	unsigned int rows = 0;

	fprintf(file, "NodeGUID, NumPorts, Status, Rank, Coordinate\n");

	for (p_map_item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     p_map_item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     p_map_item = cl_qmap_next(p_map_item)) {
		p_sw = (osm_switch_t *)p_map_item;
		p_node = p_sw->p_node;
		fprintf(file,
			"0x%016" PRIx64 ", "	/* NodeGUID */
			"%3d, "			/* NumPorts */
			"%9s, "			/* Status */
			"%3d, "			/* Rank */
			"0x%04x\n",		/* Coordinate */
			cl_ntoh64(osm_node_get_node_guid(p_node)),
			p_node->node_info.num_ports,
			(lookup_held_back_switch(p_subn, p_node) ?
			 "HELD_BACK" : "VALID"),
			p_sw->rank,
			p_sw->coord);
		rows++;
	}

	return rows;
}

static inline void dump_smdb_ports_tbl_record(FILE *file, osm_sm_t *p_sm,
					      boolean_t is_held_back_sw,
					      osm_node_t *p_node,
					      osm_port_t *p_port,
					      osm_physp_t *p_physp,
					      uint8_t special_port_type)
{
	osm_physp_t *p_default_physp;
	boolean_t is_isolated_port;
	uint8_t port_num;

	port_num = osm_physp_get_port_num(p_physp);
	is_isolated_port =
	    is_held_back_sw || lookup_isolated_port(p_sm, p_node, port_num);

	if (p_node->node_info.node_type == IB_NODE_TYPE_SWITCH)
		p_default_physp = osm_node_get_physp_ptr(p_node, 0);
	else
		p_default_physp = p_physp;

	fprintf(file,
		"0x%016" PRIx64 ", "	/* NodeGUID */
		"%3d, "			/* PortNum */
		"0x%016" PRIx64 ", "	/* PortGUID */
		"%5d, "			/* LID */
		"%d, "			/* LMC */
		"%3s, "			/* PortState */
		"%3s, "			/* LinkWidth */
		"%5s, "			/* LinkSpeed */
		"%8s, "			/* Status */
		"%7s, "			/* DiscThroughState */
		"%d, "			/* SpecialPortType */
		"%s, "			/* Timestamp */
		"%s",			/* PortLabel */
		cl_ntoh64(osm_node_get_node_guid(p_node)),
		port_num,
		cl_ntoh64(p_physp->port_guid),
		cl_ntoh16(osm_physp_get_base_lid(p_physp)),
		osm_physp_get_lmc(p_physp),
		get_physp_port_state_str(p_physp),
		get_physp_link_width_act_str(p_physp,
					     p_default_physp->port_info.capability_mask2),
		get_physp_speed_act_str(p_physp),
		is_isolated_port ? "ISOLATED" : "VALID",
		get_disc_state_str(p_port->disc_through),
		special_port_type,
		p_physp->state_timestamp_str,
		p_physp->hierarchy_info.port_label);

	if (p_sm->p_subn->is_planarized) {
		fprintf(file,
		", %3d, "		/* APort */
		"%d, "			/* PlaneNumber */
		"%d",			/* NumberOfPlanes */
		p_physp->hierarchy_info.aport, p_physp->hierarchy_info.plane_number,
		p_physp->hierarchy_info.num_of_planes);
	}

	fprintf(file, "\n");
}

static unsigned int dump_smdb_ports_tbl(FILE *file, void *context)
{
	osm_subn_t *p_subn = (osm_subn_t *)context;
	osm_sm_t *p_sm = &p_subn->p_osm->sm;
	cl_map_item_t *p_map_item;
	osm_node_t *p_node;
	osm_port_t *p_port;
	osm_physp_t *p_physp;
	osm_physp_t *p_remote_physp;
	uint8_t port, start_port, end_port;
	uint8_t special_port_type;
	boolean_t is_held_back_sw;
	unsigned int rows = 0;

	fprintf(file,
		"NodeGUID, PortNum, PortGUID, LID, LMC, PortState, "
		"LinkWidth, LinkSpeed, Status, DiscThroughState, SpecialPortType, Timestamp, PortLabel");

	if (p_subn->is_planarized) {
		fprintf(file, ", APort, PlaneNumber, NumberOfPlanes");
	}

	fprintf(file, "\n");

	for (p_map_item = cl_qmap_head(&p_subn->port_guid_tbl);
	     p_map_item != cl_qmap_end(&p_subn->port_guid_tbl);
	     p_map_item = cl_qmap_next(p_map_item)) {
		p_port = (osm_port_t *)p_map_item;
		p_node = p_port->p_node;

		if (p_node->sw) {
			/* Write switch ports */
			is_held_back_sw =
			    p_node->sw &&
			    lookup_held_back_switch(p_subn, p_node);
			start_port = 0;
			end_port = osm_node_get_num_physp(p_node);
		} else {
			/* non-switch port */
			p_physp = osm_port_get_physp(p_port);
			if (!p_physp || !osm_physp_is_valid(p_physp))
				continue;

			is_held_back_sw = FALSE;
			if (osm_hierarchy_is_planarized_physp(p_physp)) {
				start_port = 1;
				end_port = osm_node_get_num_physp(p_node);
			} else {
				start_port = p_physp->port_num;
				end_port = start_port + 1;
			}
		}

		for (port = start_port; port < end_port; port++) {
			p_physp = osm_node_get_physp_ptr(p_node, port);
			if (!p_physp || !osm_physp_is_valid(p_physp))
				continue;
			if (port && !osm_physp_get_remote(p_physp))
				/* Skip disconnect port */
				continue;

			p_remote_physp = osm_physp_get_remote(p_physp);
			if (p_remote_physp)
				special_port_type =
					ib_mlnx_epi_get_special_port_type_val(
							&p_remote_physp->ext_port_info);
			else
				special_port_type = 0;


			dump_smdb_ports_tbl_record(file, p_sm,
					is_held_back_sw,
					p_node, p_port,
					p_physp, special_port_type);
			rows++;
		}
	}


	return rows;
}

static unsigned int dump_smdb_links_tbl(FILE *file, void *context)
{
	osm_subn_t *p_subn = (osm_subn_t *)context;
	cl_map_item_t *p_map_item;
	osm_switch_t *p_sw;
	osm_node_t *p_node, *p_rem_node;
	uint64_t node_guid, rem_node_guid;
	uint8_t rem_port;
	uint8_t port;
	unsigned int rows = 0;

	fprintf(file,
		"NodeGUID1, PortNum1, NodeGUID2, PortNum2"
		"\n");

	for (p_map_item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     p_map_item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     p_map_item = cl_qmap_next(p_map_item)) {
		p_sw = (osm_switch_t *)p_map_item;
		p_node = p_sw->p_node;
		for (port = 1; port < osm_node_get_num_physp(p_node); port++) {
			p_rem_node =
			    osm_node_get_remote_node(p_node, port, &rem_port);
			if (!p_rem_node)
				continue;

			node_guid = cl_ntoh64(osm_node_get_node_guid(p_node));
			rem_node_guid = cl_ntoh64(osm_node_get_node_guid(p_rem_node));

			/* Write one direction only */
			if (osm_node_get_type(p_rem_node) == IB_NODE_TYPE_SWITCH &&
			    ((node_guid > rem_node_guid) ||
			     (node_guid == rem_node_guid && port > rem_port)))
				continue;

			fprintf(file,
				"0x%016" PRIx64 ", "	/* NodeGUID1 */
				"%3d, "			/* PortNum1 */
				"0x%016" PRIx64 ", "	/* NodeGUID2 */
				"%3d\n",		/* PortNum2 */
				node_guid,
				port,
				rem_node_guid,
				rem_port);
			rows++;
		}
	}

	return rows;
}

static
boolean_t dump_is_valid_an2an_entry(osm_subn_t *p_subn,
				   osm_db_an2an_elem_t *p_item)
{
	osm_switch_t *p_sw1 = osm_get_switch_by_guid(p_subn, p_item->guid1);
	osm_switch_t *p_sw2 = osm_get_switch_by_guid(p_subn, p_item->guid2);
	return p_sw1 && p_sw2 &&
	       osm_is_same_physical_link(&p_subn->p_osm->sm, p_sw1, p_sw2,
				         p_item->portnum1, p_item->portnum2);
}

static unsigned int dump_smdb_gpus_tbl(FILE *file, void *context)
{
	osm_subn_t *p_subn = (osm_subn_t *)context;
	cl_map_item_t *p_map_item;
	osm_node_t *p_node;
	osm_gpu_t *p_gpu;
	unsigned int rows = 0;

	fprintf(file, "NodeGUID, ALID, NumALIDs, NumPorts\n");

	for (p_map_item = cl_qmap_head(&p_subn->gpu_guid_tbl);
	     p_map_item != cl_qmap_end(&p_subn->gpu_guid_tbl);
	     p_map_item = cl_qmap_next(p_map_item)) {
		p_gpu = (osm_gpu_t *)p_map_item;
		p_node = p_gpu->p_node;
		fprintf(file,
			"0x%016" PRIx64 ", "	/* NodeGUID */
			"%5d, "			/* ALID */
			"%d, "			/* NumALIDs */
			"%3d\n",		/* NumPorts */
			cl_ntoh64(osm_node_get_node_guid(p_node)),
			osm_gpu_get_used_alid_ho(p_gpu),
			osm_gpu_get_num_alids(p_gpu),
			p_node->node_info.num_ports);
		rows++;
	}

	return rows;
}

static boolean_t is_in_engine_subn(osm_subn_t *p_engine_subn, osm_db_an2an_elem_t *p_item)
{
	return (osm_get_switch_by_guid(p_engine_subn, p_item->guid1) &&
		osm_get_switch_by_guid(p_engine_subn, p_item->guid2));
}

static unsigned int dump_smdb_an2an_tbl(FILE *file, void *context)
{
	osm_subn_t *p_subn = (osm_subn_t *)context;
	osm_subn_t *p_main_subn = p_subn->p_osm->sm.p_subn;
	cl_qlist_t entries;
	cl_list_item_t *p_item;
	osm_db_an2an_elem_t *p_elem;
	unsigned int rows = 0;

	cl_qlist_init(&entries);
	if (osm_db_an2an_get_all(p_main_subn->p_an2an, &entries)) {
		OSM_LOG(&p_main_subn->p_osm->log, OSM_LOG_ERROR, "ERR 0633: "
			"could not get an2an entry list\n");
		return 0;
	}

	fprintf(file, "NodeGUID1, PortNum1, NodeGUID2, PortNum2, Timestamp\n");
	while ((p_item = cl_qlist_remove_head(&entries)) !=
	       cl_qlist_end(&entries)) {
		p_elem = (osm_db_an2an_elem_t *)p_item;
		if ((p_main_subn == p_subn || is_in_engine_subn(p_subn, p_elem)) &&
		    dump_is_valid_an2an_entry(p_main_subn, p_elem)) {
			fprintf(file, "0x%016" PRIx64 ", %3d, 0x%016" PRIx64 ", %3d, %s\n",
				cl_ntoh64(p_elem->guid1), p_elem->portnum1,
				cl_ntoh64(p_elem->guid2), p_elem->portnum2, p_elem->timestamp_str);
			rows++;
		}
		free(p_item);
	}
	return rows;
}

static unsigned int dump_smdb_index_tbl(FILE *file, void *context)
{
	osm_subn_t *p_subn = ((smdb_tbl_context_t *)context)->p_subn;
	smdb_tbl_record_t *smdb_tbls = ((smdb_tbl_context_t *)context)->p_records;
	boolean_t dump_all = ((smdb_tbl_context_t *)context)->dump_all;
	smdb_tbl_record_t *p_tbl;
	unsigned int rows = 0;

	fprintf(file, "Name, Offset, Size, Line, Rows\n");

	for (p_tbl = &smdb_tbls[0]; p_tbl->name != NULL; p_tbl++) {
		if (!p_tbl->display_section(p_subn) ||
		    (!dump_all && !p_tbl->sub_topology_table))
			continue;
		/*
		 * Padding is required in order to ensure that table size would
		 * not change when we rewrite it.
		 */
		fprintf(file, "%-10s, %11ld, %11ld, %11ld, %11ld\n",
			p_tbl->name, p_tbl->offset, p_tbl->size, p_tbl->line,
			p_tbl->rows);
		rows++;
	}

	return rows;
}

static char *dump_get_smdb_file_name(osm_subn_t *p_subn, char *buf, size_t n, int chain_id)
{
	if (chain_id == 0)
		snprintf(buf, n, "%s/%s.dump",
			 p_subn->opt.dump_files_dir, SMDB_DUMP_FILE);
	else
		snprintf(buf, n, "%s/%s.%d.dump",
			 p_subn->opt.dump_files_dir, SMDB_DUMP_FILE, chain_id);

	return buf;
}

static inline boolean_t dump_is_routing_chains(osm_subn_t *p_subn)
{
	return osm_routing_engine_is_chain(p_subn->routing_engine_used);
}

static unsigned int dump_smdb_chains_tbl(FILE *file, void *context)
{
	char path[256];
	osm_subn_t *p_subn = (osm_subn_t *)context;
	osm_sm_t *p_sm = &p_subn->p_osm->sm;
	osm_routing_chain_mgr_t *p_rch_mgr = &p_sm->routing_chain_mgr;
	unsigned int rows = 0;
	size_t i;
	osm_routing_chain_t *p_engine;

	fprintf(file,
		"TopologyID, RoutingEngine, SMDBFile\n");

	if (!dump_is_routing_chains(p_subn))
		goto exit;

	for (i = 0; i <= MAX_ENGINES_NUMBER; i++) {
		if (p_rch_mgr->used_ids[i]) {
			p_engine = osm_rch_get_engine(p_subn->p_osm, i);
			dump_get_smdb_file_name(p_subn, path, sizeof(path),
						p_subn == p_engine->subn ? 0 :
						p_engine->id);
			fprintf(file,
				"0x%016" PRIx64 ", "	/* TopologyID */
				"%10s, "		/* RoutingEngine */
				"%s\n",			/* SMDBFile */
				p_engine->topo_id,
				p_engine->engine,
				path);
			rows++;
		}
	}

exit:
	return rows;
}

static inline void dump_smdb_tbl(FILE *file, const char *name,
				 dump_smdb_tbl_pfn_t dump_tbl_pfn,
				 void *context,
				 smdb_tbl_record_t *p_tbl,
				 unsigned int *p_line)
{
	fprintf(file, "\nSTART_%s\n", name);

	/*
	 * p_line points to the line in the file at the time dump_smdb_tbl
	 * was invoked. Set p_tbl->line to current line in the file (after
	 * writing start section token).
	 */
	p_tbl->line = *p_line + 2;
	p_tbl->offset = ftell(file);
	p_tbl->rows = dump_tbl_pfn(file, context);
	p_tbl->size = ftell(file) - p_tbl->offset;

	fprintf(file, "END_%s\n", name);

	/*
	 * Increase p_line by: section header (2) + table field names (1) +
	 * table rows (p_tbl->rows) + section footer (1)
	 */
	*p_line += p_tbl->rows + 4;
}

static void dump_smdb_tables(FILE *file, void *cxt,
			     smdb_tbl_record_t *p_smdb_tbl_records,
			     const char *name,
			     boolean_t dump_all)
{
	char date_str[MAX_STR_LEN];
	osm_subn_t *p_subn = (osm_subn_t *)cxt;
	osm_opensm_t *p_osm = p_subn->p_osm;
	smdb_tbl_record_t *p_tbl;
	smdb_tbl_record_t index_tbl;
	unsigned int line_ptr = 1;
	int result;

	smdb_tbl_context_t smdb_index_tbl_context = {
	    .p_subn = p_subn,
	    .p_records = p_smdb_tbl_records,
	    .dump_all = dump_all
	};

	/* Write file header */
	get_date_str(date_str, MAX_STR_LEN, TRUE);
	fprintf(file, "# %s, %s, %s\n", name, SMDB_FILE_VERSION, date_str);
	line_ptr++;

	/* Write index table with dummy values */
	dump_smdb_tbl(file, SMDB_INDEX_TABLE_NAME, dump_smdb_index_tbl,
		      &smdb_index_tbl_context, &index_tbl, &line_ptr);

	/* Write SMDB tables */
	for (p_tbl = &p_smdb_tbl_records[0]; p_tbl->name != NULL; p_tbl++){
		if (p_tbl->display_section(p_subn) && (dump_all || p_tbl->sub_topology_table))
			dump_smdb_tbl(file, p_tbl->name, p_tbl->dump_tbl_pfn, p_subn,
				      p_tbl, &line_ptr);
	}

	/* Rewrite index table with valid values */
	result = fseek(file, index_tbl.offset, SEEK_SET);
	if (result)
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR 0601: fseek failed with return value (%d), cannot "
			" update SMDB index table\n",
			result);
	else
		dump_smdb_index_tbl(file, &smdb_index_tbl_context);

}

static boolean_t dump_smdb_display_section(struct osm_subn * p_subn)
{
	return TRUE;
}

static boolean_t dump_smdb_display_gpu_nodes(struct osm_subn * p_subn)
{
	return osm_subn_is_nvlink_enabled(p_subn);
}

static smdb_tbl_record_t smdb_tbls_arr[] = {
	{ "SM",		dump_smdb_sm_tbl,	TRUE  ,	dump_smdb_display_section},
	{ "SM_PORTS",	dump_smdb_sm_ports_tbl,	FALSE ,	dump_smdb_display_section},
	{ "SMS",	dump_smdb_sms_tbl,	FALSE ,	dump_smdb_display_section},
	{ "NODES",	dump_smdb_nodes_tbl,	TRUE  ,	dump_smdb_display_section},
	{ "SWITCHES",	dump_smdb_switches_tbl,	TRUE  ,	dump_smdb_display_section},
	{ "GPU_NODES",	dump_smdb_gpus_tbl,	FALSE ,	dump_smdb_display_gpu_nodes},
	{ "PORTS",	dump_smdb_ports_tbl,	TRUE  ,	dump_smdb_display_section},
	{ "LINKS",	dump_smdb_links_tbl,	TRUE  ,	dump_smdb_display_section},
	{ "AN2AN",	dump_smdb_an2an_tbl,	TRUE  ,	dump_smdb_display_section},
	{ "CHAINS",	dump_smdb_chains_tbl,	FALSE ,	dump_smdb_display_section},
	{ NULL, NULL}
};

static void dump_smdb(FILE *file, void *cxt)
{

	dump_smdb_tables(file, cxt, smdb_tbls_arr, "SMDB", TRUE);
}

static void dump_smdb_sub_topology(FILE *file, void *cxt)
{

	dump_smdb_tables(file, cxt, smdb_tbls_arr, "SMDB", FALSE);
}

void osm_dump_sm_db(IN osm_opensm_t *p_osm)
{
	char path[256];
	size_t i;
	osm_sm_t *p_sm = &p_osm->sm;
	osm_routing_chain_mgr_t *p_rch_mgr = &p_sm->routing_chain_mgr;
	osm_routing_chain_t *p_engine;

	OSM_LOG_ENTER(&p_osm->log);

	snprintf(path, sizeof(path), "%s/%s.dump", p_sm->p_subn->opt.dump_files_dir,
		 SMDB_DUMP_FILE);

	osm_dump_func_to_file(p_osm, path, dump_smdb, p_sm->p_subn);

	if (!dump_is_routing_chains(p_sm->p_subn))
		goto exit;

	/* Create SMDB for sub topologies */
	for (i = 0; i <= MAX_ENGINES_NUMBER; i++) {
		if (p_rch_mgr->used_ids[i]) {
			p_engine = osm_rch_get_engine(p_osm, i);

			/* If engine run on main subnet, avoid dump */
			if (p_engine->subn == p_sm->p_subn)
				continue;

			snprintf(path, sizeof(path), "%s/%s.%zd.dump",
				 p_sm->p_subn->opt.dump_files_dir,
				 SMDB_DUMP_FILE, p_engine->id);
			osm_dump_func_to_file(p_osm, path, dump_smdb_sub_topology, p_engine->subn);
		}
	}

exit:
	OSM_LOG_EXIT(&p_osm->log);
}

static char* print_vport_state(osm_vport_t *p_vport)
{
	uint8_t state = ib_vport_info_get_vport_state(&p_vport->vport_info);

	if (state == IB_LINK_ACTIVE)
		return "ACTIVE";
	else if (state == IB_LINK_INIT)
		return "INIT";
	else if (state == IB_LINK_DOWN)
		return "DOWN";
	else
		return "UNKNOWN";
}

static unsigned int dump_virtualization_ports_tbl(FILE *file, void *context)
{
	osm_subn_t *p_subn = (osm_subn_t *)context;
	cl_map_item_t *p_map_item;
	osm_port_t *p_port;
	osm_physp_t *p_physp;
	ib_virtualization_info_t *p_vi;
	unsigned int rows = 0;

	fprintf(file, "PortGUID, LID, IndexCap, IndexTop\n");

	for (p_map_item = cl_qmap_head(&p_subn->port_guid_tbl);
	     p_map_item != cl_qmap_end(&p_subn->port_guid_tbl);
	     p_map_item = cl_qmap_next(p_map_item)) {
		p_port = (osm_port_t *)p_map_item;
		p_physp = osm_port_get_physp(p_port);

		p_vi = &p_port->virt_data.virtualization_info;
		if (!ib_virtualization_info_get_virtualization_enabled(p_vi))
			continue;

		fprintf(file,
			"0x%016" PRIx64 ", "	/* PortGUID */
			"%5d, "			/* LID */
			"%5d, "			/* IndexCap */
			"%5d\n",		/* IndexTop */
			cl_ntoh64(p_port->guid),
			cl_ntoh16(osm_physp_get_base_lid(p_physp)),
			cl_ntoh16(p_vi->vport_cap),
			cl_ntoh16(p_vi->vport_index_top));

		rows++;
	}

	return rows;
}

static unsigned int dump_virtualization_vnodes_tbl(FILE *file, void *context)
{
	osm_subn_t *p_subn = (osm_subn_t *)context;
	osm_sm_t *p_sm = &p_subn->p_osm->sm;
	osm_virt_mgr_t *p_virt_mgr = &p_sm->virt_mgr;
	cl_map_item_t *p_map_item;
	osm_vnode_t *p_vnode;

	unsigned int rows = 0;

	fprintf(file, "VNodeGUID, NumPorts, VNodeDesc\n");

	for (p_map_item = cl_qmap_head(&p_virt_mgr->vnodes);
	     p_map_item != cl_qmap_end(&p_virt_mgr->vnodes);
	     p_map_item = cl_qmap_next(p_map_item)) {
		p_vnode = (osm_vnode_t *)p_map_item;
		fprintf(file,
			"0x%016" PRIx64 ", "	/* VNodeGUID */
			"%3d, "			/* NumPorts */
			"\"%.*s\"\n",		/* VNodeDesc */
			cl_ntoh64(p_vnode->vnode_info.vnode_guid),
			p_vnode->vnode_info.num_ports,
			IB_NODE_DESCRIPTION_SIZE,
			(char *)&p_vnode->vnode_desc.description);
		rows++;
	}

	return rows;
}

static unsigned int dump_virtualization_vports_tbl(FILE *file, void *context)
{
	osm_subn_t *p_subn = (osm_subn_t *)context;
	cl_map_item_t *p_map_item;
	osm_port_t *p_port;
	ib_virtualization_info_t *p_vi;
	ib_vport_info_t *p_vpi;
	osm_vport_t *p_vport;
	uint16_t top, index;
	uint16_t vport_calc_lid, vport_lid_index;
	unsigned int rows = 0;
	uint8_t is_lid_required;

	fprintf(file,
		"PortGUID, Index, VPortGUID, VNodeGUID, VPortNum, State, "
		"IsLidRequired, VPortLidIndex, VPortLid, ActiveLid\n");

	for (p_map_item = cl_qmap_head(&p_subn->port_guid_tbl);
	     p_map_item != cl_qmap_end(&p_subn->port_guid_tbl);
	     p_map_item = cl_qmap_next(p_map_item)) {
		p_port = (osm_port_t *)p_map_item;
		p_vi = &p_port->virt_data.virtualization_info;
		if (!ib_virtualization_info_get_virtualization_enabled(p_vi))
			continue;

		top = cl_ntoh16(p_vi->vport_index_top);
		for (index = 0 ; index <= top ; index++) {
			p_vport = osm_port_get_vport(p_port, index);
			if (!p_vport)
				continue;

			p_vpi = &p_vport->vport_info;
			is_lid_required = ib_vport_info_get_lid_required(p_vpi);
			vport_lid_index = cl_ntoh16(p_vpi->lid_by_vport_index);
			vport_calc_lid = cl_ntoh16(osm_vport_calc_lid(p_vport));

			fprintf(file,
				"0x%016" PRIx64 ", "	/* PortGUID */
				"%5d, "			/* Index */
				"0x%016" PRIx64 ", "	/* VPortGUID */
				"0x%016" PRIx64 ", "	/* VNodeGUID */
				"%3d, "			/* VPortNum */
				"%s, "			/* State */
				"%d, "			/* IsLidRequired */
				"%5d, "			/* VPortLidIndex */
				"%5d, "			/* VPortLid */
				"%5d\n",		/* ActiveLid */
				cl_ntoh64(p_port->guid),
				index,
				cl_ntoh64(p_vport->vport_info.port_guid),
				cl_ntoh64(p_vport->vnode_guid),
				p_vport->vnode_port,
				print_vport_state(p_vport),
				is_lid_required,
				vport_lid_index,
				cl_ntoh16(osm_vport_get_lid(p_vport)),
				vport_calc_lid);

			rows++;
		}
	}

	return rows;
}

static void dump_virtualization(FILE *file, void *cxt)
{
	static smdb_tbl_record_t smdb_tbls[] = {
		{ "PORTS",	dump_virtualization_ports_tbl,	FALSE ,	dump_smdb_display_section},
		{ "VNODES",	dump_virtualization_vnodes_tbl,	FALSE ,	dump_smdb_display_section},
		{ "VPORTS",	dump_virtualization_vports_tbl,	FALSE ,	dump_smdb_display_section},
		{ NULL, NULL }
	};

	dump_smdb_tables(file, cxt, smdb_tbls, "VIRTUALIZATION_SMDB", TRUE);
}

void osm_dump_virtualization(osm_opensm_t * osm)
{
	char path[1024];
	if (osm_virt_mgr_is_dump_required(&osm->sm.virt_mgr) &&
	    (osm->sm.p_subn->opt.virt_enabled == OSM_CAPABILITY_ENABLED)) {
		if (osm->subn.opt.virt_dump_file)
			snprintf(path, sizeof(path), "%s", osm->subn.opt.virt_dump_file);
		else
			snprintf(path, sizeof(path), "%s/%s", osm->subn.opt.dump_files_dir, VIRT_DUMP_FILE);

		osm_dump_func_to_file(osm, path, dump_virtualization, osm->sm.p_subn);

		OSM_LOG(osm->sm.p_log, OSM_LOG_VERBOSE,"Dumped %" PRId64
			" vports to virtualization dump file\n",
			virt_dumped_port_counter);
		virt_dumped_port_counter = 0;
		osm_virt_mgr_set_dump_required(&osm->sm.virt_mgr, FALSE);
	}
}
