%{
/*
 * Copyright (c) 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Grammar of OSM Fast Recovery configuration parser.
 *
 * Author:
 *    Or Nechemia, NVIDIA
 */

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <math.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_FAST_RECOVERY_PARSER_Y_Y
#include <opensm/osm_opensm.h>

#define CHECK_PARAMS_PHYSP 	(uint8_t)1
#define CHECK_PARAMS_ALL 	(uint8_t)0xF

static void parser_tmp_struct_reset();
#define RESET_BUFFER  parser_tmp_struct_reset()

typedef struct fast_recovery_tmp_parser_struct_t_ {
	cl_list_t str_list;
	uint16_t num;
	boolean_t boolean;
} fast_recovery_tmp_parser_struct_t;

static struct osm_opensm *p_osm;
char *__fast_recovery_parser_yylval = NULL;
extern FILE * __fast_recovery_parser_in;
static fast_recovery_tmp_parser_struct_t fast_rec_tmp_parser_struct;
static cl_list_t port_group_list;
static ib_credit_watchdog_config_t credit_watchdog_parsed;
static ib_ber_config_t ber_config_parsed[IB_BER_NUM_TYPES];
uint8_t next_profile_num = 1;
static ib_ber_type_t ber_type = IB_BER_TYPE_MAX;
static ib_fast_recovery_threshold_type_t threshold_type = IB_THR_TYPE_RESERVED;
static boolean_t default_thr[IB_BER_TYPE_MAX];

int fast_recovery_column_num;
int fast_recovery_line_num;

static void parser_tmp_struct_init()
{
	cl_list_construct(&fast_rec_tmp_parser_struct.str_list);
	cl_list_init(&fast_rec_tmp_parser_struct.str_list, 10);
	cl_list_construct(&port_group_list);
	cl_list_init(&port_group_list, 10);
}

static void parser_tmp_struct_reset()
{
	cl_list_remove_all(&fast_rec_tmp_parser_struct.str_list);
	fast_rec_tmp_parser_struct.num = 0;
	fast_rec_tmp_parser_struct.boolean = 0;
}

static void parser_tmp_struct_destroy()
{
	parser_tmp_struct_reset();
	cl_list_destroy(&fast_rec_tmp_parser_struct.str_list);
	cl_list_remove_all(&port_group_list);
	cl_list_destroy(&port_group_list);
}

static boolean_t is_credit_watchdog()
{
	return ber_type == IB_BER_TYPE_MAX;
}

static void start_threshold(ib_fast_recovery_threshold_type_t thr_type)
{
	threshold_type = thr_type;
	RESET_BUFFER;
}

static void configure_profile_conf_params(osm_port_group_t * p_pgrp, uint8_t profile)
{
	cl_map_item_t *p_item;
	osm_port_group_member_t *p_grp_m;
	osm_port_t *p_port;
	osm_physp_t *p_physp;
	osm_switch_t *p_sw;
	int i;

	OSM_ASSERT(profile <= IB_PROFILE_CONFIG_MAX_PROFILE);

	/* Set to all physical switch ports of the group */
	for (p_item = cl_qmap_head(&p_pgrp->ports);
	     p_item != cl_qmap_end(&p_pgrp->ports);
	     p_item = cl_qmap_next(p_item)) {
		p_grp_m = (osm_port_group_member_t *)p_item;
		p_port = p_grp_m->port;
		p_sw = p_port->p_node->sw;
		if (!p_sw)
			continue;

		for (i = 1; i < p_port->p_node->physp_tbl_size; i++) {
			if (!p_grp_m->physp_members[i])
				continue;

			p_physp = osm_node_get_physp_ptr(p_port->p_node, i);
			if (p_physp)
				p_physp->fast_recovery_profile = profile;
		}

		/* Keep CreditWatchdogConfig values per node and profile */
		memcpy(&p_sw->fast_recovery_data.calc_credit_watchdog[profile],
		       &credit_watchdog_parsed,
		       sizeof(ib_credit_watchdog_config_t));

		for (i = 0; i < IB_BER_TYPE_MAX; i++) {
			memcpy(&p_sw->fast_recovery_data.calc_ber_config[i][profile],
			       (void *)&ber_config_parsed[i], sizeof(ib_ber_config_t));
			p_sw->fast_recovery_data.calc_default_thr[i][profile] = default_thr[i];
		}

		if (p_sw->fast_recovery_data.max_profile < profile)
			p_sw->fast_recovery_data.max_profile = profile;

	}
}

static void parser_conf_group_end()
{
	cl_list_iterator_t list_item;
	osm_port_group_t *p_cur_pgrp;

	for (list_item = cl_list_head(&port_group_list);
	     list_item != cl_list_end(&port_group_list);
	     list_item = cl_list_next(list_item)) {
		p_cur_pgrp = (osm_port_group_t *)cl_list_obj(list_item);
		if (p_cur_pgrp)
			configure_profile_conf_params(p_cur_pgrp, next_profile_num);
	}

	memset((void *)&credit_watchdog_parsed, 0, sizeof(ib_credit_watchdog_config_t));
	memset(ber_config_parsed, 0, sizeof(ber_config_parsed));

	cl_list_remove_all(&port_group_list);
	next_profile_num++;
}

static void clear_val(char **c)
{
	free(*c);
	*c = NULL;
}

static void yyerror(const char *format, ...)
{
	char s[256];
	va_list pvar;

	OSM_LOG_ENTER(&p_osm->log);

	va_start(pvar, format);
	vsnprintf(s, sizeof(s), format, pvar);
	va_end(pvar);

	OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR A107: "
		"Syntax error (line %d:%d): %s\n",
		fast_recovery_line_num, fast_recovery_column_num, s);
	OSM_LOG_EXIT(&p_osm->log);
}

int __fast_recovery_parser_wrap()
{
        return 1;
}

int __fast_recovery_parser_parse();
int __fast_recovery_parser_lex();

static void init_ber_node_fields(osm_switch_t * p_sw)
{
	memset(p_sw->fast_recovery_data.calc_default_thr, 0, sizeof(p_sw->fast_recovery_data.calc_default_thr));
	memset(p_sw->fast_recovery_data.calc_ber_config, 0, sizeof(p_sw->fast_recovery_data.calc_ber_config));
}

static void init_nodes_fast_recovery()
{
	osm_switch_t *p_sw;
	cl_map_item_t *p_next;
	int i;
	osm_physp_t *p_physp;

	/* Iterate over switches set all p_node->calc_credit_watchdog to 0 */
	p_next = cl_qmap_head(&p_osm->subn.sw_guid_tbl);
	while (p_next != cl_qmap_end(&p_osm->subn.sw_guid_tbl)) {
		p_sw = (osm_switch_t *)p_next;
		p_next = cl_qmap_next(p_next);

		for (i = 1; i < p_sw->p_node->physp_tbl_size; i++) {
			p_physp = osm_node_get_physp_ptr(p_sw->p_node, i);
			if (p_physp)
				p_physp->fast_recovery_profile = 0;
		}

		memset(&p_sw->fast_recovery_data.calc_credit_watchdog, 0,
		       (IB_PROFILE_CONFIG_MAX_PROFILE + 1) * sizeof(ib_credit_watchdog_config_t));

		init_ber_node_fields(p_sw);

		p_sw->fast_recovery_data.max_profile = 0;
	}
}

int osm_parse_fast_recovery_configuration_file(struct osm_opensm * osm_ptr)
{
	int res = 0;
	p_osm = osm_ptr;
	__fast_recovery_parser_in = NULL;
	parser_tmp_struct_init();
	fast_recovery_line_num = 0;
	fast_recovery_column_num = 0;
	next_profile_num = 1;

	init_nodes_fast_recovery();

	if (!p_osm->subn.opt.fast_recovery_conf_file) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR A101: No fast recovery configuration file provided - "
			"Quit fast recovery parser\n");
		res = 1;
		goto Exit;
	}

	__fast_recovery_parser_in = fopen(p_osm->subn.opt.fast_recovery_conf_file, "r");
	if (!__fast_recovery_parser_in) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR A102: Fast recovery configuration file specified but not found\n");
		res = 1;
		goto Exit;
	}

        res = __fast_recovery_parser_parse();
	if (res)
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR A103: Fast recovery configuration file parsing failed\n");
	else
		OSM_LOG(&p_osm->log, OSM_LOG_DEBUG, "Fast recovery configuration "
			"file parsed successfully\n");

Exit:
	if (__fast_recovery_parser_in) {
		fclose(__fast_recovery_parser_in);
		__fast_recovery_parser_in = NULL;
	}
	clear_val(&__fast_recovery_parser_yylval);
	parser_tmp_struct_destroy();
	return res;
}
%}

%token COMMENT
%token COMMA
%token NUMBER
%token NUMBER_EXP
%token TEXT
%token TK_TRUE
%token TK_FALSE

%token PORT_CONF_START
%token PORT_CONF_END
%token PORT_GROUP_NAME

%token CREDIT_WATCHDOG_START
%token CREDIT_WATCHDOG_END
%token RAW_BER_START
%token RAW_BER_END
%token EFFECTIVE_BER_START
%token EFFECTIVE_BER_END
%token SYMBOL_BER_START
%token SYMBOL_BER_END

%token GROUP_ERR_THR_ACT
%token ERR_ACT_NO_RESET
%token ERR_ACT_PHYSICAL_RESET
%token ERR_ACT_LOGICAL_RESET

%token GROUP_ENABLE_THR
%token ENABLE_NONE
%token ENABLE_ERROR
%token ENABLE_WARN_NORMAL
%token ENABLE_ALL

%token GROUP_ERR_THR
%token GROUP_WARN_THR
%token GROUP_NORM_THR

%token GROUP_TIME_WINDOW
%token GROUP_SAMPLING_RATE

%token GROUP_DEFAULT_THR

%%

fast_recovery_policy: configurations;

configurations: /* empty */
		| configurations configuration
		;

comment: COMMENT {};

configuration:    conf_groups
		| comment
		;

/*
 * Parsing port configuration groups:
 * -------------------
 * port_conf_start
 *	port_group_name: NAME1
 *	...
 *	credit_watchdog_start
 *		error_thr_action: no_reset
 *		enable_threshold: all
 *		error_threshold: 4
 *		warning_threshold: 2
 *		normal_threshold: 1
 *		time_window: 10
 *		sampling_rate: 5
 *	credit_watchdog_end
 *
 *	raw_BER_start
 *		default_threshold: true
 *		error_thr_action: logical_reset
 *		enable_threshold: all
 *		error_threshold: 1e-9
 *		warning_threshold: 1e-6
 *		normal_threshold: 1e-4
 *		// time window is RO field for BER
 *		// sampling_rate RO field for BER
 *	raw_BER_end
 *
 *	effective_BER_start
 * 	... // similar to raw_BER section
 *	effective_BER_end
 *
 *	symbol_BER_start
 * 	... // similar to raw_BER section
 *	symbol_BER_end
 * 	...
 * port_conf_end
 *
 * port_conf
 *	...
 * port_conf_end
 */

conf_groups:		conf_group
			| conf_groups conf_group
			;

conf_group: conf_group_start conf_group_entries conf_group_end;
conf_group_start: PORT_CONF_START {
		/*
		 * HACK: limit to maximum of 8 profiles.
		 * Later should limit to 8 per switch.
		 */
		if (next_profile_num > IB_PROFILE_CONFIG_MAX_PROFILE) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR A108: "
				"Fast Recovery parser supports up to %d sections. "
				"Aborting parsing\n", IB_PROFILE_CONFIG_MAX_PROFILE);
			return 1;
		}
	};
conf_group_end: PORT_CONF_END {
		parser_conf_group_end();
	};

conf_group_entries:	 /* empty */
			| conf_group_entries conf_group_entry
			;

conf_group_entry:	port_group_name
			| credit_watchdog
			| raw_ber
			| effective_ber
			| symbol_ber
			| comment
			;

credit_watchdog: credit_watchdog_start mad_fields_entries credit_watchdog_end;
credit_watchdog_start: CREDIT_WATCHDOG_START {};
credit_watchdog_end: CREDIT_WATCHDOG_END {};

raw_ber:		raw_ber_start mad_fields_entries raw_ber_end;
raw_ber_start:		RAW_BER_START {
				ber_type = IB_BER_TYPE_RAW;
			};
raw_ber_end:		RAW_BER_END {
				ber_type = IB_BER_TYPE_MAX;
			};

effective_ber:		effective_ber_start mad_fields_entries effective_ber_end;
effective_ber_start:	EFFECTIVE_BER_START {
				ber_type = IB_BER_TYPE_EFFECTIVE;
			};
effective_ber_end:	EFFECTIVE_BER_END {
				ber_type = IB_BER_TYPE_MAX;
			};

symbol_ber:		symbol_ber_start mad_fields_entries symbol_ber_end;
symbol_ber_start:	SYMBOL_BER_START {
				ber_type = IB_BER_TYPE_SYMBOL;
			};
symbol_ber_end:		SYMBOL_BER_END {
				ber_type = IB_BER_TYPE_MAX;
			};

mad_fields_entries:	/* empty*/
			| mad_fields_entries mad_fields_entry
			;

mad_fields_entry:	credit_watchdog_fields
			| ber_fields
			| mad_fields
			| comment
			;

credit_watchdog_fields:	time_window
			| sampling_rate
			;

ber_fields:		default_thr
			;

mad_fields:		err_thr_act
			| en_thr
			| error_thr
			| warning_thr
			| normal_thr
			;

port_group_name: port_group_name_start string_list {
		cl_list_iterator_t list_item;
		char *tmp_str;
		osm_port_group_t *p_port_grp;

		for (list_item = cl_list_head(&fast_rec_tmp_parser_struct.str_list);
		     list_item != cl_list_end(&fast_rec_tmp_parser_struct.str_list);
		     list_item = cl_list_next(list_item)) {
			tmp_str = (char *)cl_list_obj(list_item);
			if (tmp_str) {
				p_port_grp = osm_get_port_grp(p_osm, tmp_str);
				if (p_port_grp) {
					cl_list_insert_tail(&port_group_list, p_port_grp);
				} else {
					OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR A104: "
						"Port group with name (%s) does not exist\n",
						tmp_str);
				}
			}
			free(tmp_str);
		}
		cl_list_remove_all(&fast_rec_tmp_parser_struct.str_list);
	};

port_group_name_start: PORT_GROUP_NAME {
		RESET_BUFFER;
	};

string_list:	single_string
		| string_list COMMA single_string
		;

single_string:	TEXT {
		char *tmp_str = strdup(__fast_recovery_parser_yylval);
		if (!tmp_str) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
				"ERR A105: Couldn't allocate memory for "
				"port group name, Aborting parsing\n");
			return 1;
		}
		cl_list_insert_tail(&fast_rec_tmp_parser_struct.str_list, tmp_str);
	};

default_thr: default_thr_start bool {
		default_thr[ber_type] = fast_rec_tmp_parser_struct.boolean;
	};

default_thr_start: GROUP_DEFAULT_THR { RESET_BUFFER; }

bool: number | true | false {};

true: TK_TRUE {
		fast_rec_tmp_parser_struct.boolean = 1;
	};

false: TK_FALSE {
		fast_rec_tmp_parser_struct.boolean = 0;
	};

err_thr_act: err_thr_act_start err_thr_act_options {
		if (is_credit_watchdog())
			ib_credit_watchdog_config_set_err_threshold_action(
			    &credit_watchdog_parsed,
			    fast_rec_tmp_parser_struct.num);
		else /* BER */
			ib_ber_config_set_err_threshold_action(
			    &ber_config_parsed[ber_type],
			    fast_rec_tmp_parser_struct.num);
	};

err_thr_act_start: GROUP_ERR_THR_ACT {
		RESET_BUFFER;
	};

err_thr_act_options:	err_thr_act_no_reset
			| err_thr_act_physical_reset
			| err_thr_act_logical_reset
			;

err_thr_act_no_reset: ERR_ACT_NO_RESET {
		fast_rec_tmp_parser_struct.num = 0;
       };
err_thr_act_physical_reset: ERR_ACT_PHYSICAL_RESET {
		fast_rec_tmp_parser_struct.num = 1;
       };
err_thr_act_logical_reset: ERR_ACT_LOGICAL_RESET {
		fast_rec_tmp_parser_struct.num = 2;
       };

en_thr: en_thr_start en_thr_options {
		if (is_credit_watchdog())
			ib_credit_watchdog_config_set_en_threshold(
			    &credit_watchdog_parsed,
			    fast_rec_tmp_parser_struct.num);
		else /* BER */
			ib_ber_config_set_en_threshold(
			    &ber_config_parsed[ber_type],
			    fast_rec_tmp_parser_struct.num);
	};


en_thr_start: GROUP_ENABLE_THR {
		RESET_BUFFER;
	};

en_thr_options:		en_thr_none
			| en_thr_error
			| en_thr_warn_normal
			| en_thr_all
			;

en_thr_none: ENABLE_NONE {
		fast_rec_tmp_parser_struct.num = 0;
	  };
en_thr_error: ENABLE_ERROR {
		fast_rec_tmp_parser_struct.num = 1;
	  };
en_thr_warn_normal: ENABLE_WARN_NORMAL {
		fast_rec_tmp_parser_struct.num = 2;
	  };
en_thr_all: ENABLE_ALL {
		fast_rec_tmp_parser_struct.num = 3;
	  };

error_thr:		error_thr_start threshold;

error_thr_start:	GROUP_ERR_THR { start_threshold(IB_THR_TYPE_ERROR); };

warning_thr:		warning_thr_start threshold;

warning_thr_start:	GROUP_WARN_THR { start_threshold(IB_THR_TYPE_WARNING); };

normal_thr:		normal_thr_start threshold;

normal_thr_start:	GROUP_NORM_THR { start_threshold(IB_THR_TYPE_NORMAL); };

threshold:		threshold_number
			| threshold_exp
			;

threshold_number: number {
		if (is_credit_watchdog()) {
			uint8_t threshold = fast_rec_tmp_parser_struct.num;
			if (fast_rec_tmp_parser_struct.num > IB_CREDIT_WATCHDOG_MAX_THRESHOLD) {
				OSM_LOG(&p_osm->log, OSM_LOG_INFO,
					"CreditwatchdogConfig threshold %u is exceeding "
					"maximal value %u. "
					"Setting threshold to maximal value.\n",
					fast_rec_tmp_parser_struct.num,
					IB_CREDIT_WATCHDOG_MAX_THRESHOLD);
				threshold = IB_CREDIT_WATCHDOG_MAX_THRESHOLD;
			}

			ib_credit_watchdog_config_set_threshold(
			    &credit_watchdog_parsed, threshold_type,
			    threshold);
		} else {
			uint16_t threshold =
			    (fast_rec_tmp_parser_struct.num << IB_BER_THR_MANTISSA_SHIFT);
			ib_ber_config_set_threshold(&ber_config_parsed[ber_type],
						    threshold_type, threshold);
		}

		threshold_type = IB_THR_TYPE_RESERVED;
	};

threshold_exp: NUMBER_EXP {
		char *threshold_str;
		double threshold_double;
		uint32_t exponent = 0;
		uint32_t mantissa = 0;
		uint16_t threshold = 0;
		int n;

		if (is_credit_watchdog()) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR A106: "
				"CreditWatchdogConfig threshold must be an integer\n");
			return 1;
		}

		threshold_str = strdup(__fast_recovery_parser_yylval);
		if (!threshold_str) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
				"ERR A109: Couldn't allocate memory for "
				"exponent, Aborting parsing\n");
			return 1;
		}

		/*
		 * parse threshold specified as [1-9]e-[0-9][0-9]
		 *
		 * This section improves threshold parsing accuracy
		 * it allows setting the exact exponent and mantissa specified by
		 * the user when the string is in the format of [1-9]e-[0-9][0-9].
		 *
		 * Currently the parser reads the specified threshold into a double, and
		 * then converts it to mantissa ^ -10 exponent.
		 * Due to issues with FP representations and the conversion functions, some
		 * values might be inaccurately converted.
		 * For example, "1e-13" is converted to 9e-14 and "5e-15" is converted to 4e-15.
		 */
		n = sscanf(threshold_str, "%1ue-%2u", &mantissa, &exponent);
		if (n == 2) {
			if (mantissa == 0 || exponent > IB_BER_THR_EXPONENT_MAX) {
				OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
					"ERR A10D: Couldn't parse threshold value out of range, "
					" Aborting parsing\n");
				free(threshold_str);
				return 1;
			}
			goto set_threshold;
		}

		/* parse threshold generic floating point format */
		n = sscanf(threshold_str, "%lg", &threshold_double);
		if (n != 1) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
				"ERR A10B: Couldn't parse threshold value\n");
			free(threshold_str);
			return 1;
		}

		if (threshold_double <= 0 || threshold_double >= 1) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
				"ERR A10C: threshold value out of range (0 .. 1)\n");
			free(threshold_str);
			return 1;
		}

		while (threshold_double < 1 && exponent < IB_BER_THR_EXPONENT_MAX) {
			threshold_double = threshold_double * 10;
			exponent++;
		}

		mantissa = (uint8_t)(floor(threshold_double));

		/* mantissa should be between 1 and 9 */
		if (!mantissa)
			mantissa = 1;

	set_threshold:
		/* set value */
		threshold = exponent | (mantissa << IB_BER_THR_MANTISSA_SHIFT);
		ib_ber_config_set_threshold(&ber_config_parsed[ber_type],
					    threshold_type, threshold);

		threshold_type = IB_THR_TYPE_RESERVED;
		free(threshold_str);
       };

time_window: time_window_start number {
		if (!is_credit_watchdog()) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR A10A: "
				"Time window for BER is not configured by SM."
				"Aborting parsing.\n");
			return 1;
		}

		ib_credit_watchdog_config_set_time_window(
		    &credit_watchdog_parsed, fast_rec_tmp_parser_struct.num);
	};

time_window_start: GROUP_TIME_WINDOW {
		RESET_BUFFER;
	};

sampling_rate: sampling_rate_start number {
		ib_credit_watchdog_config_set_sampling_rate(
		    &credit_watchdog_parsed, fast_rec_tmp_parser_struct.num);
	};

sampling_rate_start: GROUP_SAMPLING_RATE {
		RESET_BUFFER;
	};

number: NUMBER {
		fast_rec_tmp_parser_struct.num = strtoull(__fast_recovery_parser_yylval, NULL, 0);
		fast_rec_tmp_parser_struct.boolean = (fast_rec_tmp_parser_struct.num > 0);
	};
%%
