/*
 * Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <ctype.h>
#include <complib/cl_debug.h>
#include <complib/cl_qmap.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_ROOT_DETECTION_C
#include <opensm/osm_switch.h>
#include <opensm/osm_opensm.h>

#define UPDN_MAX_COLOR			(256)
#define UPDN_COLORS_ARRAY_ELEMENT_SIZE	64
#define UPDN_COLORS_ARRAY_SIZE		((UPDN_MAX_COLOR)/(UPDN_COLORS_ARRAY_ELEMENT_SIZE))
#define UPDN_COLOR_ARRAY_SIZE_IN_BYTES	UPDN_COLORS_ARRAY_SIZE * sizeof(uint64_t)

#define UPDN_NO_RANK			(-1)
#define UPDN_MAX_SW_PEERS_NUM		254

#define MAX_COLOR_GROUPS		UPDN_MAX_COLOR

#define DFP_MIN_ROOT_NEIGHBOR_COLORS	2

typedef struct dfp_color_group {
	uint64_t p_own_colors_bitmask[UPDN_COLORS_ARRAY_SIZE]; /* bitmask */
	uint8_t own_colors_count;
	uint32_t potential_roots_rank;
	uint64_t p_neighbor_groups_bitmask[UPDN_COLORS_ARRAY_SIZE]; /* bitmask by group_num */
	uint8_t neighbor_groups_count;
	uint8_t group_num;

	cl_list_t new_members;
	cl_list_t potential_new_members; /* Used for reverting color by step A */
	uint64_t color_mask1[UPDN_COLORS_ARRAY_SIZE];
	uint64_t color_mask2[UPDN_COLORS_ARRAY_SIZE];
	boolean_t is_subgroup;
} dfp_color_group_t;
/*
 * FIELDS
 *	new_members
 *		List of the nodes that were added to this group on last iteration.
 *		updated on each coloring iteration to hold latest rank, until finds
 *		the potential roots rank of the group. Then, stops updating the list,
 *		so it holds the potential roots.
 *
 * 	is_subgroup
 * 		Boolean indicates whether this group was merged into another group.
 */

typedef struct updn_color_ {
	uint64_t colors[UPDN_COLORS_ARRAY_SIZE];
	dfp_color_group_t *p_color_group;
	boolean_t in_colorless_queue;
	uint8_t neighbor_group_num;
} updn_color_t;

typedef struct updn_sw_data_ {
	osm_switch_t *p_sw;
	uint32_t cas_num;
	int rank;
	updn_color_t color;
	uint32_t degree;
	uint32_t index;
	boolean_t visited;
	boolean_t need_to_color;
	boolean_t is_potential_root;
	struct updn_sw_data_ *peers[UPDN_MAX_SW_PEERS_NUM];
} updn_sw_data_t;

static inline void updn_color_add_single_color(updn_color_t *color, int color_idx)
{
	color->colors[color_idx / UPDN_COLORS_ARRAY_ELEMENT_SIZE] |=
	    (uint64_t)1 << (color_idx % UPDN_COLORS_ARRAY_ELEMENT_SIZE);
}

static void updn_color_add_colors(updn_color_t *color, updn_color_t *new_color)
{
	int i;
	int array_size = UPDN_COLORS_ARRAY_SIZE;

	for (i = 0; i < array_size; i++)
		color->colors[i] |= new_color->colors[i];
}

static void updn_color_remove_colors(uint64_t * color_arr_a,
				     uint64_t * color_arr_b)
{
	int i;
	int array_size = UPDN_COLORS_ARRAY_SIZE;

	for (i = 0; i < array_size; i++)
		color_arr_a[i] &= ~color_arr_b[i];
}

static boolean_t updn_color_get_color_num(updn_color_t * color,
					  IN uint8_t * color_num)
{
	uint32_t index, offset;
	int array_size = UPDN_COLORS_ARRAY_SIZE;
	uint64_t offset_bit;

	for (index = 0; index < array_size; index++) {
		offset_bit = (uint64_t)1;
		for (offset = 0; offset < UPDN_COLORS_ARRAY_ELEMENT_SIZE;
		     offset++) {
			if (color->colors[index] & offset_bit) {
				*color_num =
				    UPDN_COLORS_ARRAY_ELEMENT_SIZE * index + offset;
				return TRUE;
			}
			offset_bit = offset_bit << 1;
		}
	}

	return FALSE;
}

static boolean_t updn_color_is_subset_color(uint64_t * color_arr_a,
					    uint64_t * color_arr_b)
{
	int i, array_size = UPDN_COLORS_ARRAY_SIZE;
	uint64_t color_block;

	for (i = 0; i < array_size; i++) {
		color_block = color_arr_a[i] & color_arr_b[i];
		if (color_block != color_arr_b[i])
			return FALSE;
	}

	return TRUE;
}

/*
 * Function adds color_arr_b color bits to color_arr_a.
 * If color was added, return TRUE. Otherwise, FALSE.
 */
static boolean_t updn_color_add_color_to_mask(uint64_t * color_arr_a,
					      uint64_t * color_arr_b)
{
	int i, array_size = UPDN_COLORS_ARRAY_SIZE;
	uint64_t color_block;
	boolean_t color_update = FALSE;

	for (i = 0; i < array_size; i++) {
		/*
		 * At this point, subcolors are already merged.
		 * Hence, main color (a) can contain all bits of the new color (b),
		 * or none of them.
		 * Thus, if (a & b) != b, it means b holds bits that are
		 * new for a (as these are currently down in a).
		 */
		color_block = color_arr_a[i] & color_arr_b[i];
		if (color_block != color_arr_b[i]) {
			color_update = TRUE;
			color_arr_a[i] |= color_arr_b[i];
		}
	}

	return color_update;
}

static boolean_t updn_color_not_subsets_colors(uint64_t * color_arr_a,
					       uint64_t * color_arr_b)
{
	int i, count_equal_blocks_a = 0, count_equal_blocks_b = 0;
	int array_size = UPDN_COLORS_ARRAY_SIZE;
	uint64_t color_block;

	for (i = 0; i < array_size; i++) {
		color_block = color_arr_a[i] & color_arr_b[i];
		if (color_block == color_arr_b[i])
			count_equal_blocks_b++;

		if (color_block == color_arr_a[i])
			count_equal_blocks_a++;
	}

	return count_equal_blocks_b != array_size &&
	    count_equal_blocks_a != array_size;
}

static inline void updn_color_set_colors(updn_color_t *color, updn_color_t *new_color)
{
	memcpy(color->colors, new_color->colors, sizeof(uint64_t) * UPDN_COLORS_ARRAY_SIZE);
	color->p_color_group = new_color->p_color_group;
}

static inline boolean_t is_colored_arr(uint64_t * color)
{
	uint64_t test_color[UPDN_COLORS_ARRAY_SIZE];

	memset(test_color, 0, sizeof(uint64_t) * UPDN_COLORS_ARRAY_SIZE);

	return memcmp(color, test_color, UPDN_COLORS_ARRAY_SIZE);
}

static inline boolean_t updn_color_is_colored(updn_color_t *color)
{
	uint64_t test_color[UPDN_COLORS_ARRAY_SIZE];

	memset(test_color, 0, sizeof(uint64_t) * UPDN_COLORS_ARRAY_SIZE);

	return memcmp(color->colors, test_color, sizeof(uint64_t) * UPDN_COLORS_ARRAY_SIZE);
}

static inline void updn_color_erase_colors(updn_color_t *color)
{
	memset(color, 0, sizeof(updn_color_t));
}

static inline boolean_t updn_color_is_sub_color(updn_color_t *p_main_color, updn_color_t *p_color2)
{
	uint64_t main_color_mask, color_mask2;
	boolean_t is_sub_color = TRUE;
	int i;

	for (i = 0; i < UPDN_COLORS_ARRAY_SIZE; i++) {
		main_color_mask = p_main_color->colors[i];
		color_mask2 = p_color2->colors[i];
		if ((main_color_mask | color_mask2) != main_color_mask) {
			is_sub_color = FALSE;
			break;
		}
	}
	return is_sub_color;
}

static inline boolean_t updn_color_check_single_color(updn_color_t *p_color)
{
	uint64_t color_mask;
	boolean_t one_bit_set = FALSE;
	int i;

	for (i = 0; i < UPDN_COLORS_ARRAY_SIZE; i++) {
		color_mask = p_color->colors[i];
		if (!color_mask)
			continue;

		if (one_bit_set && color_mask) {
			one_bit_set = FALSE;
			break;
		}

		/* Check if only one bit set */
		if (!(color_mask & (color_mask - 1)))
			one_bit_set = TRUE;

	}

	return one_bit_set;
}

static inline boolean_t updn_color_is_same_color(updn_color_t *color1, updn_color_t *color2)
{
	return !memcmp(color1->colors, color2->colors, sizeof(uint64_t) * UPDN_COLORS_ARRAY_SIZE);
}

static inline boolean_t updn_color_is_same_color_mask(uint64_t * color_arr_a, uint64_t * color_arr_b)
{
	return !memcmp(color_arr_a, color_arr_b, sizeof(uint64_t) * UPDN_COLORS_ARRAY_SIZE);
}

static inline void set_color(updn_sw_data_t *p_sw_data, updn_color_t *color)
{
	updn_color_set_colors(&p_sw_data->color, color);
}

static inline void add_color(updn_sw_data_t *p_sw_data, updn_color_t *color)
{
	updn_color_add_colors(&p_sw_data->color, color);
}

static inline updn_color_t *get_color(updn_sw_data_t *p_sw_data)
{
	return &p_sw_data->color;
}

static inline boolean_t is_colored(updn_sw_data_t *p_sw_data)
{
	return updn_color_is_colored(&p_sw_data->color);
}

static void root_detection_add_peers_to_sw_data(OUT updn_sw_data_t *p_sw_data,
						OUT updn_sw_data_t *data_per_sw,
						OUT uint32_t *p_cas_num)
{
	osm_switch_t *p_sw = p_sw_data->p_sw;
	osm_node_t *p_node = p_sw->p_node;
	osm_node_t *p_remote_node;
	osm_physp_t *p_physp;
	uint32_t port_num;
	uint16_t remote_lid_ho;
	int cas_num = 0;

	for (port_num = 0; port_num < p_sw->num_ports; port_num++) {
		p_remote_node = osm_node_get_remote_node(p_node, port_num, NULL);
		if (!p_remote_node || p_remote_node == p_node)
			continue;

		if (osm_node_get_type(p_remote_node) == IB_NODE_TYPE_ROUTER) {
			cas_num++;
			continue;
		}

		if (!p_remote_node->sw) {
			p_physp = osm_node_get_physp_ptr(p_node, port_num);
			if (!osm_phsyp_get_special_port_type(p_physp->p_remote_physp)) {
				p_sw_data->cas_num++;
				cas_num++;
			}

			continue;
		}

		remote_lid_ho = cl_ntoh16(osm_node_get_base_lid(p_remote_node, 0));
		p_sw_data->peers[p_sw_data->degree] = &data_per_sw[remote_lid_ho];
		p_sw_data->degree++;
	}

	*p_cas_num += cas_num;
}


/************************************** UPDN section **************************************/

static void updn_mark_need_to_color(updn_sw_data_t **queue, uint32_t start_idx, uint32_t end_idx)
{
	updn_sw_data_t *p_sw_data;
	updn_sw_data_t *peer;
	updn_sw_data_t **peers;
	uint32_t j = start_idx;
	int i;

	while (j < end_idx) {
		p_sw_data = queue[j++];
		peers = p_sw_data->peers;

		for (i = 0; i < p_sw_data->degree; i++) {
			peer = peers[i];

			if (!is_colored(peer) || (p_sw_data->rank < peer->rank))
				p_sw_data->need_to_color = TRUE;
		}
	}
}

static void updn_unmark_need_to_color(updn_sw_data_t **queue, uint32_t start_idx, uint32_t end_idx)
{
	updn_sw_data_t *p_sw_data;
	uint32_t j = start_idx;

	while (j < end_idx) {
		p_sw_data = queue[j++];
		p_sw_data->need_to_color = FALSE;
	}
}

static void updn_color_marked_nodes(updn_sw_data_t **queue, uint32_t start_idx, uint32_t end_idx)
{
	updn_sw_data_t *p_sw_data;
	updn_sw_data_t *peer;
	updn_sw_data_t **peers;
	uint32_t j = start_idx;
	int i;

	while (j < end_idx) {
		p_sw_data = queue[j++];
		peers = p_sw_data->peers;

		if (!p_sw_data->need_to_color) {
			for (i = 0; i < p_sw_data->degree; i++) {
				peer = peers[i];
				if (peer->need_to_color && (p_sw_data->rank == peer->rank))
					add_color(peer, get_color(p_sw_data));
			}
		}
	}
}

static void updn_rank_root_node(OUT osm_opensm_t *p_osm,
				OUT updn_sw_data_t *p_sw_data,
				OUT unsigned *p_num_roots)
{
	p_sw_data->p_sw->rank = 0;
	OSM_LOG(&p_osm->log, OSM_LOG_VERBOSE,
		"Ranking GUID 0x%" PRIx64 " as root node\n",
		cl_ntoh64(osm_node_get_node_guid(p_sw_data->p_sw->p_node)));
	(*p_num_roots)++;
}

static boolean_t updn_is_root_reaches_all_leafs(OUT osm_opensm_t *p_osm,
						OUT updn_sw_data_t **queue,
						OUT updn_sw_data_t *p_root_data,
						OUT uint32_t *p_leafs_reached,
						IN uint32_t leafs_num)
{
	uint32_t queue_in = 0;
	uint32_t queue_out = 0;
	updn_sw_data_t **peers;
	updn_sw_data_t *p_sw_data;
	updn_sw_data_t *peer;
	boolean_t ret = TRUE;
	uint32_t leafs_reached = 0;
	int i;

	queue[queue_in++] = p_root_data;
	p_root_data->visited = TRUE;
	while (queue_out < queue_in) {
		p_sw_data = queue[queue_out++];
		peers = p_sw_data->peers;

		if (!p_sw_data->rank) {
			leafs_reached++;
			if (leafs_reached == leafs_num)
				break;
		}

		for (i = 0; i < p_sw_data->degree; i++) {
			peer = peers[i];

			if (peer->visited || (p_sw_data->rank < peer->rank))
				continue;

			peer->visited = TRUE;
			queue[queue_in++] = peer;
		}
	}

	if (leafs_reached != leafs_num) {
		OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
			"Number of leafs reached from GUID 0x%" PRIx64 ": %u, "
			"is not equal to all the leafs in the subnet: %u\n",
			cl_ntoh64(osm_node_get_node_guid(p_root_data->p_sw->p_node)),
			leafs_reached, leafs_num);
		ret = FALSE;
	}

	queue_out = 0;
	while (queue_out < queue_in) {
		p_sw_data = queue[queue_out++];
		p_sw_data->visited = FALSE;
	}

	*p_leafs_reached = leafs_reached;
	return ret;
}

static boolean_t updn_validate_roots(OUT osm_opensm_t *p_osm, IN osm_subn_t *p_subn, OUT cl_list_t *p_roots_list,
				     IN uint32_t leafs_num)
{
	cl_list_iterator_t list_iter;
	updn_sw_data_t *p_sw_data;
	boolean_t roots_valid = FALSE;
	updn_sw_data_t **queue = NULL;

	/* Use one queue for all roots to not allocate a queue for each validation */
	queue = (updn_sw_data_t **)calloc(cl_qmap_count(&p_subn->sw_guid_tbl),
					  sizeof(updn_sw_data_t *));
	if (!queue) {
		roots_valid = FALSE;
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR 8F1A: Cannot allocate a queue for validating roots list\n");
		goto Exit;
	}

	/* Roots are valid if there is a root that reaches all the leafs */
	for (list_iter = cl_list_head(p_roots_list);
	     list_iter != cl_list_end(p_roots_list);
	     list_iter = cl_list_next(list_iter)) {
		uint32_t leafs_reached = 0;
		p_sw_data = (updn_sw_data_t *)cl_list_obj(list_iter);

		if (updn_is_root_reaches_all_leafs(p_osm, queue, p_sw_data,
						   &leafs_reached, leafs_num))
			roots_valid = TRUE;;
	}

	if (!roots_valid)
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR 8F1B: Root validation failed. No roots that reach all the leafs\n");

	free(queue);
Exit:
	return roots_valid;
}

typedef struct updn_color_group_ {
	cl_list_t roots;
	uint32_t max_distanse_from_leaf;
	uint64_t min_guid;
} updn_color_group_t;

static int updn_node_color_cmp_func(const void *lhs, const void* rhs)
{
	updn_sw_data_t *p_left = *(updn_sw_data_t **)lhs;
	updn_sw_data_t *p_right = *(updn_sw_data_t **)rhs;

	return memcmp(get_color(p_left), get_color(p_right), sizeof(updn_color_t));
}

static int updn_color_group_cmp_func(const void *lhs, const void* rhs)
{
	updn_color_group_t *p_left = *(updn_color_group_t **)lhs;
	updn_color_group_t *p_right = *(updn_color_group_t **)rhs;
	uint32_t left_max_distanse_from_leaf = p_left->max_distanse_from_leaf;
	uint32_t right_max_distanse_from_leaf = p_right->max_distanse_from_leaf;
	uint64_t left_min_guid = p_left->min_guid;
	uint64_t right_min_guid = p_right->min_guid;
	uint32_t left_num_roots = cl_list_count(&p_left->roots);
	uint32_t right_num_roots = cl_list_count(&p_right->roots);

	if (left_max_distanse_from_leaf < right_max_distanse_from_leaf)
		return -1;

	if (left_max_distanse_from_leaf > right_max_distanse_from_leaf)
		return 1;

	if (left_num_roots < right_num_roots)
		return -1;

	if (left_num_roots > right_num_roots)
		return 1;

	/* Equal max_distance_from_leaf, decide by min GUID in group */
	return (left_min_guid - right_min_guid);
}

static void updn_init_color_group_lists(updn_color_group_t *color_groups)
{
	cl_list_t *p_list;
	int i;

	for (i = 0; i < UPDN_MAX_COLOR; i++) {
		p_list = &color_groups[i].roots;
		cl_list_construct(p_list);
		cl_list_init(p_list, 10);
	}
}

static void updn_destroy_color_group_lists(updn_color_group_t *color_groups)
{
	cl_list_t *p_list;
	int i;

	for (i = 0; i < UPDN_MAX_COLOR; i++) {
		p_list = &color_groups[i].roots;
		cl_list_remove_all(p_list);
		cl_list_destroy(p_list);
	}
}

static void updn_destroy_color_groups(updn_color_group_t *color_groups)
{
	updn_destroy_color_group_lists(color_groups);
	free(color_groups);
}

static void updn_color_group_find_max_hops_from_leaf(OUT updn_color_group_t *p_color_group,
						     IN uint16_t *leaf_lids_array,
						     IN uint32_t leafs_num)
{
	cl_list_t *p_roots_list = &p_color_group->roots;
	updn_sw_data_t *p_sw_data;
	cl_list_iterator_t list_iter;
	uint32_t max_hops = 0;
	uint32_t hop_val, i;

	for (list_iter = cl_list_head(p_roots_list);
	     list_iter != cl_list_end(p_roots_list);
	     list_iter = cl_list_next(list_iter)) {
		p_sw_data = (updn_sw_data_t *)cl_list_obj(list_iter);
		for (i = 0; i < leafs_num; i++) {
			hop_val =
			    osm_switch_get_least_hops(p_sw_data->p_sw, leaf_lids_array[i]);
			if (hop_val > max_hops)
				max_hops = hop_val;
		}
	}

	p_color_group->max_distanse_from_leaf = max_hops;
}

static void updn_color_group_find_min_guid(OUT updn_color_group_t *p_color_group)
{
	cl_list_t *p_roots_list = &p_color_group->roots;
	updn_sw_data_t *p_sw_data;
	osm_switch_t *p_sw;
	cl_list_iterator_t list_iter;
	uint64_t min_guid = UINT64_MAX;
	uint64_t node_guid;

	for (list_iter = cl_list_head(p_roots_list);
	     list_iter != cl_list_end(p_roots_list);
	     list_iter = cl_list_next(list_iter)) {
		p_sw_data = (updn_sw_data_t *)cl_list_obj(list_iter);
		p_sw = p_sw_data->p_sw;
		node_guid = cl_ntoh64(p_sw->p_node->node_info.node_guid);
		if (node_guid < min_guid)
			min_guid = node_guid;
	}

	p_color_group->min_guid = min_guid;
}
static ib_api_status_t
updn_sort_nodes_into_color_groups(IN osm_log_t *p_log,
				  IN cl_list_t *potential_roots_list,
				  OUT updn_color_group_t *color_groups,
				  OUT int *p_color_groups_num)
{
	int potential_roots_num = cl_list_count(potential_roots_list);
	ib_api_status_t status = IB_SUCCESS;
	updn_sw_data_t **potential_roots_array;
	updn_sw_data_t *p_sw_data;
	updn_color_t *p_prev_color;
	updn_color_t *p_color;
	cl_list_iterator_t list_iter;
	int group_idx = 0;
	int i;

	potential_roots_array =	(updn_sw_data_t **)calloc(potential_roots_num,
							  sizeof(updn_sw_data_t *));
	if (!potential_roots_array) {
		OSM_LOG(p_log, OSM_LOG_ERROR,
			"ERR 8F17: Cannot allocate memory for potential_roots_array "
			"to choose a specific color group\n");
		status = IB_ERROR;
		goto Error;
	}

	/* Fill an array of potential roots to sort by color */
	i = 0;
	for (list_iter = cl_list_head(potential_roots_list);
	     list_iter != cl_list_end(potential_roots_list);
	     list_iter = cl_list_next(list_iter)) {
		potential_roots_array[i] = (updn_sw_data_t *)cl_list_obj(list_iter);
		i++;
	}

	qsort(potential_roots_array, potential_roots_num, sizeof(updn_sw_data_t *),
	      updn_node_color_cmp_func);

	cl_list_insert_tail(&color_groups[group_idx].roots, potential_roots_array[0]);
	p_prev_color = get_color(potential_roots_array[0]);
	for (i = 1; i < potential_roots_num; i++) {
		p_sw_data = potential_roots_array[i];
		p_color = get_color(p_sw_data);
		if (!updn_color_is_same_color(p_color, p_prev_color)) {
			group_idx++;
			p_prev_color = p_color;
		}

		cl_list_insert_tail(&color_groups[group_idx].roots, p_sw_data);
	}
	*p_color_groups_num = group_idx + 1;

Error:
	free(potential_roots_array);
	return status;
}

static void updn_choose_one_color_group_roots(OUT osm_opensm_t *p_osm, cl_list_t *potential_roots_list,
					      uint16_t *leaf_lids_array, uint32_t leafs_num,
					      OUT unsigned *p_num_roots)
{
	updn_color_group_t *color_groups;
	updn_color_group_t *p_chosen_color_group;
	updn_sw_data_t *p_sw_data;
	cl_list_iterator_t list_iter;
	updn_color_group_t **pointer_color_groups;
	int i, color_groups_num;
	updn_color_t *p_chosen_color, *p_color;

	color_groups = (updn_color_group_t *)calloc(UPDN_MAX_COLOR, sizeof(updn_color_group_t));
	if (!color_groups) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR 8F18: Cannot allocate memory for color_groups array\n");
		goto Error;
	}

	updn_init_color_group_lists(color_groups);
	if (updn_sort_nodes_into_color_groups(&p_osm->log, potential_roots_list,
					      color_groups, &color_groups_num)) {
		updn_destroy_color_groups(color_groups);
		goto Error;
	}

	for (i = 0; i < color_groups_num; i++) {
		updn_color_group_t *p_color_group = &color_groups[i];
		updn_color_group_find_max_hops_from_leaf(p_color_group, leaf_lids_array,
							 leafs_num);
		updn_color_group_find_min_guid(p_color_group);
	}

	pointer_color_groups =
	    (updn_color_group_t **) malloc(color_groups_num * sizeof(updn_color_group_t *));
	if (!pointer_color_groups) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR 8F19: Cannot allocate memory for pointer_color_groups "
			"to choose a specific color group\n");
		updn_destroy_color_groups(color_groups);
		goto Error;
	}

	/* Fill an array of pointers to color groups for sorting */
	for (i = 0; i < color_groups_num; i++)
		pointer_color_groups[i] = &color_groups[i];

	qsort(pointer_color_groups, color_groups_num, sizeof(updn_color_group_t *),
	      updn_color_group_cmp_func);

	p_chosen_color_group = pointer_color_groups[0];
	OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
		"Chosen color group max_distanse_from_leaf: %d\n",
		p_chosen_color_group->max_distanse_from_leaf);

	/* Chosen color */
	list_iter = cl_list_head(&p_chosen_color_group->roots);
	p_sw_data = (updn_sw_data_t *)cl_list_obj(list_iter);
	p_chosen_color = get_color(p_sw_data);

	for (list_iter = cl_list_head(potential_roots_list);
	     list_iter != cl_list_end(potential_roots_list);
	     list_iter = cl_list_next(list_iter)) {
		p_sw_data = (updn_sw_data_t *)cl_list_obj(list_iter);
		p_color = get_color(p_sw_data);
		if (updn_color_is_sub_color(p_chosen_color, p_color)) {
			cl_list_insert_tail(&p_chosen_color_group->roots, p_sw_data);
		}
	}

	for (list_iter = cl_list_head(&p_chosen_color_group->roots);
	     list_iter != cl_list_end(&p_chosen_color_group->roots);
	     list_iter = cl_list_next(list_iter)) {
		p_sw_data = (updn_sw_data_t *)cl_list_obj(list_iter);
		updn_rank_root_node(p_osm, p_sw_data, p_num_roots);
	}

	free(pointer_color_groups);
	updn_destroy_color_groups(color_groups);
Error:
	return;
}

static int find_minimum_roots_layer(osm_opensm_t *p_osm, osm_subn_t *p_subn, cl_list_t *p_first_layer_roots_list,
				    cl_list_t *p_potential_roots_list, int current_roots_layer_rank)
{
	updn_sw_data_t **queue;
	uint32_t queue_in = 0;
	uint32_t queue_out = 0;
	uint32_t next_queue_in = 0;
	uint32_t prev_queue_out = 0;
	updn_sw_data_t **peers;
	updn_sw_data_t *p_sw_data;
	updn_sw_data_t *peer;
	boolean_t all_children_same_color;
	cl_list_iterator_t list_iter;
	int minimum_roots_layer_rank = current_roots_layer_rank;
	int i;

	queue = (updn_sw_data_t **)calloc(cl_qmap_count(&p_subn->sw_guid_tbl),
					  sizeof(updn_sw_data_t *));
	if (!queue) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR 8F1D: Cannot allocate a queue for finding minimum roots layer\n");
		return minimum_roots_layer_rank;
	}

	for (list_iter = cl_list_head(p_first_layer_roots_list);
	     list_iter != cl_list_end(p_first_layer_roots_list);
	     list_iter = cl_list_next(list_iter)) {
		p_sw_data = (updn_sw_data_t *)cl_list_obj(list_iter);

		queue[queue_in++] = p_sw_data;
		p_sw_data->visited = TRUE;
	}

	next_queue_in = queue_in;
	prev_queue_out = queue_out;
	while (queue_out < queue_in) {
		all_children_same_color = TRUE;
		p_sw_data = queue[queue_out++];
		peers = p_sw_data->peers;

		for (i = 0; i < p_sw_data->degree; i++) {
			updn_color_t *p_color = get_color(p_sw_data);
			peer = peers[i];

			if ((p_sw_data->rank > peer->rank) &&
			    (!peer->rank || !updn_color_is_same_color(p_color,
								      get_color(peer)))) {
				all_children_same_color = FALSE;
				break;
			}
		}

		if (all_children_same_color) {
			for (i = 0; i < p_sw_data->degree; i++) {
				peer = peers[i];
				if ((p_sw_data->rank > peer->rank) && peer->rank &&
				    !peer->is_potential_root && !peer->visited) {
					queue[next_queue_in++] = peer;
					peer->visited = TRUE;
				}
			}
		} else if (!p_sw_data->is_potential_root) {
			queue_out = prev_queue_out;
			break;
		}

		if (queue_out == queue_in) {
			/* Check next layer */
			minimum_roots_layer_rank--;
			queue_in = next_queue_in;
			prev_queue_out = queue_out;
		}
	}

	while (queue_out < queue_in) {
		p_sw_data = queue[queue_out++];
		if (!p_sw_data->is_potential_root) {
			cl_list_insert_tail(p_potential_roots_list, p_sw_data);
			p_sw_data->is_potential_root = TRUE;
		}
	}

	queue_out = 0;
	queue_in = next_queue_in;
	while (queue_out < queue_in) {
		p_sw_data = queue[queue_out++];
		p_sw_data->visited = FALSE;
	}

	free(queue);

	return minimum_roots_layer_rank;
}

static void updn_rank_root_nodes(IN osm_opensm_t *p_osm, IN osm_subn_t *p_subn,
				 OUT updn_sw_data_t *data_per_sw,
				 IN uint32_t max_rank, IN updn_color_t all_colors,
				 IN uint16_t *leaf_lids_array, IN uint32_t leafs_num,
				 OUT unsigned *p_num_roots)
{
	cl_list_t tmp_roots_list, potential_roots_list;
	updn_sw_data_t **peers;
	updn_sw_data_t *p_sw_data, *peer;
	boolean_t all_roots_with_same_color = TRUE;
	cl_list_iterator_t list_iter;
	int layer_rank = max_rank;
	int i, j, current_roots_layer_rank;
	int minimum_roots_layer_rank;

	OSM_LOG_ENTER(&p_osm->log);
	cl_list_construct(&tmp_roots_list);
	cl_list_init(&tmp_roots_list, 10);
	cl_list_construct(&potential_roots_list);
	cl_list_init(&potential_roots_list, 10);

	/*
	 * Traverse on all the nodes with maximal rank, and find the potential roots.
	 * If no potential roots found on the maximal rank layer,
	 * continue to look on the layers beneath.
	 * This roots can be of 2 types:
	 * 1. The node plus all its peers have all the colors.
	 * 2. The node have all the colors. In this case need to check
	 *    if in some layer beneath this node all the nodes have the same colors as well,
	 *    and take this layer instead of the node as potential roots.
	 */
	while (layer_rank && cl_is_list_empty(&tmp_roots_list)) {
		for (i = 0; i < (p_subn->max_lid + 1); i++) {
			updn_color_t same_rank_peers_color;
			updn_color_t tmp_color;
			p_sw_data = &data_per_sw[i];
			peers = p_sw_data->peers;

			if (!p_sw_data->p_sw || (p_sw_data->rank != layer_rank))
				continue;

			updn_color_erase_colors(&same_rank_peers_color);

			for (j = 0; j < p_sw_data->degree; j++) {
				peer = peers[j];

				if (p_sw_data->rank == peer->rank)
					updn_color_add_colors(&same_rank_peers_color, get_color(peer));
			}

			updn_color_set_colors(&tmp_color, get_color(p_sw_data));
			updn_color_add_colors(&tmp_color, &same_rank_peers_color);
			if (updn_color_is_same_color(get_color(p_sw_data), &all_colors)) {
				cl_list_insert_tail(&tmp_roots_list, p_sw_data);
			} else if (updn_color_is_same_color(&tmp_color, &all_colors)) {
				all_roots_with_same_color = FALSE;
				cl_list_insert_tail(&tmp_roots_list, p_sw_data);
			} else if (updn_color_is_same_color(get_color(p_sw_data), &tmp_color) &&
				   !updn_color_check_single_color(get_color(p_sw_data))) {
				all_roots_with_same_color = FALSE;
				cl_list_insert_tail(&tmp_roots_list, p_sw_data);
			}
		}
		layer_rank--;
	}

	/* Current root layer is the last we checked in the while loop before exiting */
	current_roots_layer_rank = layer_rank + 1;
	/*
	 * If all potential roots are with the same colors, need to check if there
	 * is a layer beneath that can act as roots instead of the current layer.
	 * This is done to insure the routing will be optimal.
	 */
	if (all_roots_with_same_color) {
		minimum_roots_layer_rank =
		    find_minimum_roots_layer(p_osm, p_subn, &tmp_roots_list,
					     &potential_roots_list, current_roots_layer_rank);
		if (minimum_roots_layer_rank != current_roots_layer_rank) {
			for (i = 0; i < (p_subn->max_lid + 1); i++) {
				p_sw_data = &data_per_sw[i];

				/*
				 * Add switches from the same layer to the roots list.
				 * Skip roots that were already added and switches from different layer
				 */
				if (!p_sw_data->p_sw || (p_sw_data->rank != minimum_roots_layer_rank) ||
				    p_sw_data->is_potential_root)
					continue;

				cl_list_insert_tail(&potential_roots_list, p_sw_data);
				p_sw_data->is_potential_root = TRUE;
			}
		}
	} else {
		for (list_iter = cl_list_head(&tmp_roots_list);
		     list_iter != cl_list_end(&tmp_roots_list);
		     list_iter = cl_list_next(list_iter)) {
			p_sw_data = (updn_sw_data_t *)cl_list_obj(list_iter);
			cl_list_insert_tail(&potential_roots_list, p_sw_data);
			p_sw_data->is_potential_root = TRUE;
		}
	}

	/* Apply the algorithm results only if all the roots are valid */
	if (cl_list_count(&potential_roots_list) &&
	    updn_validate_roots(p_osm, p_subn, &potential_roots_list, leafs_num))
		updn_choose_one_color_group_roots(p_osm, &potential_roots_list,
						  leaf_lids_array, leafs_num, p_num_roots);

	cl_list_remove_all(&tmp_roots_list);
	cl_list_destroy(&tmp_roots_list);
	cl_list_remove_all(&potential_roots_list);
	cl_list_destroy(&potential_roots_list);
	OSM_LOG_EXIT(&p_osm->log);
}

static void root_detection_init_data_per_sw_array(OUT updn_sw_data_t *data_per_sw,
						  OUT uint32_t *p_cas_num,
						  IN osm_subn_t *p_subn)
{
	updn_sw_data_t *p_sw_data;
	cl_map_item_t *item;
	osm_switch_t *p_sw;
	osm_node_t *p_node;
	uint16_t lid_ho;

	for (item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     item = cl_qmap_next(item)) {
		p_sw = (osm_switch_t *)item;
		p_node = p_sw->p_node;

		lid_ho = cl_ntoh16(osm_node_get_base_lid(p_node, 0));
		p_sw_data = &data_per_sw[lid_ho];
		p_sw_data->p_sw = p_sw;
		p_sw_data->rank = UPDN_NO_RANK;
		memset(&p_sw_data->color, 0, sizeof(updn_color_t));

		root_detection_add_peers_to_sw_data(p_sw_data, data_per_sw, p_cas_num);
	}

}

static void root_detection_uncolor_sw_array(OUT updn_sw_data_t *data_per_sw,
					    IN osm_subn_t *p_subn)
{
	updn_sw_data_t *p_sw_data;
	int i;

	for (i = 0; i <= p_subn->max_lid; i++) {
		p_sw_data = &data_per_sw[i];

		/* Index is not of existing switch */
		if (!p_sw_data->p_sw)
			continue;

		/* Found potential leaves rank is set to 0, hence avoid clearing it */
		if (p_sw_data->rank)
			p_sw_data->rank = UPDN_NO_RANK;
		updn_color_erase_colors(&p_sw_data->color);
		p_sw_data->visited = FALSE;
	}
}

/************************************* DF+ section **************************************/

static void init_dfp_color_group(dfp_color_group_t * p_color_group,
				 updn_color_t * color,
				 uint8_t group_num)
{
	memcpy(p_color_group->p_own_colors_bitmask, color->colors,
	       UPDN_COLORS_ARRAY_SIZE * sizeof(uint64_t));

	/* Assuming at first group creating, color is a single color */
	p_color_group->own_colors_count = 1;
	memset(p_color_group->p_neighbor_groups_bitmask, 0, UPDN_COLOR_ARRAY_SIZE_IN_BYTES);
	p_color_group->neighbor_groups_count = 0;
	p_color_group->potential_roots_rank = UPDN_NO_RANK;
	p_color_group->group_num = group_num;
	cl_list_construct(&p_color_group->new_members);
	cl_list_init(&p_color_group->new_members, 10);
	cl_list_construct(&p_color_group->potential_new_members);
	cl_list_init(&p_color_group->potential_new_members, 10);

	memset(p_color_group->color_mask1, 0, UPDN_COLOR_ARRAY_SIZE_IN_BYTES);
	memset(p_color_group->color_mask2, 0, UPDN_COLOR_ARRAY_SIZE_IN_BYTES);
	p_color_group->is_subgroup = FALSE;
}

static void destroy_dfp_color_group(dfp_color_group_t * p_color_group)
{
	cl_list_remove_all(&p_color_group->new_members);
	cl_list_destroy(&p_color_group->new_members);
	cl_list_remove_all(&p_color_group->potential_new_members);
	cl_list_destroy(&p_color_group->potential_new_members);
}

static void destroy_dfp_color_groups_list(cl_list_t * p_group_list)
{
	cl_list_iterator_t list_iter;
	dfp_color_group_t *color_group;

	for (list_iter = cl_list_head(p_group_list);
	     list_iter != cl_list_end(p_group_list);
	     list_iter = cl_list_next(list_iter)) {
		color_group = (dfp_color_group_t *)cl_list_obj(list_iter);

		destroy_dfp_color_group(color_group);
		free(color_group);
	}
	cl_list_remove_all(p_group_list);
	cl_list_destroy(p_group_list);
}

static void root_detection_clear_dfp_color_groups(cl_list_t * p_group_list,
						  dfp_color_group_t ** color_group_arr)
{
	cl_list_iterator_t list_iter;
	dfp_color_group_t *color_group;

	for (list_iter = cl_list_head(p_group_list);
	     list_iter != cl_list_end(p_group_list);
	     list_iter = cl_list_next(list_iter)) {
		color_group = (dfp_color_group_t *)cl_list_obj(list_iter);

		destroy_dfp_color_group(color_group);
		free(color_group);
	}
	cl_list_remove_all(p_group_list);

	memset(color_group_arr, 0, MAX_COLOR_GROUPS * sizeof(dfp_color_group_t *));
}

void osm_root_detection_clear_dfp_potential_roots(cl_u64_vector_t * p_dfp_potential_roots)
{
	cl_u64_vector_remove_all(p_dfp_potential_roots);
}

static void dfp_remove_color_from_potential_new_members(dfp_color_group_t * p_color_group)
{
	cl_list_iterator_t member_list_iter;
	updn_sw_data_t *p_sw_date;

	for (member_list_iter = cl_list_head(&p_color_group->potential_new_members);
	     member_list_iter != cl_list_end(&p_color_group->potential_new_members);
	     member_list_iter = cl_list_next(member_list_iter)) {
		p_sw_date = (updn_sw_data_t *)cl_list_obj(member_list_iter);
		updn_color_remove_colors(p_sw_date->color.colors,
					 p_color_group->p_own_colors_bitmask);
	}

	cl_list_remove_all(&p_color_group->potential_new_members);
}

/* Step A */
static void dfp_color_validate_next_rank_colors(cl_list_t * p_dfp_color_groups_list,
						updn_sw_data_t ** queue,
						uint32_t start_idx,
						uint32_t end_idx,
						uint32_t rank)
{
	updn_sw_data_t *p_sw_data;
	updn_sw_data_t **peers;
	uint32_t j = start_idx;
	int i;
	dfp_color_group_t *sw_color_group, *color_group;
	uint64_t *color_mask1 = NULL, *color_mask2 = NULL, *peer_colors;
	cl_list_iterator_t list_iter;

	while (j < end_idx) {
		p_sw_data = queue[j++];
		peers = p_sw_data->peers;
		sw_color_group = p_sw_data->color.p_color_group;

		for (i = 0; i < p_sw_data->degree; i++) {
			if (peers[i]->rank <= p_sw_data->rank)
				continue;

			/*
			 * Meaning group was marked as an island by previous
			 * coloring iteration OR by this function.
			 */
			if (sw_color_group->potential_roots_rank != UPDN_NO_RANK) {
				updn_color_remove_colors(peers[i]->color.colors,
							 sw_color_group->p_own_colors_bitmask);
				continue;
			}

			if (updn_color_is_subset_color(sw_color_group->p_own_colors_bitmask,
						       peers[i]->color.colors))
				continue;

			/*
			 * At this point node's color is not subset of color group.
			 * Hence potential new member
			 */
			cl_list_insert_head(&sw_color_group->potential_new_members,
					    (void *)peers[i]);

			peer_colors = peers[i]->color.colors;
			color_mask1 = sw_color_group->color_mask1;
			color_mask2 = sw_color_group->color_mask2;

			if (!is_colored_arr(color_mask1) ||
			    updn_color_is_subset_color(peer_colors, color_mask1)) {
				memcpy(color_mask1, peer_colors, UPDN_COLOR_ARRAY_SIZE_IN_BYTES);

				if (is_colored_arr(color_mask2) &&
				    updn_color_is_subset_color(peer_colors, color_mask2))
					memset(color_mask2, 0, UPDN_COLOR_ARRAY_SIZE_IN_BYTES);

			} else if (is_colored_arr(color_mask2) &&
				   updn_color_is_subset_color(peer_colors, color_mask2)) {
				memcpy(color_mask2, peer_colors, UPDN_COLOR_ARRAY_SIZE_IN_BYTES);
			}
		}
	}

	/*
	 * Iterate over color groups. If found 2 colors that are not subcolors
	 * (each is a part of another group) - remove color from all 'new members'
	 */
	for (list_iter = cl_list_head(p_dfp_color_groups_list);
	     list_iter != cl_list_end(p_dfp_color_groups_list);
	     list_iter = cl_list_next(list_iter)) {
		color_group = (dfp_color_group_t *)cl_list_obj(list_iter);
		color_mask1 = color_group->color_mask1;
		color_mask2 = color_group->color_mask2;

		if (!cl_list_count(&color_group->potential_new_members))
			continue;

		/* If group has roots rank, would not have colored next rank */
		if (color_group->potential_roots_rank != UPDN_NO_RANK)
			continue;

		if (color_mask1 && color_mask2 &&
		    is_colored_arr(color_mask2) &&
		    updn_color_not_subsets_colors(color_mask1, color_mask2)) {
			color_group->potential_roots_rank = rank;
			/* Notice we empty new members list inside this function */
			dfp_remove_color_from_potential_new_members(color_group);
		} else {
			/*
			 * If got here, group has real members of next rank.
			 * Hence, empty list here so will be updated at
			 * inserting to group (next step of this iteration).
			 */
			cl_list_remove_all(&color_group->new_members);
		}
	}
}

/* NOTICE!!!! This merge function avoids updating potential roots rank */
static void dfp_color_merge_colors(dfp_color_group_t * color_group_a,
				   dfp_color_group_t * color_group_b,
				   dfp_color_group_t ** color_group_arr)
{
	int i, j;
	int array_size = UPDN_COLORS_ARRAY_SIZE;
	uint64_t *arr_a, *arr_b, *peer_bitmask;
	uint64_t added[UPDN_COLORS_ARRAY_SIZE], group_a_bit, group_b_bit;
	uint8_t a_group_num, b_group_num, a_block_idx, b_block_idx,  count = 0;
	dfp_color_group_t *peer_group, *curr_group;

	/* Note that own_colors holds all the colors that were merged into the color group
	 * Hence, all point to the same group by color_group_arr */
	arr_a = color_group_a->p_own_colors_bitmask;
	arr_b = color_group_b->p_own_colors_bitmask;
	b_group_num = color_group_b->group_num;
	for (i = 0; i < array_size; i++) {
		if (arr_b[i] != 0 && arr_a[i] != arr_b[i]) {
			/* Count those that are added from b to a */
			added[i] = arr_b[i] & ~(arr_a[i]);
			for (j = 0; j < UPDN_COLORS_ARRAY_ELEMENT_SIZE; j++) {
				if (added[i] & ((uint64_t)1 << j)) {
					count++;
					/* set group a as new color's color group */
					curr_group = color_group_arr[UPDN_COLORS_ARRAY_ELEMENT_SIZE * i + j];
					curr_group->group_num = color_group_a->group_num;
					color_group_arr[UPDN_COLORS_ARRAY_ELEMENT_SIZE * i + j] =
					    color_group_a;
				}
			}
			arr_a[i] |= arr_b[i];
		}
	}
	color_group_a->own_colors_count += count;

	/* Merge neighbors */
	arr_a = color_group_a->p_neighbor_groups_bitmask;
	arr_b = color_group_b->p_neighbor_groups_bitmask;
	for (i = 0; i < array_size; i++) {
		if (arr_b[i] != 0 && arr_a[i] != arr_b[i]) {
			/* Count neighbors that are added from b to a */
			added[i] = arr_b[i] & ~(arr_a[i]);
			for (j = 0; j < UPDN_COLORS_ARRAY_ELEMENT_SIZE; j++) {
				if (added[i] & ((uint64_t)1 << j))
					color_group_a->neighbor_groups_count++;
			}
			arr_a[i] |= arr_b[i];
		}
	}

	/* Remove groups b and a from neighbors of merged group */
	a_group_num = color_group_a->group_num;
	group_a_bit = (uint64_t)1 << (a_group_num % UPDN_COLORS_ARRAY_ELEMENT_SIZE);
	a_block_idx = a_group_num / UPDN_COLORS_ARRAY_ELEMENT_SIZE;
	group_b_bit = (uint64_t)1 << (b_group_num % UPDN_COLORS_ARRAY_ELEMENT_SIZE);
	b_block_idx = b_group_num / UPDN_COLORS_ARRAY_ELEMENT_SIZE;
	if (arr_a[b_block_idx] & group_b_bit) {
		arr_a[b_block_idx] &= ~group_b_bit;
		color_group_a->neighbor_groups_count--;
	}
	if (arr_a[a_block_idx] & group_a_bit) {
		arr_a[a_block_idx] &= ~group_a_bit;
		color_group_a->neighbor_groups_count--;
	}

	/*
	 * For each neighbor of b, remove b from neighbors and mark a
	 * as a neighbor
	 */
	for (i = 0; i < array_size; i++) {
		for (j = 0; j < UPDN_COLORS_ARRAY_ELEMENT_SIZE; j++) {
			if (!(arr_b[i] & ((uint64_t)1 << j)))
				continue;

			peer_group =
			    color_group_arr[i * UPDN_COLORS_ARRAY_ELEMENT_SIZE + j];
			peer_bitmask = peer_group->p_neighbor_groups_bitmask;
			if (peer_bitmask[a_block_idx] & group_a_bit) {
				peer_bitmask[a_block_idx] |= group_a_bit;
				peer_group->neighbor_groups_count++;
			}

			if (peer_bitmask[b_block_idx] & group_b_bit) {
				peer_bitmask[b_block_idx] &= ~group_b_bit;
				peer_group->neighbor_groups_count--;
			}
		}
	}

	/* Update roots rank - later on, when finding multiple neighbor colors */
}

static void dfp_validate_color_groups(cl_list_t * p_dfp_color_groups_list,
				      dfp_color_group_t ** color_group_arr,
				      uint32_t rank)
{
	dfp_color_group_t *p_color_group, *other_group;
	cl_list_iterator_t list_iter;
	uint64_t *neighbor_arr;
	int i, j;

	for (list_iter = cl_list_head(p_dfp_color_groups_list);
	     list_iter != cl_list_end(p_dfp_color_groups_list);
	     list_iter = cl_list_next(list_iter)) {
		p_color_group = (dfp_color_group_t *)cl_list_obj(list_iter);

		if (p_color_group->is_subgroup)
			continue;

		if (p_color_group->neighbor_groups_count == 1) {
			/*
			 * A color group that has only 1 other color group as a
			 * neighbor is in fact a subcolor of that neighbor.
			 * Hence, merge them.
			 */
			neighbor_arr = p_color_group->p_neighbor_groups_bitmask;
			for (i = 0; i < UPDN_COLORS_ARRAY_SIZE; i++) {
				if (neighbor_arr[i] == 0)
					continue;

				/* Find the single neighbor group and merge them */
				for (j = 0; j < UPDN_COLORS_ARRAY_ELEMENT_SIZE; j++) {
					if (!(neighbor_arr[i] & ((uint64_t)1 << j)))
						continue;
					other_group = color_group_arr[UPDN_COLORS_ARRAY_ELEMENT_SIZE * i + j];

					dfp_color_merge_colors(other_group, p_color_group,
							       color_group_arr);
					p_color_group->potential_roots_rank = UPDN_NO_RANK;
					p_color_group->is_subgroup = TRUE;

					i = UPDN_COLORS_ARRAY_SIZE;
					break;
				}
			}
		} else if (p_color_group->potential_roots_rank == UPDN_NO_RANK) {
			if (p_color_group->neighbor_groups_count > 1) {
				/*
				 * A color group that is connected to multiple
				 * other color groups is an island.
				 */
				/*
				 * Note: Assuming up to 3 levels topology!
				 * In order to support deeper topology, consider separating
				 * this step to another iteration, before the 'single color
				 * neighbors' check (above), and make sure neighbors are also
				 * multiple colored.
				 */
				p_color_group->potential_roots_rank = rank;
			} else {
				/*
				 * If color got no potential rank at the end of
				 * the iteration, these new members are not its
				 * roots, hence no need to save them.
				 */
				cl_list_remove_all(&p_color_group->new_members);
			}
		}
	}
}

/* NOTICE!!!! This merge function avoids updating potential roots rank */
static void dfp_set_groups_as_neighbors(dfp_color_group_t *p_color_group_a,
					dfp_color_group_t *p_color_group_b)
{
	uint64_t *neighbors_bitmask, group_bit;
	uint8_t block_idx;

	/* Set b to a's neighbors */
	neighbors_bitmask = p_color_group_a->p_neighbor_groups_bitmask;
	block_idx = p_color_group_b->group_num / UPDN_COLORS_ARRAY_ELEMENT_SIZE;
	group_bit = (uint64_t)1 << (p_color_group_b->group_num % UPDN_COLORS_ARRAY_ELEMENT_SIZE);
	if (!(neighbors_bitmask[block_idx] & group_bit)) {
		neighbors_bitmask[block_idx] |= group_bit;
		p_color_group_a->neighbor_groups_count++;
	}

	/* Set a to b's neighbors */
	neighbors_bitmask = p_color_group_b->p_neighbor_groups_bitmask;
	block_idx = p_color_group_a->group_num / UPDN_COLORS_ARRAY_ELEMENT_SIZE;
	group_bit = (uint64_t)1 << (p_color_group_a->group_num % UPDN_COLORS_ARRAY_ELEMENT_SIZE);
	if (!(neighbors_bitmask[block_idx] & group_bit)) {
		neighbors_bitmask[block_idx] |= group_bit;
		p_color_group_b->neighbor_groups_count++;
	}
}

static void dfp_insert_to_color_groups(cl_list_t * p_dfp_color_groups_list,
				       dfp_color_group_t ** color_group_arr,
				       updn_sw_data_t ** queue,
				       uint32_t start_idx,
				       uint32_t end_idx)
{
	updn_sw_data_t *p_sw_data;
	updn_sw_data_t *peer;
	updn_sw_data_t **peers;
	uint32_t i, j = start_idx, offset;
	uint8_t sw_color_num = 0;
	uint64_t color_bit, *curr_group_color_bitmask;
	dfp_color_group_t *sw_color_group, *other_group;

	while (j < end_idx) {
		p_sw_data = queue[j++];
		if (!updn_color_get_color_num(&p_sw_data->color, &sw_color_num))
			continue;

		sw_color_group = color_group_arr[sw_color_num];
		curr_group_color_bitmask = sw_color_group->p_own_colors_bitmask;

		/*
		 * Node may hold multiple colors, hence iterates over
		 * the whole color array.
		 */
		for (i = 0; i < UPDN_COLORS_ARRAY_SIZE; i++) {
			if (p_sw_data->color.colors[i] == 0 ||
			    curr_group_color_bitmask[i] == p_sw_data->color.colors[i])
				continue;

			for (offset = 0; offset < UPDN_COLORS_ARRAY_ELEMENT_SIZE;
			     offset++) {
				color_bit = (uint64_t)1 << offset;
				/*
				 * If node has colors that are not part of its
				 * current color group yet, merge each of them
				 * into its group.
				 */
				if ((curr_group_color_bitmask[i] & color_bit) ||
				    !(p_sw_data->color.colors[i] & color_bit))
					continue;

				other_group =
				    color_group_arr[UPDN_COLORS_ARRAY_ELEMENT_SIZE *
						    i + offset];
				dfp_color_merge_colors(sw_color_group,
						       other_group,
						       color_group_arr);
				other_group->is_subgroup = TRUE;
			}
		}
	}

	/* Add peers of newly colored as group's neighbors */
	j = start_idx;
	while (j < end_idx) {
		p_sw_data = queue[j++];
		peers = p_sw_data->peers;
		if (!updn_color_get_color_num(&p_sw_data->color, &sw_color_num))
			continue;

		sw_color_group = color_group_arr[sw_color_num];

		/* Notice: if color was reverted, it got a potential roots rank */
		if (sw_color_group->potential_roots_rank == UPDN_NO_RANK)
			cl_list_insert_head(&sw_color_group->new_members, p_sw_data);

		for (i = 0; i < p_sw_data->degree; i++) {
			peer = peers[i];
			if (!is_colored_arr(peer->color.colors) ||
			    peer->color.p_color_group == sw_color_group)
				continue;

			/*
			 * Update neighbor_groups of both curr and peer.
			 * At this point - avoids updating potential_roots_rank.
			 */
			dfp_set_groups_as_neighbors(color_group_arr[sw_color_group->group_num],
						    color_group_arr[peer->color.p_color_group->group_num]);
		}
	}
}

static boolean_t dfp_validate_islands(cl_list_t * p_dfp_color_groups_list)
{
	cl_list_iterator_t list_iter;
	dfp_color_group_t *color_group;
	uint8_t num_groups = 0;

	for (list_iter = cl_list_head(p_dfp_color_groups_list);
	     list_iter != cl_list_end(p_dfp_color_groups_list);
	     list_iter = cl_list_next(list_iter)) {
		color_group = (dfp_color_group_t *)cl_list_obj(list_iter);
		if (!color_group->is_subgroup) {
			num_groups++;
		}
	}
	/*
	 * Assuming it is sufficient to have one island connected to all the
	 * others, since islands are not necessarily a clique
	 */
	for (list_iter = cl_list_head(p_dfp_color_groups_list);
	     list_iter != cl_list_end(p_dfp_color_groups_list);
	     list_iter = cl_list_next(list_iter)) {
		color_group = (dfp_color_group_t *)cl_list_obj(list_iter);
		if (color_group->is_subgroup)
			continue;
		if (color_group->neighbor_groups_count == (num_groups - 1))
			return TRUE;
	}

	return FALSE;
}

static uint32_t root_detection_dfp_uncolor_new_members(osm_subn_t * p_subn,
						       cl_list_t * new_members,
						       updn_sw_data_t ** colorless_queue,
						       uint64_t * group_color)
{
	updn_sw_data_t * p_sw_data, **peers, **queue = NULL;
	cl_list_iterator_t member_list_iter;
	uint32_t queue_in = 0, queue_out = 0, i, prev_queue_in;
	uint32_t colorless_count = 0;

	queue = (updn_sw_data_t **)calloc(cl_qmap_count(&p_subn->sw_guid_tbl),
					  sizeof(updn_sw_data_t *));
	if (!queue) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR 8F04: "
			"Cannot allocate memory for updn_sw_data Queue\n");
		goto Exit;
	}

	/* Insert new members to queue */
	for (member_list_iter = cl_list_head(new_members);
	     member_list_iter != cl_list_end(new_members);
	     member_list_iter = cl_list_next(member_list_iter))
		queue[queue_in++] = (updn_sw_data_t *)cl_list_obj(member_list_iter);

	while (queue_out < queue_in) {
		p_sw_data = queue[queue_out++];
		peers = p_sw_data->peers;

		prev_queue_in = queue_in;
		for (i = 0; i < p_sw_data->degree; i++) {
			if (peers[i]->rank < p_sw_data->rank ||
			    !is_colored_arr(peers[i]->color.colors))
				continue;

			/*
			 * Iterate over same color and uncolor nodes until find sub-color peer.
			 * Hence, add peer of the same color to queue.
			 */
			if (updn_color_is_same_color_mask(peers[i]->color.colors, group_color)) {
				queue[queue_in++] = peers[i];
			} else if (updn_color_is_subset_color(group_color, peers[i]->color.colors)) {
				/* Add to colorless queue only if found peers of subcolor */
				colorless_queue[colorless_count++] = p_sw_data;

				/*
				 * If current switch is found as colorless, avoid iterating on its
				 * colored peers.
				 */
				queue_in = prev_queue_in;
				break;
			}
		}

		/* Uncolor once done checking its peers */
		updn_color_erase_colors(&p_sw_data->color);
	}

	free(queue);
Exit:
	return colorless_count;
}

static void dfp_iterate_group_member_neighbors(osm_subn_t * p_subn,
					       cl_list_t * p_dfp_color_groups_list,
					       dfp_color_group_t ** color_group_arr,
					       updn_sw_data_t ** colorless_queue,
					       dfp_color_group_t * color_group,
					       updn_sw_data_t * p_member,
					       OUT uint32_t * p_colorless_count,
					       OUT cl_u64_vector_t * dfp_potential_roots)
{
	updn_sw_data_t **member_peers, **potential_colorless_peers;
	int i, j;
	uint64_t node_guid;
	boolean_t is_root = FALSE;

	member_peers = p_member->peers;

	if (p_member->rank == 1) {
		node_guid = cl_ntoh64(p_member->p_sw->p_node->node_info.node_guid);
		cl_u64_vector_insert(dfp_potential_roots, node_guid, NULL);
		is_root = TRUE;
	} else {
		for (i = 0; i < p_member->degree; i++) {
			/*
			 * Node is a root if it has potential root rank, and it is
			 * connected to at least one other color group.
			 */
			if (is_colored(member_peers[i]) &&
			    color_group_arr[member_peers[i]->color.p_color_group->group_num] !=
			    color_group_arr[color_group->group_num]) {
				node_guid = cl_ntoh64(p_member->p_sw->p_node->
						          node_info.node_guid);
				cl_u64_vector_insert(dfp_potential_roots, node_guid, NULL);
				is_root = TRUE;
				break;
			}
		}
	}

	if (!colorless_queue || !is_root)
		return;

	/* If found a root, check its colorless peers */
	for (i = 0; i < p_member->degree; i++) {
		/* Already added to colorless_queue, or colored */
		if (member_peers[i]->color.in_colorless_queue ||
		    is_colored_arr(member_peers[i]->color.colors))
			continue;

		/*
		 * Add colorless node to colorless queue in case it has at least
		 * two peers of different colors, meaning it is a colorless root.
		 *
		 * If both leaf and its root are colorless, root will be
		 * in colorless queue, and the other node will be detected through
		 * the root during 'root_detection_dfp_find_potential_leaves' BFS.
		 * This way, leaf node will be found to be a leaf.
		 */
		potential_colorless_peers = member_peers[i]->peers;
		for (j = 0; j < member_peers[i]->degree; j++) {
			/*
			 * Colorless node is a  potential root - connected to
			 * at least 2 color group (color_group and another one).
			 */
			if (is_colored(potential_colorless_peers[j]) &&
			    color_group_arr[potential_colorless_peers[j]->
				color.p_color_group->group_num] !=
				color_group_arr[color_group->group_num]) {

				colorless_queue[(*p_colorless_count)++] = member_peers[i];
				member_peers[i]->color.in_colorless_queue = TRUE;
				break;
			}
		}
	}
}


static uint32_t dfp_rank_root_nodes(osm_subn_t * p_subn,
				    cl_list_t * p_dfp_color_groups_list,
				    dfp_color_group_t ** color_group_arr,
				    updn_sw_data_t ** colorless_queue,
				    OUT cl_u64_vector_t * dfp_potential_roots)
{
	cl_list_iterator_t list_iter, member_list_iter;
	dfp_color_group_t *p_color_group;
	updn_sw_data_t *p_member;
	uint32_t colorless_count = 0;

	/*
	 * Assuming it is sufficient to have one island connected to all the
	 * others, since islands are not necessarily a clique
	 */
	for (list_iter = cl_list_head(p_dfp_color_groups_list);
	     list_iter != cl_list_end(p_dfp_color_groups_list);
	     list_iter = cl_list_next(list_iter)) {
		p_color_group = (dfp_color_group_t *)cl_list_obj(list_iter);
		if (p_color_group->is_subgroup)
			continue;

		/*
		 * Avoid this step if not on first coloring attempt (colorless
		 * queue is NULL).
		 * We assume islands are up to 3 level. Hence, when rank > 2, it
		 * means color group (that should represent an island) got beyond
		 * the roots of the island.
		 */
		if (colorless_queue && p_color_group->potential_roots_rank > 2 &&
		    p_color_group->potential_roots_rank != UPDN_NO_RANK) {
			colorless_count += root_detection_dfp_uncolor_new_members(
					       p_subn, &p_color_group->new_members,
					       colorless_queue,
					       p_color_group->p_own_colors_bitmask);
			continue;
		}

		for (member_list_iter= cl_list_head(&p_color_group->new_members);
		     member_list_iter != cl_list_end(&p_color_group->new_members);
		     member_list_iter = cl_list_next(member_list_iter)) {
			p_member = (updn_sw_data_t *)cl_list_obj(member_list_iter);
			dfp_iterate_group_member_neighbors(
			    p_subn, p_dfp_color_groups_list, color_group_arr,
			    colorless_queue, p_color_group, p_member, &colorless_count,
			    dfp_potential_roots);
		}
	}
	return colorless_count;
}

static void root_detection_insert_leafs_to_queue(osm_subn_t * p_subn,
						 updn_sw_data_t ** queue,
						 updn_sw_data_t * data_per_sw,
						 uint32_t * p_num_of_leaves,
						 uint32_t * p_queue_in,
						 uint8_t topo)
{
	osm_switch_t *p_sw;
	cl_map_item_t *item;
	uint8_t max_cas_on_spine = (topo == OSM_TOPOLOGY_TYPE_DFP) ?
		p_subn->opt.dfp_max_cas_on_spine : p_subn->opt.max_cas_on_spine;
	uint16_t lid_ho;
	updn_sw_data_t *p_sw_data;

	/* Find leafs and put them in the queue */
	for (item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     item = cl_qmap_next(item)) {
		p_sw = (osm_switch_t *)item;
		lid_ho = cl_ntoh16(osm_node_get_base_lid(p_sw->p_node, 0));
		p_sw_data = &data_per_sw[lid_ho];

		if ((p_sw_data->cas_num > max_cas_on_spine) &&
		    p_sw_data->degree) {
			p_sw_data->rank = 0;
			queue[(*p_queue_in)++] = p_sw_data;
			(*p_num_of_leaves)++;
		}
	}
}

static void root_detection_color_nodes(osm_subn_t *p_subn,
				       updn_color_t * all_colors,
				       dfp_color_group_t ** color_group_arr,
				       cl_list_t * dfp_color_groups_list,
				       updn_sw_data_t ** queue,
				       uint32_t queue_in,
				       uint32_t * p_max_rank,
				       uint8_t topo)
{
	updn_sw_data_t *p_sw_data;
	updn_sw_data_t **peers;
	updn_color_t current_color, peers_color;
	uint8_t next_color_group_idx = 0;
	uint32_t next_color_idx = 0, i;
	uint32_t queue_out = 0, queue_prev_out;
	uint32_t next_queue_in = 0;

	/*
	 * Traverse the graph by layers while coloring the nodes in the following manner:
	 * 1. First traverse the nodes with hosts (Potential leaves).
	 *	a. Give a new color to the leaf if its neighbors aren't colored,
	 *	   otherwise, color the leaf with the neighbor's colors.
	 *	b. If a new color is given to the leaf, color with it
	 *	   the uncolored neighbors too.
	 * 2. A node colors its neighbor if the neighbor doesn't have a color
	 *    or, the neighbor has a bigger rank than the node.
	 *    In DF+ topologies, avoid color same rank peers.
	 * 3. A special case is when a node needs to pass its color to a neighbor on
	 *    the same rank.
	 *    + In UPDN, it only happens if the node can't pass its color to any
	 *      other node while the neighbor needs to pass its color to some node.
	 *    + In DF+, this can happen in two cases:
	 *      a. All peers of equal/larger rank are of the same color.
	 *         Hence this is a subcolor of its peers color group.
	 *      b. Node is a spine, hence other nodes of same rank and color have
	 *         equal/larger rank peers of multiple colors, that do not respond
	 *         to case a.
	 *         Hence node is a potential root, and so are the other nodes of
	 *         same rank and color.
	 * 4. Another special case, relevant to DF+ topologies in which islands
	 *    are of different levels.
	 *    Coloring step may color forward from low level islands to the roots
	 *    of the higher lever islands. Hence, revert coloring forward if
	 *    found to color nodes that belong to different colors - Step A.
	 */
	updn_color_erase_colors(all_colors);

	next_queue_in = queue_in;
	queue_prev_out = queue_out;
	while (queue_out < queue_in) {
		p_sw_data = queue[queue_out++];
		peers = p_sw_data->peers;

		updn_color_erase_colors(&current_color);
		updn_color_erase_colors(&peers_color);

		if (p_sw_data->rank > *p_max_rank)
			*p_max_rank = p_sw_data->rank;

		for (i = 0; i < p_sw_data->degree; i++) {
			if (topo == OSM_TOPOLOGY_TYPE_DFP) {
				/*
				 * In DF+ topologies, should:
				 * 1. Avoid color by same rank peers
				 * 2. Maintain color groups
				 */
				if (peers[i]->rank == p_sw_data->rank)
					continue;

				if (!peers_color.p_color_group)
					peers_color.p_color_group =
					    peers[i]->color.p_color_group;
			} 
			updn_color_add_colors(&peers_color, get_color(peers[i]));
		}

		/* Color current by its peers color, or give it a new color */
		if (!is_colored(p_sw_data)) {
			if (updn_color_is_colored(&peers_color)) {
				updn_color_add_colors(&current_color,
						      &peers_color);
				current_color.p_color_group = peers_color.p_color_group;
			} else {
				/* current_color is empty at this point */
				updn_color_add_single_color(&current_color, next_color_idx);
				updn_color_add_single_color(all_colors, next_color_idx);
				next_color_idx++;

				if (topo == OSM_TOPOLOGY_TYPE_DFP) {
					current_color.p_color_group =
					    (dfp_color_group_t *)malloc(sizeof(dfp_color_group_t));

					init_dfp_color_group(current_color.p_color_group,
							     &current_color, next_color_group_idx++);
					cl_list_insert_head(dfp_color_groups_list,
							    (void *)current_color.p_color_group);

					color_group_arr[next_color_group_idx - 1] =
					    current_color.p_color_group;
				}
			}

			/* Set new color */
			set_color(p_sw_data, &current_color);

			/*
			 * For DF+, avoid coloring forward from a node that its color
			 * group is already found to be an island (at the end of
			 * previous iteration)
			 */
			if (!(topo == OSM_TOPOLOGY_TYPE_DFP &&
			     (!p_sw_data->color.p_color_group ||
			      p_sw_data->color.p_color_group->potential_roots_rank ==
			      p_sw_data->rank))) {
				/* Color peers and add them to the queue */
				for (i = 0; i < p_sw_data->degree; i++) {
					updn_sw_data_t *peer = peers[i];

					if (!is_colored(peer) && (peer->rank == UPDN_NO_RANK)) {
						set_color(peer, get_color(p_sw_data));
						peer->rank = p_sw_data->rank + 1;
						queue[next_queue_in++] = peer;
					}
				}
			}
		} else if (!(topo == OSM_TOPOLOGY_TYPE_DFP &&
			     (p_sw_data->color.p_color_group->potential_roots_rank != UPDN_NO_RANK ||
			      p_sw_data->color.p_color_group->is_subgroup))) {
			/*
			 * For DF+, avoid coloring forward from a node that its
			 * color group is already found to be an island
			 */
			for (i = 0; i < p_sw_data->degree; i++) {
				updn_sw_data_t *peer = peers[i];

				/* Color peers and add them to the queue */
				if (!is_colored(peer) && (peer->rank == UPDN_NO_RANK)) {
					set_color(peer, get_color(p_sw_data));
					peer->rank = p_sw_data->rank + 1;
					queue[next_queue_in++] = peer;
				} else if (p_sw_data->rank < peer->rank) {
					add_color(peer, get_color(p_sw_data));
				}
			}
		}

		if (queue_out != queue_in)
			continue;

		if (topo == OSM_TOPOLOGY_TYPE_TREE) {
			updn_mark_need_to_color(queue, queue_out, next_queue_in);
			updn_color_marked_nodes(queue, queue_out, next_queue_in);
			updn_unmark_need_to_color(queue, queue_out, next_queue_in);
		} else if (topo == OSM_TOPOLOGY_TYPE_DFP) {
			/*
			 * If a current node has larger rank peers of different
			 * colors, it is a potential root and those peers are of
			 * other islands. Hence, remove color of current node from
			 * those peers and mark its rank as root rank of its
			 * color.
			 */
			/* Step A */
			dfp_color_validate_next_rank_colors(dfp_color_groups_list,
							    queue, queue_prev_out,
							    queue_out, *p_max_rank);
			/*
			 * Merge groups if node is multicolored.
			 * Insert newly colored nodes to their color groups,
			 * and add to group's new_members list, if root rank
			 * haven't been found yet.
			 */
			dfp_insert_to_color_groups(dfp_color_groups_list,
						   color_group_arr,
						   queue, queue_out,
						   next_queue_in);

			/*
			 * Iterate over color groups, group with multiple
			 * neighbor colors is an island, whereas a group with
			 * a single neighbor color is a sub color.
			*/
			dfp_validate_color_groups(dfp_color_groups_list,
						  color_group_arr, (*p_max_rank) + 1);
		}

		queue_in = next_queue_in;
		if (queue_out < queue_in)
			queue_prev_out = queue_out;
	}
}

static boolean_t root_detection_dfp_find_potential_leaves(IN osm_subn_t * p_subn,
							  IN dfp_color_group_t ** color_group_arr,
							  IN uint32_t colorless_count,
							  OUT updn_sw_data_t ** colorless_queue,
							  OUT updn_sw_data_t ** queue,
							  OUT uint32_t * p_queue_in,
							  OUT uint32_t * p_num_of_leaves,
							  OUT cl_u64_vector_t * dfp_potential_roots)
{
	updn_sw_data_t *p_sw_data;
	updn_sw_data_t **peers;
	uint32_t colorless_queue_out = 0, colorless_queue_in = colorless_count, i;
	boolean_t is_potential_leaf, found_potential_leaves = FALSE;
	uint8_t num_neighbor_colors, neighbor_color_num;
	uint64_t neighbor_colors[UPDN_COLORS_ARRAY_SIZE];
	uint64_t *peer_color;
	uint64_t node_guid;

	while (colorless_queue_out < colorless_queue_in) {
		p_sw_data = colorless_queue[colorless_queue_out++];
		peers = p_sw_data->peers;
		num_neighbor_colors = 0;
		memset(neighbor_colors, 0, sizeof(uint64_t) * UPDN_COLORS_ARRAY_SIZE);

		is_potential_leaf = TRUE;
		for (i = 0; i < p_sw_data->degree; i++) {
			if (!is_colored_arr(peers[i]->color.colors) &&
			    !peers[i]->visited) {
				if (!peers[i]->color.in_colorless_queue) {
					colorless_queue[colorless_queue_in++] = peers[i];
					peers[i]->color.in_colorless_queue = TRUE;
				}
				/* Has unvisited colorless peers, hence not a leaf switch */
				is_potential_leaf = FALSE;
			} else if (num_neighbor_colors < DFP_MIN_ROOT_NEIGHBOR_COLORS &&
				   is_colored_arr(peers[i]->color.colors) &&
				   updn_color_get_color_num(&peers[i]->color, &neighbor_color_num)) {
				peer_color =
				    color_group_arr[neighbor_color_num]->p_own_colors_bitmask;

				/*
				 * In case there is a root disconnected from all its leaves,
				 * it will be colorless. However should avoid adding it as a
				 * potential leaf.
				 * Hence, count neighbor colors, and if connected to more than 1,
				 * it is colorless root, not a leaf
				 */
				if (updn_color_add_color_to_mask(neighbor_colors, peer_color))
					num_neighbor_colors++;
			}
		}

		if (num_neighbor_colors >= DFP_MIN_ROOT_NEIGHBOR_COLORS) {
			is_potential_leaf = FALSE;
			/* Add to potential roots */
			node_guid = cl_ntoh64(p_sw_data->p_sw->p_node->node_info.node_guid);
			cl_u64_vector_insert(dfp_potential_roots, node_guid, NULL);
		}

		if (is_potential_leaf && !p_sw_data->visited) {
			/*
			 * Add potential leaves to the main queue, which is used
			 * for coloring, starting from the leaves upwards.
			 */
			queue[(*p_queue_in)++] = p_sw_data;
			p_sw_data->rank = 0;
			(*p_num_of_leaves)++;
			found_potential_leaves = TRUE;
		}

		p_sw_data->visited = TRUE;
	}

	return found_potential_leaves;
}

static ib_api_status_t root_detection_dfp_validate_missing_leaves(osm_subn_t *p_subn,
								  updn_sw_data_t * data_per_sw,
								  updn_color_t * all_colors,
								  dfp_color_group_t ** color_group_arr,
								  cl_list_t * dfp_color_groups_list,
								  updn_sw_data_t ** queue,
								  uint32_t * p_colorless_count,
								  uint32_t * p_max_rank,
								  uint8_t topo,
								  OUT cl_u64_vector_t * dfp_potential_roots)
{
	updn_sw_data_t ** colorless_queue = NULL;
	uint32_t num_of_leaves = 0;
	uint32_t queue_in = 0;
	ib_api_status_t status = IB_SUCCESS;

	colorless_queue = (updn_sw_data_t **)calloc(cl_qmap_count(&p_subn->sw_guid_tbl),
						    sizeof(updn_sw_data_t *));
	if (!colorless_queue) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR 8F05: "
			"Cannot allocate memory for updn_sw_data Queue\n");
		status = IB_INSUFFICIENT_MEMORY;
		goto Exit;
	}

	/*
	 * In case there are islands without hosts, we may:
	 *     1. The islands with no hosts will not get colored.
	 *        Hence need to find potential leaf switches.
	 *     2. Get a single island with more than 3 levels.
	 *        Due to asymmetricity, uncolor nodes as if coloring stopped
	 *        at the point before colors were merged into one.
	 *        Then, find potential leaves and recolor as in case 1.
	 */
	*p_colorless_count =
	    dfp_rank_root_nodes(p_subn, dfp_color_groups_list, color_group_arr,
				colorless_queue, dfp_potential_roots);

	/*
	 * If found colorless nodes that are connected to multiple colors,
	 * these are potential roots, meaning there are leaf switches that
	 * were not marked as leaves (too few hosts).
	 */
	if (*p_colorless_count) {
		/* Iterate over colorless nodes and insert potential leaves to queue */
		if (!root_detection_dfp_find_potential_leaves(p_subn,
							      color_group_arr,
							      *p_colorless_count,
							      colorless_queue, queue,
							      &queue_in, &num_of_leaves,
							      dfp_potential_roots))
			goto Exit;

		/* Insert leaves according to max_cas_on_spine count */
		root_detection_insert_leafs_to_queue(p_subn, queue, data_per_sw,
						     &num_of_leaves, &queue_in, topo);
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_INFO,
			"Recoloring with updated leaves_num %u\n", num_of_leaves);

		/*
		 * Rerun coloring algorithm,
		 * this time with potential leaf switches that have no hosts
		 */
		root_detection_uncolor_sw_array(data_per_sw, p_subn);
		osm_root_detection_clear_dfp_potential_roots(dfp_potential_roots);
		root_detection_clear_dfp_color_groups(dfp_color_groups_list,
						      color_group_arr);
		(*p_max_rank) = 0;
		root_detection_color_nodes(p_subn, all_colors, color_group_arr,
					   dfp_color_groups_list, queue, queue_in,
					   p_max_rank, topo);
	}

Exit:
	free(colorless_queue);
	return status;
}

/* Find Root nodes automatically by coloring */
void osm_root_detection_find_root_nodes_by_coloring(OUT osm_opensm_t * p_osm,
						    IN osm_subn_t *p_subn,
						    IN uint8_t topo,
						    OUT unsigned * p_num_roots,
						    OUT cl_u64_vector_t * dfp_potential_roots)
{
	osm_switch_t *p_sw;
	uint32_t i, cas_num = 0;
	uint16_t lid_ho;
	uint16_t *leaf_lids_array = NULL;

	updn_sw_data_t *data_per_sw;
	updn_sw_data_t *p_sw_data;
	updn_color_t all_colors;
	uint32_t max_rank = 0;
	uint32_t num_of_leaves = 0;

	uint32_t queue_in = 0;
	updn_sw_data_t **queue = NULL;

	dfp_color_group_t **color_group_arr;
	cl_list_t dfp_color_groups_list;
	uint64_t node_guid;

	OSM_LOG_ENTER(&p_osm->log);

	lid_ho = p_subn->max_lid + 1;

	color_group_arr =
	   (dfp_color_group_t **)malloc(MAX_COLOR_GROUPS * sizeof(dfp_color_group_t *));
	if (!color_group_arr) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 8F02: "
			"Cannot allocate memory for data per color group array\n");
		goto Exit;
	}

	data_per_sw = calloc(lid_ho, sizeof(*data_per_sw));
	if (!data_per_sw) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 8F14: "
			"Cannot allocate memory for data per switch array\n");
		goto Exit;
	}

	queue = (updn_sw_data_t **)calloc(cl_qmap_count(&p_subn->sw_guid_tbl),
					  sizeof(updn_sw_data_t *));
	if (!queue) {
		free(data_per_sw);
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 8F15: "
			"Cannot allocate memory for updn_sw_data Queue\n");
		goto Exit;
	}

	root_detection_init_data_per_sw_array(data_per_sw, &cas_num, p_subn);

	root_detection_insert_leafs_to_queue(p_subn, queue, data_per_sw,
					     &num_of_leaves, &queue_in, topo);

	OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
		"Found %u CAs and RTRs, %u SWs in the subnet. leafs_num %u\n",
		cas_num, cl_qmap_count(&p_subn->sw_guid_tbl), num_of_leaves);

	/* Special case: single switch topology */
	if (cl_qmap_count(&p_subn->sw_guid_tbl) == 1) {
		p_sw = (osm_switch_t *)cl_qmap_head(&p_subn->sw_guid_tbl);
		lid_ho = cl_ntoh16(osm_node_get_base_lid(p_sw->p_node, 0));
		p_sw_data = &data_per_sw[lid_ho];
		if (topo == OSM_TOPOLOGY_TYPE_TREE)
			updn_rank_root_node(p_osm, p_sw_data, p_num_roots);
		else if (topo == OSM_TOPOLOGY_TYPE_DFP) {
			node_guid = cl_ntoh64(p_sw->p_node->node_info.node_guid);
			cl_u64_vector_insert(dfp_potential_roots, node_guid, NULL);
		}
		free(queue);
		free(data_per_sw);
		goto Exit;
	}

	if (topo == OSM_TOPOLOGY_TYPE_TREE) {
		leaf_lids_array = malloc(num_of_leaves * sizeof(uint16_t));
		if (!leaf_lids_array) {
			free(queue);
			free(data_per_sw);
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 8F16: "
				"Cannot allocate memory for leaf leads array\n");
			goto Exit;
		}

		for (i = 0; i < num_of_leaves; i++) {
			p_sw_data = queue[i];

			lid_ho = cl_ntoh16(osm_node_get_base_lid(p_sw_data->p_sw->p_node, 0));
			leaf_lids_array[i] = lid_ho;
		}
	} else if (topo == OSM_TOPOLOGY_TYPE_DFP) {
		cl_list_construct(&dfp_color_groups_list);
		cl_list_init(&dfp_color_groups_list, 10);
	}

	/* Color the node starting from leaf nodes already in queue */
	root_detection_color_nodes(p_subn, &all_colors, color_group_arr,
				   &dfp_color_groups_list, queue, queue_in,
				   &max_rank, topo);

	/* Choosing the roots and validate correctness */
	if (topo == OSM_TOPOLOGY_TYPE_TREE) {
		updn_rank_root_nodes(p_osm, p_subn, data_per_sw, max_rank, all_colors,
				     leaf_lids_array, num_of_leaves, p_num_roots);
		free(leaf_lids_array);
	} else if (topo == OSM_TOPOLOGY_TYPE_DFP) {
		uint32_t colorless_count = 0;

		if (!root_detection_dfp_validate_missing_leaves(p_subn, data_per_sw,
								&all_colors, color_group_arr,
								&dfp_color_groups_list, queue,
								&colorless_count,
								&max_rank, topo,
								dfp_potential_roots)) {

			/* Choosing the roots and validate correctness */
			if (!dfp_validate_islands(&dfp_color_groups_list)) {
				OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 8F03: "
					"DF+ islands are not valid. Failed to find "
					"roots\n");
			} else {
				if (colorless_count)
					dfp_rank_root_nodes(p_subn, &dfp_color_groups_list, color_group_arr,
							    NULL, dfp_potential_roots);
			}
		}

		destroy_dfp_color_groups_list(&dfp_color_groups_list);
	}

	free(queue);
	free(data_per_sw);
Exit:
	if (color_group_arr)
		free(color_group_arr);

	OSM_LOG_EXIT(&p_osm->log);
	return;
}
