/*
 * Copyright (c) 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <time.h>
#include <dlfcn.h>
#include <stdint.h>
#include <linux/limits.h>
#include <complib/cl_qmap.h>
#include <complib/cl_passivelock.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_SUBSCRIBER_C
#include <opensm/osm_version.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_log.h>
#include <opensm/osm_subscriber.h>

#define ERROR01 "01: Invalid type %d\n"
#define ERROR02 "02: Invalid callbacks\n"
#define ERROR03 "03: Failed to allocate memory\n"
#define ERROR04 "04: NULL subscriber\n"

#define SUBSCRIBER_ERROR(P_OSM, FORMAT, ...)				\
        do {								\
                subscriber_error(P_OSM, FORMAT, ## __VA_ARGS__);	\
        } while (0)

/* ---------------------------------------------------------------------------------------------- */
/**
 * Internal functions
 */
static void subscriber_error(IN osm_opensm_t *p_osm, IN const char *format, IN...)
{
	osm_subn_t	*p_subn = &p_osm->subn;
	osm_log_t	*p_log = &p_subn->p_osm->log;
	va_list		args;
	char		buffer[1024];

	va_start(args, format);
	vsnprintf(buffer, sizeof(buffer), format, args);
	OSM_LOG(p_log, OSM_LOG_ERROR, "ERR %02X%s\n", FILE_ID, buffer);
	va_end(args);
}

static cl_qlist_t *subscriber_type_to_qlist(IN osm_opensm_t *p_osm, IN osm_subscriber_type_t type)
{
	cl_qlist_t	*p_qlist = NULL;
	osm_subn_t	*p_subn = &p_osm->subn;

	switch (type) {
	case osm_subscriber_type_event:
        	p_qlist = &p_subn->event_subscriber_list;
		break;
	case osm_subscriber_type_routing:
        	p_qlist = &p_subn->routing_subscriber_list;
		break;
	default:
		break;
	}

	return p_qlist;
}

/* ---------------------------------------------------------------------------------------------- */

static int verify_event_callbacks(IN void *p_callbacks)
{
	osm_event_callbacks_t	*p_event_callbacks = (osm_event_callbacks_t *)p_callbacks;

	if (p_event_callbacks == NULL) {
		return 0;
	} else if (p_event_callbacks->report == NULL) {
		return 0;
	}

	return 1;
}

static int verify_routing_callbacks(IN void *p_callbacks)
{
	osm_routing_callbacks_t	*p_routing_callbacks = (osm_routing_callbacks_t *)p_callbacks;

	if (p_routing_callbacks == NULL) {
		return 0;
	}

	return 1;
}

static int (*subscriber_callback_table[osm_subscriber_type_end])(void *) = {
	[ osm_subscriber_type_event ]   = verify_event_callbacks,
	[ osm_subscriber_type_routing ] = verify_routing_callbacks,
};

static int verify_callbacks(IN osm_subscriber_type_t type, IN void *p_callbacks)
{
	return subscriber_callback_table[type](p_callbacks);
}

/* ---------------------------------------------------------------------------------------------- */
/**
 * API functions
 */
osm_subscriber_t *osm_subscriber_create(IN osm_opensm_t *p_osm,
					IN const char *name,
					IN osm_subscriber_type_t type,
					IN void *callbacks,
					IN void *context)
{
        osm_subscriber_t	*p_subscriber;
	cl_qlist_t		*p_qlist;

	/*
	 * Find the proper subscriber list.
	 */
	if ((p_qlist = subscriber_type_to_qlist(p_osm, type)) == NULL) {
		SUBSCRIBER_ERROR(p_osm, ERROR01, type);
		return FALSE;
	}

	/*
 	 * Verify the callbacks.
	 */
	if (verify_callbacks(type, callbacks) == 0) {
		SUBSCRIBER_ERROR(p_osm, ERROR02);
		return NULL;
	}

	/*
	 * Create the subscriber interface.
	 */
	if ((p_subscriber = calloc(1, sizeof(osm_subscriber_t))) == NULL) {
		SUBSCRIBER_ERROR(p_osm, ERROR03);
		return NULL;
	}

	/*
	 * Setup the subscriber struct for events.
	 */
	p_subscriber->name = strdup(name);
	p_subscriber->type = type;
	p_subscriber->context = context;
	p_subscriber->callbacks = callbacks;
	p_subscriber->qlist = p_qlist;

	cl_qlist_insert_tail(p_qlist, &p_subscriber->list_item);
	return p_subscriber;
}

void osm_subscriber_destroy(IN osm_opensm_t *p_osm,
			    IN osm_subscriber_t *p_subscriber)
{

	/*
	 * Verify input.
	 */
	if (p_subscriber == NULL) {
		SUBSCRIBER_ERROR(p_osm, ERROR04);
		return;
	}

	/*
	 * Remove subscriber from list.
	 */
	cl_qlist_remove_item(p_subscriber->qlist, &p_subscriber->list_item);

	/*
	 * Cleanup.
	 */
	free(p_subscriber->name);
	free(p_subscriber);
}
