/*
 * Copyright (c) 2021-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * Copyright (c) 2009 Sun Microsystems, Inc. All rights reserved.
 * Copyright (c) 2004-2009 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2009 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 * Copyright (c) 2009 HNR Consulting. All rights reserved.
 * Copyright (c) 2013 Oracle and/or its affiliates. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of osm_state_mgr_t.
 * This file implements the State Manager object.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <iba/ib_types.h>
#include <complib/cl_passivelock.h>
#include <complib/cl_debug.h>
#include <complib/cl_qmap.h>
#include <complib/cl_crc.h>
#include <complib/cl_vector.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_STATE_MGR_C
#include <opensm/osm_sm.h>
#include <opensm/osm_madw.h>
#include <opensm/osm_switch.h>
#include <opensm/osm_router.h>
#include <opensm/osm_log.h>
#include <opensm/osm_subnet.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_msgdef.h>
#include <opensm/osm_node.h>
#include <opensm/osm_port.h>
#include <vendor/osm_vendor_api.h>
#include <opensm/osm_inform.h>
#include <opensm/osm_inform_mgr.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_congestion_control.h>
#include <opensm/osm_db.h>
#include <opensm/osm_db_pack.h>
#include <opensm/osm_service.h>
#include <opensm/osm_guid.h>
#include <opensm/osm_unhealthy_ports.h>
#include <opensm/osm_port_group.h>
#include <opensm/osm_routing_chain.h>
#include <opensm/osm_virt_mgr.h>
#include <opensm/osm_an2an.h>
#include <opensm/osm_vendor_specific.h>
#include <opensm/osm_n2n.h>
#include <opensm/osm_fast_recovery.h>
#include <opensm/osm_router_mgr.h>
#include <opensm/osm_nvlink.h>

#define NODES_DISTANCE_VECTOR_MIN_SIZE	8
#define NODES_DISTANCE_VECTOR_GROW_SIZE	8
/*
 * Note we use COLOR_B implicitly once unvisited peer is colored
 * by opposite color of current node
 */
#define COLOR_A				0

extern void osm_drop_mgr_process(IN osm_sm_t * sm);
extern int osm_qos_init(IN osm_opensm_t * p_osm);
extern int osm_qos_setup(IN osm_opensm_t * p_osm,
			 IN boolean_t enhanced_qos_sweep_only);
extern int osm_pkey_mgr_process(IN osm_opensm_t * p_osm);
extern int osm_mcast_mgr_process(IN osm_sm_t * sm, boolean_t config_all);
extern int osm_link_mgr_process(IN osm_sm_t * sm, IN uint8_t state, boolean_t process_endports);
extern void osm_guid_mgr_process(IN osm_sm_t * sm);
extern void osm_inform_mgr_process(IN osm_sm_t * sm);
extern void osm_link_mgr_preprocess_links(IN osm_sm_t * sm);
extern int osm_set_port_state(osm_sm_t *sm, osm_physp_t *p, uint8_t const phys_state,
                          uint8_t const port_state);
extern int osm_parse_device_configuration_file(struct osm_opensm * osm_ptr);

extern boolean_t osm_link_mgr_is_init_state_expected(osm_sm_t * sm,
						     osm_physp_t * p_physp);

typedef struct osm_sw_data {
	osm_switch_t *p_sw;
	boolean_t visited;
	boolean_t color;
	uint32_t degree;
	struct osm_sw_data *peers[OSM_MAX_SW_PEERS_NUM];
} osm_sw_data_t;

static void state_mgr_up_msg(IN const osm_sm_t * sm)
{
	/*
	 * This message should be written only once - when the
	 * SM moves to Master state and the subnet is up for
	 * the first time.
	 */
	osm_log_v2(sm->p_log, sm->p_subn->first_time_master_sweep ?
		   OSM_LOG_SYS : OSM_LOG_INFO, FILE_ID, "SUBNET UP\n");

	OSM_LOG_MSG_BOX(sm->p_log, OSM_LOG_VERBOSE,
			sm->p_subn->opt.sweep_interval ?
			"SUBNET UP" : "SUBNET UP (sweep disabled)");
}

static void state_mgr_reset_node_count(IN cl_map_item_t * p_map_item,
				       IN void *context)
{
	uint8_t port_num;
	boolean_t first_discovery = *((boolean_t *) context);

	osm_node_t *p_node = (osm_node_t *) p_map_item;

	p_node->discovery_count = 0;

	/* Reset previous discovered ports in IGNORED state */
	for (port_num = 0; port_num < p_node->physp_tbl_size; port_num++) {
		/* this is the first discovery of the heavy sweep
		   Need to cleanup the need_update flag */
		if (first_discovery)
			p_node->physp_table[port_num].need_update = 0;

		if (p_node->physp_ignored[port_num] == OSM_PHYSP_IGNORE_EXISTING)
			p_node->physp_ignored[port_num] = OSM_PHYSP_IGNORE_NEW;

		p_node->remote_ignore_type[port_num] = OSM_IGNORE_PORT_NONE;
	}
	memset(p_node->physp_discovered, 0, sizeof(uint8_t) * p_node->physp_tbl_size);
}

static void state_mgr_reset_port_data(IN cl_map_item_t * p_map_item,
				      IN void *context)
{
	osm_port_t *p_port = (osm_port_t *) p_map_item;
	osm_physp_t* p_physp = osm_port_get_physp(p_port);

	p_port->discovery_count = 0;

	if (p_port->disc_through != OSM_DISC_IGNORED)
		p_port->disc_through = OSM_NOT_DISC;

	osm_physp_set_num_dr_tried(p_physp, 0);
}

static void state_mgr_reset_switch_count(IN cl_map_item_t * p_map_item,
					 IN void *context)
{
	osm_switch_t *p_sw = (osm_switch_t *) p_map_item;

	/*
	 * During discovery, need_update == 1 might be lowered to 0.
	 * need_update set to 2 means switch needs to be updated irregardless
	 */
	if ((p_sw->max_lid_ho != 0) && (p_sw->need_update != 2))
		p_sw->need_update = 1;
}

static void state_mgr_reset_switch_need_update(IN cl_map_item_t * p_map_item,
					       IN void *context)
{
	osm_switch_t *p_sw = (osm_switch_t *) p_map_item;

	p_sw->need_update = 0;
}

static void state_mgr_update_hm_no_resp_state(IN cl_fmap_item_t * obj, IN void *context)
{
	osm_unhealthy_port_t * p_unhealthy_port = (osm_unhealthy_port_t *) obj;
	osm_sm_t *sm = context;
	uint32_t mask = 0;

	OSM_ASSERT(p_unhealthy_port);

	if (p_unhealthy_port->no_resp_hist == 0)
		return;

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
		"Updating unhealthy NoResp status for 0x%016" PRIx64 " port %d\n",
		cl_ntoh64(p_unhealthy_port->peer_guid),
		p_unhealthy_port->peer_port_num);

	mask = ~(1 << (sm->p_subn->opt.hm_num_no_resp_sweeps_window + 1));
	p_unhealthy_port->no_resp_hist = (p_unhealthy_port->no_resp_hist << 1 & mask);

}

static void state_mgr_update_hm_set_err_state(IN cl_fmap_item_t * obj, IN void *context)
{
	osm_unhealthy_port_t * p_unhealthy_port = (osm_unhealthy_port_t *) obj;
	osm_sm_t *sm = context;
	uint32_t mask = 0;

	OSM_ASSERT(p_unhealthy_port);

	if (p_unhealthy_port->set_err_hist == 0)
		return;

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
		"Updating unhealthy SetErr status for 0x%016" PRIx64 " port %d\n",
		cl_ntoh64(p_unhealthy_port->peer_guid),
		p_unhealthy_port->peer_port_num);

	mask = ~(1 << (sm->p_subn->opt.hm_num_set_err_sweeps_window + 1));
	p_unhealthy_port->set_err_hist = (p_unhealthy_port->set_err_hist << 1 & mask);
}

static void state_mgr_update_hm_flapping_state(IN cl_fmap_item_t * obj, IN void *context)
{
	osm_unhealthy_port_t * p_unhealthy_port = (osm_unhealthy_port_t *) obj;
	osm_sm_t *sm = context;
	uint32_t mask = 0;

	OSM_ASSERT(p_unhealthy_port);

	if (p_unhealthy_port->flapping_hist == 0)
		return;

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
		"Updating unhealthy Flapping status for 0x%016" PRIx64 " port %d\n",
		cl_ntoh64(p_unhealthy_port->peer_guid),
		p_unhealthy_port->peer_port_num);

	mask = ~(1 << (sm->p_subn->opt.hm_num_flapping_sweeps_window + 1));
	p_unhealthy_port->flapping_hist = (p_unhealthy_port->flapping_hist << 1 & mask);
}
/* Clean up MC membership for removed pkeys */
static void osm_align_mcast_to_pkeys(IN osm_opensm_t *p_osm)
{
	osm_mgrp_t *mgrp, *next_mgrp;
	cl_qmap_t *p_tbl;
	cl_fmap_t *p_mgrp_tbl;
	osm_alias_guid_t *p_alias_guid;
	osm_mcm_alias_guid_t *mcm_alias_guid, *next_mcm_alias_guid;
	boolean_t mgrp_deleted = FALSE;
	const osm_pkey_tbl_t *pkey_tbl;
	ib_net16_t pkey;

	OSM_LOG_ENTER(&p_osm->log);

	OSM_ASSERT(p_osm);

	CL_PLOCK_EXCL_ACQUIRE(&p_osm->lock);

	p_tbl = &p_osm->subn.alias_port_guid_tbl;
	p_mgrp_tbl = &p_osm->subn.mgrp_mgid_tbl;
	next_mgrp = (osm_mgrp_t *) cl_fmap_head(p_mgrp_tbl);

	while (next_mgrp != (osm_mgrp_t *) cl_fmap_end(p_mgrp_tbl)) {
		mgrp_deleted = FALSE;
		mgrp = next_mgrp;
		next_mgrp = (osm_mgrp_t *) cl_fmap_next(&next_mgrp->map_item);
		next_mcm_alias_guid = (osm_mcm_alias_guid_t *)
				       cl_qmap_head(&mgrp->mcm_alias_port_tbl);
		while (!mgrp_deleted &&
			next_mcm_alias_guid != (osm_mcm_alias_guid_t *)
						cl_qmap_end(&mgrp->mcm_alias_port_tbl)) {

			mcm_alias_guid = next_mcm_alias_guid;
			next_mcm_alias_guid = (osm_mcm_alias_guid_t *)
				cl_qmap_next(&next_mcm_alias_guid->map_item);

			p_alias_guid = (osm_alias_guid_t *)
					cl_qmap_get(p_tbl,
						    mcm_alias_guid->alias_guid);

			if (p_alias_guid == (osm_alias_guid_t*)cl_qmap_end(p_tbl))
				continue;
			if (!(pkey_tbl = osm_alias_guid_get_pkey_tbl(p_alias_guid)))
				continue;

			pkey = mgrp->mcmember_rec.pkey;
			/* skip MCG with default pkey */
			if (ib_pkey_get_base(pkey) == IB_DEFAULT_PARTIAL_PKEY)
				continue;

			/* still shares the pkey */
			if (osm_pkey_tbl_has_pkey(&p_osm->log,
						  pkey,
						  pkey_tbl,
						  p_osm->subn.opt.allow_both_pkeys))
				continue;
			mgrp_deleted = osm_mgrp_remove_mcm_alias_guid(&p_osm->subn,
								      &p_osm->log,
								      mgrp,
								      mcm_alias_guid);
		}
	}
	CL_PLOCK_RELEASE(&p_osm->lock);
	OSM_LOG_EXIT(&p_osm->log);
}
static void state_mgr_get_sw_info(IN cl_map_item_t * p_object, IN void *context)
{
	osm_node_t *p_node;
	osm_physp_t *p_physp;
	osm_dr_path_t *p_dr_path;
	osm_madw_context_t mad_context;
	osm_switch_t *const p_sw = (osm_switch_t *) p_object;
	osm_sm_t *sm = context;
	ib_api_status_t status;

	OSM_LOG_ENTER(sm->p_log);

	p_node = p_sw->p_node;
	p_physp = osm_node_get_physp_ptr(p_node, 0);
	p_dr_path = osm_physp_get_dr_path_ptr(sm, p_physp);

	memset(&mad_context, 0, sizeof(mad_context));

	mad_context.si_context.node_guid = osm_node_get_node_guid(p_node);
	mad_context.si_context.set_method = FALSE;
	mad_context.si_context.light_sweep = TRUE;
	mad_context.si_context.lft_top_change = FALSE;

	status = osm_req_get(sm, p_dr_path, IB_MAD_ATTR_SWITCH_INFO, 0,
			     FALSE, ib_port_info_get_m_key(&p_physp->port_info),
			     0, OSM_MSG_LIGHT_SWEEP_FAIL, &mad_context);
	if (status != IB_SUCCESS)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3304: "
			"Request for SwitchInfo from 0x%" PRIx64 " failed (%s)\n",
			cl_ntoh64(osm_node_get_node_guid(p_node)),
			ib_get_err_str(status));

	OSM_LOG_EXIT(sm->p_log);
}

void state_mgr_get_rt_info(IN cl_map_item_t * p_object, IN void *context)
{
	osm_physp_t *p_physp;
	osm_madw_context_t mad_context;
	osm_router_t *const p_router = (osm_router_t *) p_object;
	osm_sm_t *sm = context;
	ib_api_status_t status;

	OSM_LOG_ENTER(sm->p_log);

	if (p_router->is_router_info_set_error) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, 
			"Skip RouterInfo GET router with port GUID 0x%" PRIx64
			" as it previous RouterInfo SET failed.\n",
			cl_ntoh64(osm_port_get_guid(p_router->p_port)));
		OSM_LOG_EXIT(sm->p_log);
		return;
	}

	p_physp = p_router->p_port->p_physp;

	memset(&mad_context, 0, sizeof(mad_context));
	mad_context.ri_context.port_guid = osm_physp_get_port_guid(p_physp);
	mad_context.ri_context.set_method = FALSE;
	mad_context.ri_context.sweep_type = OSM_SWEEP_CONTEXT_LIGHT;
	status = osm_req_get(sm, osm_physp_get_dr_path_ptr(sm, p_physp),
			     IB_MAD_ATTR_ROUTER_INFO, 0,
			     FALSE,
			     ib_port_info_get_m_key(&p_physp->port_info),
			     0, CL_DISP_MSGID_NONE, &mad_context);
	if (status != IB_SUCCESS)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3353: "
			"Request for RouterInfo from 0x%" PRIx64 " failed (%s)\n",
			cl_ntoh64(osm_physp_get_port_guid(p_physp)),
			ib_get_err_str(status));
	OSM_LOG_EXIT(sm->p_log);
}

/**********************************************************************
 Initiate a remote port info request for the given physical port
 **********************************************************************/
static void state_mgr_get_remote_port_info(IN osm_sm_t * sm,
					   IN osm_physp_t * p_physp)
{
	osm_dr_path_t *p_dr_path;
	osm_dr_path_t rem_node_dr_path;
	osm_madw_context_t mad_context;
	osm_bad_mkey_context_t bad_mkey_context;
	uint64_t heavy_sweep_count;
	size_t index;
	ib_net64_t m_key = 0;
	boolean_t find_m_key = TRUE;
	ib_api_status_t status;

	OSM_LOG_ENTER(sm->p_log);

	/* generate a dr path leaving on the physp to the remote node */
	p_dr_path = osm_physp_get_dr_path_ptr(sm, p_physp);
	memcpy(&rem_node_dr_path, p_dr_path, sizeof(osm_dr_path_t));
	if (osm_dr_path_extend(&rem_node_dr_path, osm_physp_get_port_num(p_physp))) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 332D: "
			"DR path with hop count %d couldn't be extended "
			"so skipping PortInfo query\n",
			p_dr_path->hop_count);
		goto Exit;
	}

	memset(&mad_context, 0, sizeof(mad_context));

	mad_context.pi_context.node_guid =
	    osm_node_get_node_guid(osm_physp_get_node_ptr(p_physp));
	mad_context.pi_context.port_guid = p_physp->port_guid;
	mad_context.pi_context.set_method = FALSE;
	mad_context.pi_context.light_sweep = TRUE;
	mad_context.pi_context.active_transition = FALSE;
	mad_context.pi_context.client_rereg = FALSE;
	mad_context.pi_context.port_num = osm_physp_get_port_num(p_physp);

	if (sm->p_subn->opt.m_key_per_port) {
		index = cl_vector_get_size(&sm->p_subn->bad_mkey_context_list);

		/* Update this event only per main port, as it is not port related */
		heavy_sweep_count = sm->p_subn->p_osm->stats_counters[0][HEAVY_SWEEP_COUNT];
		bad_mkey_context.m_key = cl_hton64((heavy_sweep_count << 32) | (uint32_t)index);
		bad_mkey_context.p_physp = p_physp;

		if (cl_vector_set(&sm->p_subn->bad_mkey_context_list, index,
				  &bad_mkey_context) == CL_SUCCESS) {
			m_key = bad_mkey_context.m_key;
			find_m_key = FALSE;
		} else {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 332F: "
				"Failed to set MKEY context to request for "
				"remote PortInfo from 0x%" PRIx64 "\n",
				cl_ntoh64(p_physp->port_guid));
		}
	}

	/* note that with some negative logic - if the query failed it means
	 * that there is no point in going to heavy sweep */
	status = osm_req_get(sm, &rem_node_dr_path, IB_MAD_ATTR_PORT_INFO, 0,
			     find_m_key, m_key, 0, CL_DISP_MSGID_NONE, &mad_context);

	if (status != IB_SUCCESS)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 332E: "
			"Request for remote PortInfo from 0x%" PRIx64 " failed (%s)\n",
			cl_ntoh64(p_physp->port_guid), ib_get_err_str(status));

Exit:
	OSM_LOG_EXIT(sm->p_log);
}

static boolean_t is_sm_port_valid(IN osm_sm_t * sm, uint8_t port_index)
{
	ib_port_attr_t port_attr;
	boolean_t is_sm_port_valid = FALSE;

	OSM_LOG_ENTER(sm->p_log);

	if (sm->mad_ctrl.h_bind[port_index] != OSM_BIND_INVALID_HANDLE) {
		osm_vendor_get_port_attr(sm->p_subn->p_osm->p_vendor[port_index], &port_attr);
		if (port_attr.link_state > IB_LINK_DOWN)
			is_sm_port_valid = TRUE;
	}

	OSM_LOG_EXIT(sm->p_log);
	return is_sm_port_valid;
}

/**********************************************************************
 Initiates a thorough sweep of the subnet.
 Used when there is suspicion that something on the subnet has changed.
**********************************************************************/
static ib_api_status_t state_mgr_sweep_hop_0(IN osm_sm_t * sm,
					     IN boolean_t first_discovery)
{
	ib_api_status_t status;
	osm_dr_path_t dr_path;
	osm_bind_handle_t h_bind;
	uint8_t path_array[IB_SUBNET_PATH_HOPS_MAX];

	OSM_LOG_ENTER(sm->p_log);

	memset(path_array, 0, sizeof(path_array));

	/*
	 * First, get the bind handle.
	 */
	h_bind = osm_sm_mad_ctrl_get_bind_handle(&sm->mad_ctrl);
	if (h_bind != OSM_BIND_INVALID_HANDLE) {
		OSM_LOG_MSG_BOX(sm->p_log, OSM_LOG_VERBOSE,
				"INITIATING HEAVY SWEEP");
		/*
		 * Start the sweep by clearing the port counts, then
		 * get our own NodeInfo at 0 hops.
		 */
		CL_PLOCK_ACQUIRE(sm->p_lock);

		cl_qmap_apply_func(&sm->p_subn->node_guid_tbl,
				   state_mgr_reset_node_count, &first_discovery);

		cl_qmap_apply_func(&sm->p_subn->port_guid_tbl,
				   state_mgr_reset_port_data, sm);

		cl_qmap_apply_func(&sm->p_subn->sw_guid_tbl,
				   state_mgr_reset_switch_count, sm);

		cl_fmap_apply_func(&sm->p_subn->unhealthy_ports_tbl,
				   state_mgr_update_hm_set_err_state, sm);

		cl_fmap_apply_func(&sm->p_subn->unhealthy_ports_tbl,
				   state_mgr_update_hm_flapping_state, sm);

		cl_fmap_apply_func(&sm->p_subn->unhealthy_ports_tbl,
				   state_mgr_update_hm_no_resp_state, sm);

		/* Set the in_sweep_hop_0 flag in subn to be TRUE.
		 * This will indicate the sweeping not to continue beyond the
		 * the current node.
		 * This is relevant for the case of SM on switch, since in the
		 * switch info we need to signal somehow not to continue
		 * the sweeping. */
		sm->p_subn->in_sweep_hop_0 = TRUE;

		/* Clear bad_mkey_context_list */
		cl_vector_set_size(&sm->p_subn->bad_mkey_context_list, 0);

		CL_PLOCK_RELEASE(sm->p_lock);

		osm_dr_path_init(&dr_path, 0, path_array);
		CL_PLOCK_ACQUIRE(sm->p_lock);
		status = osm_req_get(sm, &dr_path, IB_MAD_ATTR_NODE_INFO, 0,
				     TRUE, 0, 0, CL_DISP_MSGID_NONE, NULL);
		CL_PLOCK_RELEASE(sm->p_lock);
		if (status != IB_SUCCESS)
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3305: "
				"Request for NodeInfo failed (%s)\n",
				ib_get_err_str(status));
	} else {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"No bound ports. Deferring sweep...\n");
		status = IB_INVALID_STATE;
	}

	OSM_LOG_EXIT(sm->p_log);
	return status;
}

/**********************************************************************
 Clear out all existing port lid assignments
**********************************************************************/
static ib_api_status_t state_mgr_clean_known_lids(IN osm_sm_t * sm)
{
	ib_api_status_t status = IB_SUCCESS;
	cl_ptr_vector_t *p_vec = &(sm->p_subn->port_lid_tbl);
	uint32_t i;

	OSM_LOG_ENTER(sm->p_log);

	/* we need a lock here! */
	CL_PLOCK_ACQUIRE(sm->p_lock);

	for (i = 0; i < cl_ptr_vector_get_size(p_vec); i++)
		cl_ptr_vector_set(p_vec, i, NULL);

	CL_PLOCK_RELEASE(sm->p_lock);

	OSM_LOG_EXIT(sm->p_log);
	return status;
}

/**********************************************************************
 Clear out all virtual ports
**********************************************************************/
static void stat_mgr_clear_virtual_port(IN cl_map_item_t * p_map_item,
					IN void *context)
{
	osm_port_t *p_port = (osm_port_t *)p_map_item;
	osm_sm_t *sm = context;

	osm_port_delete_virtualization(sm, p_port);
}

static ib_api_status_t state_mgr_clear_virtual_ports(IN osm_sm_t * sm)
{
	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);
	cl_qmap_apply_func(&sm->p_subn->port_guid_tbl,
			   stat_mgr_clear_virtual_port, sm);
	virt_mgr_handle_empty_vnodes(sm, FALSE, TRUE);
	CL_PLOCK_RELEASE(sm->p_lock);

	return IB_SUCCESS;
}

/**********************************************************************
 Clear out non-manual unhealthy ports
**********************************************************************/
static ib_api_status_t state_mgr_clean_non_manual_unhealthy_ports(IN osm_sm_t * sm)
{
	cl_fmap_item_t *p_item, *p_item_next;
	cl_fmap_t *unhealthy_ports_tbl = &sm->p_subn->unhealthy_ports_tbl;
	osm_unhealthy_port_t *p_unhealthy_port;

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);
	p_item_next = cl_fmap_head(unhealthy_ports_tbl);
	while (p_item_next != cl_fmap_end(unhealthy_ports_tbl)) {
		p_item = p_item_next;
		p_item_next = cl_fmap_next(p_item);

		p_unhealthy_port = (osm_unhealthy_port_t *) p_item;
		if (!(p_unhealthy_port->cond_mask & (1 << OSM_HM_COND_MANUAL))) {
			cl_fmap_remove_item(unhealthy_ports_tbl, p_item);
			osm_unhealthy_port_remove(sm, &p_unhealthy_port);
		}
	}

	CL_PLOCK_RELEASE(sm->p_lock);
	return IB_SUCCESS;
}

/**********************************************************************
 Clear out all revocered unhealthy ports
**********************************************************************/
static void state_mgr_clean_recovered_unhealthy_ports(IN osm_sm_t * sm)
{
	cl_fmap_item_t *p_item, *p_item_next;
	cl_fmap_t *unhealthy_ports_tbl = &sm->p_subn->unhealthy_ports_tbl;
	osm_unhealthy_port_t *p_unhealthy_port;

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);
	p_item_next = cl_fmap_head(unhealthy_ports_tbl);
	while(p_item_next != cl_fmap_end(unhealthy_ports_tbl)) {
		p_item = p_item_next;
		p_item_next = cl_fmap_next(p_item);

		p_unhealthy_port = (osm_unhealthy_port_t *) p_item;
		if (osm_unhealthy_port_is_healthy(sm, p_unhealthy_port)) {
			cl_fmap_remove_item(unhealthy_ports_tbl, p_item);
			osm_unhealthy_port_remove(sm, &p_unhealthy_port);
		}

	}

	CL_PLOCK_RELEASE(sm->p_lock);
}

/**********************************************************************
 Clear SA cache
**********************************************************************/
static ib_api_status_t state_mgr_sa_clean(IN osm_sm_t * sm)
{
	ib_api_status_t status = IB_SUCCESS;
	cl_qmap_t *p_port_guid_tbl;
	osm_assigned_guids_t *p_assigned_guids, *p_next_assigned_guids;
	osm_alias_guid_t *p_alias_guid, *p_next_alias_guid;
	osm_mcm_port_t *mcm_port;
	osm_subn_t * p_subn;
	osm_port_t *p_port;
	osm_infr_t *p_infr;
	osm_svcr_t *p_svcr;

	OSM_LOG_ENTER(sm->p_log);

	p_subn = sm->p_subn;

	/* Flush any queued traps and reports */
	osm_flush_traps(p_subn);
	osm_flush_reports(p_subn);

	/* we need a lock here! */
	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);

	if (p_subn->opt.drop_event_subscriptions) {
		/* Clean InformInfo records */
		p_infr = (osm_infr_t *) cl_qlist_remove_head(&p_subn->sa_infr_list);
		while (p_infr !=
		       (osm_infr_t *) cl_qlist_end(&p_subn->sa_infr_list)) {
			osm_infr_delete(p_infr);
			p_infr = (osm_infr_t *) cl_qlist_remove_head(&p_subn->sa_infr_list);
		}

		/* For now, treat Service Records in same category as InformInfos */
		/* Clean Service records */
		p_svcr = (osm_svcr_t *) cl_qlist_remove_head(&p_subn->sa_sr_list);
		while (p_svcr !=
		       (osm_svcr_t *) cl_qlist_end(&p_subn->sa_sr_list)) {
			osm_svcr_delete(p_svcr);
			p_svcr = (osm_svcr_t *) cl_qlist_remove_head(&p_subn->sa_sr_list);
		}
	}

	/* Clean Multicast member list on each port */
	p_port_guid_tbl = &p_subn->port_guid_tbl;
	for (p_port = (osm_port_t *) cl_qmap_head(p_port_guid_tbl);
	     p_port != (osm_port_t *) cl_qmap_end(p_port_guid_tbl);
	     p_port = (osm_port_t *) cl_qmap_next(&p_port->map_item)) {
		while (!cl_is_qlist_empty(&p_port->mcm_list)) {
			mcm_port = cl_item_obj(cl_qlist_head(&p_port->mcm_list),
					       mcm_port, list_item);
			osm_mgrp_delete_port(p_subn, sm->p_log, mcm_port->mgrp,
					     p_port);
		}
		/* Hack - clean alias guid table from physp */
		free(p_port->p_physp->p_guids);
		p_port->p_physp->p_guids = NULL;
	}

	/* Clean Alias Guid work objects */
	while (cl_qlist_count(&p_subn->alias_guid_list))
		osm_guid_work_obj_delete((osm_guidinfo_work_obj_t *)
			cl_qlist_remove_head(&p_subn->alias_guid_list));

	/* Clean Assigned GUIDs table */
	p_next_assigned_guids = (osm_assigned_guids_t *)
				cl_qmap_head(&p_subn->assigned_guids_tbl);
	while (p_next_assigned_guids !=
	       (osm_assigned_guids_t *) cl_qmap_end(&p_subn->assigned_guids_tbl)) {
		p_assigned_guids = p_next_assigned_guids;
		p_next_assigned_guids = (osm_assigned_guids_t *)
					cl_qmap_next(&p_assigned_guids->map_item);
		cl_qmap_remove_item(&p_subn->assigned_guids_tbl,
				    &p_assigned_guids->map_item);
		osm_assigned_guids_delete(&p_assigned_guids);
        }

	/* Clean Alias GUIDs table */
	p_next_alias_guid = (osm_alias_guid_t *)
			    cl_qmap_head(&p_subn->alias_port_guid_tbl);
	while (p_next_alias_guid !=
	       (osm_alias_guid_t *) cl_qmap_end(&p_subn->alias_port_guid_tbl)) {
		p_alias_guid = p_next_alias_guid;
		p_next_alias_guid = (osm_alias_guid_t *)
				    cl_qmap_next(&p_alias_guid->map_item);
		if (osm_alias_guid_get_alias_guid(p_alias_guid) !=
		    osm_alias_guid_get_base_guid(p_alias_guid)) {
			/* Clean if it's not base port GUID */
			cl_qmap_remove_item(&p_subn->alias_port_guid_tbl,
					    &p_alias_guid->map_item);
			osm_alias_guid_delete(&p_alias_guid);
		}
	}

	p_subn->p_osm->sa.dirty = TRUE;

	CL_PLOCK_RELEASE(sm->p_lock);

	/* Flush any queued traps and reports due to cleaning SA subscriptions */
	osm_flush_traps(p_subn);
	osm_flush_reports(p_subn);

	OSM_LOG_EXIT(sm->p_log);
	return status;
}

/**********************************************************************
 Clear out all vports from virt manager queue
**********************************************************************/
static ib_api_status_t state_mgr_clear_virt_mgr_queue(IN osm_sm_t * sm)
{
	OSM_LOG_ENTER(sm->p_log);

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);
	osm_virt_mgr_empty_queue(&sm->virt_mgr);
	CL_PLOCK_RELEASE(sm->p_lock);

	OSM_LOG_EXIT(sm->p_log);
	return IB_SUCCESS;;
}

/**********************************************************************
 Notifies the transport layer that the local LID has changed,
 which give it a chance to update address vectors, etc..
**********************************************************************/
static ib_api_status_t state_mgr_notify_lid_change(IN osm_sm_t * sm)
{
	ib_api_status_t status;
	osm_bind_handle_t h_bind;

	OSM_LOG_ENTER(sm->p_log);

	/*
	 * First, get the bind handle.
	 */
	h_bind = osm_sm_mad_ctrl_get_bind_handle(&sm->mad_ctrl);
	if (h_bind == OSM_BIND_INVALID_HANDLE) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3306: "
			"No bound ports\n");
		status = IB_ERROR;
		goto Exit;
	}

	/*
	 * Notify the transport layer that we changed the local LID.
	 */
	status = osm_vendor_local_lid_change(h_bind);
	if (status != IB_SUCCESS)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3307: "
			"Vendor LID update failed (%s)\n",
			ib_get_err_str(status));

Exit:
	OSM_LOG_EXIT(sm->p_log);
	return status;
}

/**********************************************************************
 This function clears sm port when it is down
**********************************************************************/
static void state_mgr_clear_sm_port_when_down(IN osm_sm_t * sm)
{
	ib_net64_t port_guid;
	osm_port_t *p_port;

	OSM_LOG_ENTER(sm->p_log);

	port_guid = sm->p_subn->sm_port_guid;

	/*
	 * If we don't know our own port guid yet, assume the port is down.
	 */
	if (port_guid == 0) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 331F: "
			"SM port GUID unknown\n");
		return;
	}

	CL_PLOCK_ACQUIRE(sm->p_lock);
	p_port = osm_get_port_by_guid(sm->p_subn, port_guid);
	if (!p_port) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3320: "
			"SM port with GUID:%016" PRIx64 " is unknown\n",
			cl_ntoh64(port_guid));
		CL_PLOCK_RELEASE(sm->p_lock);
		return;
	}

	osm_port_delete_virtualization(sm, p_port);
	p_port->p_node->gi_updated = FALSE;
	p_port->p_node->ext_node_info_updated = FALSE;
	if (osm_node_get_type(p_port->p_node) == IB_NODE_TYPE_SWITCH &&
	    p_port->p_node->sw)
		osm_ar_invalidate_ar_cache_params(p_port->p_node->sw);
	CL_PLOCK_RELEASE(sm->p_lock);
}

/**********************************************************************
 Returns true if the SM port is down.
 The SM's port object must exist in the port_guid table.
**********************************************************************/
static boolean_t state_mgr_is_sm_port_down(IN osm_sm_t * sm)
{
	ib_net64_t port_guid;
	osm_port_t *p_port;
	uint8_t state;

	OSM_LOG_ENTER(sm->p_log);

	port_guid = sm->p_subn->sm_port_guid;

	/*
	 * If we don't know our own port guid yet, assume the port is down.
	 */
	if (port_guid == 0) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3308: "
			"SM port GUID unknown\n");
		state = IB_LINK_DOWN;
		goto Exit;
	}

	CL_PLOCK_ACQUIRE(sm->p_lock);
	p_port = osm_get_port_by_guid(sm->p_subn, port_guid);
	if (!p_port) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3309: "
			"SM port with GUID:%016" PRIx64 " is unknown\n",
			cl_ntoh64(port_guid));
		state = IB_LINK_DOWN;
		CL_PLOCK_RELEASE(sm->p_lock);
		goto Exit;
	}

	if (p_port->p_node->sw &&
	    !ib_switch_info_is_enhanced_port0(&p_port->p_node->sw->switch_info))
		state = IB_LINK_ACTIVE;	/* base SP0 */
	else
		state = osm_physp_get_port_state(p_port->p_physp);

	if (!p_port->discovery_count) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR 330A: Failed to discover SM port\n");
		state = IB_LINK_DOWN;
	}

	CL_PLOCK_RELEASE(sm->p_lock);

Exit:
	OSM_LOG_EXIT(sm->p_log);
	return (state == IB_LINK_DOWN);
}

static void state_clear_switch_mft(cl_map_item_t * const p_map_item,
				   void *context)
{
	osm_switch_t *p_sw = (osm_switch_t *)p_map_item;
	osm_mcast_tbl_reset_mask(&p_sw->mcast_tbl);
}

/**********************************************************************
 Sweeps the node 1 hop away.
 This sets off a "chain reaction" that causes discovery of the subnet.
 Used when there is suspicion that something on the subnet has changed.
**********************************************************************/
static ib_api_status_t state_mgr_sweep_hop_1(IN osm_sm_t * sm)
{
	ib_api_status_t status = IB_SUCCESS;
	osm_madw_context_t context;
	osm_node_t *p_node;
	osm_port_t *p_port;
	osm_dr_path_t hop_1_path;
	ib_net64_t port_guid;
	uint8_t port_num;
	uint8_t path_array[IB_SUBNET_PATH_HOPS_MAX];
	uint8_t num_ports;
	osm_physp_t *p_ext_physp;

	OSM_LOG_ENTER(sm->p_log);

	/*
	 * First, get our own port and node objects.
	 */
	port_guid = sm->p_subn->sm_port_guid;

	OSM_ASSERT(port_guid);

	/* Set the in_sweep_hop_0 flag in subn to be FALSE.
	 * This will indicate the sweeping to continue beyond the
	 * the current node.
	 * This is relevant for the case of SM on switch, since in the
	 * switch info we need to signal that the sweeping should
	 * continue through the switch. */
	sm->p_subn->in_sweep_hop_0 = FALSE;

	p_port = osm_get_port_by_guid(sm->p_subn, port_guid);
	if (!p_port) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3310: "
			"No SM port object\n");
		status = IB_ERROR;
		goto Exit;
	}

	p_node = p_port->p_node;
	OSM_ASSERT(p_node);

	port_num = ib_node_info_get_local_port_num(&p_node->node_info);

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
		"Probing hop 1 on local port %u\n", port_num);

	memset(path_array, 0, sizeof(path_array));
	/* the hop_1 operations depend on the type of our node.
	 * Currently - legal nodes that can host SM are SW and CA */
	switch (osm_node_get_type(p_node)) {
	case IB_NODE_TYPE_CA:
	case IB_NODE_TYPE_ROUTER:
		memset(&context, 0, sizeof(context));
		context.ni_context.node_guid = osm_node_get_node_guid(p_node);
		context.ni_context.port_num = port_num;
		context.ni_context.node_type = osm_node_get_type(p_node);
		context.ni_context.disc_through = OSM_DISC_NORMAL;

		path_array[1] = port_num;

		osm_dr_path_init(&hop_1_path, 1, path_array);
		CL_PLOCK_ACQUIRE(sm->p_lock);
		status = osm_req_get(sm, &hop_1_path, IB_MAD_ATTR_NODE_INFO, 0,
				     TRUE, 0, 0, CL_DISP_MSGID_NONE, &context);
		CL_PLOCK_RELEASE(sm->p_lock);
		if (status != IB_SUCCESS)
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3311: "
				"Request for NodeInfo failed (%s)\n",
				ib_get_err_str(status));
		break;

	case IB_NODE_TYPE_SWITCH:
		/* Need to go over all the ports of the switch, and send a
		 * node_info from them. This doesn't include the port 0 of the
		 * switch, which hosts the SM.
		 * Note: We'll send another switchInfo on port 0, since if no
		 * ports are connected, we still want to get some response, and
		 * have the subnet come up.
		 */
		num_ports = osm_node_get_num_physp(p_node);
		for (port_num = 1; port_num < num_ports; port_num++) {
			if (lookup_isolated_port(sm, p_node, port_num))
				continue;
			/* go through the port only if the port is not DOWN */
			p_ext_physp = osm_node_get_physp_ptr(p_node, port_num);
			if (p_ext_physp && ib_port_info_get_port_state
			    (&(p_ext_physp->port_info)) > IB_LINK_DOWN) {
				memset(&context, 0, sizeof(context));
				context.ni_context.node_guid =
				    osm_node_get_node_guid(p_node);
				context.ni_context.port_num = port_num;
				context.ni_context.disc_through =
				    OSM_DISC_NORMAL;
				path_array[1] = port_num;
				osm_dr_path_init(&hop_1_path, 1, path_array);
				CL_PLOCK_ACQUIRE(sm->p_lock);
				status = osm_req_get(sm, &hop_1_path,
						     IB_MAD_ATTR_NODE_INFO, 0,
						     TRUE, 0, 0,
						     CL_DISP_MSGID_NONE,
						     &context);
				CL_PLOCK_RELEASE(sm->p_lock);
				if (status != IB_SUCCESS)
					OSM_LOG(sm->p_log, OSM_LOG_ERROR,
						"ERR 3312: "
						"Request for NodeInfo failed (%s)\n",
						ib_get_err_str(status));
			}
		}
		break;

	default:
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR 3313: Unknown node type %d (%s)\n",
			osm_node_get_type(p_node), p_node->print_desc);
	}

Exit:
	OSM_LOG_EXIT(sm->p_log);
	return status;
}

static ib_api_status_t state_mgr_send_pi_to_ignored_port(osm_sm_t * sm,
							 osm_madw_context_t context,
							 osm_port_t *p_port)
{
	ib_api_status_t status;
	osm_node_t *p_node = p_port->p_node;
	osm_physp_t *p_physp = p_port->p_physp;

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
		"Sending GetPortInfo to ignored port 0x%" PRIx64 "\n",
		cl_ntoh64(osm_port_get_guid(p_port)));

	memset(&context, 0, sizeof(context));
	context.pi_context.node_guid = osm_node_get_node_guid(p_node);
	context.pi_context.port_guid = osm_physp_get_port_guid(p_physp);
	context.pi_context.set_method = FALSE;
	context.pi_context.light_sweep = FALSE;
	context.pi_context.active_transition = FALSE;

	status = osm_req_get_pi(sm, p_physp, p_physp->port_num, TRUE, 0, &context, FALSE);
	if (status != IB_SUCCESS)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3334: "
			"Failure initiating PortInfo request (%s)\n",
			ib_get_err_str(status));
	return status;
}

static ib_api_status_t state_mgr_sweep_ignored_iter(IN osm_sm_t * sm)
{
	ib_api_status_t status = IB_SUCCESS;
	osm_madw_context_t context;
	osm_node_t *p_node;
	osm_port_t *p_port;
	osm_physp_t *p_physp;
	unsigned port;
	ib_net64_t guid;
	osm_dr_path_t *p_path;

	OSM_LOG_ENTER(sm->p_log);

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);
	while (cl_list_count(&sm->p_subn->ignored_nodes_list)) {
		p_port = (osm_port_t *) cl_list_remove_head(&sm->p_subn->ignored_nodes_list);

		p_node = p_port->p_node;
		p_physp = p_port->p_physp;
		p_path = osm_physp_get_dr_path_ptr(sm, p_physp);
		/* If port was discovered only through ignored port:
		 *     If Switch:
		 *         - Send Get(SwitchInfo)
		 *     If HCA/RTR:
		 *         - Send Get(PortInfo)
		 * Else if a switch:
		 *     - Go over all ports and for the ignored
		 *       ports send Get(PortInfo)
		 */
		if (p_port->disc_through == OSM_DISC_IGNORED) {
			if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH) {
				OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
					"Sending GetSwitchInfo to ignored "
					"switch 0x%" PRIx64 "\n",
					cl_ntoh64(osm_node_get_node_guid(p_node)));
				context.si_context.node_guid =
				    osm_node_get_node_guid(p_node);
				context.si_context.set_method = FALSE;
				context.si_context.light_sweep = FALSE;
				context.si_context.lft_top_change = FALSE;

				status = osm_req_get(sm, p_path,
				    IB_MAD_ATTR_SWITCH_INFO, 0, TRUE, 0,
				    0, CL_DISP_MSGID_NONE, &context);
					if (status != IB_SUCCESS)
					/* continue despite error */
					OSM_LOG(sm->p_log, OSM_LOG_ERROR,
						"ERR 3332: Failure initiating"
						" SwitchInfo request (%s)\n",
						ib_get_err_str(status));
				continue;
			} else {
				if (lookup_no_discover_port(sm, p_node,
							    p_physp->port_num)) {
					guid = osm_node_get_node_guid(p_node);
					OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
						"Skipping PI for no_discover "
						"port: guid 0x%" PRIx64
						", port %d\n", cl_ntoh64(guid),
						p_physp->port_num);
					continue;
				}
				status = state_mgr_send_pi_to_ignored_port(sm, context, p_port);
				continue;
			}

		} else if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH) {
			memset(&context, 0, sizeof(context));
			context.pi_context.node_guid =
			    osm_node_get_node_guid(p_node);
			context.pi_context.port_guid =
			    osm_physp_get_port_guid(p_physp);
			context.pi_context.set_method = FALSE;
			context.pi_context.light_sweep = FALSE;
			context.pi_context.active_transition = FALSE;

			for (port = 1; port < p_node->physp_tbl_size;
			     port++) {
				if (p_node->physp_ignored[port] == OSM_PHYSP_NORMAL)
					continue;
				if (lookup_no_discover_port(sm, p_node, port)) {
					OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
						"Skipping PI for no_discover "
						"sw port: guid 0x%" PRIx64
						", port %d\n",
						cl_ntoh64(context.pi_context.node_guid),
						port);
					continue;
				}
				OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
					"Sending GetPortInfo to ignored "
					"switch 0x%" PRIx64 ", port %u\n",
					cl_ntoh64(osm_node_get_node_guid(p_node)),
					port);

				status = osm_req_get_pi(sm, p_physp, port, TRUE,
							0, &context, TRUE);
				if (status != IB_SUCCESS)
					OSM_LOG(sm->p_log, OSM_LOG_ERROR,
						"ERR 3334: Failure initiating"
						" PortInfo request (%s)\n",
						ib_get_err_str(status));
			}
		} else {
			if (lookup_no_discover_port(sm, p_node,
						    p_physp->port_num)) {
				guid = osm_node_get_node_guid(p_node);
				OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
					"Skipping PI for no_discover "
					"port: guid 0x%" PRIx64
					", port %d\n", cl_ntoh64(guid),
					p_physp->port_num);
				continue;
			}
			status = state_mgr_send_pi_to_ignored_port(sm, context, p_port);
			continue;
		}
	}
	CL_PLOCK_RELEASE(sm->p_lock);
	OSM_LOG_EXIT(sm->p_log);

	return status;
}

static void query_sm_info(cl_map_item_t * item, void *cxt)
{
	osm_madw_context_t context;
	osm_remote_sm_t *r_sm = cl_item_obj(item, r_sm, map_item);
	osm_sm_t *sm = cxt;
	ib_api_status_t ret;
	osm_port_t *p_port;

	p_port= osm_get_port_by_guid(sm->p_subn, r_sm->smi.guid);
	if (p_port == NULL) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3340: "
			"No port object on given sm object\n");
		return;
        }

	context.smi_context.port_guid = r_sm->smi.guid;
	context.smi_context.set_method = FALSE;
	context.smi_context.light_sweep = TRUE;

	ret = osm_req_get(sm, osm_physp_get_dr_path_ptr(sm, p_port->p_physp),
			  IB_MAD_ATTR_SM_INFO, 0, FALSE,
			  ib_port_info_get_m_key(&p_port->p_physp->port_info),
			  0, CL_DISP_MSGID_NONE, &context);
	if (ret != IB_SUCCESS)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3314: "
			"Failure requesting SMInfo (%s)\n",
			ib_get_err_str(ret));
}

static void state_mgr_reset_state_change_bit(IN cl_map_item_t * obj,
					     IN void *context)
{
	osm_madw_context_t mad_context;
	osm_switch_t *p_sw = (osm_switch_t *) obj;
	osm_sm_t *sm = context;
	osm_node_t *p_node;
	osm_physp_t *p_physp;
	osm_dr_path_t *p_path;
	ib_api_status_t status;
	ib_switch_info_t si;

	OSM_LOG_ENTER(sm->p_log);

	OSM_ASSERT(p_sw);

	p_node = p_sw->p_node;

	OSM_ASSERT(p_node);

	p_physp = osm_node_get_physp_ptr(p_node, 0);
	p_path = osm_physp_get_dr_path_ptr(sm, p_physp);

	if (!ib_switch_info_get_state_change(&p_sw->switch_info))
		goto exit;

	si = p_sw->switch_info;

	ib_switch_info_set_state_change(&si);

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
		"Resetting PortStateChange on switch GUID 0x%016" PRIx64 "\n",
		cl_ntoh64(osm_node_get_node_guid(p_node)));

	mad_context.si_context.light_sweep = FALSE;
	mad_context.si_context.node_guid = osm_node_get_node_guid(p_node);
	mad_context.si_context.set_method = TRUE;
	mad_context.si_context.lft_top_change = FALSE;

	status = osm_req_set(sm, p_path, p_physp, (uint8_t *) &si,
			     sizeof(si), IB_MAD_ATTR_SWITCH_INFO,
			     0, FALSE, ib_port_info_get_m_key(&p_physp->port_info),
			     0, CL_DISP_MSGID_NONE, &mad_context);

	if (status != IB_SUCCESS)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 332A: "
			"Sending SwitchInfo attribute for 0x%" PRIx64 " failed (%s)\n",
			cl_ntoh64(osm_node_get_node_guid(p_node)),
			ib_get_err_str(status));

exit:
	OSM_LOG_EXIT(sm->p_log);
}

static void state_mgr_update_node_desc(IN cl_map_item_t * obj, IN void *context)
{
	osm_madw_context_t mad_context;
	osm_node_t *p_node = (osm_node_t *) obj;
	osm_sm_t *sm = context;
	osm_physp_t *p_physp = NULL;
	unsigned i, num_ports;
	ib_api_status_t status;

	OSM_LOG_ENTER(sm->p_log);

	OSM_ASSERT(p_node);

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
		"Updating NodeDesc for 0x%016" PRIx64 "\n",
		cl_ntoh64(osm_node_get_node_guid(p_node)));

	/* get a physp to request from. */
	num_ports = osm_node_get_num_physp(p_node);
	for (i = 0; i < num_ports; i++)
		if ((p_physp = osm_node_get_physp_ptr(p_node, i)))
			break;

	if (!p_physp) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 331C: "
			"Failed to find any valid physical port object.\n");
		goto exit;
	}

	mad_context.nd_context.node_guid = osm_node_get_node_guid(p_node);

	status = osm_req_get(sm, osm_physp_get_dr_path_ptr(sm, p_physp),
			     IB_MAD_ATTR_NODE_DESC, 0, FALSE,
			     ib_port_info_get_m_key(&p_physp->port_info),
			     0, CL_DISP_MSGID_NONE, &mad_context);
	if (status != IB_SUCCESS)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR 331B: Failure initiating NodeDescription request "
			"(%s) to 0x%016" PRIx64 "\n", ib_get_err_str(status),
			cl_ntoh64(osm_node_get_node_guid(p_node)));

exit:
	OSM_LOG_EXIT(sm->p_log);
}

void osm_reset_switch_state_change_bit(IN osm_opensm_t *osm)
{
	CL_PLOCK_ACQUIRE(&osm->lock);
	cl_qmap_apply_func(&osm->subn.sw_guid_tbl, state_mgr_reset_state_change_bit,
			   &osm->sm);
	CL_PLOCK_RELEASE(&osm->lock);
}

void osm_update_node_desc(IN osm_opensm_t *osm)
{
	CL_PLOCK_ACQUIRE(&osm->lock);
	cl_qmap_apply_func(&osm->subn.node_guid_tbl, state_mgr_update_node_desc,
			   &osm->sm);
	CL_PLOCK_RELEASE(&osm->lock);
}

/**********************************************************************
 During a light sweep, check each node to see if the node description
 is valid and if not issue a ND query.
**********************************************************************/
static void state_mgr_get_node_desc(IN cl_map_item_t * obj, IN void *context)
{
	osm_node_t *p_node = (osm_node_t *) obj;
	osm_sm_t *sm = context;

	OSM_LOG_ENTER(sm->p_log);

	OSM_ASSERT(p_node);

	if (p_node->print_desc
	    && strcmp(p_node->print_desc, OSM_NODE_DESC_UNKNOWN))
		/* if ND is valid, do nothing */
		goto exit;

	OSM_LOG(sm->p_log, OSM_LOG_ERROR,
		"ERR 3319: Unknown node description for node GUID "
		"0x%016" PRIx64 ".  Reissuing ND query\n",
		cl_ntoh64(osm_node_get_node_guid(p_node)));

	state_mgr_update_node_desc(obj, context);

exit:
	OSM_LOG_EXIT(sm->p_log);
}

/**********************************************************************
 Initiates a lightweight sweep of the subnet.
 Used during normal sweeps after the subnet is up.
**********************************************************************/
static ib_api_status_t state_mgr_light_sweep_start(IN osm_sm_t * sm)
{
	ib_api_status_t status = IB_SUCCESS;
	osm_bind_handle_t h_bind;
	cl_qmap_t *p_sw_tbl;
	cl_map_item_t *p_next;
	osm_node_t *p_node;
	osm_physp_t *p_physp;
	uint8_t port_num;
	boolean_t discover_remote, discover_port;
	boolean_t topoconfig_discover_port;
	osm_neighbors_info_t *p_neighbors_info = NULL;

	OSM_LOG_ENTER(sm->p_log);

	p_sw_tbl = &sm->p_subn->sw_guid_tbl;

	/*
	 * First, get the bind handle.
	 */
	h_bind = osm_sm_mad_ctrl_get_bind_handle(&sm->mad_ctrl);
	if (h_bind == OSM_BIND_INVALID_HANDLE) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"No bound ports. Deferring sweep...\n");
		status = IB_INVALID_STATE;
		goto _exit;
	}

	OSM_LOG_MSG_BOX(sm->p_log, OSM_LOG_VERBOSE, "INITIATING LIGHT SWEEP");
	CL_PLOCK_ACQUIRE(sm->p_lock);
	cl_qmap_apply_func(p_sw_tbl, state_mgr_get_sw_info, sm);
	CL_PLOCK_RELEASE(sm->p_lock);

	CL_PLOCK_ACQUIRE(sm->p_lock);
	sm->router_mgr.ri_routers_changes_bitmask = 0;
	cl_qmap_apply_func(&sm->p_subn->rtr_guid_tbl, state_mgr_get_rt_info, sm);
	CL_PLOCK_RELEASE(sm->p_lock);

	CL_PLOCK_ACQUIRE(sm->p_lock);
	cl_qmap_apply_func(&sm->p_subn->node_guid_tbl, state_mgr_get_node_desc,
			   sm);
	CL_PLOCK_RELEASE(sm->p_lock);

	CL_PLOCK_ACQUIRE(sm->p_lock);
	cl_fmap_apply_func(&sm->p_subn->unhealthy_ports_tbl,
			    state_mgr_update_hm_no_resp_state, sm);
	CL_PLOCK_RELEASE(sm->p_lock);

	/* now scan the list of physical ports that were not down but have no remote port */
	CL_PLOCK_ACQUIRE(sm->p_lock);

	/* Clear bad_mkey_context_list */
	cl_vector_set_size(&sm->p_subn->bad_mkey_context_list, 0);

	p_next = cl_qmap_head(&sm->p_subn->node_guid_tbl);
	while (p_next != cl_qmap_end(&sm->p_subn->node_guid_tbl)) {
		p_node = (osm_node_t *) p_next;
		p_next = cl_qmap_next(p_next);

		p_neighbors_info = NULL;
		if (sm->p_subn->topo_config_enabled)
			p_neighbors_info = osm_topoconfig_lookup(sm->p_subn->p_topoconfig_hash_tbl,
								 osm_node_get_node_guid(p_node));

		for (port_num = 1; port_num < osm_node_get_num_physp(p_node);
		     port_num++) {
			p_physp = osm_node_get_physp_ptr(p_node, port_num);

			discover_remote =
			    (p_node->remote_ignore_type[port_num] != OSM_IGNORE_PORT_NO_DISCOVER &&
			     p_node->remote_ignore_type[port_num] != OSM_IGNORE_PORT_TOPOCONFIG_MISMATCH);


			discover_port = !lookup_no_discover_port(sm, p_node,
								port_num);
			topoconfig_discover_port = TRUE;
			if (p_neighbors_info && p_neighbors_info->neighbors[port_num].state != OSM_TOPOCONFIG_ACTIVE)
				topoconfig_discover_port = FALSE;

			if (p_physp && (osm_physp_get_port_state(p_physp) !=
					IB_LINK_DOWN)
			    && !osm_physp_get_remote(p_physp) && discover_remote
			    && discover_port && topoconfig_discover_port) {
				OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3315: "
					"Unknown remote side for node 0x%016"
					PRIx64
					" (%s) port %u. Adding to light sweep sampling list\n",
					cl_ntoh64(osm_node_get_node_guid
						  (p_node)),
					p_node->print_desc, port_num);

				osm_dump_dr_path_v2(sm->p_log,
						    osm_physp_get_dr_path_ptr
						    (sm, p_physp), FILE_ID, OSM_LOG_ERROR);

				state_mgr_get_remote_port_info(sm, p_physp);
			}
		}
	}

	cl_qmap_apply_func(&sm->p_subn->sm_guid_tbl, query_sm_info, sm);

	CL_PLOCK_RELEASE(sm->p_lock);

_exit:
	OSM_LOG_EXIT(sm->p_log);
	return status;
}

/**********************************************************************
 * Go over all the remote SMs (as updated in the sm_guid_tbl).
 * Find if there is a remote sm that is a master SM.
 * If there is a remote master SM - return a pointer to it,
 * else - return NULL.
 **********************************************************************/
static osm_remote_sm_t *state_mgr_exists_other_master_sm(IN osm_sm_t * sm)
{
	cl_qmap_t *p_sm_tbl;
	osm_remote_sm_t *p_sm;
	osm_remote_sm_t *p_sm_res = NULL;
	osm_node_t *p_node;
	uint8_t other_sm_priority;

	OSM_LOG_ENTER(sm->p_log);

	p_sm_tbl = &sm->p_subn->sm_guid_tbl;

	/* go over all the remote SMs */
	for (p_sm = (osm_remote_sm_t *) cl_qmap_head(p_sm_tbl);
	     p_sm != (osm_remote_sm_t *) cl_qmap_end(p_sm_tbl);
	     p_sm = (osm_remote_sm_t *) cl_qmap_next(&p_sm->map_item)) {
		/* If the sm is in MASTER state - return a pointer to it */
		p_node = osm_get_node_by_guid(sm->p_subn, p_sm->smi.guid);
		if (ib_sminfo_get_state(&p_sm->smi) == IB_SMINFO_STATE_MASTER) {
			other_sm_priority = ib_sminfo_get_priority(&p_sm->smi);
			OSM_LOG(sm->p_log, OSM_LOG_INFO,
				"Found remote master SM with guid:0x%016" PRIx64
				" priority %d (node %s)\n", cl_ntoh64(p_sm->smi.guid),
				other_sm_priority, p_node ? p_node->print_desc : "UNKNOWN");
			p_sm_res = p_sm;
			goto Exit;
		}
	}

Exit:
	OSM_LOG_EXIT(sm->p_log);
	return p_sm_res;
}

/**********************************************************************
 * Go over all remote SMs (as updated in the sm_guid_tbl).
 * Find the one with the highest priority and lowest guid.
 * Compare this SM to the local SM. If the local SM is higher -
 * return NULL, if the remote SM is higher - return a pointer to it.
 **********************************************************************/
static osm_remote_sm_t *state_mgr_get_highest_sm(IN osm_sm_t * sm)
{
	cl_qmap_t *p_sm_tbl;
	osm_remote_sm_t *p_sm = NULL;
	osm_remote_sm_t *p_highest_sm;
	uint8_t highest_sm_priority;
	ib_net64_t highest_sm_guid;
	osm_node_t *p_node;

	OSM_LOG_ENTER(sm->p_log);

	p_sm_tbl = &sm->p_subn->sm_guid_tbl;

	/* Start with the local sm as the standard */
	p_highest_sm = NULL;
	highest_sm_priority = sm->p_subn->active_sm_priority;
	highest_sm_guid = sm->p_subn->sm_port_guid;

	/* go over all the remote SMs */
	for (p_sm = (osm_remote_sm_t *) cl_qmap_head(p_sm_tbl);
	     p_sm != (osm_remote_sm_t *) cl_qmap_end(p_sm_tbl);
	     p_sm = (osm_remote_sm_t *) cl_qmap_next(&p_sm->map_item)) {

		/* If the sm is in NOTACTIVE state - continue */
		if (ib_sminfo_get_state(&p_sm->smi) ==
		    IB_SMINFO_STATE_NOTACTIVE)
			continue;

		if (osm_sm_is_greater_than(ib_sminfo_get_priority(&p_sm->smi),
					   p_sm->smi.guid, highest_sm_priority,
					   highest_sm_guid)) {
			/* the new p_sm is with higher priority - update the highest_sm */
			/* to this sm */
			p_highest_sm = p_sm;
			highest_sm_priority =
			    ib_sminfo_get_priority(&p_sm->smi);
			highest_sm_guid = p_sm->smi.guid;
		}
	}

	if (p_highest_sm != NULL) {
		p_node = osm_get_node_by_guid(sm->p_subn, p_highest_sm->smi.guid);
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"Found higher priority SM with guid: %016" PRIx64 " (node %s)\n",
			cl_ntoh64(p_highest_sm->smi.guid),
			p_node ? p_node->print_desc : "UNKNOWN");
	}
	OSM_LOG_EXIT(sm->p_log);
	return p_highest_sm;
}

/**********************************************************************
 * Send SubnSet(SMInfo) SMP with HANDOVER attribute to the
 * remote_sm indicated.
 **********************************************************************/
static void state_mgr_send_handover(IN osm_sm_t * sm, IN osm_remote_sm_t * p_sm)
{
	uint8_t payload[IB_SMP_DATA_SIZE];
	ib_sm_info_t *p_smi = (ib_sm_info_t *) payload;
	osm_madw_context_t context;
	const osm_port_t *p_port;
	ib_api_status_t status;

	OSM_LOG_ENTER(sm->p_log);

	/*
	 * Send a query of SubnSet(SMInfo) HANDOVER to the remote sm given.
	 */

	memset(&context, 0, sizeof(context));
	p_port = osm_get_port_by_guid(sm->p_subn, p_sm->smi.guid);
	if (p_port == NULL) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3316: "
			"No port object on given remote_sm object\n");
		goto Exit;
	}

	/* update the master_guid in the sm_state_mgr object according to */
	/* the guid of the port where the new Master SM should reside. */
	OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
		"Handing over mastership. Updating sm_state_mgr master_guid: %016"
		PRIx64 " (node %s)\n", cl_ntoh64(p_port->guid),
		p_port->p_node ? p_port->p_node->print_desc : "UNKNOWN");
	sm->master_sm_guid = p_port->guid;

	context.smi_context.port_guid = p_port->guid;
	context.smi_context.set_method = TRUE;

	memset(payload, 0, sizeof(payload));
	p_smi->guid = sm->p_subn->sm_port_guid;
	p_smi->act_count = cl_hton32(sm->p_subn->p_osm->stats[0].qp0_mads_sent);
	p_smi->pri_state = (uint8_t) (sm->p_subn->sm_state |
				      sm->p_subn->active_sm_priority << 4);
	p_smi->sm_key = sm->p_subn->opt.sm_key;

	CL_PLOCK_ACQUIRE(sm->p_lock);
	status = osm_req_set(sm, osm_physp_get_dr_path_ptr(sm, p_port->p_physp),
			     p_port->p_physp,
			     payload, sizeof(payload), IB_MAD_ATTR_SM_INFO,
			     IB_SMINFO_ATTR_MOD_HANDOVER, FALSE,
			     ib_port_info_get_m_key(&p_port->p_physp->port_info),
			     0, CL_DISP_MSGID_NONE, &context);
	CL_PLOCK_RELEASE(sm->p_lock);

	if (status != IB_SUCCESS)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3317: "
			"Failure requesting SMInfo (%s), remote SM at 0x%"
			PRIx64 " (node %s)\n",
			ib_get_err_str(status), cl_ntoh64(p_port->guid),
			p_port->p_node ? p_port->p_node->print_desc : "UNKNOWN");

Exit:
	OSM_LOG_EXIT(sm->p_log);
}

/**********************************************************************
 * Send Trap 64 on all new ports.
 **********************************************************************/
static void state_mgr_report_new_ports(IN osm_sm_t * sm)
{
	ib_gid_t port_gid;
	ib_mad_notice_attr_t notice;
	ib_api_status_t status;
	ib_net64_t port_guid;
	cl_map_item_t *p_next;
	osm_port_t *p_port;
	uint16_t min_lid_ho;
	uint16_t max_lid_ho;

	OSM_LOG_ENTER(sm->p_log);

	CL_PLOCK_ACQUIRE(sm->p_lock);
	p_next = cl_qmap_head(&sm->p_subn->port_guid_tbl);
	while (p_next != cl_qmap_end(&sm->p_subn->port_guid_tbl)) {
		p_port = (osm_port_t *) p_next;
		p_next = cl_qmap_next(p_next);

		if (!p_port->is_new)
			continue;

		port_guid = osm_port_get_guid(p_port);
		/* issue a notice - trap 64 */

		/* details of the notice */
		notice.generic_type = 0x83;	/* is generic subn mgt type */
		ib_notice_set_prod_type_ho(&notice, 4);	/* A Class Manager generator */
		/* endport becomes reachable */
		notice.g_or_v.generic.trap_num = CL_HTON16(64);
		/* The sm_base_lid is saved in network order already. */
		notice.issuer_lid = sm->p_subn->sm_base_lid;
		/* following C14-72.1.1 and table 119 p739 */
		/* we need to provide the GID */
		port_gid.unicast.prefix = sm->p_subn->opt.subnet_prefix;
		port_gid.unicast.interface_id = port_guid;
		memcpy(&(notice.data_details.ntc_64_67.gid), &(port_gid),
		       sizeof(ib_gid_t));

		/* According to page 653 - the issuer gid in this case of trap
		 * is the SM gid, since the SM is the initiator of this trap. */
		notice.issuer_gid.unicast.prefix =
		    sm->p_subn->opt.subnet_prefix;
		notice.issuer_gid.unicast.interface_id =
		    sm->p_subn->sm_port_guid;

		status = osm_report_notice(sm->p_log, sm->p_subn, &notice,
					   (uint64_t) cl_atomic_inc(&sm->p_subn->tid) & (uint64_t)(0xFFFFFFFF),
					   0);
		if (status != IB_SUCCESS)
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3318: "
				"Error sending trap reports on GUID:0x%016"
				PRIx64 " (%s)\n", port_gid.unicast.interface_id,
				ib_get_err_str(status));

		osm_port_get_lid_range_ho(p_port, &min_lid_ho, &max_lid_ho);
		OSM_LOG(sm->p_log, OSM_LOG_INFO,
			"Discovered new port with GUID:0x%016" PRIx64
			" LID range [%u,%u] of node: %s\n",
			cl_ntoh64(port_gid.unicast.interface_id),
			min_lid_ho, max_lid_ho,
			p_port->p_node ? p_port->p_node->
			print_desc : "UNKNOWN");

		p_port->is_new = 0;
	}
	CL_PLOCK_RELEASE(sm->p_lock);

	OSM_LOG_EXIT(sm->p_log);
}

/**********************************************************************
 * Make sure that the lid_port_tbl of the subnet has only the ports
 * that are recognized, and in the correct lid place. There could be
 * errors if we wanted to assign a certain port with lid X, but that
 * request didn't reach the port. In this case port_lid_tbl will have
 * the port under lid X, though the port isn't updated with this lid.
 * We will run a new heavy sweep (since there were errors in the
 * initialization), but here we'll clean the database from incorrect
 * information.
 **********************************************************************/
static void state_mgr_check_tbl_consistency(IN osm_sm_t * sm)
{
	cl_qmap_t *p_port_guid_tbl;
	osm_port_t *p_port;
	osm_port_t *p_next_port;
	cl_ptr_vector_t *p_port_lid_tbl;
	size_t max_lid, ref_size, curr_size, lid;
	osm_port_t *p_port_ref, *p_port_stored;
	cl_ptr_vector_t ref_port_lid_tbl;
	uint16_t min_lid_ho;
	uint16_t max_lid_ho;
	uint16_t lid_ho;

	OSM_LOG_ENTER(sm->p_log);

	if (sm->lid_mgr.dirty == FALSE)
		goto Exit;

	sm->lid_mgr.dirty = FALSE;

	cl_ptr_vector_construct(&ref_port_lid_tbl);
	cl_ptr_vector_init(&ref_port_lid_tbl,
			   cl_ptr_vector_get_size(&sm->p_subn->port_lid_tbl),
			   OSM_SUBNET_VECTOR_GROW_SIZE);

	p_port_guid_tbl = &sm->p_subn->port_guid_tbl;

	/* Let's go over all the ports according to port_guid_tbl,
	 * and add the port to a reference port_lid_tbl. */
	p_next_port = (osm_port_t *) cl_qmap_head(p_port_guid_tbl);
	while (p_next_port != (osm_port_t *) cl_qmap_end(p_port_guid_tbl)) {
		p_port = p_next_port;
		p_next_port =
		    (osm_port_t *) cl_qmap_next(&p_next_port->map_item);

		osm_port_get_lid_range_ho(p_port, &min_lid_ho, &max_lid_ho);
		for (lid_ho = min_lid_ho; lid_ho <= max_lid_ho; lid_ho++)
			cl_ptr_vector_set(&ref_port_lid_tbl, lid_ho, p_port);
	}

	p_port_lid_tbl = &sm->p_subn->port_lid_tbl;

	ref_size = cl_ptr_vector_get_size(&ref_port_lid_tbl);
	curr_size = cl_ptr_vector_get_size(p_port_lid_tbl);
	/* They should be the same, but compare it anyway */
	max_lid = (ref_size > curr_size) ? ref_size : curr_size;

	for (lid = 1; lid < max_lid; lid++) {
		p_port_ref = NULL;
		p_port_stored = NULL;
		cl_ptr_vector_at(p_port_lid_tbl, lid, (void *)&p_port_stored);
		cl_ptr_vector_at(&ref_port_lid_tbl, lid, (void *)&p_port_ref);

		if (p_port_stored == p_port_ref)
			/* This is the "good" case - both entries are the
			 * same for this lid. Nothing to do. */
			continue;

		if (p_port_ref == NULL) {
			/* There is an object in the subnet database for this
			 * lid, but no such object exists in the reference
			 * port_list_tbl. This can occur if we wanted to assign
			 * a certain port with some lid (different than the one
			 * pre-assigned to it), and the port didn't get the
			 * PortInfo Set request. Due to this, the port is
			 * updated with its original lid in our database
			 * rather than the new lid we wanted to give it. */
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3322: "
				"lid %zu is wrongly assigned to port 0x%016"
				PRIx64 " (\'%s\' port %u) in port_lid_tbl\n",
				lid,
				cl_ntoh64(osm_port_get_guid(p_port_stored)),
				p_port_stored->p_node->print_desc,
				p_port_stored->p_physp->port_num);
			cl_ptr_vector_set(p_port_lid_tbl, lid, NULL);
		} else if (p_port_stored == NULL)
			/* There is an object in the new database, but no
			 * object in our subnet database. This is the matching
			 * case of the prior check - the port still has its
			 * original lid. */
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3323: "
				"port 0x%016" PRIx64 " (\'%s\' port %u)"
				" exists in new port_lid_tbl under lid %zu,"
				" but missing in subnet port_lid_tbl db\n",
				cl_ntoh64(osm_port_get_guid(p_port_ref)),
				p_port_ref->p_node->print_desc,
				p_port_ref->p_physp->port_num, lid);
		else {
			/* if we reached here then p_port_stored != p_port_ref.
			 * We were trying to set a lid to p_port_stored, but
			 * it didn't reach it, and p_port_ref also didn't get
			 * the lid update. */
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3324: "
				"lid %zu has port 0x%016" PRIx64
				" (\'%s\' port %u) in new port_lid_tbl db, "
				"and port 0x%016" PRIx64 " (\'%s\' port %u)"
				" in subnet port_lid_tbl db\n", lid,
				cl_ntoh64(osm_port_get_guid(p_port_ref)),
				p_port_ref->p_node->print_desc,
				p_port_ref->p_physp->port_num,
				cl_ntoh64(osm_port_get_guid(p_port_stored)),
				p_port_stored->p_node->print_desc,
				p_port_stored->p_physp->port_num);
			cl_ptr_vector_set(p_port_lid_tbl, lid, NULL);
		}

		/*
		 * Clear the lid of the port in order to ignore it
		 *  in routing phase
		 */
		if (p_port_stored) {
			OSM_LOG(sm->p_log, OSM_LOG_INFO, "Clearing Lid for "
				"port 0x%016" PRIx64 "\n",
				cl_ntoh64(osm_port_get_guid(p_port_stored)));
			osm_port_clear_base_lid(p_port_stored);
		}

		/* Make sure we'll do another heavy sweep. */
		sm->p_subn->subnet_initialization_error = TRUE;
	}

	cl_ptr_vector_destroy(&ref_port_lid_tbl);

Exit:
	OSM_LOG_EXIT(sm->p_log);
}

static void check_switch_lft(cl_map_item_t * item, void *context)
{
	osm_switch_t *sw = (osm_switch_t *) item;
	osm_sm_t *p_sm = (osm_sm_t *) context;

	/* Do not send LFT, block is AR configured */
	if (sw->ar_configured == OSM_SW_AR_CONF_PLUGIN) {
		OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG,
			"AR configured on 0x%016" PRIx64
			", no need to check switch lft.\n",
			cl_ntoh64(sw->p_node->node_info.node_guid));
		return;
	}

	if (!sw->new_lft)
		return;

	if (lookup_held_back_switch(p_sm->p_subn, sw->p_node))
		return;

	if (sw->ar_configured == OSM_SW_AR_CONF_DISABLED)
		sw->ar_configured = OSM_SW_AR_CONF_NONE;

	if (sw->ar_configured == OSM_SW_AR_CONF_NONE &&
	    memcmp(sw->lft, sw->new_lft, sw->max_lid_ho + 1))
		   osm_log_v2(p_sm->p_log, OSM_LOG_ERROR, FILE_ID, "ERR 331D: "
			      "LFT of switch 0x%016" PRIx64 " (%s) is not up to date\n",
			      cl_ntoh64(sw->p_node->node_info.node_guid),
			      sw->p_node->print_desc);

	if (sw->ar_configured == OSM_SW_AR_CONF_CONFIGURED &&
	    osm_ar_cmp_ar_lft(&sw->ar_data))
		   osm_log_v2(p_sm->p_log, OSM_LOG_ERROR, FILE_ID, "ERR 331E: "
			      "AR LFT of switch 0x%016" PRIx64 " (%s) is not up to date\n",
			      cl_ntoh64(sw->p_node->node_info.node_guid),
			      sw->p_node->print_desc);
}

int wait_for_pending_transactions(osm_opensm_t * p_osm)
{
	uint8_t num_ports = p_osm->subn.opt.num_guids;
	uint8_t port_idx;
	osm_stats_t * p_stats;
	boolean_t repeat = TRUE;

	while (repeat) {
		repeat = FALSE;
		for (port_idx = 0; port_idx < num_ports; port_idx ++) {
			p_stats = &p_osm->stats[port_idx];
#ifdef HAVE_LIBPTHREAD
			pthread_mutex_lock(&p_stats->mutex);
			while (p_stats->qp0_mads_outstanding && !osm_exit_flag) {
				pthread_cond_wait(&p_stats->cond, &p_stats->mutex);
				/*
				 * if we have MADs on this port, need to check
				 * other ports again, since MADs could be sent on them,
				 * while this port is checked.
				 */
				repeat = TRUE;
			}
			pthread_mutex_unlock(&p_stats->mutex);
#else
			while (1) {
				unsigned count = p_stats->qp0_mads_outstanding;
				if (!count || osm_exit_flag)
					break;
				cl_event_wait_on(&p_stats->event, EVENT_NO_TIMEOUT, TRUE);
			}
#endif
		}
	}

	if (osm_exit_flag) {
		CL_PLOCK_EXCL_ACQUIRE(p_osm->sm.p_lock);
		osm_subn_clear_ports_mads_queue(&p_osm->subn);
		CL_PLOCK_RELEASE(p_osm->sm.p_lock);
	} else {
		osm_subn_clear_ports_mads_queue(&p_osm->subn);
	}

	return osm_exit_flag;
}

static void dump_node_guids(cl_map_item_t *item, FILE *file, void *cxt)
{
	osm_sm_t *sm = (osm_sm_t *) cxt;

	if (osm_get_node_by_guid(sm->p_subn, cl_qmap_key(item)))
		fprintf(file, "0x%016" PRIx64 "\n",
			cl_ntoh64(cl_qmap_key(item)));
}

/**********************************************************************
 Sweeps the ignored ports.
 We want to find all ports that resides only behind ignored ports
 and weren't discovered earlier.
**********************************************************************/
static ib_api_status_t state_mgr_sweep_ignored(IN osm_sm_t * sm)
{
	ib_api_status_t status = IB_SUCCESS;
	OSM_LOG_ENTER(sm->p_log);
	while (cl_list_count(&sm->p_subn->ignored_nodes_list)) {
		if (state_mgr_sweep_ignored_iter(sm) != IB_SUCCESS ||
		    wait_for_pending_transactions(sm->p_subn->p_osm)) {
			status = IB_ERROR;
			goto Exit;
		}
	}
Exit:
	OSM_LOG_EXIT(sm->p_log);
	return status;
}

/*
	Iterate over configured held back switches,
	for each one mark all ports as unhealthy
*/
static void osm_handle_held_back_switches(osm_sm_t *sm)
{
	char heldback_file_path[1024];
	if (OSM_LOG_IS_ACTIVE_V2(sm->p_log, OSM_LOG_VERBOSE)) {
		snprintf(heldback_file_path, sizeof(heldback_file_path),
			 "%s/opensm-heldback-switch.dump",
			 sm->p_subn->p_osm->subn.opt.dump_files_dir);
		osm_dump_qmap_to_file(sm->p_subn->p_osm,
				      heldback_file_path,
				      &sm->p_subn->held_back_sw_tbl,
				      dump_node_guids, sm);
	}
}

static void osm_mark_rebooted_ports(osm_sm_t *sm)
{
	cl_map_item_t *p_item;
	osm_node_t *p_node;
	osm_physp_t *p_physp;
	uint8_t init_down_ports, port_num;

	if (sm->p_subn->first_time_master_sweep == TRUE)
		return;

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);
	for (p_item = cl_qmap_head(&sm->p_subn->node_guid_tbl);
	     p_item != cl_qmap_end(&sm->p_subn->node_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_node = (osm_node_t *) p_item;
		init_down_ports = 0;
		for (port_num = 1; port_num < osm_node_get_num_physp(p_node);
		     port_num++) {
			p_physp = osm_node_get_physp_ptr(p_node, port_num);
			if (!p_node->physp_discovered[port_num] || !p_physp ||
			    ib_port_info_get_port_state(&p_physp->port_info) ==
							IB_LINK_DOWN) {
				init_down_ports++;
				continue;
			} else if (ib_port_info_get_port_state(&p_physp->port_info) ==
							       IB_LINK_INIT) {
				if (!osm_link_mgr_is_init_state_expected(sm, p_physp))
					osm_hm_set_by_physp(sm, p_physp, OSM_HM_COND_FLAPPING);

				init_down_ports++;
			}
		}
		/* We don't check for Switch's port 0 */
		if (init_down_ports == osm_node_get_num_physp(p_node) - 1)
			osm_hm_set_by_node(sm, p_node, OSM_HM_COND_REBOOT);
	}
	CL_PLOCK_RELEASE(sm->p_lock);

}

static void state_mgr_restore_node_link(IN cl_map_item_t * p_map_item,
					IN void *context)
{
	uint8_t port, remote_port;
	osm_node_t *p_node, *p_rem_node;
	osm_sm_t * sm = context;

	p_node = (osm_node_t *) p_map_item;

	for (port = 1; port < p_node->physp_tbl_size; port++) {
		if (!osm_node_relink(p_node, port))
			continue;

		p_rem_node = osm_node_get_remote_node(p_node, port,
						      &remote_port);
		OSM_ASSERT(p_rem_node);
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"Restored links between 0x%" PRIx64
			", port %u and 0x%" PRIx64 "%u\n",
			osm_node_get_node_guid(p_node), port,
			osm_node_get_node_guid(p_rem_node),
			remote_port);
	}
}

static void state_mgr_remove_fnm_ring_link(IN cl_map_item_t * p_map_item,
				           IN void *context)
{
	osm_sm_t *sm = (osm_sm_t *) context;
	osm_switch_t *p_sw = (osm_switch_t *) p_map_item;
	osm_node_t *p_node, *p_remote_node;
	osm_physp_t *p_physp, *p_remote_physp;
	uint8_t port_num, remote_port_num;
	boolean_t rem_is_sw;
	osm_hierarchy_info_t *p_hierarchy, *p_remote_hierarchy;
	boolean_t is_prism_sw;

	p_node = p_sw->p_node;
	is_prism_sw = osm_switch_is_prism_switch(p_sw, NULL);

	for (port_num = 1; port_num < osm_node_get_num_physp(p_node); port_num++) {
		p_physp = osm_node_get_physp_ptr(p_node, port_num);
		if (!p_physp)
			continue;
		p_remote_physp = osm_physp_get_remote(p_physp);
		if (!p_remote_physp)
			continue;

		rem_is_sw = (osm_node_get_type(p_remote_physp->p_node) ==
				IB_NODE_TYPE_SWITCH);

		p_hierarchy = &p_physp->hierarchy_info;
		p_remote_hierarchy = &p_remote_physp->hierarchy_info;

		/*
		 * Look at FNM link between switches.
		 * On the prism switch the FNM link between asics is not for routing.
		 * For non prism switch (ex. Black Mamba) remove one link on the FNM ring
		 * from routing (between first asic and last) to prevent credit loop.
		 */
		if (rem_is_sw && ib_mlnx_epi_get_is_fnm_port(&p_physp->ext_port_info) &&
		    ib_mlnx_epi_get_is_fnm_port(&p_remote_physp->ext_port_info) &&
		    (is_prism_sw ||
		     (p_hierarchy->asic > 0 && p_remote_hierarchy->asic > 0 &&
		      (p_remote_hierarchy->asic + 1 < p_hierarchy->asic)))) {
			p_remote_node =
			    osm_physp_get_node_ptr(p_remote_physp);
			remote_port_num =
			    osm_physp_get_port_num(p_remote_physp);
			OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
				"Ignored link: "
				"Unlinking fnm ring%s, local node 0x%" PRIx64
				", port %u (asic %u)"
				"\n\t\t\t\tand remote node 0x%" PRIx64
				", port %u (asic %u) \n",
				is_prism_sw ? " on prism switch" : "",
				cl_ntoh64(osm_node_get_node_guid
					(p_node)), port_num, p_hierarchy->asic,
				cl_ntoh64(osm_node_get_node_guid
					(p_remote_node)),
				remote_port_num, p_remote_hierarchy->asic);

			osm_node_unlink(p_node, port_num,
					p_remote_node,
					remote_port_num,
					TRUE);
			sm->p_subn->restore_links = TRUE;
		}
	}
}

static void state_mgr_remove_port_link(IN cl_map_item_t * p_map_item,
				       IN void *context)
{
	osm_sm_t *sm = (osm_sm_t *) context;
	osm_port_t *p_port = (osm_port_t *) p_map_item;
	osm_node_t *p_node, *p_remote_node;
	osm_physp_t *p_physp, *p_remote_physp;
	uint8_t port_num, remote_port_num;
	boolean_t is_held_back, rem_is_sw, is_reset_required;

	p_node = p_port->p_node;
	p_physp = p_port->p_physp;

	if (osm_node_get_type(p_node) != IB_NODE_TYPE_SWITCH)
		return;

	is_held_back = lookup_held_back_switch(sm->p_subn, p_node);
	for (port_num = 1; port_num < p_node->physp_tbl_size; port_num++) {
		p_physp = osm_node_get_physp_ptr(p_node, port_num);
		if (!p_physp)
			continue;
		p_remote_physp = osm_physp_get_remote(p_physp);
		if (!p_remote_physp)
			continue;

		rem_is_sw = (osm_node_get_type(p_remote_physp->p_node) ==
				IB_NODE_TYPE_SWITCH);

		/*
		 * If remote is ca we do not want to remove the link
		 * because when ca is back we want lfts to include it
		 * so we won't need to rerun routing
		 */
		is_reset_required = (p_physp->port_reset_required && rem_is_sw );

		if (is_held_back || is_reset_required ||
		    lookup_isolated_port(sm, p_node, port_num)) {
			p_remote_node =
			    osm_physp_get_node_ptr(p_remote_physp);
			remote_port_num =
			    osm_physp_get_port_num(p_remote_physp);
			OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
				"Ignored link: "
				"Unlinking local node 0x%" PRIx64
				", port %u"
				"\n\t\t\t\tand remote node 0x%" PRIx64
				", port %u\n",
				cl_ntoh64(osm_node_get_node_guid
					(p_node)), port_num,
				cl_ntoh64(osm_node_get_node_guid
					(p_remote_node)),
				remote_port_num);
			osm_node_unlink(p_node, (uint8_t) port_num,
					p_remote_node,
					(uint8_t) remote_port_num,
					TRUE);
			sm->p_subn->restore_links = TRUE;
		}
	}
}

static void state_mgr_remove_aport_link(IN cl_map_item_t * p_map_item,
				       IN void *context)
{
	osm_sm_t *sm = (osm_sm_t *) context;
	osm_system_t *p_system = (osm_system_t *) p_map_item;
	osm_aport_t *p_aport;
	uint8_t aport_num;
	osm_physp_t *p_physp, *p_remote_physp;
	uint8_t plane_number;
	osm_node_t *p_node, *p_remote_node;
	uint8_t port_num, remote_port_num;

	for (aport_num = 1; aport_num <= p_system->max_aport_available; aport_num++) {
		p_aport = &p_system->aport_table[aport_num];

		/* go over physps of aport */
		if (osm_hierarchy_is_asymmetric_aport(p_aport)) {
			for (plane_number = 1; plane_number <= p_aport->max_found_plane_number;
			     plane_number++) {
				p_physp = p_aport->plane_ports[plane_number];
				if (!p_physp)
					continue;

				p_remote_physp = osm_physp_get_remote(p_physp);
				if (!p_remote_physp)
					continue;

				p_node = osm_physp_get_node_ptr(p_physp);
				port_num = osm_physp_get_port_num(p_physp);
				p_remote_node =
					osm_physp_get_node_ptr(p_remote_physp);
				remote_port_num =
					osm_physp_get_port_num(p_remote_physp);

				OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
					"Ignored link: "
					"Unlinking local aport %u of system 0x%" PRIx64 " node 0x%" PRIx64
					", port %u and remote node 0x%" PRIx64
					", port %u\n",
					aport_num, cl_ntoh64(p_system->system_guid),
					cl_ntoh64(osm_node_get_node_guid(p_node)),
					osm_physp_get_port_num(p_physp),
					cl_ntoh64(osm_node_get_node_guid(p_remote_node)),
					remote_port_num);

				osm_node_unlink(p_node, port_num,
						p_remote_node,
						remote_port_num,
						TRUE);
				sm->p_subn->restore_links = TRUE;
			}
		}
	}
}

static void osm_remove_links(osm_sm_t * sm)
{
	if (sm->p_subn->found_ignored_port || sm->p_subn->found_link_for_reset)
		cl_qmap_apply_func(&sm->p_subn->port_guid_tbl,
				   state_mgr_remove_port_link, sm);

	if (sm->p_subn->is_planarized) {
		cl_qmap_apply_func(&sm->p_subn->sw_guid_tbl,
				state_mgr_remove_fnm_ring_link, sm);
		/* go over aports */
		cl_qmap_apply_func(&sm->p_subn->system_guid_tbl, state_mgr_remove_aport_link, sm);
	}
}

static void osm_restore_links(osm_sm_t * sm)
{
	if (sm->p_subn->restore_links == FALSE)
		return;

	cl_qmap_apply_func(&sm->p_subn->node_guid_tbl,
			   state_mgr_restore_node_link, sm);
}

static void osm_disable_ar_for_non_reordering_devices(osm_sm_t * sm)
{
	uint16_t lids[IB_LID_UCAST_END_HO + 1];
	osm_subn_t *p_subn = sm->p_subn;
	osm_port_t *p_port;
	osm_switch_t *p_sw;
	cl_map_item_t *p_item;
	uint16_t lid, num_lids = 0;


	/* Create list of lids that does not support packet reordering */
	for (lid = 1; lid <= p_subn->max_lid; lid++) {
		p_port = osm_get_port_by_lid_ho(p_subn, lid);
		if (!p_port)
			continue;

		if (ib_general_info_is_cap_enabled(&p_port->p_node->general_info,
						   IB_GENERAL_INFO_IS_GLOBAL_OOO_SUPPORTED_BIT))
			continue;

		lids[num_lids++] = lid;
	}

	if (!num_lids)
		return;

	/* Disable AR toward LIDs that does not support packet reordering */
	for (p_item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     p_item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_sw = (osm_switch_t *)p_item;
		if (p_sw->ar_configured != OSM_SW_AR_CONF_CONFIGURED)
			continue;

		osm_ar_lft_set_static_state_for_dlids(&p_sw->ar_data, lids, num_lids);
	}
}

/* This reachability is based on the routing tables of the switch connected to SM */
static void state_mgr_check_reachability(osm_opensm_t * p_osm)
{
	osm_sm_t *sm = &p_osm->sm;
	osm_port_t *p_sm_port;
	osm_physp_t *p_physp;
	osm_switch_t *p_sw;
	osm_node_t *p_sm_node;
	osm_node_t *p_sw_node;
	osm_node_t *p_dest_node;
	osm_port_t *p_port;
	cl_map_item_t *p_item;
	uint16_t lid;
	uint8_t port_num;
	boolean_t previously_reachable;

	p_sm_port = osm_get_port_by_guid(sm->p_subn, sm->p_subn->sm_port_guid);
	if (!p_sm_port) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3354: "
			"Could not check reachability, sm port was not found\n");
		return;
	}

	/* Subnet is either single port (SM's port) or host-to-host setup */
	if (!cl_qmap_count(&p_osm->subn.sw_guid_tbl)) {
		OSM_ASSERT(cl_qmap_count(&p_osm->subn.port_guid_tbl) <= 2);
		for (p_item = cl_qmap_head(&p_osm->subn.port_guid_tbl);
		     p_item != cl_qmap_end(&p_osm->subn.port_guid_tbl);
		     p_item = cl_qmap_next(p_item)) {
			p_port = (osm_port_t *)p_item;
			p_port->is_reachable = TRUE;
		}
		return;
	}

	p_sm_node = p_sm_port->p_node;

	if (osm_node_get_type(p_sm_node) == IB_NODE_TYPE_SWITCH) {
		p_sw_node = p_sm_node;
	} else {
		port_num = osm_physp_get_port_num(osm_port_get_physp(p_sm_port));
		p_sw_node = osm_node_get_remote_node(p_sm_node, port_num, NULL);
	}

	/*
	 * SM's port was already found, thus subnet contains
	 * other ports than SM's, and SM should have a remote
	 */
	if (!p_sw_node || !p_sw_node->sw) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3355: "
			"Could not check reachability, sm node was not found\n");
		return;
	}

	p_sw = p_sw_node->sw;

	for (p_item = cl_qmap_head(&p_osm->subn.port_guid_tbl);
	     p_item != cl_qmap_end(&p_osm->subn.port_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_port = (osm_port_t *)p_item;
		lid = cl_ntoh16(osm_port_get_base_lid(p_port));
		if (!lid) {
			OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
				"Could not check reachability of port "
				"0x%" PRIx64 ", lid is 0\n",
				cl_ntoh64(p_port->guid));
			p_port->is_reachable = FALSE;
			continue;
		}

		previously_reachable = p_port->is_reachable;
		p_port->is_reachable =
		    (osm_switch_get_port_by_lid(p_sw, lid, OSM_NEW_LFT) != OSM_NO_PATH);

		/* Update ports that became reachable */
		if (!previously_reachable &&
		    p_port->is_reachable &&
		    (p_physp = osm_port_get_physp(p_port))) {
			p_physp->need_update = 2;

			/* If port is port 0 of switch, update switch */
			p_dest_node = osm_physp_get_node_ptr(p_physp);
			if (osm_physp_get_port_num(p_physp) == 0 &&
			    osm_node_get_type(p_dest_node) == IB_NODE_TYPE_SWITCH &&
			    p_dest_node->sw) {
				p_dest_node->sw->need_update = 2;
			}
		}
	}
}

static void add_peers_to_sw_data(OUT osm_sw_data_t * p_sw_data,
				 OUT osm_sw_data_t * data_per_sw)
{
	osm_switch_t *p_sw = p_sw_data->p_sw;
	osm_node_t *p_node = p_sw->p_node;
	osm_node_t *p_remote_node;
	uint32_t port_num;
	uint16_t remote_lid_ho;

	for (port_num = 0; port_num < p_sw->num_ports; port_num++) {
		p_remote_node = osm_node_get_remote_node(p_node, port_num, NULL);
		if (!p_remote_node || p_remote_node == p_node ||
		    osm_node_get_type(p_remote_node) != IB_NODE_TYPE_SWITCH)
			continue;

		remote_lid_ho = cl_ntoh16(osm_node_get_base_lid(p_remote_node, 0));
		p_sw_data->peers[p_sw_data->degree] = &data_per_sw[remote_lid_ho];
		p_sw_data->degree++;
	}
}

static void init_data_per_sw_array(IN osm_subn_t * p_subn,
				   OUT osm_sw_data_t * data_per_sw)
{
	osm_sw_data_t *p_sw_data;
	cl_map_item_t *item;
	osm_switch_t *p_sw;
	osm_node_t *p_node;
	uint16_t lid_ho;

	for (item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     item = cl_qmap_next(item)) {
		p_sw = (osm_switch_t *)item;
		p_node = p_sw->p_node;

		lid_ho = cl_ntoh16(osm_node_get_base_lid(p_node, 0));
		p_sw_data = &data_per_sw[lid_ho];
		p_sw_data->p_sw = p_sw;
		add_peers_to_sw_data(p_sw_data, data_per_sw);
	}
}

static uint16_t osm_topo_get_first_sw_lid(osm_subn_t * p_subn)
{
	cl_map_item_t *item = cl_qmap_head(&p_subn->sw_guid_tbl);
	osm_switch_t *p_sw = (osm_switch_t *)item;

	if (item == cl_qmap_end(&p_subn->sw_guid_tbl))
		return 0;

	return cl_ntoh16(osm_node_get_base_lid(p_sw->p_node, 0));
}

static osm_topology_type_enum detect_topology_type(osm_subn_t * p_subn)
{
	osm_opensm_t *p_osm = p_subn->p_osm;
	osm_sw_data_t *sw_data_arr = NULL;
	osm_sw_data_t *p_sw_data = NULL, *peer, *p_curr_sw;
	osm_sw_data_t **queue = NULL;
	uint16_t lid_ho, first_sw_lid;
	uint32_t queue_in = 0, queue_out = 0, i;
	boolean_t is_bipartite = TRUE;
	osm_topology_type_enum result = OSM_TOPOLOGY_TYPE_UNKNOWN;


	OSM_LOG_ENTER(&p_osm->log);

	 if (!cl_qmap_count(&p_subn->sw_guid_tbl))
		goto Exit;

	lid_ho = p_subn->max_lid + 1;
	sw_data_arr = calloc(lid_ho, sizeof(osm_sw_data_t));
	if (!sw_data_arr) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 3301: "
			"Cannot allocate memory for data per switch array\n");
		goto Exit;
	}

	queue = (osm_sw_data_t **)calloc(cl_qmap_count(&p_subn->sw_guid_tbl),
				     sizeof(osm_sw_data_t *));
	if (!queue) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 3302: "
			"Cannot allocate memory for sw_data Queue\n");
		goto Exit;
	}
	init_data_per_sw_array(p_subn, sw_data_arr);

	/* Start BFS from the first switch on sw_guid_tbl */
	first_sw_lid = osm_topo_get_first_sw_lid(p_subn);
	p_sw_data = &sw_data_arr[first_sw_lid];
	p_sw_data->color = COLOR_A;
	p_sw_data->visited = TRUE;
	queue[queue_in++] = p_sw_data;

	/* BFS the queue until no next element */
	while (queue_out < queue_in && is_bipartite) {
		p_curr_sw = queue[queue_out++];

		for (i = 0; i < p_curr_sw->degree; i++) {
			peer = p_curr_sw->peers[i];
			if (!peer->visited) {
				peer->color = !p_curr_sw->color;
				peer->visited = TRUE;
				queue[queue_in++] = peer;
			} else if (peer->color == p_curr_sw->color) {
				is_bipartite = FALSE;
				break;
			}
		}
	}

	if (is_bipartite)
		result = OSM_TOPOLOGY_TYPE_TREE;
	else
		result = OSM_TOPOLOGY_TYPE_DFP;

Exit:
	if (sw_data_arr)
		free(sw_data_arr);
	if (queue)
		free(queue);

	switch (result) {
	case OSM_TOPOLOGY_TYPE_TREE:
		OSM_LOG(&p_osm->log, OSM_LOG_INFO, "Fabric topology is tree topology\n");
		break;
	case OSM_TOPOLOGY_TYPE_DFP:
		OSM_LOG(&p_osm->log, OSM_LOG_INFO, "Fabric topology is DF+ topology\n");
		break;
	default:
		OSM_LOG(&p_osm->log, OSM_LOG_INFO, "Fabric topology is unidentified\n");

	}

	OSM_LOG_EXIT(&p_osm->log);

	return result;
}

/* send routing tables to switches and routers */
static void state_mgr_set_routing_tables(osm_sm_t *sm, osm_node_filter_t filter)
{
	osm_switch_t *p_sw;

	osm_ucast_mgr_set_fwd_tables(&sm->ucast_mgr, filter);

	/*
	 * Set ar_configured indicator to TRUE when Adaptive Routing was configured
	 * on the fabric.
	 * Note: SHIELD does not affect this indicator.
	 */
	sm->ucast_mgr.ar_configured = FALSE;
	if (sm->p_subn->ar_conf_params.ar_mode != OSM_ADV_ROUTING_MODE_DISABLED ||
	    sm->p_subn->is_ar_plugin_active) {
		for (p_sw = (osm_switch_t *) cl_qmap_head(&sm->p_subn->sw_guid_tbl);
		     p_sw != (osm_switch_t *) cl_qmap_end(&sm->p_subn->sw_guid_tbl);
		     p_sw = (osm_switch_t *) cl_qmap_next(&p_sw->map_item)) {
			if (p_sw->ar_configured == OSM_SW_AR_CONF_CONFIGURED ||
			    p_sw->ar_configured == OSM_SW_AR_CONF_PLUGIN) {
				sm->ucast_mgr.ar_configured = TRUE;
				break;
			}
		}
	}

	/* routing completed successfully */
	if (sm->ucast_mgr.p_subn->opt.use_ucast_cache) {
		sm->ucast_mgr.cache_valid = TRUE;
		sm->ucast_mgr.vlid_cache_valid = TRUE;
	}

	/*
	 * if FLIDs feature is enabled, set router lid table tables,
	 * that may be changed during routing
	 */
	if (osm_subn_is_flids_enabled(sm->p_subn) && osm_local_flid_enabled(sm->p_subn))
		osm_router_mgr_set_rtr_lid_tables(&sm->router_mgr);
}

static void state_mgr_activate_links(osm_sm_t *sm, boolean_t process_endports)
{
	/*
	 * The LINK_PORTS state is required since we cannot count on
	 * the port state change MADs to succeed. This is an artifact
	 * of the spec defining state change from state X to state X
	 * as an error. The hardware then is not required to process
	 * other parameters provided by the Set(PortInfo) Packet.
	 */

	osm_link_mgr_process(sm, IB_LINK_NO_CHANGE, process_endports);
	if (wait_for_pending_transactions(sm->p_subn->p_osm))
		return;

	OSM_LOG_MSG_BOX(sm->p_log, OSM_LOG_VERBOSE,
			process_endports ?
			"ENDPORT LINKS CONFIGURED - SET LINKS TO ARMED STATE" :
			"SWITCH LINKS CONFIGURED - SET LINKS TO ARMED STATE");

	osm_link_mgr_process(sm, IB_LINK_ARMED, process_endports);
	if (wait_for_pending_transactions(sm->p_subn->p_osm))
		return;

	OSM_LOG_MSG_BOX(sm->p_log, OSM_LOG_VERBOSE,
			process_endports ?
			"ENDPORT LINKS ARMED - SET LINKS TO ACTIVE STATE" :
			"SWITCH LINKS ARMED - SET LINKS TO ACTIVE STATE");

	osm_link_mgr_process(sm, IB_LINK_ACTIVE, process_endports);
	if (wait_for_pending_transactions(sm->p_subn->p_osm))
		return;
}

static boolean_t osm_topo_routing(osm_sm_t * sm, boolean_t * rerouting_done)
{
	boolean_t ret = TRUE;
	osm_opensm_t *p_osm = sm->p_subn->p_osm;
	ib_api_status_t previous_result;
	osm_routing_chain_mgr_t *rch_mgr = &sm->routing_chain_mgr;
	size_t i;
	boolean_t enable_queries_during_routing =
	    sm->p_subn->opt.enable_queries_during_routing;

	/* Reset ignored links restore flag */
	sm->p_subn->restore_links = FALSE;

	CL_PLOCK_EXCL_ACQUIRE(&p_osm->routing_lock);
	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);

	/* Reinitialize the ucast mgr thread pool */
	if (sm->ucast_mgr.variables_reinit_thread_pool) {
		osm_ucast_mgr_thread_pool_init(&sm->ucast_mgr);
		sm->ucast_mgr.variables_reinit_thread_pool = FALSE;
	}

	if (osm_is_crc_invalidates_cache(sm)) {
		OSM_LOG(sm->p_log, OSM_LOG_INFO, "Invalidating ucast cache "
			"due to crc files changes\n");
		sm->ucast_mgr.variables_invalidate_cache = TRUE;
		sm->p_subn->force_multicast_rerouting = TRUE;
	}

	if (sm->p_subn->is_auto_engine_used)
		sm->p_subn->detected_topo = detect_topology_type(sm->p_subn);

	osm_remove_links(sm);

	/*
	 * Main lock is released at this point in order to allow other threads
	 * (i.e. SA queries) to work in parallel with routing calculation.
	 * Threads that access routing structures are locked by routing_lock.
	 */
	if (enable_queries_during_routing)
		CL_PLOCK_RELEASE(sm->p_lock);

	previous_result = sm->topology_mgr.creation_result;
	if ((osm_topo_mgr_create(p_osm) != IB_SUCCESS) &&
	    (previous_result == IB_SUCCESS))
			sm->ucast_mgr.variables_invalidate_cache = TRUE;

	if (sm->ucast_mgr.variables_invalidate_cache) {
		sm->p_subn->ignore_existing_lfts = TRUE;
		if (sm->ucast_mgr.p_subn->opt.use_ucast_cache)
			osm_ucast_cache_invalidate(&sm->ucast_mgr);
	}

	/*
	 * Proceed with unicast forwarding table configuration; if it fails
	 * return early to wait for a trap or the next sweep interval.
	 */
	if (rerouting_done)
		*rerouting_done = FALSE;

	if ((sm->ucast_mgr.cache_valid == FALSE ||
	    osm_ucast_cache_process(&sm->ucast_mgr))) {
		for (i = 0; i <= MAX_ENGINES_NUMBER; i++)
			rch_mgr->used_ids[i] = FALSE;

		osm_ucast_mgr_update_groupid_maps(&sm->ucast_mgr);
		if (osm_ucast_mgr_process(&sm->ucast_mgr)) {
			osm_ucast_cache_invalidate(&sm->ucast_mgr);
			ret = FALSE;
		} else {
			if (rerouting_done)
				*rerouting_done = TRUE;
		}
	} else
		osm_topo_mgr_revert_prev(p_osm);

	/*
	 * Notice that at this point routing_lock is still acquired and links
	 * are not restored yet.
	 */
	if (ret) {
		if (sm->p_subn->opt.sharp_enabled)
			osm_validate_an2an_links(sm);

		if (sm->p_subn->opt.enable_ar_by_device_cap)
			osm_disable_ar_for_non_reordering_devices(sm);

		if (sm->ucast_mgr.vlid_cache_valid == FALSE)
			osm_subn_apply_vlids_to_lfts(sm->p_subn);

		/* hbf and whbf related calculations */
		if (sm->ucast_mgr.cache_valid == FALSE) {
			cl_qmap_apply_func(&sm->p_subn->sw_guid_tbl,
					   osm_ar_calculate_hbf, sm);
		}

		/* pFRN */
		osm_ar_calculate_pfrn(sm->p_subn);

		/* calculate FLIDs if the feature is enabled and the routing engine supports it */
		if (osm_subn_is_flids_enabled(sm->p_subn) &&
		    sm->p_subn->routing_engine_used->calculate_flids) {
			sm->p_subn->routing_engine_used->calculate_flids(
					sm->p_subn->routing_engine_used->context);
		}

		/*
		 * In planarized fabric, build correct routing from prism switches
		 * towards HCAs connected to non prism switch leaf systems.
		 */
		if (sm->p_subn->is_planarized)
			osm_build_system_ar_group(sm->p_subn);
	}

	/*
	 * When enable_queries_during_routing is enabled, main lock is not
	 * acquired at this point. Thus need to acquire before accessing to
	 * topology structures.
	 */
	if (enable_queries_during_routing)
		CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);

	osm_restore_links(sm);

	CL_PLOCK_RELEASE(sm->p_lock);
	CL_PLOCK_RELEASE(&p_osm->routing_lock);

	sm->ucast_mgr.variables_invalidate_cache = FALSE;

	state_mgr_check_reachability(sm->p_subn->p_osm);


	return ret;
}

static uint32_t osm_reset_node_changed_links(osm_sm_t * sm,
					     osm_node_t * p_node, int hops)
{
	uint8_t payload[IB_SMP_DATA_SIZE];
	ib_port_info_t *p_pi = (ib_port_info_t *) payload;
	osm_madw_context_t context;
	osm_physp_t *p_physp, *p_remote_physp;
	osm_dr_path_t * dr_path;
	ib_api_status_t status;
	uint8_t num_ports, port;
	uint8_t link_reset_mode;
	uint32_t num_link_reset = 0;

	num_ports = osm_node_get_num_physp(p_node);
	for (port = 1; port < num_ports; port++) {
		p_physp = osm_node_get_physp_ptr(p_node, port);
		if (!p_physp)
			continue;

		p_remote_physp = osm_physp_get_remote(p_physp);
		if (!p_remote_physp)
			continue;

		dr_path = osm_physp_get_dr_path_ptr(sm, p_remote_physp);
		if (dr_path->hop_count < hops)
			continue;

		link_reset_mode = (p_physp->port_reset_required >
				   p_remote_physp->port_reset_required) ?
				  p_physp->port_reset_required :
				  p_remote_physp->port_reset_required;

		if (!link_reset_mode)
			continue;

		/* Create mad */
		memcpy(p_pi, &p_physp->port_info, sizeof(ib_port_info_t));

		if (link_reset_mode == OSM_PORT_RESET_TYPE_PHYSICAL) {
			OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
				"Performing physical reset of port %d of node "
				"0x%016" PRIx64 "\n",
				port, cl_ntoh64(osm_node_get_node_guid(p_node)));
			ib_port_info_set_port_phys_state(
			    IB_PORT_PHYS_STATE_POLLING, p_pi);
			ib_port_info_set_port_state(p_pi, IB_LINK_NO_CHANGE);
		} else {
			/* link_reset_mode == OSM_PORT_RESET_TYPE_LOGICAL */
			OSM_LOG(sm->p_log, OSM_LOG_INFO,
				"Performing logical reset of port %d of node "
				"0x%016" PRIx64 "\n",
				port, cl_ntoh64(osm_node_get_node_guid(p_node)));
			ib_port_info_set_port_state(p_pi, IB_LINK_DOWN);
			ib_port_info_set_port_phys_state(
			    IB_PORT_PHYS_STATE_NO_CHANGE, p_pi);
		}

		memset(&context, 0, sizeof(context));
		context.pi_context.active_transition = FALSE;
		context.pi_context.node_guid = osm_node_get_node_guid(p_node);
		context.pi_context.port_guid = osm_physp_get_port_guid(p_physp);
		context.pi_context.set_method = TRUE;
		context.pi_context.light_sweep = FALSE;
		context.pi_context.client_rereg = FALSE;

		status = osm_req_set_pi(sm, p_physp, port, payload, FALSE,
					p_pi->m_key, &context, TRUE);
		if (status == IB_SUCCESS) {
			/* Reset port_reset_required indicators */
			p_physp->port_reset_required = OSM_PORT_RESET_TYPE_NONE;
			p_remote_physp->port_reset_required = OSM_PORT_RESET_TYPE_NONE;

			/*
			 * If this is sw to sw link we remove the link so when
			 * the link will come back the ucast cache will be
			 * invalidated
			 */
			if ((osm_node_get_type(p_physp->p_node) ==
				IB_NODE_TYPE_SWITCH) &&
			    (osm_node_get_type(p_remote_physp->p_node) ==
				IB_NODE_TYPE_SWITCH))
				osm_physp_unlink(p_physp, p_remote_physp);

			num_link_reset++;
		} else
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3350: "
				"Failed to send port reset request to port %d "
				"of node 0x%016" PRIx64 "\n",
				port,
				cl_ntoh64(osm_node_get_node_guid(p_node)));
	}

	return num_link_reset;
}

static void add_node_to_distance_list(osm_log_t * p_log,
					cl_ptr_vector_t * p_lst_vector,
					osm_node_t * p_node,
					uint8_t hop_count)
{
	cl_list_t *p_list;

	if (cl_ptr_vector_at(p_lst_vector, hop_count, (void *)&p_list) ||
	    !p_list) {
		p_list = malloc(sizeof(cl_list_t));
		if (!p_list) {
			OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 3351: "
				"Could not allocate memory for nodes distances "
				"list\n");
			return;
		}

		cl_list_construct(p_list);
		if (cl_list_init(p_list, 10)) {
			OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 3352: "
				"Failed to initialize nodes distances list\n");
			free(p_list);
			return;
		}

		cl_ptr_vector_set(p_lst_vector, hop_count, p_list);
	}

	if (cl_list_insert_tail(p_list, p_node))
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 3353: "
			"Failed to insert node to nodes distances list\n");
}

static void clear_auxiliary_ports_dr_paths(osm_sm_t *sm, osm_physp_t *p_physp)
{
	size_t dr_paths_array_size =
	    (OSM_MAX_BINDING_PORTS - 1) * sizeof(osm_dr_path_t);
	uint8_t sm_port_idx = sm->p_subn->sm_port_idx;

	if (sm_port_idx == 0) {
		memset(&p_physp->dr_path[1], 0, dr_paths_array_size);
	} else if (sm_port_idx == OSM_MAX_BINDING_PORTS - 1) {
		memset(&p_physp->dr_path[0], 0, dr_paths_array_size);
	} else {
		memset(&p_physp->dr_path[0], 0, sm_port_idx * sizeof(osm_dr_path_t));
		memset(&p_physp->dr_path[sm_port_idx  + 1], 0,
		       (OSM_MAX_BINDING_PORTS - sm_port_idx - 1) * sizeof(osm_dr_path_t));
	}
}

static void clear_unhealthy_auxiliary_ports_paths(osm_sm_t * sm)
{
	cl_qmap_t *p_port_guid_tbl;
	osm_physp_t *p_physp;
	osm_node_t *p_node;
	osm_port_t *p_port;
	cl_map_item_t *p_item;
	int i;

	p_port_guid_tbl = &sm->p_subn->port_guid_tbl;
	for (p_item = cl_qmap_head(p_port_guid_tbl);
	     p_item != cl_qmap_end(p_port_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_port = (osm_port_t *)p_item;
		p_node = p_port->p_node;

		if (p_port->disc_through == OSM_DISC_NORMAL)
			continue;

		if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH) {
			for (i = 0; i < p_node->physp_tbl_size; i++) {
				p_physp = &p_node->physp_table[i];
				if (!p_physp->port_guid)
					continue;

				clear_auxiliary_ports_dr_paths(sm, p_physp);
			}
		} else {
			p_physp = p_port->p_physp;
			clear_auxiliary_ports_dr_paths(sm, p_physp);
		}
	}
}

typedef struct bfs_data {
	void *prev_priv;
	uint8_t bfs_port_count[IB_NODE_NUM_PORTS_MAX + 1];
	uint8_t visited;
} bfs_data_t;

static void delete_bfs_data(osm_switch_t *p_sw,
			    bfs_data_t *p_bfs_data)
{
	p_sw->priv = p_bfs_data->prev_priv;
	free(p_bfs_data);
}

static bfs_data_t *create_bfs_data(osm_switch_t *p_sw)
{
	bfs_data_t *p_bfs_data;

	p_bfs_data = calloc(1, sizeof(bfs_data_t));
	if (!p_bfs_data)
		return NULL;

	p_bfs_data->prev_priv = p_sw->priv;

	return p_bfs_data;
}

static void init_bfs_data(osm_sm_t *p_sm)
{
	cl_map_item_t *item;
	osm_switch_t *p_sw;

	for (item = cl_qmap_head(&p_sm->p_subn->sw_guid_tbl);
	     item != cl_qmap_end(&p_sm->p_subn->sw_guid_tbl);
	     item = cl_qmap_next(item)) {
		p_sw = (osm_switch_t *)item;
		p_sw->priv = create_bfs_data(p_sw);
		if (!p_sw->priv) {
			OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 3336: "
				"Cannot create bfs data for switch "
				"GUID: 0x%016" PRIx64 "\n",
				cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)));
			return;
		}
	}
}

static void destroy_bfs_data(osm_sm_t *p_sm)
{
	cl_map_item_t *item;
	osm_switch_t *p_sw;

	for (item = cl_qmap_head(&p_sm->p_subn->sw_guid_tbl);
	     item != cl_qmap_end(&p_sm->p_subn->sw_guid_tbl);
	     item = cl_qmap_next(item)) {
		p_sw = (osm_switch_t *) item;
		delete_bfs_data(p_sw, p_sw->priv);
	}
}

static void update_remote_switch_dr_path(osm_dr_path_t *path, osm_node_t *p_node, uint8_t path_idx)
{
	int remote_port_num;
	osm_physp_t *p_remote_physp;

	for (remote_port_num = 0; remote_port_num < p_node->physp_tbl_size; remote_port_num++) {
		p_remote_physp = osm_node_get_physp_ptr(p_node, remote_port_num);
		if (!p_remote_physp)
			continue;

		osm_dr_path_init(osm_physp_get_dr_path_ptr_by_index(p_remote_physp, path_idx),
				 path->hop_count, path->path);
	}
}

/****s* OpenSM: sort_port_t
* NAME
*  bfs_port_t
*
* DESCRIPTION
*  Manages the sorted port data for the bfs node.
*
* SYNOPSIS
*/
typedef struct sort_port {
	uint8_t port_num;
	uint8_t port_use_count;
} sort_port_t;
/*
* FIELDS
*	port_num
*		port number.
*
*	port_use_count
*		Number of times a port was used in previous dr_paths calculations.
*
* SEE ALSO
*	bfs_node_t object
*********/

/****s* OpenSM: bfs_node_t
* NAME
*  bfs_node_t
*
* DESCRIPTION
*  Manages the node data to be inserted to a bfs_list.
*
* SYNOPSIS
*/
typedef struct bfs_node {
	cl_list_item_t list_item;
	osm_node_t *p_node;
	uint8_t start_port_idx;
	sort_port_t ports_data[IB_NODE_NUM_PORTS_MAX + 1];
} bfs_node_t;
/*
* FIELDS
*	list_item
*		Linkage structure for cl_qlist. Used by bfs_list.
*	p_node
*		Pointer to the node object for this bfs_node.
*
*	start_port_idx
*		An index in a sorted array of ports (ports_data),
*		that gives us the index in ports_data array to look
*		for the next unvisited neighbor from,
*		on the next Traverse of the BFS list
*
* 	ports_data
* 		Stores the sorted data of how many times each port was
* 		used in previous dr_paths calculations
*********/

static int cmp_func_by_port_use(const void * lhs, const void * rhs)
{
	sort_port_t *p_lhs = (sort_port_t *) lhs;
	sort_port_t *p_rhs = (sort_port_t *) rhs;

	return (p_lhs->port_use_count - p_rhs->port_use_count);
}

static bfs_node_t *create_bfs_node(osm_sm_t *p_sm, osm_node_t * p_node)
{
	int port_num;
	bfs_node_t *p_bfs_node = calloc(1, sizeof(bfs_node_t));

	if (!p_bfs_node) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 3335: "
			"Memory allocation failed for bfs_node "
			"of node 0x%016" PRIx64 "\n",
			cl_ntoh64(osm_node_get_node_guid(p_node)));
		return NULL;
	}

	p_bfs_node->p_node = p_node;

	/* Set to 0: the first index in the sorted array of ports (ports_data) */
	p_bfs_node->start_port_idx = 0;

	/*
	 * Copy the number of times the ports where used in previous
	 * dr_paths calculations, to sort it by the use counter
	 */
	for (port_num = 0; port_num < p_node->physp_tbl_size; port_num++) {
		bfs_data_t *p_bfs_data = (bfs_data_t *)p_node->sw->priv;

		p_bfs_node->ports_data[port_num].port_num = port_num;
		p_bfs_node->ports_data[port_num].port_use_count =
		    p_bfs_data->bfs_port_count[port_num];
	}

	/*
	 * Make sure the undefined ports will be the last
	 * after the sort to avoid using them
	 */
	for (; port_num < IB_NODE_NUM_PORTS_MAX + 1; port_num++)
		p_bfs_node->ports_data[port_num].port_use_count = UINT8_MAX;

	/*
	 * Make sure port 0 will be the last
	 * after the sort to avoid using it
	 */
	p_bfs_node->ports_data[0].port_use_count = UINT8_MAX;

	qsort(p_bfs_node->ports_data, p_node->physp_tbl_size,
	      sizeof(sort_port_t), cmp_func_by_port_use);

	return p_bfs_node;
}

static void set_visited(osm_switch_t *p_sw, uint8_t path_index)
{
	bfs_data_t *p_bfs_data = (bfs_data_t *)p_sw->priv;

	/*
	 * Color the visited switches with the Opensm port index,
	 * the current BFS started from + 1, to prevent the clearing
	 * of visited parameter each BFS calculation.
	 */
	p_bfs_data->visited = path_index + 1;
}

static inline boolean_t is_sw_visited(osm_switch_t *p_sw, uint8_t port_index)
{
	bfs_data_t *p_bfs_data = (bfs_data_t *)p_sw->priv;

	return (p_bfs_data->visited == port_index + 1);
}

static ib_api_status_t calculate_dr_path_process_node(osm_sm_t *p_sm, cl_qlist_t *p_bfs_list,
						      osm_physp_t *p_remote_physp,
						      osm_physp_t *p_physp, uint8_t path_index,
						      boolean_t *p_is_node_inserted)
{
	ib_api_status_t status = IB_SUCCESS;
	osm_dr_path_t path = *osm_physp_get_dr_path_ptr_by_index(p_physp, path_index);
	osm_node_t *p_remote_node = p_remote_physp->p_node;
	bfs_data_t *p_bfs_data, *p_remote_bfs_data;
	bfs_node_t *p_bfs_node;

	*p_is_node_inserted = FALSE;

	if (osm_dr_path_extend(&path, osm_physp_get_port_num(p_physp))) {
		OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG,
			"ERR 3333: DR path via SM port index %u to "
			"Port GUID: 0x%016" PRIx64 ", couldn't be extended to "
			"Port GUID: 0x%016" PRIx64 "\n",
			path_index,
			cl_ntoh64(p_physp->port_guid),
			cl_ntoh64(p_remote_physp->port_guid));
		osm_dump_dr_path(p_sm->p_log, &path, OSM_LOG_ERROR);
		status = IB_ERROR;
		goto Exit;
	}

	if (osm_node_get_type(p_remote_node) == IB_NODE_TYPE_SWITCH) {
		p_bfs_node = create_bfs_node(p_sm, p_remote_node);
		if (!p_bfs_node) {
			status = IB_INSUFFICIENT_RESOURCES;
			goto Exit;
		}

		update_remote_switch_dr_path(&path, p_remote_node, path_index);

		cl_qlist_insert_tail(p_bfs_list, &p_bfs_node->list_item);
		*p_is_node_inserted = TRUE;
		set_visited(p_remote_node->sw, path_index);

		/* Increase the use counter of the exit port on the switch */
		if (osm_node_get_type(p_physp->p_node) == IB_NODE_TYPE_SWITCH) {
			p_bfs_data = (bfs_data_t *)p_physp->p_node->sw->priv;
			p_bfs_data->bfs_port_count[p_physp->port_num]++;
		}

		/*
		 * Increase the use counter of the entrance port on the remote switch.
		 * On the next step of the BFS it doesn't use this port.
		 */
		p_remote_bfs_data = (bfs_data_t *)p_remote_node->sw->priv;
		p_remote_bfs_data->bfs_port_count[p_remote_physp->port_num]++;
	} else {
		osm_dr_path_init(osm_physp_get_dr_path_ptr_by_index(p_remote_physp, path_index),
				 path.hop_count, path.path);
	}

Exit:
	return status;
}

static void print_paths(osm_sm_t *p_sm, uint8_t port_index)
{
	cl_map_item_t *p_item;

	for (p_item = cl_qmap_head(&p_sm->p_subn->port_guid_tbl);
	     p_item != cl_qmap_end(&p_sm->p_subn->port_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		osm_port_t *p_port = (osm_port_t *) p_item;

		OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG,
			"Direct path via SM port index %u to Node GUID: 0x%016" PRIx64
			", Port GUID: 0x%016" PRIx64 "\n",
			port_index,
			cl_ntoh64(osm_node_get_node_guid(p_port->p_node)),
			cl_ntoh64(p_port->guid));
		osm_dump_dr_path(p_sm->p_log,
				 osm_physp_get_dr_path_ptr_by_index(p_port->p_physp, port_index),
				 OSM_LOG_DEBUG);
	}
}

static ib_api_status_t osm_calculate_dr_path(osm_sm_t *p_sm, ib_net64_t start_port_guid,
					     uint8_t path_index)
{
	ib_api_status_t status = IB_ERROR;
	cl_qlist_t bfs_list1, bfs_list2;
	cl_qlist_t *p_bfs_list, *p_other_bfs_list, *p_tmp_bfs_list;
	osm_node_t *p_node, *p_remote_node;
	osm_physp_t *p_physp, *p_remote_physp;
	osm_port_t *p_start_port;
	int port_num, port_idx;
	int processed_nodes_num = 0;
	bfs_node_t *p_bfs_node;
	cl_list_item_t *p_list_item;
	boolean_t is_node_inserted = FALSE;

	OSM_LOG_ENTER(p_sm->p_log);

	p_start_port = osm_get_port_by_guid(p_sm->p_subn, start_port_guid);
	if (!p_start_port)
		goto Exit;

	/* currently send MADs only through first plane */
	p_physp = osm_port_get_physp_by_plane(p_start_port, 1);

	if (!p_physp || !osm_link_is_healthy(p_physp))
		goto Exit;

	cl_qlist_init(&bfs_list1);
	cl_qlist_init(&bfs_list2);

	p_bfs_list = &bfs_list1;
	p_other_bfs_list = &bfs_list2;

	/*
	 * Add the first switch to the BFS queue.
	 * In case OpenSM runs on a switch, add this switch.
	 * Otherwise add the remote node if it's a switch
	 */
	if (osm_node_get_type(p_start_port->p_node) == IB_NODE_TYPE_SWITCH) {
		p_bfs_node = create_bfs_node(p_sm, p_start_port->p_node);
		if (!p_bfs_node)
			goto Exit;

		set_visited(p_start_port->p_node->sw, path_index);
		cl_qlist_insert_tail(p_bfs_list, &p_bfs_node->list_item);
	} else {
		p_remote_physp = osm_physp_get_remote(p_physp);
		if (!p_remote_physp)
			goto Exit;

		if ((p_physp->p_node->physp_ignored[p_physp->port_num] ==
		     OSM_PHYSP_IGNORE_EXISTING) ||
		    (p_remote_physp->p_node->physp_ignored[p_remote_physp->port_num] ==
		     OSM_PHYSP_IGNORE_EXISTING))
			goto Exit;

		calculate_dr_path_process_node(p_sm, p_bfs_list, p_remote_physp, p_physp,
					       path_index, &is_node_inserted);
	}

	while (!cl_is_qlist_empty(p_bfs_list)) {
		processed_nodes_num = 0;

		/*
		 * Traverse the BFS list and for each node try to find
		 * one unvisited peer switch to add to the second list.
		 */
		for (p_list_item = cl_qlist_head(p_bfs_list);
		     p_list_item != cl_qlist_end(p_bfs_list);
		     p_list_item = cl_qlist_next(p_list_item)) {
			p_bfs_node = (bfs_node_t *) p_list_item;
			p_node = p_bfs_node->p_node;

			/*
			 * Traverse the switch ports until found an unvisited
			 * peer switch to add to the second BFS list
			 */
			for (port_idx = p_bfs_node->start_port_idx;
			     port_idx < p_node->physp_tbl_size - 1; port_idx++) {
				processed_nodes_num++;

				/*
				 * port_num takes the least "weighted" port from the sorted array,
				 * through which to continue the discovery
				 */
				port_num = p_bfs_node->ports_data[port_idx].port_num;
				p_physp = osm_node_get_physp_ptr(p_node, port_num);
				if (!p_physp || !osm_link_is_healthy(p_physp) || !port_num)
					continue;

				p_remote_physp = osm_physp_get_remote(p_physp);
				if (!p_remote_physp)
					continue;

				p_remote_node = osm_physp_get_node_ptr(p_remote_physp);
				if (!p_remote_node)
					continue;

				/*
				 * Skip links that were discovered through
				 * unhealthy no_discover/isolated links,
				 * and heldback links
				 */
				if ((p_node->physp_ignored[port_num] ==
				     OSM_PHYSP_IGNORE_EXISTING) ||
				    (p_remote_node->physp_ignored[p_remote_physp->port_num] ==
				     OSM_PHYSP_IGNORE_EXISTING))
					continue;

				/*
				 * Check if the port is visited. In case of a HOST,
				 * need to check only that the port isn't the starting port.
				 * In case of a SWITCH, check if it is visited.
				 */
				if (((osm_node_get_type(p_remote_node) != IB_NODE_TYPE_SWITCH) &&
				     (osm_physp_get_port_guid(p_remote_physp) != start_port_guid)) ||
				    ((osm_node_get_type(p_remote_node) == IB_NODE_TYPE_SWITCH) &&
				     p_remote_node->sw && !is_sw_visited(p_remote_node->sw, path_index))) {
					calculate_dr_path_process_node(p_sm, p_other_bfs_list,
								       p_remote_physp, p_physp,
								       path_index,
								       &is_node_inserted);
					if (is_node_inserted)
						break;
				}
			}
			p_bfs_node->start_port_idx = port_idx + 1;
		}

		/*
		 * If there are no unvisited peers left to process
		 * flush current BFS list and move to process the second BFS list
		 */
		if (processed_nodes_num == 0) {
			while (!cl_is_qlist_empty(p_bfs_list)) {
				p_bfs_node = (bfs_node_t *)cl_qlist_remove_head(p_bfs_list);
				free(p_bfs_node);
			}

			p_tmp_bfs_list = p_bfs_list;
			p_bfs_list = p_other_bfs_list;
			p_other_bfs_list = p_tmp_bfs_list;
		}
	}

	if (osm_log_get_level(p_sm->p_log) >= OSM_LOG_DEBUG)
		print_paths(p_sm, path_index);

	status = IB_SUCCESS;
Exit:
	OSM_LOG_EXIT(p_sm->p_log);

	return status;
}


static boolean_t osm_reset_changed_links(osm_sm_t * sm)
{
	cl_ptr_vector_t nodes_list_vector;
	cl_map_item_t *p_item;
	osm_switch_t * p_sw;
	osm_node_t * p_node;
	osm_physp_t * p_physp;
	osm_dr_path_t * dr_path;
	osm_port_t * p_port;
	cl_list_t * p_list;
	cl_list_iterator_t iter;
	int hops;
	uint32_t num_link_reset = 0;

	OSM_LOG_ENTER(sm->p_log);

	cl_ptr_vector_construct(&nodes_list_vector);
	cl_ptr_vector_init(&nodes_list_vector,
			   NODES_DISTANCE_VECTOR_MIN_SIZE,
			   NODES_DISTANCE_VECTOR_GROW_SIZE);

	/* sort switches by direct path from sm node */
	for (p_item = cl_qmap_head(&sm->p_subn->sw_guid_tbl);
	     p_item != cl_qmap_end(&sm->p_subn->sw_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_sw = (osm_switch_t *)p_item;
		p_node = p_sw->p_node;
		p_physp = osm_node_get_physp_ptr(p_node, 0);
		OSM_ASSERT(p_physp &&(p_physp->dr_path[sm->p_subn->sm_port_idx].hop_count <=
				      IB_SUBNET_PATH_HOPS_MAX));
		dr_path = osm_physp_get_dr_path_ptr(sm, p_physp);

		add_node_to_distance_list(sm->p_log, &nodes_list_vector,
					  p_node, dr_path->hop_count);
	}

	/* handle sm port when running from HCA */
	if (sm->p_subn->opt.allow_sm_port_reset) {
		p_port = osm_get_port_by_guid(sm->p_subn,
					      sm->p_subn->sm_port_guid);
		OSM_ASSERT(p_port && p_port->p_node);
		if (osm_node_get_type(p_port->p_node) != IB_NODE_TYPE_SWITCH)
			add_node_to_distance_list(sm->p_log,
						  &nodes_list_vector,
						  p_port->p_node, 0);
	}

	/* reset ports starting from the farthest */
	for (hops = cl_ptr_vector_get_size(&nodes_list_vector) - 1; hops >= 0;
	     hops--) {
		p_list = cl_ptr_vector_get(&nodes_list_vector, hops);
		if (!p_list)
			continue;

		for (iter = cl_list_head(p_list);
		     iter != cl_list_end(p_list);
		     iter = cl_list_next(iter)) {
			p_node = (osm_node_t *)cl_list_obj(iter);
			num_link_reset += osm_reset_node_changed_links(sm, p_node, hops);
		}
		(void) wait_for_pending_transactions(sm->p_subn->p_osm);
		cl_ptr_vector_remove(&nodes_list_vector, hops);
		cl_list_remove_all(p_list);
		cl_list_destroy(p_list);
		free(p_list);
	}

	cl_ptr_vector_destroy(&nodes_list_vector);

	OSM_LOG_EXIT(sm->p_log);

	return num_link_reset;
}

/*
 *  Handle hup signal and return an indication whether heavy sweep is required
 *  due to the hup signal or not
 */
static boolean_t handle_hup_signal(osm_sm_t *p_sm)
{
	boolean_t sweep = FALSE;
	boolean_t pkey_change = FALSE, qos_change = FALSE;

	OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG, "Handling HUP signal\n");

	CL_PLOCK_EXCL_ACQUIRE(p_sm->p_lock);
	p_sm->p_subn->hup_signal = FALSE;
	CL_PLOCK_RELEASE(p_sm->p_lock);

	if (osm_two_crc_exclusive_changed(p_sm,
					  OSM_CRC_PARTITION_FILE,
					  OSM_CRC_ENHANCED_QOS_POLICY_FILE,
					  &pkey_change, &qos_change)) {
		if (pkey_change) {
			OSM_LOG_MSG_BOX(p_sm->p_log, OSM_LOG_VERBOSE,
					"INITIATING PKEY SWEEP");
			OSM_LOG(p_sm->p_log, OSM_LOG_INFO,
				"Only partitions file changed, "
				"Running pkey sweep (pkey mgr only)\n");
			osm_pkey_mgr_process(p_sm->p_subn->p_osm);
			if (wait_for_pending_transactions(
				p_sm->p_subn->p_osm))
				goto EXIT;
			osm_align_mcast_to_pkeys(p_sm->p_subn->p_osm);
			OSM_LOG_MSG_BOX(p_sm->p_log, OSM_LOG_VERBOSE,
					"PKEY SWEEP COMPLETE");
		}
		if (qos_change) {
			OSM_LOG_MSG_BOX(p_sm->p_log, OSM_LOG_VERBOSE,
					"INITIATING ENHANCED QOS SWEEP");
			OSM_LOG(p_sm->p_log, OSM_LOG_INFO,
				"Only enhanced qos policy file changed, "
				"Running enhanced qos sweep\n");
			osm_qos_setup(p_sm->p_subn->p_osm, TRUE);
			if (wait_for_pending_transactions(
				p_sm->p_subn->p_osm))
				goto EXIT;
			OSM_LOG_MSG_BOX(p_sm->p_log, OSM_LOG_VERBOSE,
					"ENHANCED QOS SWEEP COMPLETE");
		}
	} else {
		if (p_sm->crc_changed) {
			OSM_LOG(p_sm->p_log, OSM_LOG_VERBOSE, "Files other than "
				"partition file and enhanced qos policy "
				"were changed, suggesting heavy sweep\n");
			sweep = TRUE;

		} else {
			sweep = p_sm->p_subn->opt.sweep_every_hup_signal;
			OSM_LOG(p_sm->p_log, OSM_LOG_VERBOSE, "No configuration "
				"changes, heavy sweep suggestion: %s\n",
				sweep ? "TRUE" : "FALSE");
		}
	}
EXIT:
	return sweep;
}

static unsigned handle_conf_changes(osm_sm_t * sm)
{
	unsigned parsed = 0;

	/*
	 * Need to check first if main configuration file changed
	 * because it specifies the names of all other crc files,
	 * If it changed need to rescan it before recalculating crc
	 */
	if (osm_calc_files_crc(sm, OSM_CRC_CONFIG_FILE, FALSE)) {
		if (osm_subn_rescan_conf_files(sm->p_subn) < 0)
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 331A: "
				"osm_subn_rescan_conf_file failed\n");
		else
			parsed = 1;

		/*
		 * We need to calculate crc even though conf file parsing
		 * failed because when it fails we still perform the sweep
		 * with the old values that are already in the opt structure
		 */
		osm_calc_files_crc(sm, OSM_CRC_MAX, TRUE);
	} else {
		osm_calc_files_crc(sm, OSM_CRC_MAX, TRUE);

		if (sm->p_subn->force_heavy_sweep || sm->crc_changed ) {
			if (osm_subn_rescan_conf_files(sm->p_subn) < 0)
				OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 331A: "
					"osm_subn_rescan_conf_file failed\n");
			else
				parsed = 1;
		}
	}
	return parsed;
}

/*
 * If we change the physp state we might need to set it back so it might be
 * important for the user of this function to know whether a MAD was sent to
 * change physp state or not, therefore this function return an indication as
 * for whether a MAD sent to set requested state or not.
 */
static boolean_t osm_state_mgr_set_sm_physp(osm_sm_t *sm,
					    uint8_t const phys_state)
{
	boolean_t res = FALSE;
	osm_port_t *p_port;
	osm_physp_t *p_physp;
	p_port = osm_get_port_by_guid(sm->p_subn, sm->p_subn->sm_port_guid);

	if (!p_port) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3329: "
			"SM port with GUID:%016" PRIx64 " is unknown\n",
			cl_ntoh64(sm->p_subn->sm_port_guid));
                goto Exit;
	}
	p_physp = osm_port_get_physp(p_port);

	if (ib_port_info_get_port_phys_state(&p_physp->port_info) == phys_state)
                goto Exit;

	if (osm_set_port_state(sm, p_physp, phys_state, IB_LINK_NO_CHANGE)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3331: "
                        "Could not change physical port state of the SM port\n");
		goto Exit;
	}
	res = TRUE;
	(void) wait_for_pending_transactions(sm->p_subn->p_osm);
Exit:
	return res;
}

static void do_sweep_standalone(osm_sm_t * sm, boolean_t standalone_heavy_sweep)
{
	osm_port_t *sm_port;
	boolean_t disabled;

	if (sm->p_subn->sm_state == IB_SMINFO_STATE_DISCOVERING)
		osm_sm_state_mgr_process(sm, OSM_SM_SIGNAL_DISCOVERY_COMPLETED);

	osm_drop_mgr_process(sm);
	if (!sm->p_subn->first_time_master_sweep && !standalone_heavy_sweep &&
	    !sm->p_subn->last_sm_port_state) {
		if ((sm->p_subn->opt.sa_db_dump ||
		     OSM_LOG_IS_ACTIVE_V2(sm->p_log, OSM_LOG_VERBOSE)) &&
		    !osm_sa_db_file_dump(sm->p_subn->p_osm))
			osm_opensm_report_event(sm->p_subn->p_osm,
						OSM_EVENT_ID_SA_DB_DUMPED,
						NULL);

		CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);
		sm_port = osm_get_port_by_guid(sm->p_subn, sm->p_subn->sm_port_guid);
		if (sm_port && osm_port_is_virt_supported(sm_port))
			osm_virt_mgr_enqueue_port(sm, sm_port);
		CL_PLOCK_RELEASE(sm->p_lock);

		osm_dump_virtualization(sm->p_subn->p_osm);

		OSM_LOG_MSG_BOX(sm->p_log, OSM_LOG_VERBOSE,
				"LIGHT SWEEP COMPLETE");
		return;
	}
	sm->p_subn->last_sm_port_state = 0;
	OSM_LOG_MSG_BOX(sm->p_log, OSM_LOG_VERBOSE,
			"INITIATING HEAVY SWEEP STANDALONE");
	osm_opensm_report_event(sm->p_subn->p_osm,
				OSM_EVENT_ID_HEAVY_SWEEP_DONE, NULL);
	sm->p_subn->first_time_master_sweep = FALSE;
	disabled = osm_state_mgr_set_sm_physp(sm, IB_PORT_PHYS_STATE_DISABLED);

	osm_pkey_mgr_process(sm->p_subn->p_osm);

	/* try to restore SA DB (this should be before lid_mgr
	   because we may want to disable clients reregistration
	   when SA DB is restored) */
	osm_sa_db_file_load(sm->p_subn->p_osm);

	if (wait_for_pending_transactions(sm->p_subn->p_osm))
		return;
	osm_align_mcast_to_pkeys(sm->p_subn->p_osm);

	OSM_LOG_MSG_BOX(sm->p_log, OSM_LOG_VERBOSE,
			"PKEY setup completed - STARTING SM LID CONFIG");

	osm_lid_mgr_process_sm(&sm->lid_mgr);
	if (wait_for_pending_transactions(sm->p_subn->p_osm))
		return;

	OSM_LOG_MSG_BOX(sm->p_log, OSM_LOG_VERBOSE,
			"SM LID ASSIGNMENT COMPLETE - STARTING SUBNET LID CONFIG");
	state_mgr_notify_lid_change(sm);
	/* At this point we need to check the consistency of
	 * the port_lid_tbl under the subnet. There might be
	 * errors in it if PortInfo Set requests didn't reach
	 * their destination. */

	sm->p_subn->force_client_rereg = FALSE;

	OSM_LOG_MSG_BOX(sm->p_log, OSM_LOG_VERBOSE,
			"LID ASSIGNMENT COMPLETE - STARTING PORT SPEEDS CONFIG");

	state_mgr_up_msg(sm);
	/* Write a new copy of our persistent guid2mkey database */
	osm_db_store(sm->p_subn->p_g2m, sm->p_subn->opt.fsync_high_avail_files);
	osm_db_store(sm->p_subn->p_neighbor,
		     sm->p_subn->opt.fsync_high_avail_files);
	osm_db_store(sm->p_subn->p_an2an,
		     sm->p_subn->opt.fsync_high_avail_files);

	if (sm->p_subn->opt.cc_key_enable != OSM_CAPABILITY_IGNORED)
		osm_db_store(sm->p_subn->p_g2cckey,
			     sm->p_subn->opt.fsync_high_avail_files);

	if (sm->p_subn->opt.vs_key_enable != OSM_CAPABILITY_IGNORED)
		osm_db_store(sm->p_subn->p_g2vskey,
			     sm->p_subn->opt.fsync_high_avail_files);

	if (sm->p_subn->opt.n2n_key_enable != OSM_CAPABILITY_IGNORED) {
		osm_db_store(sm->p_subn->p_g2n2n_mgrkey,
			     sm->p_subn->opt.fsync_high_avail_files);
		osm_db_store(sm->p_subn->p_g2n2nkey,
			     sm->p_subn->opt.fsync_high_avail_files);
	}

	if (disabled)
		osm_state_mgr_set_sm_physp(sm, IB_PORT_PHYS_STATE_POLLING);

	osm_opensm_report_event(sm->p_subn->p_osm, OSM_EVENT_ID_SUBNET_UP,
				NULL);
}

static void clear_nodes(cl_map_item_t * const p_item, void *context)
{
	osm_node_t *p_node = (osm_node_t *)p_item;

	/* Clear MADs updated flags */
	p_node->gi_updated = FALSE;
	p_node->ext_node_info_updated = FALSE;
	if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH && p_node->sw)
		osm_ar_invalidate_ar_cache_params(p_node->sw);
}

static void osm_port_set_remote_sw(cl_map_item_t * const p_item, void *context)
{
	osm_port_t *p_port = (osm_port_t *)p_item;
	osm_node_t *p_node = p_port->p_node;
	osm_node_t *p_rem_node;
	uint8_t port_num;
	osm_sm_t *p_sm = (osm_sm_t*)context;

	/* the remote switch node of the end port of switch, is the switch itself */
	if (p_node->sw) {
		p_port->p_remote_sw_node = p_node;
		p_port->remote_sw_port = 0;
	} else {
		/* get remote node and remote port */
		port_num = osm_physp_get_port_num(osm_port_get_physp(p_port));
		p_rem_node =
			osm_node_get_remote_node(p_node, port_num, &p_port->remote_sw_port);
		p_port->p_remote_sw_node = p_rem_node;
		/* remote node doesn't exist or host to host connection */
		if (!p_rem_node || !p_rem_node->sw) {
			p_port->p_remote_sw_node = NULL;
			p_port->remote_sw_port = 0;
		}
	}

	if (p_port->p_remote_sw_node)
		OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG,
			"port with GUID: %016" PRIx64 " has remote switch GUID: %016" PRIx64 "\n",
			cl_ntoh64(osm_port_get_guid(p_port)),
			cl_ntoh64(osm_node_get_node_guid(p_port->p_remote_sw_node)));
}

static void set_remote_port(osm_sm_t *sm)
{
	osm_port_t *sm_port;

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);

	sm_port = osm_get_port_by_guid(sm->p_subn, sm->p_subn->sm_port_guid);
	/*
	 * If we don't know our own port guid yet, assume the port is down.
	 */
	if (!sm_port) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 330C: "
			"SM port with GUID: 0x%016" PRIx64 " not found\n",
			cl_ntoh64(sm->p_subn->sm_port_guid));
		goto EXIT;
	}
	cl_qmap_apply_func(&sm->p_subn->port_guid_tbl, osm_port_set_remote_sw, sm);

EXIT:
	CL_PLOCK_RELEASE(sm->p_lock);
}

static void osm_send_query_hierarchy_info(cl_map_item_t * const p_item, void *context)
{
	osm_port_t *p_port = (osm_port_t *)p_item;
	osm_sm_t *p_sm = (osm_sm_t*)context;
	osm_node_t *p_node;
	uint32_t port_num;
	osm_physp_t *p_physp;
	uint8_t type;

	if (!p_port || !p_port->p_node)
		return;

	p_node = p_port->p_node;
	type = osm_node_get_type(p_node);

	/* query all ports of switch or all plane ports of planarized HCA port */
	if (type == IB_NODE_TYPE_SWITCH ||
	    (type == IB_NODE_TYPE_CA &&
	     p_node->ext_node_info.ext_node_type == IB_EXTENDED_NODE_TYPE_MULTI_PLANE_HCA)) {
		for (port_num = 1; port_num < p_node->physp_tbl_size; port_num++) {
			p_physp = osm_node_get_physp_ptr(p_node, port_num);

			if (!p_physp)
				continue;

			/* get Hierarchical Info for ports in any state */
			osm_get_hierarchy_info(p_sm, p_node, p_port, p_physp, 0);
		}
	} else {
		p_physp = osm_port_get_physp(p_port);

		if (!p_physp)
			return;

		/* get Hierarchical Info for ports in any state */
		osm_get_hierarchy_info(p_sm, p_node, p_port, p_physp, 0);
	}

	osm_port_send_next_mad(p_sm, p_port);
}

/*
 * This function adds the port to the virt mgr queue in the following cases:
 *
 * 1) first_virt_discovery flag on - Virt still not discovered for the port
 * 2) port lid = 0 - Implies that port was rebooted or not treated correctly
 * 3) need_update  - Implies that port was rebooted or not treated correctly
 * 4) sm lid in port not equal sm lid - Implies this is first time we meet
 *    this port.
 */
static void add_port_to_virt_mgr(cl_map_item_t * const p_item, void *context)
{
	osm_port_t *p_port = (osm_port_t *)p_item;
	osm_port_t *sm_port;
	osm_node_t *p_node = p_port->p_node;
	osm_sm_t *p_sm = (osm_sm_t*)context;
	ib_net16_t sm_lid, sm_lid_by_port;

	sm_port = osm_get_port_by_guid(p_sm->p_subn, p_sm->p_subn->sm_port_guid);

	/*
	 * If we don't know our own port guid yet, assume the port is down.
	 */
	if (!sm_port) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 330B: "
			"SM port with GUID:%016" PRIx64 " not found\n",
			cl_ntoh64(p_sm->p_subn->sm_port_guid));
		goto EXIT;
	}

	if (p_node->node_info.node_type == IB_NODE_TYPE_SWITCH)
		goto EXIT;

	if (!osm_cap_mgr_is_enabled(&p_sm->cap_mgr, GI, p_port->p_node))
		goto EXIT;

	if (!p_node->gi_updated) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 332B: "
			"GeneralInfo not updated for node  0x%" PRIx64
			", Can't add port 0x%" PRIx64 " to virt manager\n",
			cl_ntoh64(p_port->guid),
			cl_ntoh64(osm_node_get_node_guid(p_node)));
		goto EXIT;
	}

	sm_lid = osm_port_get_base_lid(sm_port);
	sm_lid_by_port = p_port->p_physp->port_info.master_sm_base_lid;
	if (ib_general_info_is_cap_enabled(&p_node->general_info,
		       IB_GENERAL_INFO_VIRTUALIZATION_SUPPORTED_BIT)) {
		if (p_sm->p_subn->first_time_master_sweep ||
		    !osm_port_get_base_lid(p_port)        ||
		    (sm_lid_by_port != sm_lid)) {
			p_port->virt_data.first_virt_discovery = TRUE;
		}

		if (p_port->p_physp->need_update || p_port->virt_data.first_virt_discovery)
			osm_virt_mgr_enqueue_port(p_sm, p_port);
	}

EXIT:
	return;
}

static void send_get_pfrn_required_info(osm_sm_t * sm)
{
	ib_api_status_t status;
	cl_map_item_t *p_map_item;
	boolean_t need_update = FALSE;

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);
	for (p_map_item = cl_qmap_head(&sm->p_subn->sw_guid_tbl);
	     p_map_item != cl_qmap_end(&sm->p_subn->sw_guid_tbl);
	     p_map_item = cl_qmap_next(p_map_item)) {
		osm_switch_t *p_sw = (osm_switch_t *)p_map_item;
		need_update = p_sw->need_update || sm->p_subn->need_update;

		if (osm_ar_info_is_pfrn_supported(&p_sw->ar_data.ar_info) &&
		    (!p_sw->ar_data.pfrn_config_updated || need_update)) {
			status = osm_ar_get_pfrn_config(sm, p_sw);
			if (status != IB_SUCCESS)
				OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3615:"
					" Failed to send PFRNConfig request to switch"
					" 0x%" PRIx64 " (%s)\n",
					cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)),
					ib_get_err_str(status));
		}
	}
	CL_PLOCK_RELEASE(sm->p_lock);
}

static void send_get_ar_required_info(osm_sm_t * sm)
{
	ib_api_status_t status;
	cl_map_item_t *p_map_item;
	boolean_t need_update = FALSE;

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);
	for (p_map_item = cl_qmap_head(&sm->p_subn->sw_guid_tbl);
	     p_map_item != cl_qmap_end(&sm->p_subn->sw_guid_tbl);
	     p_map_item = cl_qmap_next(p_map_item)) {
		osm_switch_t *p_sw = (osm_switch_t *)p_map_item;
		need_update = p_sw->need_update || sm->p_subn->need_update;

		/* Reset the status of AR MADs */
		p_sw->ar_data.status = AR_STATUS_SUCCESS;
		p_sw->ar_data.pending_mads = 0;

		if (ib_general_info_is_cap_enabled(&p_sw->p_node->general_info,
			IB_GENERAL_INFO_ADAPTIVE_ROUTING_REV_SUPPORTED_BIT) &&
		    (!p_sw->ar_data.ar_info_updated || need_update)) {
			/*
			 * If need_update is true, clear capabilities cache.
			 * This is required in case switch configuration was changed.
			 */
			if (need_update)
				osm_ar_capabilities_clear_all(&p_sw->ar_data);

			status = osm_ar_get_ar_info(sm, p_sw);
			if (status != IB_SUCCESS)
				OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3612:"
					" Failed to send ARInfo request to switch"
					" 0x%" PRIx64 " (%s)\n",
					cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)),
					ib_get_err_str(status));
		}

		/* send PLFTInfo and ExternalNodeInfo MADs get */
		if (ib_general_info_is_cap_enabled(&p_sw->p_node->general_info,
			IB_GENERAL_INFO_PRIVET_LINEAR_SUPPORTED_BIT) &&
		    (!p_sw->ar_data.plft_info_updated || need_update)) {
			status = osm_ar_get_plft_info(sm, p_sw);
			if (status != IB_SUCCESS)
				OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3613:"
					" Failed to send PLFTInfo request to switch"
					" 0x%" PRIx64 " (%s)\n",
					cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)),
					ib_get_err_str(status));
		}
	}
	CL_PLOCK_RELEASE(sm->p_lock);

	/* pFRN support bit is in ARInfo, hence wait for responses before pFRN step */
	if (wait_for_pending_transactions(sm->p_subn->p_osm))
		return;

	send_get_pfrn_required_info(sm);
}

static void set_active_sm_ports(osm_sm_t * sm)
{
	uint8_t port_index;
	osm_port_t *p_port;
	ib_net64_t guid;
	osm_bind_handle_t bind;

	for (port_index = 0; port_index < OSM_MAX_BINDING_PORTS; port_index++) {
		/* lookup OpenSM port guid */
		guid = sm->p_subn->opt.guid[port_index];
		p_port = osm_get_port_by_guid(sm->p_subn, guid);
		bind = sm->mad_ctrl.h_bind[port_index];

		/*
		 * no bind and no port -> no port with such guid in the fabric
		 * no bind and port exists - probably the guid on another machine
		 * in both cases we reported when bind failed
		 */
		if (!bind) {
			sm->p_subn->active_sm_guid[port_index] = FALSE;
		} else {
			/* bind and a healthy port exists - report if port changed state */
			if (p_port && p_port->disc_through == OSM_DISC_NORMAL
			    && !lookup_isolated_port(sm, p_port->p_node, p_port->p_physp->port_num)) {
				if (sm->p_subn->active_sm_guid[port_index] == FALSE) {
					/* report that port is up */
					OSM_LOG(sm->p_log, OSM_LOG_INFO,
						"OpenSM port guid 0x%016" PRIx64
						" at port index %u is UP.\n",
						cl_ntoh64(guid), port_index);
					sm->p_subn->active_sm_guid[port_index] = TRUE;
				}
			} else {
			/* bind and no port - report if port changed state */
				if (sm->p_subn->active_sm_guid[port_index] == TRUE) {
					/* report that port is down */
					OSM_LOG(sm->p_log, OSM_LOG_INFO,
						"OpenSM port guid 0x%016" PRIx64
						" at port index %u is DOWN%s.\n",
						cl_ntoh64(guid), port_index,
						(p_port && p_port->disc_through != OSM_DISC_NORMAL)?
						" and unhealthy/behind unhealthy port" : "");
					sm->p_subn->active_sm_guid[port_index] = FALSE;
				}
			}
		}
	}
}

static void osm_calculate_ports_paths(osm_sm_t * sm)
{
	uint8_t port_idx;
	uint8_t sm_port_idx = sm->p_subn->sm_port_idx;

	init_bfs_data(sm);

	/*
	 * Calculate only the healthy paths.
	 * For unhealthy paths, leave a path to node only via main openSM port.
	 * For index 0: use the discovery dr_paths for nodes
	 * that were discovered in the ignored stage of the discovery.
	 * For the auxiliary ports:
	 * Remove any dr_paths containing unhealthy links from previous calculations.
	 * Nodes that can be reached only through an unhealthy path will not have a path.
	 */
	clear_unhealthy_auxiliary_ports_paths(sm);

	/*
	 * Override the main SM port dr_paths, that were built during discovery.
	 * Calculate the main SM port separately, because in case the first guid
	 * in the guid list is zero, the SM will run only with the main port,
	 * that is saved in sm_port_guid(if the first guid is not 0, sm_port_guid equals to it).
	 */
	osm_calculate_dr_path(sm, sm->p_subn->sm_port_guid, sm_port_idx);

	for (port_idx = 0; port_idx < sm->p_subn->opt.num_guids; port_idx++) {
		if (!sm->p_subn->active_sm_guid[port_idx] || (port_idx == sm_port_idx))
			continue;

		osm_calculate_dr_path(sm, sm->p_subn->opt.guid[port_idx], port_idx);
	}

	destroy_bfs_data(sm);
}

static ib_api_status_t change_sm_port(osm_sm_t * sm, uint8_t port_index)
{
	ib_api_status_t status;
	osm_opensm_t *p_osm = sm->p_subn->p_osm;

	OSM_LOG_ENTER(sm->p_log);
	sm->p_subn->sm_port_idx = port_index;
	sm->p_subn->sm_port_guid = sm->p_subn->opt.guid[port_index];

	status = osm_sm_mad_ctrl_bind(&sm->mad_ctrl, sm->p_subn->opt.guid[port_index], port_index, TRUE);
	if (status != IB_SUCCESS)
		goto Exit;

	status = osm_sa_bind(&p_osm->sa, sm->p_subn->sm_port_guid);
	if (status != IB_SUCCESS)
		goto Exit;

	status = osm_vendor_specific_bind(&p_osm->vs, sm->p_subn->sm_port_guid);
	if (status != IB_SUCCESS)
		goto Exit;

	status = osm_congestion_control_bind(&p_osm->cc, sm->p_subn->sm_port_guid);
	if (status != IB_SUCCESS)
		goto Exit;

#ifdef ENABLE_OSM_PERF_MGR
	status = osm_perfmgr_bind(&p_osm->perfmgr, sm->p_subn->sm_port_guid);
	if (status != IB_SUCCESS)
		goto Exit;
#endif				/* ENABLE_OSM_PERF_MGR */

	status = osm_n2n_bind(&p_osm->n2n, sm->p_subn->sm_port_guid);
	if (status != IB_SUCCESS)
		goto Exit;

	osm_vendor_set_sm(sm->mad_ctrl.h_bind[port_index], TRUE);
	p_osm->sm.ucast_mgr.variables_invalidate_cache = TRUE;

Exit:
	OSM_LOG_EXIT(sm->p_log);
	return status;
}

static inline boolean_t eni_is_need_update(osm_sm_t * sm, osm_node_t * p_node)
{
	if (p_node->sw)
		return sm->p_subn->need_update || p_node->sw->need_update;

	return sm->p_subn->need_update;
}

static void state_mgr_get_extended_node_info(IN cl_map_item_t * p_map_item, IN void *context)
{
	osm_node_t *p_node = (osm_node_t *)p_map_item;
	osm_sm_t *sm = (osm_sm_t *)context;
	osm_physp_t *p_physp;
	osm_dr_path_t *p_dr_path;
	osm_madw_context_t mad_context;
	ib_api_status_t status;
	unsigned port_num = 0;

	OSM_LOG_ENTER(sm->p_log);

	/* Avoid querying nodes that were not discovered on current sweep */
	if (!p_node->discovery_count)
		return;

	/* Validate node support ExtendedNodeInfo */
	if (!ib_general_info_is_cap_enabled(&p_node->general_info,
		IB_GENERAL_INFO_EXTENDED_NODE_INFO_SUPPORTED_BIT))
		return;

	/*
	 * For switch nodes, send ENI only if SwitchInfo was received and
	 * switch was created on internal structures.
	 */
	if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH && !p_node->sw)
		return;

	/*
	 * Validate ExtendedNodeInfo was not received from that node yet.
	 * If was received, should be resent only if node or subnet need update.
	 */
	if (p_node->ext_node_info_updated && !eni_is_need_update(sm, p_node))
		return;

	if (!p_node->sw)
		port_num = ib_node_info_get_local_port_num(&p_node->node_info);

	p_physp = osm_node_get_physp_ptr(p_node, port_num);

	p_dr_path = osm_physp_get_dr_path_ptr(sm, p_physp);

	memset(&mad_context, 0, sizeof(mad_context));

	mad_context.eni_context.node_guid = osm_node_get_node_guid(p_node);
	mad_context.eni_context.set_method = FALSE;

	p_node->ext_node_info_updated = FALSE;
	status = osm_req_get(sm, p_dr_path, IB_MAD_ATTR_MLNX_EXTENDED_NODE_INFO,
			     0, FALSE,
			     ib_port_info_get_m_key(&p_physp->port_info),
			     0, CL_DISP_MSGID_NONE, &mad_context);

	if (status != IB_SUCCESS)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 330D: "
			"Request get for ExtendedNodeInfo from 0x%" PRIx64 " failed (%s)\n",
			cl_ntoh64(osm_node_get_node_guid(p_node)),
			ib_get_err_str(status));

	OSM_LOG_EXIT(sm->p_log);
}

static void do_sweep(osm_sm_t * sm)
{
	ib_api_status_t status;
	osm_remote_sm_t *p_remote_sm;
	cl_list_item_t *p_list_item, *p_rem_item;
	unsigned config_parsed = 0;
	unsigned rediscovery_count = 0;
	uint32_t ca_num_ports;
	uint32_t num_link_reset;
	char *prev_routing_engine_name = NULL;
	struct osm_routing_engine *re;
	boolean_t standalone_heavy_sweep = FALSE;
	boolean_t hup_signal;
	boolean_t rerouting_done = FALSE;
	sm->sharp_nodes_disabled = FALSE;
	uint8_t port_index, i;
	uint8_t prev_sm_port_idx = sm->p_subn->sm_port_idx;
	uint8_t num_guids = sm->p_subn->opt.num_guids;
	boolean_t performance_test_in_progress = FALSE;
	boolean_t performance_test_discovery_in_progress = FALSE;

	if (sm->p_subn->opt.enable_sm_port_failover) {
		if (!is_sm_port_valid(sm, prev_sm_port_idx) &&
		    (sm->p_subn->opt.stand_alone == OSM_STANDALONE_NA)) {
			for (i = 0, port_index = (prev_sm_port_idx + 1) % num_guids; i < num_guids;
			     i++, port_index = (port_index + 1) % num_guids) {
				osm_opensm_t *p_osm = sm->p_subn->p_osm;
				osm_unhealthy_port_t *p_unhealthy_port;

				if (port_index == prev_sm_port_idx ||
				    !is_sm_port_valid(sm, port_index))
					continue;

				/* Don't failover to axillary port if unhealthy isolate/no_discover */
				/* HACK: in port index 0 we will have port for both SMI and GSI interface */
				p_unhealthy_port =
				    osm_hm_get_unhealthy_port_by_guid(sm, osm_vendor_get_node_guid(p_osm->p_vendor[port_index]),
								      p_osm->p_vendor[port_index]->umad_port[0].portnum);
				if (p_unhealthy_port &&
				    (p_unhealthy_port->isolate || p_unhealthy_port->no_discover))
					continue;


				osm_vendor_set_sm(sm->mad_ctrl.h_bind[prev_sm_port_idx], FALSE);
				osm_vendor_unbind(sm->mad_ctrl.h_server_bind[prev_sm_port_idx]);
				osm_sa_mad_ctrl_unbind(&p_osm->sa.mad_ctrl, prev_sm_port_idx);
				osm_vendor_unbind(p_osm->cc.bind_handle[prev_sm_port_idx]);
				osm_vendor_unbind(p_osm->vs.bind_handle[prev_sm_port_idx]);
#ifdef ENABLE_OSM_PERF_MGR
				osm_vendor_unbind(p_osm->perfmgr.bind_handle[prev_sm_port_idx]);
#endif				/* ENABLE_OSM_PERF_MGR */

				osm_vendor_unbind(p_osm->n2n.bind_handle[prev_sm_port_idx]);
				osm_vl15_shutdown(&p_osm->super_vl15.vl15[prev_sm_port_idx], &p_osm->mad_pool);
				CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);
				osm_vendor_close_ports(&p_osm->p_server_vendor[prev_sm_port_idx]);
				osm_vendor_close_ports(&p_osm->p_gmp_vendor[prev_sm_port_idx]);

				change_sm_port(sm, port_index);

				CL_PLOCK_RELEASE(sm->p_lock);
				OSM_LOG(sm->p_log, OSM_LOG_INFO,
					"Forcing Heavy Sweep due "
					"to change in SM port\n");
				sm->p_subn->force_heavy_sweep = TRUE;
				OSM_LOG(sm->p_log, OSM_LOG_INFO,
					"Setting IS_SM on port 0x%016" PRIx64 "\n",
					cl_ntoh64(sm->p_subn->opt.guid[port_index]));
				sm->p_subn->sm_port_changed = TRUE;
				break;
			}
		}
	}

	if (sm->p_subn->force_first_time_master_sweep) {
		sm->p_subn->force_heavy_sweep = TRUE;
		sm->p_subn->coming_out_of_standby = TRUE;
		sm->p_subn->first_time_master_sweep = TRUE;
		sm->p_subn->force_multicast_rerouting = TRUE;
		sm->p_subn->force_first_time_master_sweep = FALSE;
	}

	/* after subnet initialization error, run heavy sweep */
	if (sm->p_subn->subnet_initialization_error)
		sm->p_subn->force_heavy_sweep = TRUE;

	/* Force heavy sweep */
	if (sm->p_subn->light_sweep_counter >=
	    sm->p_subn->opt.force_heavy_sweep_window) {
		if (sm->p_subn->opt.virt_enabled == OSM_CAPABILITY_ENABLED)
			osm_virt_mgr_vport_state_change_check(sm);
		sm->p_subn->force_heavy_sweep = TRUE;
		OSM_LOG(sm->p_log, OSM_LOG_INFO, "Sanity heavy sweep is "
			"triggered\n");
	}

	/*
	 * In case of hup signal and chain engine is used
	 * we would like to perform heavy sweep regardless to
	 * configuration changes
	 */

	CL_PLOCK_ACQUIRE(sm->p_lock);
	hup_signal = sm->p_subn->hup_signal;
	CL_PLOCK_RELEASE(sm->p_lock);

	re = sm->p_subn->routing_engine_used;
	if (hup_signal && osm_routing_engine_is_chain(re)) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Forcing Heavy Sweep due "
			"to chain engine or health policy usage\n");
		sm->p_subn->force_heavy_sweep = TRUE;
	}

	if (sm->p_subn->force_heavy_sweep || hup_signal)
		config_parsed = handle_conf_changes(sm);

	if (sm->p_subn->sm_state != IB_SMINFO_STATE_MASTER &&
	    sm->p_subn->sm_state != IB_SMINFO_STATE_DISCOVERING)
		goto Exit;

	if (hup_signal) {
		if (handle_hup_signal(sm)){
			sm->p_subn->force_heavy_sweep = TRUE;
			standalone_heavy_sweep = TRUE;
		}

		if (!sm->p_subn->force_heavy_sweep)
			goto Exit;
	}

	if (sm->p_subn->coming_out_of_standby) {
		/*
		 * Need to force re-write of sm_base_lid to all ports
		 * to do that we want all the ports to be considered
		 * foreign
		 */
		state_mgr_clean_known_lids(sm);

		/*
		 * Need to clean virt manager queue
		 */
		state_mgr_clear_virt_mgr_queue(sm);

		/*
		 * Need to clear all virtual ports and vnodes information
		 */
		state_mgr_clear_virtual_ports(sm);

		/*
		 * Need to clean SA cache when state changes to STANDBY
		 * after handover.
		 */
		state_mgr_sa_clean(sm);

		/*
		 * Need to clean non-manual unhealthy ports across the fabric
		 */
		state_mgr_clean_non_manual_unhealthy_ports(sm);

		/* Clear multicast trees information */
		osm_subn_clear_mtrees(sm->p_subn);

		/* Clear multicast forwarding tables */
		cl_qmap_apply_func(&sm->p_subn->sw_guid_tbl,
				   state_clear_switch_mft, NULL);

		/*
		 * Need to reconfigure LFTs, PKEYs, and QoS on all switches
		 * when coming out of STANDBY
		 */
		sm->p_subn->need_update = 1;

		cl_qmap_apply_func(&sm->p_subn->node_guid_tbl, clear_nodes,
				   NULL);
	} else
		/* Clean information of recovered ports */
		state_mgr_clean_recovered_unhealthy_ports(sm);

	sm->master_sm_found = 0;

	/*
	 * If we already have switches, then try a light sweep.
	 * Otherwise, this is probably our first discovery pass
	 * or we are connected in loopback. In both cases do a
	 * heavy sweep.
	 * Note: If we are connected in loopback we want a heavy
	 * sweep, since we will not be getting any traps if there is
	 * a lost connection.
	 */
	/*  if we are in DISCOVERING state - this means it is either in
	 *  initializing or wake up from STANDBY - run the heavy sweep */
	if (cl_qmap_count(&sm->p_subn->sw_guid_tbl)
	    && sm->p_subn->sm_state != IB_SMINFO_STATE_DISCOVERING
	    && sm->p_subn->opt.force_heavy_sweep == FALSE
	    && sm->p_subn->force_heavy_sweep == FALSE
	    && sm->p_subn->force_reroute == FALSE
	    && sm->p_subn->subnet_initialization_error == FALSE
	    && (state_mgr_light_sweep_start(sm) == IB_SUCCESS)) {
		struct timeval curr_time;
		if (wait_for_pending_transactions(sm->p_subn->p_osm))
			goto Exit;
		if (!sm->p_subn->force_heavy_sweep) {
			/* If only FLIDs have been changed, signal FLID sweep */
			if (sm->router_mgr.ri_routers_changes_bitmask == IB_ROUTER_RLTC)
				osm_sm_signal(sm, OSM_SIGNAL_FLID_ROUTING_SWEEP);

			if ((sm->p_subn->opt.sa_db_dump ||
			     OSM_LOG_IS_ACTIVE_V2(sm->p_log, OSM_LOG_VERBOSE)) &&
			    !osm_sa_db_file_dump(sm->p_subn->p_osm))
				osm_opensm_report_event(sm->p_subn->p_osm,
							OSM_EVENT_ID_SA_DB_DUMPED,
							NULL);

			osm_dump_virtualization(sm->p_subn->p_osm);

			if (sm->p_subn->opt.force_heavy_sweep_window > 0)
				sm->p_subn->light_sweep_counter++;

			if (sm->p_subn->opt.offsweep_balancing_enabled == TRUE) {
				gettimeofday(&curr_time, NULL);
				if ((sm->ucast_mgr.rebalancing_required == TRUE) &&
				    (curr_time.tv_sec -
				     sm->ucast_mgr.rerouting_timestamp.tv_sec
				    >= sm->p_subn->opt.offsweep_balancing_window))
					osm_sm_signal(sm,
						      OSM_SIGNAL_IDLE_TIME_REBALANCE_ROUTING);
			}

			/* Currently N2N Class is relevant only for pFRN */
			if (sm->p_subn->pfrn_supported)
				osm_n2n_configure(sm->p_subn->p_osm);

			if (sm->p_subn->opt.mlnx_congestion_control &&
			    osm_congestion_control_setup(sm->p_subn->p_osm, FALSE))
				goto Exit;

			if (sm->p_subn->opt.vs_key_enable &&
			    osm_vendor_specific_configure_subnet_keys(sm->p_subn->p_osm))
				goto Exit;

			if (sm->hm_dirty)
				osm_hm_mgr_process(sm);

			OSM_LOG_MSG_BOX(sm->p_log, OSM_LOG_VERBOSE,
					"LIGHT SWEEP COMPLETE");

			/* Update this event only per main port, as it is not port related */
			sm->p_subn->p_osm->stats_counters[0][LIGHT_SWEEP_COUNT]++;
			goto Exit;
		}
	}

	/*
	 * Unicast cache should be invalidated when subnet re-route is
	 * requested, and when OpenSM comes out of standby state.
	 */
	if (sm->p_subn->opt.use_ucast_cache &&
	    (sm->p_subn->force_reroute || sm->p_subn->coming_out_of_standby))
		osm_ucast_cache_invalidate(&sm->ucast_mgr);

	/*
	 * If we don't need to do a heavy sweep and we want to do a reroute,
	 * just reroute only.
	 */
	if (cl_qmap_count(&sm->p_subn->sw_guid_tbl)
	    && sm->p_subn->sm_state != IB_SMINFO_STATE_DISCOVERING
	    && sm->p_subn->opt.force_heavy_sweep == FALSE
	    && sm->p_subn->force_heavy_sweep == FALSE
	    && sm->p_subn->force_reroute == TRUE
	    && sm->p_subn->subnet_initialization_error == FALSE) {
		/* Reset flag */
		sm->p_subn->force_reroute = FALSE;

		/* Reset ignored links restore flag */
		sm->p_subn->restore_links = FALSE;

		/* Re-program the switches fully */
		sm->p_subn->ignore_existing_lfts = TRUE;
		if (!osm_topo_routing(sm, NULL)) {
			OSM_LOG_MSG_BOX(sm->p_log, OSM_LOG_VERBOSE,
					"REROUTE FAILED");
			sm->p_subn->subnet_initialization_error = TRUE;
			goto Exit;
		}
		state_mgr_set_routing_tables(sm, OSM_NODE_FILTER_ALL);

		/* Currently N2N Class is relevant only for pFRN */
		if (sm->p_subn->pfrn_supported)
			osm_n2n_configure(sm->p_subn->p_osm);

		osm_qos_setup(sm->p_subn->p_osm, FALSE);

		/* Reset flag */
		if (sm->p_subn->opt.offsweep_balancing_enabled == TRUE)
			sm->ucast_mgr.rebalancing_required = FALSE;

		sm->p_subn->ignore_existing_lfts = FALSE;

		if (wait_for_pending_transactions(sm->p_subn->p_osm))
			goto Exit;

		osm_vendor_specific_setup(sm->p_subn->p_osm);
		if (sm->p_subn->opt.mlnx_congestion_control) {
			osm_congestion_control_setup(sm->p_subn->p_osm, FALSE);

			if (osm_congestion_control_wait_pending_transactions(sm->p_subn->p_osm))
				goto Exit;
		}

		if (!sm->p_subn->subnet_initialization_error) {
			OSM_LOG_MSG_BOX(sm->p_log, OSM_LOG_VERBOSE,
					"REROUTE COMPLETE");
			osm_opensm_report_event(sm->p_subn->p_osm,
						OSM_EVENT_ID_UCAST_ROUTING_DONE,
						(void *) UCAST_ROUTING_REROUTE);
			goto Exit;
		}
	}

	sm->p_subn->in_heavy_sweep = TRUE;

	OSM_LOG_MSG_BOX(sm->p_log, OSM_LOG_INFO, "HEAVY SWEEP START");
	OSM_LOG(sm->p_log, OSM_LOG_INFO, "Entering heavy sweep with flags: "
		"force_heavy_sweep %u, "
		"coming out of standby %u, "
		"subnet initialization error %u, "
		"sm port change %u\n",
		sm->p_subn->force_heavy_sweep,
		sm->p_subn->coming_out_of_standby,
		sm->p_subn->subnet_initialization_error,
		sm->p_subn->sm_port_changed);

	osm_opensm_report_event(sm->p_subn->p_osm,
				OSM_EVENT_ID_HEAVY_SWEEP_START, NULL);

	osm_subn_print_conf_changes(sm->p_subn);

	LOGPERF_START(sm->p_subn->p_osm, OSM_PERFLOG_MAIN_EVENT, "");
	performance_test_in_progress = TRUE;
	/* go to heavy sweep */
repeat_discovery:
	LOGPERF_START(sm->p_subn->p_osm, "DISCOVERY", "");
	performance_test_discovery_in_progress = TRUE;

	/* First of all - unset all flags */
	sm->p_subn->force_heavy_sweep = FALSE;
	sm->p_subn->force_reroute = FALSE;
	sm->p_subn->subnet_initialization_error = FALSE;
	sm->p_subn->found_ignored_port = FALSE;
	sm->p_subn->found_link_for_reset = FALSE;

	/* Reset tracking values in case limiting component got removed
	 * from fabric. */
	sm->p_subn->min_ca_mtu = IB_MAX_MTU;
	sm->p_subn->min_ca_rate = IB_PATH_RECORD_RATE_600_GBS;
	sm->p_subn->min_data_vls = IB_MAX_NUM_VLS - 1;
	sm->p_subn->min_sw_data_vls = IB_MAX_NUM_VLS - 1;

	/* rescan configuration updates */
	if (!config_parsed)
		config_parsed = handle_conf_changes(sm);

	osm_cap_mgr_process(&sm->cap_mgr, sm);

	if (sm->p_subn->sm_state != IB_SMINFO_STATE_MASTER)
		sm->p_subn->need_update = 1;

	status = state_mgr_sweep_hop_0(sm, (rediscovery_count == 0)? TRUE : FALSE);
	if (status != IB_SUCCESS ||
	    wait_for_pending_transactions(sm->p_subn->p_osm))
		goto Exit;

	if (state_mgr_is_sm_port_down(sm) == TRUE) {
		if (sm->p_subn->opt.stand_alone != OSM_STANDALONE_NA) {
			do_sweep_standalone(sm, standalone_heavy_sweep);
			goto Exit;
		}
		if (sm->p_subn->last_sm_port_state) {
			sm->p_subn->last_sm_port_state = 0;
			state_mgr_clear_virtual_ports(sm);
			state_mgr_sa_clean(sm);
			osm_log_v2(sm->p_log, OSM_LOG_SYS, FILE_ID, "SM port is down\n");
			osm_log_v2(sm->p_log, OSM_LOG_SYS, FILE_ID,
				   "Check SM is configured to use a physical port\n");
			OSM_LOG_MSG_BOX(sm->p_log, OSM_LOG_VERBOSE,
					"SM PORT DOWN");
		}

		/* Remove any event subscriptions */
		if (sm->p_subn->opt.drop_event_subscriptions) {
			OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
				"InformInfo list size %d\n",
				cl_qlist_count(&sm->p_subn->sa_infr_list));
			CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);
			p_list_item = cl_qlist_head(&sm->p_subn->sa_infr_list);
			while (p_list_item !=
			       cl_qlist_end(&sm->p_subn->sa_infr_list)) {
				p_rem_item = p_list_item;
				p_list_item = cl_qlist_next(p_list_item);
				osm_infr_remove_from_db(sm->p_subn, sm->p_log,
								(osm_infr_t *)p_rem_item);
			}
			CL_PLOCK_RELEASE(sm->p_lock);
		}

		/* Clear sm port */
		state_mgr_clear_sm_port_when_down(sm);

		/* Flush any queued traps and reports */
		osm_flush_traps(sm->p_subn);
		osm_flush_reports(sm->p_subn);

		/* clear virt mgr queue */
		state_mgr_clear_virt_mgr_queue(sm);

		/* Run the drop manager - we want to clear all records */
		osm_drop_mgr_process(sm);

		/* Move to DISCOVERING state */
		if (sm->p_subn->sm_state != IB_SMINFO_STATE_DISCOVERING)
			osm_sm_state_mgr_process(sm, OSM_SM_SIGNAL_DISCOVER);
		osm_opensm_report_event(sm->p_subn->p_osm,
					OSM_EVENT_ID_STATE_CHANGE, NULL);
		goto Exit;
	} else {
		if (sm->p_subn->opt.stand_alone == OSM_STANDALONE_SA) {
			sm->p_subn->last_sm_port_state = 1;
			if (sm->p_subn->sm_state != IB_SMINFO_STATE_DISCOVERING) {
                                osm_sm_state_mgr_process(sm, OSM_SM_SIGNAL_DISCOVER);
	                        osm_opensm_report_event(sm->p_subn->p_osm,
        	                                        OSM_EVENT_ID_STATE_CHANGE, NULL);
			}
			OSM_LOG_MSG_BOX(sm->p_log, OSM_LOG_VERBOSE,
					"ENTERING STANDALONE STANDBY STATE");
			sm->ucast_mgr.variables_invalidate_cache = TRUE;
			goto Exit;
                }
		if (!sm->p_subn->last_sm_port_state) {
			sm->p_subn->last_sm_port_state = 1;
			osm_inform_mgr_send_signal(&sm->p_subn->p_osm->sa);
			osm_log_v2(sm->p_log, OSM_LOG_SYS, FILE_ID, "SM port is up\n");
			OSM_LOG_MSG_BOX(sm->p_log, OSM_LOG_VERBOSE,
					"SM PORT UP");
		}
	}
	status = state_mgr_sweep_hop_1(sm);
	if (status != IB_SUCCESS ||
	    wait_for_pending_transactions(sm->p_subn->p_osm))
		goto Exit;

	status = state_mgr_sweep_ignored(sm);
	if (status != IB_SUCCESS ||
	    wait_for_pending_transactions(sm->p_subn->p_osm))
		goto Exit;

	LOGPERF_END(sm->p_subn->p_osm, "DISCOVERY", "");
	performance_test_discovery_in_progress = FALSE;
	/* discovery completed - check other sm presence */
	if (sm->master_sm_found) {
		/*
		 * Call the sm_state_mgr with signal
		 * MASTER_OR_HIGHER_SM_DETECTED_DONE
		 */
		osm_sm_state_mgr_process(sm,
					 OSM_SM_SIGNAL_MASTER_OR_HIGHER_SM_DETECTED);
		OSM_LOG_MSG_BOX(sm->p_log, OSM_LOG_VERBOSE,
				"ENTERING STANDBY STATE");
		/* notify master SM about us */
		osm_send_trap144(sm, 0);
		osm_opensm_report_event(sm->p_subn->p_osm,
					OSM_EVENT_ID_STATE_CHANGE, NULL);
		goto Exit;
	}

	/* if new sweep requested - don't bother with the rest */
	if (sm->p_subn->force_heavy_sweep) {
		boolean_t skip_discovery = FALSE;
		/* skip if discovery repeated max_seq_redisc sequential times */
		if (sm->p_subn->opt.max_seq_redisc) {
			rediscovery_count++;
			if (rediscovery_count >= sm->p_subn->opt.max_seq_redisc)
				skip_discovery = TRUE;
		}
		else
			rediscovery_count = 1;

		if (skip_discovery == FALSE) {
			config_parsed = 0;
			goto repeat_discovery;
		}
	}

	osm_opensm_report_event(sm->p_subn->p_osm,
				OSM_EVENT_ID_HEAVY_SWEEP_DONE, NULL);

	sm->p_subn->light_sweep_counter = 0;

	OSM_LOG_MSG_BOX(sm->p_log, OSM_LOG_VERBOSE, "HEAVY SWEEP COMPLETE");

	/* Update this event only per main port, as it is not port related */
	sm->p_subn->p_osm->stats_counters[0][HEAVY_SWEEP_COUNT]++;

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);
	/* Required for GPU discovery */
	cl_qmap_apply_func(&sm->p_subn->node_guid_tbl, state_mgr_get_extended_node_info, sm);
	CL_PLOCK_RELEASE(sm->p_lock);
	if (wait_for_pending_transactions(sm->p_subn->p_osm))
		goto Exit;

	if (osm_subn_is_nvlink_enabled(sm->p_subn)) {
		osm_nvlink_discovery(sm->p_subn);
		if (status != IB_SUCCESS ||
		    wait_for_pending_transactions(sm->p_subn->p_osm))
			goto Exit;
	}

	osm_drop_mgr_process(sm);

	set_remote_port(sm);

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);
	cl_qmap_apply_func(&sm->p_subn->port_guid_tbl, osm_send_query_hierarchy_info, sm);
	CL_PLOCK_RELEASE(sm->p_lock);

	if (wait_for_pending_transactions(sm->p_subn->p_osm))
		goto Exit;

	set_active_sm_ports(sm);

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);
	osm_calculate_ports_paths(sm);
	CL_PLOCK_RELEASE(sm->p_lock);

	/*
	 * After discovery and drop manager it's time to add all SRIOV
	 * supporting ports to virt manager queue
	 */
	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);
	cl_qmap_apply_func(&sm->p_subn->port_guid_tbl, add_port_to_virt_mgr, sm);
	CL_PLOCK_RELEASE(sm->p_lock);

	osm_handle_held_back_switches(sm);
	osm_port_grp_mgr_create(sm->p_subn->p_osm);

	if (sm->p_subn->opt.device_configuration_file)
		osm_parse_device_configuration_file(sm->p_subn->p_osm);

	/*
	 * Now that the heavy sweep is completed, need to check
	 * whether the SM is allowed to manage this fabric.
	 */
	ca_num_ports = cl_qmap_count(&sm->p_subn->port_guid_tbl) -
		       cl_qmap_count(&sm->p_subn->sw_guid_tbl);
	if (sm->p_subn->opt.ca_port_soft_limit > 0 &&
	    sm->p_subn->opt.ca_port_soft_limit < ca_num_ports) {
		osm_opensm_report_event(sm->p_subn->p_osm,
					OSM_EVENT_ID_SOFT_CA_PORT_NUM_THRESHOLD,
					(void *)(uintptr_t)ca_num_ports);
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3321: "
			"Fabric has %d CA ports - "
			"CA port soft limit (%d) exceeded\n",
			ca_num_ports, sm->p_subn->opt.ca_port_soft_limit);
		OSM_LOG(sm->p_log, OSM_LOG_SYS,
			"CA port number soft limit exceeded\n");
	}

	if (sm->p_subn->opt.ca_port_num_limit > 0 &&
	    sm->p_subn->opt.ca_port_num_limit < ca_num_ports) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3325: "
			"Fabric has %d CA ports - "
			"CA port hard limit (%d) exceeded\n",
			ca_num_ports, sm->p_subn->opt.ca_port_num_limit);
		OSM_LOG(sm->p_log, OSM_LOG_SYS,
			"CA port number hard limit exceeded\n");
		osm_exit_flag = TRUE;
		goto Exit;
	}

	/* If we are MASTER - get the highest remote_sm, and
	 * see if it is higher than our local sm.
	 */
	if (sm->p_subn->sm_state == IB_SMINFO_STATE_MASTER) {
		p_remote_sm = state_mgr_get_highest_sm(sm);
		if (p_remote_sm != NULL) {
			sm->p_subn->in_heavy_sweep = FALSE;
			osm_inform_mgr_send_signal(&sm->p_subn->p_osm->sa);

			/* report new ports (trap 64) before leaving MASTER */
			state_mgr_report_new_ports(sm);

			osm_inform_mgr_process(sm);

			/* need to handover the mastership
			 * to the remote sm, and move to standby */
			state_mgr_send_handover(sm, p_remote_sm);
			osm_sm_state_mgr_process(sm,
						 OSM_SM_SIGNAL_HANDOVER_SENT);
			goto Exit;
		} else {
			/* We are the highest sm - check to see if there is
			 * a remote SM that is in master state. */

			/*
			 * If polling_sm_guid == OSM_INVALID_GUID, it indicates
			 * that we either received handover or had polling
			 * timeout and it should be cleared.
			 */
			CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);
			if (sm->polling_sm_guid == OSM_INVALID_GUID)
				sm->polling_sm_guid = 0;
			CL_PLOCK_RELEASE(sm->p_lock);

			p_remote_sm = state_mgr_exists_other_master_sm(sm);
			if (p_remote_sm != NULL) {
				/* There is a remote SM that is master.
				 * need to wait for that SM to relinquish control
				 * of its portion of the subnet. C14-60.2.1.
				 * Also - need to start polling on that SM. */
				CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);
				sm->polling_sm_guid = p_remote_sm->smi.guid;
				CL_PLOCK_RELEASE(sm->p_lock);
				osm_sm_state_mgr_process(sm,
							 OSM_SM_SIGNAL_WAIT_FOR_HANDOVER);
				goto Exit;
			} else if (sm->polling_sm_guid) {
				/* Stop polling SM if it's not found */
				osm_sm_state_mgr_process(sm,
							 OSM_SM_SIGNAL_POLLING_TIMEOUT);
				goto Exit;
			}
		}
	}

	/*
	 * If we are not MASTER already - this means that we are
	 * in discovery state. call osm_sm_state_mgr with signal
	 * DISCOVERY_COMPLETED
	 */
	if (sm->p_subn->sm_state == IB_SMINFO_STATE_DISCOVERING)
		osm_sm_state_mgr_process(sm, OSM_SM_SIGNAL_DISCOVERY_COMPLETED);

	osm_build_aports(sm);

	/* analyze aports for asymmetry */
	if (sm->p_subn->is_planarized)
		osm_analyze_systems_aports(sm);

	osm_set_active_sm_priority(sm);

	osm_mark_rebooted_ports(sm);

	osm_reset_switch_state_change_bit(sm->p_subn->p_osm);
	if (wait_for_pending_transactions(sm->p_subn->p_osm))
		goto Exit;

	osm_router_mgr_process(&sm->router_mgr);

	/* Discover virtual ports */
	if (sm->p_subn->opt.virt_enabled &&
	    (sm->p_subn->first_time_master_sweep || sm->p_subn->coming_out_of_standby ||
	     sm->p_subn->sm_port_changed)) {
		osm_virt_mgr_sweep(sm, OSM_VIRT_MGR_SWEEP_STAGE_DISCOVERY);
	}

	osm_pkey_mgr_process(sm->p_subn->p_osm);

	osm_qos_init(sm->p_subn->p_osm);

	/* try to restore SA DB (this should be before lid_mgr
	   because we may want to disable clients reregistration
	   when SA DB is restored) */
	osm_sa_db_file_load(sm->p_subn->p_osm);

	if (wait_for_pending_transactions(sm->p_subn->p_osm))
		goto Exit;

	osm_align_mcast_to_pkeys(sm->p_subn->p_osm);

	OSM_LOG_MSG_BOX(sm->p_log, OSM_LOG_VERBOSE,
			"PKEY setup completed - STARTING SM LID CONFIG");

	osm_lid_mgr_process_sm(&sm->lid_mgr);
	if (wait_for_pending_transactions(sm->p_subn->p_osm))
		goto Exit;

	OSM_LOG_MSG_BOX(sm->p_log, OSM_LOG_VERBOSE,
			"SM LID ASSIGNMENT COMPLETE - STARTING SUBNET LID CONFIG");
	state_mgr_notify_lid_change(sm);

	osm_lid_mgr_process_subnet(&sm->lid_mgr);
	if (wait_for_pending_transactions(sm->p_subn->p_osm))
		goto Exit;

	/* Assign LIDs virtual ports */
	if (sm->p_subn->opt.virt_enabled == OSM_CAPABILITY_ENABLED &&
	    (sm->p_subn->first_time_master_sweep || sm->p_subn->coming_out_of_standby ||
	     sm->p_subn->sm_port_changed)) {
		osm_virt_mgr_sweep(sm, OSM_VIRT_MGR_SWEEP_STAGE_SETUP);
	}

	osm_guid_mgr_process(sm);
	if (wait_for_pending_transactions(sm->p_subn->p_osm))
		goto Exit;
	OSM_LOG_MSG_BOX(sm->p_log, OSM_LOG_VERBOSE, "ALIAS GUIDS CONFIGURED");

	/* At this point we need to check the consistency of
	 * the port_lid_tbl under the subnet. There might be
	 * errors in it if PortInfo Set requests didn't reach
	 * their destination. */

	state_mgr_check_tbl_consistency(sm);

	sm->p_subn->force_client_rereg = FALSE;

	OSM_LOG_MSG_BOX(sm->p_log, OSM_LOG_VERBOSE,
			"LID ASSIGNMENT COMPLETE - STARTING PORT SPEEDS CONFIG");

	/*
	 * Preprocess links without sending MADs.
	 */

	osm_link_mgr_preprocess_links(sm);

	OSM_LOG_MSG_BOX(sm->p_log, OSM_LOG_VERBOSE,
			"PORTS SPEEDS CONFIGURED - STARTING SWITCH TABLE CONFIG");

	/*
	 * Remember the last used routing engine, in order to force multicast
	 * recalculation if routing engine changed.
	 */
        prev_routing_engine_name = strdup(osm_routing_engine_name(sm->p_subn->routing_engine_used));

	if (sm->p_subn->first_time_master_sweep)
		sm->p_subn->ignore_existing_lfts = TRUE;
	else if (sm->p_subn->opt.offsweep_balancing_enabled) {
		sm->p_subn->ignore_existing_lfts = FALSE;
	}

	send_get_ar_required_info(sm);

	if (wait_for_pending_transactions(sm->p_subn->p_osm))
		goto Exit;

	LOGPERF_START(sm->p_subn->p_osm, "ROUTING", "");

	/*
	 * Determine whether hbf_reroute_optimization can be used
	 * This optimization allow SM to minimize number of forwarding table
	 * changes when HBF is enabled on all SLs and supported on all switches.
	 */
	sm->p_subn->hbf_reroute_optimization =
	    !sm->p_subn->first_time_master_sweep &&
	    (sm->p_subn->opt.hbf_sl_mask == 0xffff);

	if (osm_topo_routing(sm, &rerouting_done) == FALSE) {
		sm->p_subn->subnet_initialization_error = TRUE;
		LOGPERF_END(sm->p_subn->p_osm, "ROUTING", "");
		goto Exit;
	}
	sm->p_subn->hbf_reroute_optimization = FALSE;
	LOGPERF_END(sm->p_subn->p_osm, "ROUTING", "");

	/*
	 * Update flags according to routing decision
	 */

	if (sm->p_subn->routing_engine_used &&
	    !osm_routing_engine_is_named(sm->p_subn->routing_engine_used, prev_routing_engine_name))
		sm->p_subn->force_multicast_rerouting = TRUE;

	/* rebalancing performed */
	if (sm->p_subn->opt.offsweep_balancing_enabled) {
		/*
		* if rerouting was performed with ignore_existing_lfts = TRUE
		*  no need to rebalance
		*/
		if (rerouting_done) {
			gettimeofday(&sm->ucast_mgr.rerouting_timestamp, NULL);
			if (sm->p_subn->ignore_existing_lfts == TRUE) {
				sm->ucast_mgr.rebalancing_required = FALSE;
			} else {
				sm->ucast_mgr.rebalancing_required = TRUE;
			}
		}
	}

	sm->p_subn->ignore_existing_lfts = FALSE;


	/* Deploy calculated routing tables */
	state_mgr_activate_links(sm, FALSE);
	state_mgr_set_routing_tables(sm, OSM_NODE_FILTER_CHANGED);
	state_mgr_set_routing_tables(sm, OSM_NODE_FILTER_UNCHANGED);

	LOGPERF_START(sm->p_subn->p_osm, "QOS", "");
	osm_qos_setup(sm->p_subn->p_osm, FALSE);

	if (wait_for_pending_transactions(sm->p_subn->p_osm)) {
		LOGPERF_END(sm->p_subn->p_osm, "QOS", "");
		goto Exit;
	}
	LOGPERF_END(sm->p_subn->p_osm, "QOS", "");


	OSM_LOG_MSG_BOX(sm->p_log, OSM_LOG_VERBOSE,
			"SWITCHES CONFIGURED FOR UNICAST");
	osm_opensm_report_event(sm->p_subn->p_osm,
				OSM_EVENT_ID_UCAST_ROUTING_DONE,
				(void *) UCAST_ROUTING_HEAVY_SWEEP);

	if (!sm->p_subn->opt.disable_multicast) {
		osm_subn_parse_mc_roots_file(sm->p_subn);
		osm_mcast_mgr_process(sm, TRUE);
		if (wait_for_pending_transactions(sm->p_subn->p_osm))
			goto Exit;
		OSM_LOG_MSG_BOX(sm->p_log, OSM_LOG_VERBOSE,
				"SWITCHES CONFIGURED FOR MULTICAST");

		/*
		 * When incremental multicast routing feature is disabled, we
		 * set p_subn->force_multicast_rerouting to TRUE in order to
		 * force recalculation of the multicast trees.
		 */
		sm->p_subn->force_multicast_rerouting =
			!sm->p_subn->opt.enable_inc_mc_routing;
	}

	state_mgr_activate_links(sm, TRUE);

	/* deactivate unhealthy / isolated links */
	osm_link_mgr_process(sm, IB_LINK_DOWN, TRUE);
	if (wait_for_pending_transactions(sm->p_subn->p_osm))
		goto Exit;

	/* Activate virtual ports */
	if (sm->p_subn->opt.virt_enabled == OSM_CAPABILITY_ENABLED &&
	    (sm->p_subn->first_time_master_sweep || sm->p_subn->coming_out_of_standby ||
	     sm->p_subn->sm_port_changed)) {
		osm_virt_mgr_sweep(sm, OSM_VIRT_MGR_SWEEP_STAGE_ACTIVATION);
	}

	/*
	 * Get NodeDescriptions after SMLID set in port, and drop manager
	 * and routing completed to eliminate possible race condition
	 *
	 * If sharp nodes were just disabled we would like to avoid timeouts
	 * on node description from TCA that was removed
	 */
	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);
	sm->p_subn->in_heavy_sweep = FALSE;
	CL_PLOCK_RELEASE(sm->p_lock);

	if (!sm->sharp_nodes_disabled)
		osm_update_node_desc(sm->p_subn->p_osm);
	if (wait_for_pending_transactions(sm->p_subn->p_osm))
		goto Exit;

	/*
	 * The sweep completed!
	 */

	/* Currently N2N Class is relevant only for pFRN */
	if (sm->p_subn->pfrn_supported)
		osm_n2n_configure(sm->p_subn->p_osm);

	osm_vendor_specific_setup(sm->p_subn->p_osm);

	/* Fast Recovery configuration (including setting feature profiles) */
	if (sm->p_subn->opt.fast_recovery_enabled)
		osm_fast_recovery_configure(sm->p_subn);

	/* Now do GSI configuration */
	if (sm->p_subn->opt.mlnx_congestion_control) {
		LOGPERF_START(sm->p_subn->p_osm, "CC_MGR", "");
		osm_congestion_control_setup(sm->p_subn->p_osm, TRUE);

		if (osm_congestion_control_wait_pending_transactions(sm->p_subn->p_osm)) {
			LOGPERF_END(sm->p_subn->p_osm, "CC_MGR", "");
			goto Exit;
		}
		LOGPERF_END(sm->p_subn->p_osm, "CC_MGR", "");
	}

	/* reset links that were marked to be reset */
	if (sm->p_subn->found_link_for_reset) {
		num_link_reset = osm_reset_changed_links(sm);
		if (num_link_reset) {
			sm->p_subn->force_heavy_sweep = TRUE;
			OSM_LOG(sm->p_log, OSM_LOG_INFO,
				"Forcing heavy sweep due to %d link resets\n",
				num_link_reset);
		}
	}

	osm_inform_mgr_send_signal(&sm->p_subn->p_osm->sa);

	/*
	 * Send trap 64 on newly discovered endports
	 */
	state_mgr_report_new_ports(sm);
	osm_inform_mgr_process(sm);

	/* check switch lft buffers assignments */
	cl_qmap_apply_func(&sm->p_subn->sw_guid_tbl, check_switch_lft,
			   sm);
	osm_hm_mgr_process(sm);

	/* in any case we zero this flag */
	sm->p_subn->coming_out_of_standby = FALSE;
	sm->p_subn->sm_port_changed = FALSE;

	sm->p_subn->first_time_master_sweep = FALSE;

	/* If there were errors - then the subnet is not really up */
	if (sm->p_subn->subnet_initialization_error == TRUE) {
		osm_log_v2(sm->p_log, OSM_LOG_SYS, FILE_ID,
			   "Errors during initialization\n");
		OSM_LOG_MSG_BOX(sm->p_log, OSM_LOG_ERROR,
				"ERRORS DURING INITIALIZATION");
	} else {
		sm->p_subn->need_update = 0;
		state_mgr_up_msg(sm);

		if ((OSM_LOG_IS_ACTIVE_V2(sm->p_log, OSM_LOG_VERBOSE) ||
		    sm->p_subn->opt.sa_db_dump) &&
		    !osm_sa_db_file_dump(sm->p_subn->p_osm))
			osm_opensm_report_event(sm->p_subn->p_osm,
						OSM_EVENT_ID_SA_DB_DUMPED,
						NULL);
	}

	cl_qmap_apply_func(&sm->p_subn->sw_guid_tbl,
			   state_mgr_reset_switch_need_update, NULL);

	/*
	 * Finally signal the subnet up event
	 */
	cl_event_signal(&sm->subnet_up_event);

	/* if we got a signal to force heavy sweep or errors
	 * in the middle of the sweep - try another sweep. */
	if (sm->p_subn->force_heavy_sweep)
		osm_sm_signal(sm, OSM_SIGNAL_SWEEP);

	/* Write a new copy of our persistent guid2mkey database */
	osm_db_store(sm->p_subn->p_g2m, sm->p_subn->opt.fsync_high_avail_files);
	osm_db_store(sm->p_subn->p_neighbor,
		     sm->p_subn->opt.fsync_high_avail_files);
	osm_db_store(sm->p_subn->p_an2an,
		     sm->p_subn->opt.fsync_high_avail_files);

	if (sm->p_subn->opt.cc_key_enable != OSM_CAPABILITY_IGNORED)
		osm_db_store(sm->p_subn->p_g2cckey,
			     sm->p_subn->opt.fsync_high_avail_files);

	if (sm->p_subn->opt.vs_key_enable != OSM_CAPABILITY_IGNORED)
		osm_db_store(sm->p_subn->p_g2vskey,
			     sm->p_subn->opt.fsync_high_avail_files);

	if (sm->p_subn->opt.n2n_key_enable != OSM_CAPABILITY_IGNORED) {
		osm_db_store(sm->p_subn->p_g2n2n_mgrkey,
			     sm->p_subn->opt.fsync_high_avail_files);
		osm_db_store(sm->p_subn->p_g2n2nkey,
			     sm->p_subn->opt.fsync_high_avail_files);
	}

	/* Dump SM database, and if subnet initialized, dump additional files */
	osm_dump_subn_all(sm->p_subn->p_osm, sm->p_subn, "",
			  sm->p_subn->subnet_initialization_error ?
			  OSM_DUMP_FLAG_SMDB : OSM_DUMP_FLAG_ALL);

	/* Store the group IDs in persistent file */
	osm_ucast_mgr_group_id_db_store(&sm->ucast_mgr);

	osm_opensm_report_event(sm->p_subn->p_osm, OSM_EVENT_ID_SUBNET_UP,
				NULL);
Exit:
	free(prev_routing_engine_name);

	if (performance_test_discovery_in_progress)
		LOGPERF_END(sm->p_subn->p_osm, "DISCOVERY", "");
	if (performance_test_in_progress)
		LOGPERF_END(sm->p_subn->p_osm, OSM_PERFLOG_MAIN_EVENT, "");
	return;
}

static void do_process_mgrp_queue(osm_sm_t * sm)
{
	if (!sm->p_subn->opt.disable_multicast) {
		osm_mcast_mgr_process(sm, sm->p_subn->reroute_all_mlids);
		(void) wait_for_pending_transactions(sm->p_subn->p_osm);
	}

	if (sm->p_subn->force_heavy_sweep ||
	    sm->p_subn->subnet_initialization_error)
		osm_sm_signal(sm, OSM_SIGNAL_SWEEP);
}

static void do_process_guid_queue(osm_sm_t *sm)
{
	osm_guid_mgr_process(sm);
	(void) wait_for_pending_transactions(sm->p_subn->p_osm);
}

void osm_state_mgr_process(IN osm_sm_t * sm, IN osm_signal_t signal)
{
	boolean_t disabled;
	OSM_ASSERT(sm);

	OSM_LOG_ENTER(sm->p_log);

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Received signal %s in state %s\n",
		osm_get_sm_signal_str(signal),
		osm_get_sm_mgr_state_str(sm->p_subn->sm_state));

	/*
	 * If polling_sm_guid is not 0, there is another master SM in the fabric
	 * with lower priority. Consecutively the SM is polling this other SM
	 * for handover or timeout. In this case and the case the SM is not in
	 * master state the SM needs to stop processing events except from OSM_SIGNAL_SWEEP.
	 */
	if (signal != OSM_SIGNAL_SWEEP) {
		if ((sm->p_subn->sm_state != IB_SMINFO_STATE_MASTER) ||
		    sm->polling_sm_guid) {
			OSM_LOG(sm->p_log, OSM_LOG_VERBOSE, "Ignoring signal %s in state %s\n",
				osm_get_sm_signal_str(signal),
				osm_get_sm_mgr_state_str(sm->p_subn->sm_state));
			return;
		}
	}

	switch (signal) {
	case OSM_SIGNAL_SWEEP:
		if (!sm->p_subn->sweeping_enabled) {
			OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "sweeping disabled - "
				"ignoring signal %s in state %s\n",
				osm_get_sm_signal_str(signal),
				osm_get_sm_mgr_state_str(sm->p_subn->sm_state));
		} else {
			cl_timer_stop(&sm->sweep_timer);
			do_sweep(sm);
			if (sm->p_subn->opt.sweep_interval)
				cl_timer_start(&sm->sweep_timer,
					sm->p_subn->opt.sweep_interval * 1000);
		}
		break;
	case OSM_SIGNAL_IDLE_TIME_PROCESS_REQUEST:
		do_process_mgrp_queue(sm);
		break;
	case OSM_SIGNAL_GUID_PROCESS_REQUEST:
		do_process_guid_queue(sm);
		if (sm->p_subn->force_client_rereg) {
			sm->p_subn->force_heavy_sweep = TRUE;
			osm_sm_signal(sm, OSM_SIGNAL_SWEEP);
		}
		break;
	case OSM_SIGNAL_VIRTMGR_PROCESS_REQUEST:
		disabled = FALSE;
		if (sm->p_subn->opt.stand_alone == OSM_STANDALONE_SA &&
                    state_mgr_is_sm_port_down(sm) == TRUE)
			disabled = osm_state_mgr_set_sm_physp(sm,
						IB_PORT_PHYS_STATE_DISABLED);
		osm_virt_mgr_process(sm);
		if (sm->p_subn->opt.stand_alone == OSM_STANDALONE_SA &&
                    state_mgr_is_sm_port_down(sm) == TRUE && disabled)
			osm_state_mgr_set_sm_physp(sm, IB_PORT_PHYS_STATE_POLLING);
		break;
	case OSM_SIGNAL_INFORM_PROCESS_REQUEST:
		osm_inform_mgr_process(sm);
		break;
	case OSM_SIGNAL_IDLE_TIME_REBALANCE_ROUTING:
		if (sm->ucast_mgr.rebalancing_required == FALSE)
			break;

		/* Re-program the switches fully */
		 OSM_LOG(sm->p_log, OSM_LOG_INFO,
			 "Offsweep rebalancing started\n");
		if (sm->ucast_mgr.p_subn->opt.use_ucast_cache)
			osm_ucast_cache_invalidate(&sm->ucast_mgr);

		sm->p_subn->ignore_existing_lfts = TRUE;

		if (!osm_topo_routing(sm, NULL)) {
			OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
				"REROUTE FAILED\n");
			sm->p_subn->subnet_initialization_error = TRUE;
			osm_sm_signal(sm, OSM_SIGNAL_SWEEP);
			break;
		}
		state_mgr_set_routing_tables(sm, OSM_NODE_FILTER_ALL);

		osm_qos_setup(sm->p_subn->p_osm, FALSE);

		/* Reset flags */
		sm->ucast_mgr.rebalancing_required = FALSE;

		sm->p_subn->ignore_existing_lfts = FALSE;

		if (wait_for_pending_transactions(sm->p_subn->p_osm))
			break;

		if (!sm->p_subn->opt.disable_multicast) {
			sm->p_subn->force_multicast_rerouting = TRUE;
			osm_mcast_mgr_process(sm, TRUE);
			if (wait_for_pending_transactions(sm->p_subn->p_osm))
				return;
			OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
				"SWITCHES CONFIGURED FOR MULTICAST\n");

			/*
			 * When incremental multicast routing feature is disabled, we
			 * set p_subn->force_multicast_rerouting to TRUE in order to
			 * force recalculation of the multicast trees.
			*/
			sm->p_subn->force_multicast_rerouting =
				!sm->p_subn->opt.enable_inc_mc_routing;
		}

		OSM_LOG(sm->p_log, OSM_LOG_INFO,
				"Offsweep rebalancing finished\n");
		break;

	case OSM_SIGNAL_DUMP_STATISTIC:
		osm_dump_statistic(sm->p_subn->p_osm);
		break;
	case OSM_SIGNAL_FLID_ROUTING_SWEEP:
		osm_router_mgr_process_changes(&sm->router_mgr);
		break;
	default:
		OSM_ASSERT(FALSE);
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3320: "
			"Invalid SM signal %u\n", signal);
		break;
	}

	OSM_LOG_EXIT(sm->p_log);
}
